#ifndef LANG_CORE_TYPES_HPP
#define LANG_CORE_TYPES_HPP

#include <memory>
#include <boost/exception/all.hpp>

namespace leaf
{

typedef char src_char_t;

} //eon leaf

using std::unique_ptr;
using std::shared_ptr;
using std::make_shared;

namespace error
{
	typedef boost::error_info<struct error_reason,std::string> tag_error_reason;
	typedef boost::error_info<struct error_condition,std::string> tag_error_condition;
	
	/*
		The variadic template allows a very simple reusable macro that can be used with
		just a condition, with a message, and optional additional exception parameters.
	*/
	
	template<typename X>
	void extend_exception( X & /*ex*/ )
	{ }

	//I'm really hoping somebody can help me cleanup these 3 string functions:
	//http://stackoverflow.com/questions/14241495/variadic-template-trouble-matching-const-and-non-const-stdstring
	template<typename X>
	void extend_exception( X & ex, std::string && value )
	{ ex << tag_error_reason( value ); }
	template<typename X>
	void extend_exception( X & ex, std::string const & value )
	{ ex << tag_error_reason( value ); }
	template<typename X>
	void extend_exception( X & ex, std::string & value )
	{ ex << tag_error_reason( value ); }

	template<typename X>
	void extend_exception( X & ex, char const * value )
	{
		ex << tag_error_reason( value );
	}

	template<typename X, typename T>
	void extend_exception( X & ex, T&& value )
	{
		ex << value;
	}

	template<typename X, typename TF, typename ... TR>
	void extend_exception( X & ex, TF&& f, TR&& ... rest )
	{
		extend_exception( ex, f );
		extend_exception( ex, rest... );
	}

	template<typename X, typename ... TA>
	X create_exception( TA ... all )
	{
		X ex;
		extend_exception( ex, all... );
		return ex;
	}
} //eon error
	
/**
	Recoverable errors are thrown when the internal state has been corrupted
	and the throwing object chain is no longer valid.
	
	Refer to:
	http://mortoray.com/2013/01/03/how-to-handle-an-error/
*/
struct recoverable_error : virtual boost::exception, virtual std::exception
{
	recoverable_error() { }
	recoverable_error( char const * msg )
	{
		(*this) << error::tag_error_reason( msg );
	}
	recoverable_error( std::string const & msg )
	{
		(*this) << error::tag_error_reason( msg );
	}
	recoverable_error( recoverable_error const & o )
		: boost::exception( o ), std::exception( o )
	{ }
};

/**
	Unwindable errors are thrown when the internal state has not been changed
	and the objects can continue to be used.
	
	It is derived from a recoverable_error since a recovery catch can deal with these
	whereas an unwindable catch cannot deal with recoverable.
*/
struct unwindable_error : public recoverable_error
{
	unwindable_error() { }
	unwindable_error( char const * msg )
	{
		(*this) << error::tag_error_reason( msg );
	}
	unwindable_error( std::string const & msg )
	{
		(*this) << error::tag_error_reason( msg );
	}
	unwindable_error( unwindable_error const & o )
		: boost::exception( o ), std::exception( o ), recoverable_error()
	{ }
};

#define STATE_CHECK( cond_, ... )	\
	if( !(cond_) ) { BOOST_THROW_EXCEPTION( \
		error::create_exception<recoverable_error>( __VA_ARGS__ ) \
		<< error::tag_error_condition( #cond_ ) ); }
#define STATE_FAIL( ... )	\
	BOOST_THROW_EXCEPTION( \
		error::create_exception<recoverable_error>( __VA_ARGS__ )  )
	
#define PRE_CHECK( cond_, ... )	\
	if( !(cond_) ) { BOOST_THROW_EXCEPTION( \
		error::create_exception<unwindable_error>( __VA_ARGS__ ) \
		<< error::tag_error_condition( #cond_ ) ); }
#define PRE_FAIL( ... )	\
	BOOST_THROW_EXCEPTION( \
		error::create_exception<unwindable_error>( __VA_ARGS__ )  )
	
#endif