#ifndef LEAF_IR_INSTRUCTION_HPP
#define LEAF_IR_INSTRUCTION_HPP

#include "fwd.hpp"

namespace ir
{

struct instruction
{
	enum type_t
	{
		t_acquire_shared,
		t_alloc_array,
		t_alloc_shared,
		t_array_offset,
		t_array_ptr_offset,
		t_binary,
		t_bind_func,
		t_bind_func_explicit,
		t_bind_raw_func,
		t_cast,
		t_clone,
		t_comment,
		t_fail,
		t_free_array,
		t_funccall,
		t_loop,
		t_loop_flow,
		t_loss,
		t_malloc,
		t_pack_tuple,
		t_promote,
		t_release_shared,
		t_resolve,
		t_resume,
		t_return,
		t_special_field,
		t_store,
		t_struct_pointer,
		t_switch,
		t_unary,
		t_void_context,
	} type;
	
	virtual ~instruction() {};
	
	/**
		Checks that the state of the isntruction is valid. This will throw a recoverable_error
		if something is wrong (corruption).
	*/
	virtual void validate() const { }
protected:
	instruction( type_t type )
		: type( type )
	{ }
};

/**
	Stores data at the given location. The left types supported are:
		raw_pointer<T> : in which case the "right" value T is stored at that location
		shared_pointer<T> : in which case the "right" value T is stored in the shared value location
*/
struct store_inst : public instruction
{
	value_ref left, right;
	
	store_inst()
		: instruction( t_store )
	{}
	
	void validate() const;
};

/**	
	Exposes the results of the funciton in the "returns" values.
*/
struct funccall_inst : public instruction
{
	std::vector<value_ref> returns, params;
	//which function to call
	value_ref call;
	//provide an explicit context for a context function
	boost::optional<value_ref> context;
	
	funccall_inst()
		: instruction( t_funccall )
	{ }
	
	void validate() const;
};

struct pack_tuple_inst : public instruction
{
	std::vector<value_ref> sub;
	value_defn result;
	
	pack_tuple_inst()
		: instruction( t_pack_tuple )
	{ }
	
	void validate() const;
};

/**
	Allows both a value returning and non-value returning conditional series.
*/
struct switch_inst : public instruction
{
	struct conditional_block
	{
		shared_ptr<block> cond_block, expr_block; //init'd in ctor
		value_ref expr_value;
		boost::optional<value_ref> cond_value; //if not set then assumed true
		
		conditional_block( block & parent );
	};
	std::vector<shared_ptr<conditional_block>> blocks;
	value_defn result; //a phi of the individual expr_value's

	//true if expr_value and result are defined, otherwise no result values
	bool with_result;
	
	switch_inst( bool with_result )
		: instruction( t_switch )
		, with_result( with_result )
	{ }
	
	void validate() const;
};

/**
	Does iteration over a block so long as a condition remains true.
*/
struct loop_inst : public instruction
{
	shared_ptr<block> cond_block, body_block; //init'd in ctor
	value_ref cond_value; //must be produced by cond_block
	
	loop_inst( block & parent);
	
	void validate() const;
};

/**
	Reduces an addressing value_ref to a value_defn.
*/
struct resolve_inst : public instruction
{
	value_ref source;
	value_defn result;
	
	resolve_inst( )
		: instruction( t_resolve )
	{ }
	
	void validate() const;
};

/**
	Obtains a pointer to a structure field given the current pointer.
*/
struct struct_pointer_inst : public instruction
{
	value_ref struct_ptr;
	size_t index;
	
	value_defn result;
	
	struct_pointer_inst()
		: instruction( t_struct_pointer )
	{ }
	
	void validate() const { /*TODO:*/ };
};

/**
	Allocate memory on the heap.
*/
struct malloc_inst : public instruction
{
	//allocate an area to hold this type
	shared_ptr<ir::type const> direct_type;
	//the resulting memory, this type should be a pointer to direct_type
	value_defn result;
	
	malloc_inst()
		: instruction( t_malloc )
	{ }
	
	void validate() const;
};

/**
	Contructs a bound function from a raw function.
*/
struct bind_raw_func_inst : public instruction
{
	value_ref raw_func;
	value_defn result;
	
	bind_raw_func_inst()
		: instruction( t_bind_raw_func )
	{ }
	
	void validate() const;
};

struct shared_object_defn
{
	//allocate an are to hold a shared object of this type
	shared_ptr<ir::type const> direct_type;
	//the resulting memory, this type should be a pointer to direct_type (see users for ptr type)
	value_defn result;
	//which function accepts a raw_pointer to direct_type and destroys it
	value_ref destructor;
	//destructor is not specified, will be provided at time of release
	bool static_destructor;
	
	shared_object_defn()
		: static_destructor( false )
	{ }
	
	void validate() const;
};

struct alloc_shared_inst : public instruction
{
	//result is a shared_ptr
	shared_object_defn sod;
	
	alloc_shared_inst()
		: instruction( t_alloc_shared )
	{ }
	
	void validate() const;
};

/**
	Builds a context to call a function. The context is a shared object.
*/
struct bind_func_inst : public instruction
{
	//the ctx_func to call
	value_ref ctx_func;
	//the context which is bound
	shared_object_defn ctx_data;
	
	//the resulting bound function
	value_defn result;
	
	bind_func_inst()
		: instruction( t_bind_func )
	{ }
	
	void validate() const;
};

/**
	Binds the provided context to the unbound function to produce a
	bound function.
	
	TODO: the naming of this and bind_func_inst should be clarified, and if possible
	merged. This one is more logically binding, the other is creating the context.
*/
struct bind_func_explicit_inst : public instruction 
{
	value_ref ctx_func;
	value_ref ctx_data;
	
	value_defn result;
	
	bind_func_explicit_inst()
		: instruction( t_bind_func_explicit )
	{ }
	
	void validate() const;
};

/**
	Builds a ctx_function which will result in an exception at runtime
	if ever called. There is no need for the resulting object to be unique.
*/
struct void_context_inst : public instruction
{
	value_defn result;
	
	void_context_inst()
		: instruction( t_void_context )
	{ }
	
	void validate() const { /*TODO:*/ }
};

/**
	Release a shared object
*/
struct release_shared_inst : public instruction
{
	enum destroy_how_t
	{
		//no destructor is invoked
		dh_trivial,
		//destructor was registered at creation time
		dh_dynamic,
	} destroy_how;
	
	value_ref object;
	
	release_shared_inst()
		: instruction( t_release_shared )
	{ }
	
	void validate() const;
};

struct acquire_shared_inst : public instruction
{
	value_ref object;
	
	acquire_shared_inst()
		: instruction( t_acquire_shared )
	{ }
	
	void validate() const;
};

struct return_inst : public instruction
{
	return_inst()
		: instruction( t_return )
	{ }
	
	void validate() const { };
};

struct alloc_array_inst : public instruction
{
	alloc_array_inst()
		: instruction( t_alloc_array )
	{ }
	
	value_defn result;
	shared_ptr<ir::type const> item_type;
	value_ref length;
	
	void validate() const;
};

struct free_array_inst : public instruction
{
	free_array_inst()
		: instruction( t_free_array )
	{ }
	
	value_ref array;
	
	void validate() const;
};

struct array_offset_inst : public instruction
{
	array_offset_inst()
		: instruction( t_array_offset )
	{ }
	
	value_ref array, offset;
	value_defn result;
	
	void validate() const;
};

struct array_ptr_offset_inst : public instruction
{
	array_ptr_offset_inst()
		: instruction( t_array_ptr_offset )
	{ }
	
	bool lvalue;
	value_ref array, offset;
	value_defn result;
	
	void validate() const;
};


struct special_field_inst : public instruction
{
	enum field_t
	{
		//size_t result, object must be array
		f_array_len,
	} field;
	
	special_field_inst()
		: instruction( t_special_field )
	{}
	
	value_ref object;
	value_defn result;
	
	void validate() const;
};

struct promote_loss_inst : public instruction
{
protected:
	promote_loss_inst( type_t t ) 
		: instruction( t ) {
	}

public:
	enum what_t {
		w_integer,
		w_float,
		w_binary,
	} to_what, from_what;
	unsigned to_bitsize;
	
	value_ref source;
	value_defn result;
	
	void validate() const;
};

struct promote_inst : public promote_loss_inst
{
	promote_inst()
		: promote_loss_inst( t_promote )
	{}
};
struct loss_inst : public promote_loss_inst
{
	loss_inst()
		: promote_loss_inst( t_loss )
	{}
};

struct fail_inst : public instruction
{
	fail_inst()
		: instruction( t_fail )
	{ }
	
	//at this point we just use an integer level (all handled the same)
	int level;
};

struct resume_inst : public instruction
{
	resume_inst()
		: instruction( t_resume )
	{ }
};

struct loop_flow_inst : public instruction
{
	loop_flow_inst()
		: instruction( t_loop_flow )
	{ }
	
	enum flow_t {
		f_break,
		f_next,
	} flow;
	void validate() const;
};

/**
	A fundamental binary operation. These will be limited to generally those
	that would benefit from a close-to-the-host compilation.
*/
struct binary_inst : public instruction
{
	binary_inst()
		: instruction( t_binary )
	{ }
	
	enum op_t {
		o_add, 
		o_sub,
		o_mul,
		o_div,
		
		o_mod,
		
		o_bit_shl,
		o_bit_shr,
		o_bit_or,
		o_bit_and,
		o_bit_xor,
		
		o_eq,
		o_neq,
		o_lt,
		o_lte,
		o_gt,
		o_gte,
	} op;
	static const op_t min_op_ = o_add;
	static const op_t max_op_ = o_gte;
	bool is_boolean() const {
		return op >= o_eq && op <= o_gte;
	}
	
	value_ref a, b;
	value_defn result;
	
	void validate() const;
};

struct unary_inst : public instruction
{
	unary_inst()
		: instruction( t_unary )
	{ }

	enum op_t {
		o_neg, 
		o_abs,
		
		o_bit_cpl,
	} op;
	static const op_t min_op_ = o_neg;
	static const op_t max_op_ = o_bit_cpl;

	value_ref param;
	value_defn result;
	
	void validate() const;
};

/**
	Allows adding comments to the IR. Has no affect on the execution.
*/
struct comment_inst : public instruction
{
	comment_inst()
		: instruction( t_comment )
	{ }

	std::string comment;
};

/**
	Creates a temporary copy of the pointer-to-value and returns a
	pointer to that temporary.
*/
struct clone_inst : public instruction
{
	clone_inst()
		: instruction( t_clone )
	{ 
		source_value = false;
	}
	
	//the source is not a pointer, the result should be a pointer to a copy of the source value
	bool source_value;
	value_ref source;
	value_defn result;
	
	void validate() const;
};

/**
	Treat one type as another: the size must be the same.
*/
struct cast_inst : public instruction
{
	cast_inst()
		: instruction( t_cast )
	{ }
	
	value_ref source;
	value_defn result;
	
	void validate() const;
};

} //eon ir

#endif