#ifndef LEAF_IR_LLVM_GEN_HPP
#define LEAF_IR_LLVM_GEN_HPP

#include "ir/fwd.hpp"

#include <deque>
#include <map>

//these are all forward declared so a user of this class doesn't need to include LLVM headers
namespace llvm
{
	class LLVMContext;
	class Module;
	class Value;
	class Function;
	class BasicBlock;
	class GlobalValue;
	class Linker;
	class ExecutionEngine;
	class DataLayout;
	class Type;
	class PointerType;
	class FunctionType;
}

namespace ir_llvm
{

class function_builder;
class type_manager;

class block_data;
typedef shared_ptr<block_data> block_ptr;

class gen
{
public:
	struct options
	{
		//enable backend(LLVM) optimizations
		bool llvm_opt;
		//distinct exception flow (instead of combined)
		bool distinct_exception_flow;
		//try to pre-reduce branching constructs to only those needed
		bool reduce_branch;
		//try to reduce blocks used in flow
		bool reduce_flow;
		//generate PIC objects
		bool pic;
		
		options()
			: llvm_opt( false )
			, distinct_exception_flow( false )
			, reduce_branch( true )
			, reduce_flow( true )
			, pic( false )
		{ }
	};
	gen( std::string const & base_ll, options gopts = options() );
	~gen();
	
	int execute( ir::module const & m, std::string const & show_code );
	void emit_object( ir::module const & m, std::string const & base_path,
		std::string const & show_code );
private:
	void show_code( std::string const & sc );
	
	options gopts;
	
	unique_ptr<llvm::LLVMContext> llvm_context;
	unique_ptr<llvm::Module> module, base_module;
	unique_ptr<llvm::DataLayout> data_layout;
	unique_ptr<llvm::Linker> linker;
	unique_ptr<type_manager> types;
	std::multimap<std::string,llvm::Function*> func_map;

	//used to do initiliazation of global values
	llvm::Function * startup_func;
	unique_ptr<function_builder> startup_func_builder;
	
	//related to exception handling
	llvm::Value * except_personality;
	llvm::Type * except_type;
	
	
	void link();
	llvm::ExecutionEngine * create_engine();
	
	void create_std_functions();
	llvm::Value * gen_malloc( llvm::Value * size, function_builder & bb );
	llvm::Value * gen_malloc( size_t size, function_builder & bb );
	llvm::Value * heap_alloc_type( llvm::Type * t, function_builder & bb );
	
	void gen_free( llvm::Value * ptr, function_builder & bb );
	
	void build_module( ir::module const & m );
	void build_wrapped_entry();
	void build_block( ir::block const & bk, function_builder & bb );
	void build_block_part( ir::block_seq_item const & bk, function_builder & bb );
	void setup_function( ir::function const & func );
	void build_function( ir::function const & func );
	
	void build_instruction( ir::instruction const & i, function_builder & bb );
	void build_global_variable( ir::value const & gv );
	
	#define LLVM_GEN_BUILD_INSTRUCTIONS \
		BINST(acquire_shared) \
		BINST(alloc_array) \
		BINST(alloc_shared) \
		BINST(array_offset) \
		BINST(array_ptr_offset) \
		BINST(binary) \
		BINST(bind_func) \
		BINST(bind_func_explicit) \
		BINST(bind_raw_func) \
		BINST(cast) \
		BINST(clone) \
		BINST(comment) \
		BINST(fail) \
		BINST(free_array) \
		BINST(funccall) \
		BINST(loop) \
		BINST(loop_flow) \
		BINST(loss) \
		BINST(malloc) \
		BINST(pack_tuple) \
		BINST(promote) \
		BINST(release_shared) \
		BINST(resolve) \
		BINST(resume) \
		BINST(return) \
		BINST(special_field) \
		BINST(store) \
		BINST(struct_pointer) \
		BINST(switch) \
		BINST(unary) \
		BINST(void_context)
	
	#define BINST(type) \
		void build_##type##_inst( ir::type##_inst const & i, function_builder & bb);
		LLVM_GEN_BUILD_INSTRUCTIONS
	#undef BINST

	llvm::Value* obtain_value( ir::value_ref const & vr, function_builder & bb );
	
	enum landing_type_t : int {
		lt_cleanup = -1,
		lt_all = 0,
		//positive values are level to catch
	};
 	block_ptr create_landing_block( block_ptr block, block_ptr next, function_builder & bb, 
 		int landing_type );
	std::map<int,llvm::Value*> clause_level_globals;
	
	//internal tracking of symbols
	struct symbol_cref
	{
		llvm::Value * value;
		
		enum ref_type_t
		{
			rt_value,
			rt_pointer,
			rt_global
		} ref_type;
	};
	typedef std::deque<symbol_cref> symbol_deque_t;
	symbol_deque_t symbol_deque;
	void create_symbol_cref( intptr_t & track, llvm::Value * v, symbol_cref::ref_type_t rt );
};

} //eon ir

#endif