#ifndef LEAF_IR_LLVM_TYPE_MANAGER_HPP
#define LEAF_IR_LLVM_TYPE_MANAGER_HPP

#include "ir/fwd.hpp"

#include <map>

namespace llvm
{
	class Type;
	class FunctionType;
	class PointerType;
	class LLVMContext;
	class DataLayout;
	class IntegerType;
	class StructType;
}

namespace ir_llvm
{

//fields of the generic context wrappers
enum context_field : unsigned
{
	cf_ref_count = 0,
	//function to call on deletion
	cf_destructor,
	//a pointer to the function to call (unbound function)
	cf_func_ptr,
	//the data to pass as the first argument
	cf_ctx_ptr,
	//this *must* be last, and it is actaully the start of the data, not a pointer itself (not always used)
	cf_ctx_offset,
	
	cf_size
};

enum shared_field : unsigned
{
	sf_ref_count = 0,
	sf_destructor,
	sf_data,
	
	sf_size
};

//shared objects share the same basic headers
static_assert( unsigned(sf_ref_count) == unsigned(cf_ref_count), "" );
static_assert( unsigned(sf_destructor) == unsigned(cf_destructor), "" );

/**
	Separates logic dealing with types from the gen class to make it clearer.
*/
class type_manager
{
	llvm::LLVMContext & context;
	llvm::DataLayout & data_layout;
	
	llvm::PointerType * generic_ptr_;
	llvm::Type * void_;
	llvm::FunctionType * destructor_;
	llvm::StructType * extype_type_;
public:
	type_manager( llvm::LLVMContext & context, llvm::DataLayout & data_layout );
	
	llvm::PointerType * generic_ptr()
	{ return generic_ptr_; }
	llvm::Type * type_void()
	{ return void_; }
	llvm::IntegerType * f_intx( unsigned bits );
	llvm::IntegerType * f_natural_int();
	llvm::IntegerType * f_abi_char();
	llvm::PointerType * ptr_to( llvm::Type * t );
	llvm::FunctionType * destructor()
	{ return destructor_; }
	llvm::Type * f_float( unsigned bits );
	llvm::StructType * extype_type()
	{ return extype_type_; }

	llvm::IntegerType* main_argc();
	llvm::PointerType* main_argv();
	
	size_t alloc_size( llvm::Type * t );
	size_t max_align();
	
	//types
	struct type_ref {
		ir::type const * ir_type;
		bool ctx, shared;
		
		type_ref()
			: ctx( false )
			, shared( false )
		{ }
		
		bool operator<( type_ref const & o ) const;
	};
	struct type_entry {
		llvm::Type * ll_type;
		//can it be used directly as a return value
		bool direct_return;
		
		type_entry()
			: ll_type( nullptr )
			, direct_return( false )
		{ }
	};
	typedef std::map<type_ref, type_entry> type_map_t;
	type_map_t type_map;
	
	//The type when used as a typical value
	type_entry value( ir::type const & type ); //assume safe to hold pointer
	
	struct func_type_entry {
		//the full function type
		llvm::FunctionType * ll_func_type;
		//the return type (logical, whether returned or as parameter)
		llvm::Type * ll_ret_type;
		//true if returned directly from funciton, false if as parameter
		bool is_direct_return;
		//true if this function uses a context
		bool is_context;
	};
	typedef std::map<type_ref, func_type_entry> func_type_map_t;
	func_type_map_t func_type_map;
	func_type_entry true_function( ir::type const & type );
	
	//ll_type is a StructType
	type_entry true_ctx_function( ir::type const & t );
	
	type_entry true_shared( ir::type const & t );

	/**
		Generalizes lookup and caching code.
		
		LLVM's asserts check types by pointer equality thus we must ensure identical
		types get the exact same type object back. I have no idea how this will work
		accross modules! 
	*/
	template<typename Map, typename TypeEntry>
	struct type_map_helper_ {
		type_map_helper_( Map & tmap, ir::type const & gtype )
			: tmap( tmap ) { 
			tr.ir_type = &gtype;
		}
		
		Map & tmap;
		type_ref tr;
		
		bool has() {
			return tmap.find( tr ) != tmap.end();
		}
		TypeEntry get() {
			return tmap.find( tr )->second;
		}
		void set( TypeEntry te ) {
			tmap[tr] = te;
		}
	};

	type_map_helper_<type_map_t,type_entry> type_map_helper( ir::type const & gtype ) {
		return type_map_helper_<type_map_t,type_entry>( type_map, gtype );
	}
	type_map_helper_<func_type_map_t,func_type_entry> func_type_map_helper( ir::type const & gtype ) {
		return type_map_helper_<func_type_map_t,func_type_entry>( func_type_map, gtype );
	}
};

} //eon ir_llvm

#endif