#ifndef LEAF_IR_TYPE_HPP
#define LEAF_IR_TYPE_HPP

#include "../core_types.hpp"
#include "value.hpp"

#include <vector>

namespace util {
	struct validation_context;
}

namespace ir
{

struct type
{
	enum fundamental_t
	{
		f_function_ptr = 1, //sub=returns, params=params
		f_raw_pointer, //raw pointer to sub[0]
		f_shared_pointer, //shared object of sub[0]
		f_tuple, //sub for parts
		f_array, //array to sub[0]
		f_optional, //sub[0] = boolean, sub[1] = part (behaves as tuple for LLVM)
		
		//trivial types
		f_integer,
		f_binary,
		f_float,
		f_boolean,
	} fundamental;
	
	//for integer/float
	int trivial_bitsize;
	//for function_ptr
	enum access_t {
		a_bound = 0,
		a_raw,
		a_unbound
	} access;

	bool has_trivial_bitsize() const
	{ return fundamental == f_integer || fundamental == f_float || fundamental == f_binary; }
	
	//list item types, or returns for function, sub[0] for pointer types
	std::vector<shared_ptr<type const>> sub;
	//params for function
	std::vector<shared_ptr<type const>> params;
	//for functions an explicit context type, or null if private
	shared_ptr<type const> context;
	
	//name of the variable (empty if unnamed/irrelevant)
	std::string name;
	
	//custom pre-destruction code
	boost::optional<value_defn> destroy;
	
	//packing for a tuple type
	enum pack_t {
		p_normal = 0,
		p_abi,
		p_byte,
	} pack;
	
	shared_ptr<type> clone( bool with_name = true ) const;
	
	/**
		Compares the types. This does not consider the name.
	*/
	bool operator<(type const & o) const
	{ return compare(o) < 0; }
	bool operator==(type const &o) const
	{ return compare(o) == 0; }
	
	int compare( type const & o ) const;
	//compare just the sub and params components
	int sub_compare( type const & o ) const;
	
	type( fundamental_t f )
		: fundamental( f )
	{ complete_init(); }
	type( access_t access )
		: fundamental( f_function_ptr )
		, access( access )
	{ complete_init(); }
	type( type const & o ) = delete;
	type() = delete;
	void operator=(type const &) = delete;
	
	void validate() const;
	void validate_impl( util::validation_context & vctx ) const;
	
	bool is_integer() const
	{ return fundamental == f_integer; }
	bool is_binary() const
	{ return fundamental == f_binary; }
	bool is_float() const
	{ return fundamental == f_float; }
	bool is_scalar() const
	{ return is_integer() || is_float() || is_binary(); }
	bool is_boolean() const
	{ return fundamental == f_boolean; }
	bool is_shared_pointer() const
	{ return fundamental == f_shared_pointer; }
 	bool is_raw_pointer() const
 	{ return fundamental == f_raw_pointer; }
	bool is_shared_object_pointer() const
	{ return fundamental == f_shared_pointer 
		|| (fundamental == f_function_ptr && access == a_bound); }
	bool is_callable() const
	{ return fundamental == f_function_ptr; }
	bool is_array() const
	{ return fundamental == f_array; }
	bool is_optional() const
	{ return fundamental == f_optional; }

	//for types with items returns that item type
	shared_ptr<type const> item_type() const
	{ STATE_CHECK( sub.size() == 1 ); return sub[0]; }

	shared_ptr<type const> sub_type( size_t ndx ) const
	{ STATE_CHECK( ndx < sub.size() ); return sub[ndx]; }
	
	/**
		Obtains a name of this type suitable for use in exported symbols. It will be
		unique per type.
	*/
	std::string const & get_export_name() const;
	
	/**
		Checks if this type will require a destructor when it is being released.
		@return true indicates the value can be trivially released and no part release
			needs to be performed
	*/
	bool is_trivial_destruction() const;
	
	static shared_ptr<type> get_raw_pointer( shared_ptr<type const> t );
	static shared_ptr<type const> get_deref( shared_ptr<type const> t );
	static shared_ptr<type const> get_deref_shared( shared_ptr<type const> t );
	static shared_ptr<type> get_shared_pointer( shared_ptr<type const> t );
	static shared_ptr<type> get_raw_shared( shared_ptr<type const> t );
	static shared_ptr<type> get_array( shared_ptr<type const> t );
	
private:
	void complete_init() {
	}
	
	//cache the export name
	mutable std::string export_name;
	
	typedef std::pair<type const*, type const*> type_pair;
	typedef std::vector<type_pair> type_pair_stack;
	
	int sub_compare( type const & o, type_pair_stack & tps ) const;
	int compare( type const & o, type_pair_stack & tps ) const;
	int common_sub_compare( std::vector<shared_ptr<type const>> const & a, 
		std::vector<shared_ptr<type const>> const & o, type_pair_stack & tps ) const;
};

} //eon ir

#endif