#ifndef EXPRESSION_HPP
#define EXPRESSION_HPP

#include "scope_ref.hpp"
#include "statement.hpp"
#include "intr_type.hpp"

#include "util/number.hpp"
#include "util/unicode.hpp"

#include <boost/optional.hpp>

#include <vector>

namespace leaf
{

class statement_block;
class var_statement;
struct expression;

/**
	A statement which evalutes a single expression. This needs to be distinct from the
	expression itself since it has properties only a statement can have (like error handling)
*/
struct expression_statement : public error_statement
{
	statement_form_t statement_form() const
	{ return sf_expression_statement; }
	
	shared_ptr<expression> expr;
	
	virtual void validate() const;
};

#define EXPRESSION_FORMS \
	EXPRESSION_FORM(binary) \
	EXPRESSION_FORM(block) \
	EXPRESSION_FORM(conditional) \
	EXPRESSION_FORM(const_number) \
	EXPRESSION_FORM(default) \
	EXPRESSION_FORM(field) \
	EXPRESSION_FORM(funccall) \
	EXPRESSION_FORM(funcdefn) \
	EXPRESSION_FORM(new_func) \
	EXPRESSION_FORM(special_field) \
	EXPRESSION_FORM(subscript) \
	EXPRESSION_FORM(switch) \
	EXPRESSION_FORM(symbol) \
	EXPRESSION_FORM(text) \
	EXPRESSION_FORM(transform) \
	EXPRESSION_FORM(tuple) \
	EXPRESSION_FORM(unary) \
	EXPRESSION_FORM(value_list)

/**
	Expressions are a series of variables, operators and functions and evaluate to a particular
	value.
	
	SUBSCRIPT: This is not treated like a function call since it has special typing rules and
	special considerations for constants and tuples.
*/
struct expression
{
	enum form_t
	{
		#define EXPRESSION_FORM(form) f_##form,
		EXPRESSION_FORMS
		#undef EXPRESSION_FORM
	} form;

	static char const * name_of( form_t f );

	//the resulting type of this expression
	extr_type_ref type;
	//the specified type of this expression (from parsing)
	shared_ptr<type_spec> spec_type;
	
	//a possible root name for this expression (may be useful for backend generation)
	std::string sug_name;
	
	virtual ~expression() { }
	
	virtual void validate() const;
	
	/**
		If the expression has an output and its type is constructable from the inputs
		then do that now. By default this throws since it shouldn't be called unless it
		is known safe to do so.
	*/
	virtual void construct_result_type()
	{ STMT_FAIL(*this); }

	/**
		If the expression is merely a transform of an inner expression this returns
		the numbe of those inner expressions. If the inner expression is not strictly
		transformed then this should return 0
	*/
	virtual size_t num_inner_trans() const
	{ return 0; }
	virtual shared_ptr<expression> get_inner_trans( size_t /*ndx*/ ) 
	{ STMT_FAIL(*this); }
	virtual void set_inner_trans( size_t /*ndx*/, shared_ptr<expression> /*expr*/ )
	{ STMT_FAIL(*this); }
	
	/**
		@return true if the expression could have side-effects and/or evaluate differently
			each time it is evaluated.
	*/
	virtual bool may_have_side_effects() const
	{ return true; /*safe response*/ }
	
	/**
		If set indicates the source location which best represents the location
		of this statement.
	*/
	boost::optional<util::source_location> src_loc;
	
	bool complete;
	bool is_complete() const {
		return complete && type.is_complete();
	}
	
	//for the compiler to use
	mutable intptr_t cref;
protected:
	expression( form_t form )
		: form( form )
	{ 
		cref = 0;
		complete = true;
	}
};

struct expr_const_number : public expression
{
	expr_const_number()
		: expression( f_const_number )
	{ }
	
	 bool may_have_side_effects() const
	 { return false; }
	
	enum src_form_t 
	{
		//pre-typing
		sf_rational,
		sf_boolean,
		//post-typing types
		sf_integer,
		sf_binary,
		sf_float,
	} src_form;
	
	long double float_value() const
	{ return value.as_double(); } //TODO: need long_double
	int64_t integer_value() const
	{ return value.as_int(); }
	bool boolean_value() const
	{ return val_boolean; }
		
	void set_integer_value( int64_t val )
	{ value = util::number::integer(val); }
	
	util::number value;
	bool val_boolean;
	
	void validate() const;
};

struct expr_symbol : public expression
{
	expr_symbol()
		: expression( f_symbol )
	{ }
	
	std::string symbol;
	
	bool may_have_side_effects() const
	{ return false; }

	//for symbol (POST-TYPING), to which var_statement this symbol refers
	scope_ref sref;
	
	void validate() const;
};

/**
	A constant textual expression.
*/
struct expr_text : public expression
{
	expr_text()
		: expression( f_text )
	{ }
	
	enum what_t {
		w_string,
		w_char,
	} what;
	
	unicode::codestr text;
};

struct expr_with_block : public expression
{
	shared_ptr<statement_block> block;

	void validate() const;
protected:
	expr_with_block( form_t form  )
		: expression( form )
	{ }
};

struct expr_block : public expr_with_block
{
	expr_block()
		: expr_with_block( f_block )
	{ }
};

struct expr_funcdefn : public expr_with_block
{
	expr_funcdefn()
		: expr_with_block( f_funcdefn )
	{ }
	
	//an entry in param_defaults is set for each parameter which has a default value
	std::map<size_t,shared_ptr<expression>> param_defaults;
};

/**
	Array indexing/offsetting. 
	
	On input this may also contain field access, since the parser cannot know if the variable
	is an array. Typing however should convert such accesses to "expr_field" and expr_subscript
	is strictly for indexing.
*/
struct expr_subscript : public expression
{
	expr_subscript()
		: expression( f_subscript )
	{ }
	
	shared_ptr<expression> base, offset;
	
	void validate() const;
};

/**
	Access to a field of a tuple or custom type.
*/
struct expr_field : public expression
{
	expr_field()
		: expression( f_field )
	{
		bind = false;
	}
	
	shared_ptr<expression> base;
	//the name of the field
	std::string field;
	//a field_index overrides name saying exactly which item
	boost::optional<size_t> field_index;
	//a scope_ref is a dependent field in a custom type
	boost::optional<scope_ref> sref;
	
	//binds the base to the result, such as a bound functiopn
	bool bind;
	
	void validate() const;
};

/**
	Accesses a special field/attribute of a given object.
	
	NOTE: Though the form is basically the same as "transform" or "unary" these classes
	are kept distinct since they convey a different meaning.
*/
struct expr_special_field : public expression
{
	expr_special_field()
		: expression( f_special_field )
	{ }
	
	enum field_t
	{
		f_array_size,
		f_optional_set,
	} field;
	shared_ptr<expression> base;
	
	void validate() const;
};

struct expr_switch : public expression
{
	expr_switch()
		: expression( f_switch )
	{ }
	
	struct expr_list_item
	{
		shared_ptr<expression> cond; //null for an else clause
		shared_ptr<expression> value;
		
		void validate() const;
	};
	typedef std::vector<expr_list_item> expr_list_t;
	expr_list_t expr_list;
	
	size_t num_inner_trans() const
	{ return expr_list.size(); }
	shared_ptr<expression> get_inner_trans( size_t ndx )
	{ return expr_list[ndx].value; }
	void set_inner_trans( size_t ndx, shared_ptr<expression> expr )
	{ expr_list[ndx].value = expr; }
	
	bool may_have_side_effects() const;
	void construct_result_type();
	void validate() const;
};

struct expr_with_sub : public expression
{
	//C++: unique_ptr would actually be more correct, but needing to always call std::move just
	//reduces code clarity. The standard is really missing a non-thread-safe shared_ptr (I'll
	//have to create one)
	std::vector<shared_ptr<expression> > sub;

	void validate() const;
protected:
	expr_with_sub( form_t form )
		: expression( form )
	{ }
};

struct expr_value_list : public expr_with_sub
{
	expr_value_list()
		: expr_with_sub( f_value_list )
	{ }
};

struct expr_tuple : public expr_with_sub
{
	expr_tuple()
		: expr_with_sub( f_tuple )
	{ }
};

struct expr_funccall : public expression
{
	expr_funccall()
		: expression( f_funccall )
	{ }
	
	shared_ptr<expression> func, params;
	//for explicit context/member functions
	shared_ptr<expression> context;
	
	void validate() const;
};

/**
	Transforms change the type of the variable without changing the underlying semantic
	value.
*/
struct expr_transform : public expression
{
	expr_transform()
		: expression( f_transform )
	{ }
	
	enum transform_t
	{
		t_deref,
		t_drop_lvalue,
		//indicates that "param" exists multiple times in this statement but should be evaluated only once
		t_shared_expr,
		//indicates a basic promotions/conversion of a fundmental from param->type to this->type
		t_promote,
		//indicates a basic loss conversion of a fundamental from param->type to this->type
		t_loss,
		//create an optional from "param"
		t_make_optional,
		//deref an optional
		t_deref_optional,
		//wraps a raw function as a bound function
		t_bind_raw_func,
		//binds an unbind function
		t_bind_func,
		//get the value_ptr to value, or array_ptr of array
		t_addr_of,
		//integer of char code point
		t_code_val,
		//char code point of integer
		t_char_val,
	} transform;
	
	shared_ptr<expression> param;

	size_t num_inner_trans() const
	{ return transform == t_promote ? 0 : 1; }
	shared_ptr<expression> get_inner_trans( size_t ndx ) { 
		STMT_CHECK( *this, ndx == 0 ); 
		return param; 
	}
	void set_inner_trans( size_t ndx, shared_ptr<expression> expr ) {
		STMT_CHECK( *this, ndx == 0 );
		 param = expr; 
	}
	
	void construct_result_type();
	void validate() const;
};

/**
	Provides access to a new function for a given type.
*/
struct expr_new_func : public expression
{
	expr_new_func()
		: expression( f_new_func )
	{
	}
	
	extr_type_ref new_type;
	shared_ptr<type_spec> new_spec_type;
	
	void validate() const;
};

/**
	A default value of the given type.
*/
struct expr_default : public expression 
{
	expr_default()
		: expression( f_default )
	{ }
};

/**
	A family of conditional expressions. These are guaranteed lazy-evaluated of
	unneeded components.
*/
struct expr_conditional : public expression
{
	expr_conditional()
		: expression( f_conditional )
	{ }
	
	enum op_t {
		//if 'cond' is set then a, otherwise 'b'
		o_optional_or,
		//if 'cond' is true then optional set to a, otherwise unset optional
		o_optional_cond,
	} op;
	static const op_t min_op_ = o_optional_or;
	static const op_t max_op_ = o_optional_cond;
	
	shared_ptr<expression> cond, a, b;
	
	size_t num_inner_trans() const;
	shared_ptr<expression> get_inner_trans( size_t ndx );
	void set_inner_trans( size_t ndx, shared_ptr<expression> expr );
	
	void construct_result_type();
	void validate() const;
};

/**
	A fundamental scalar binary operation.
	
	Operations are marked to work on (i)integers, (b)binary, (f)floats, and (x)boolean on all
	sizes. The type of both arguments and the result will be exactly
	the same.
*/
struct expr_binary : public expression
{
	expr_binary()
		: expression( f_binary )
	{ }
	
	enum op_t {
		o_add, //ifb
		o_sub, //ifb
		o_mul, //ifb
		o_div, //ifb
		
		o_mod, //i
		
		o_bit_shl, //b
		o_bit_shr, //b
		o_bit_or, //b
		o_bit_and, //b
		o_bit_xor, //b
		
		//boolean results
		o_eq, //ifbx
		o_neq, //ifbx
		o_lt, //ifbx
		o_lte, //ifbx
		o_gt, //ifbx
		o_gte, //ifbx
	} op;
	static const op_t min_op_ = o_add;
	static const op_t max_op_ = o_gte;
	bool is_boolean() const {
		return op >= o_eq && op <= o_gte;
	}
	
	shared_ptr<expression> a, b;
	
	
	void validate() const;
};

/**
	A fundamental scalar unary operation.
*/
struct expr_unary : public expression
{
	expr_unary()
		: expression( f_unary )
	{ }
	
	enum op_t {
		o_neg, //if
		o_abs, //if

		o_bit_cpl, //b
	} op;
	static const op_t min_op_ = o_neg;
	static const op_t max_op_ = o_bit_cpl;
	
	shared_ptr<expression> param;
	
	void validate() const;
};

} //eon leaf

#endif