#ifndef intr_type_HPP
#define intr_type_HPP

#include "../core_types.hpp"
#include "platform/platform.hpp"
#include "fwd.hpp"
#include "scope_ref.hpp"

#include "util/spec_optional.hpp"

#include <boost/optional.hpp>

#include <vector>

namespace util {
	struct validation_context;
}

/*
	CPP: This hierarchy is a very good example of the problem of creating a virtual type hierarchy
	in C++. It's a real pain to provide clone functions, comparisons, calling super, etc. Most of the
	code here is boilerplate.
*/
namespace leaf
{

class serial_out;
class serial_in;
struct scope;

#define intr_type_CLONE(t) \
	virtual shared_ptr<intr_type> clone() const \
	{ return make_shared<t>( *this ); } \
	shared_ptr<t> clone_derived() const \
	{ return std::static_pointer_cast<t>( clone() ); }

/**
	Defines an intrinsic type. This represents the data which is actually being stored, the value of
	a variable.
	
	IMMUTABLE: Post-init intr_type is assumed to be immutable by all other users. This allows
	"clone" to do a shallow copy and avoids several needs of "clone" itself.
*/
struct intr_type
{
	friend class extr_type_ref;
	
	//used to repeat boilerplatr functions
	#define INTR_TYPE_FUN_CLASSES \
		FUN_CLASS_TYPE(unknown) \
		FUN_CLASS_TYPE(tuple) \
		FUN_CLASS_TYPE(value_list) \
		FUN_CLASS_TYPE(function) \
		FUN_CLASS_TYPE(rational) \
		FUN_CLASS_TYPE(integer) \
		FUN_CLASS_TYPE(binary) \
		FUN_CLASS_TYPE(float) \
		FUN_CLASS_TYPE(boolean) \
		FUN_CLASS_TYPE(array) \
		FUN_CLASS_TYPE(char) \
		FUN_CLASS_TYPE(custom) \
		FUN_CLASS_TYPE(module)
		
	enum fun_class_t : int8_t
	{
		#define FUN_CLASS_TYPE(s) f_##s,
		INTR_TYPE_FUN_CLASSES
		#undef FUN_CLASS_TYPE
	};
	
	fun_class_t get_fun_class() const
	{ return fun_class; }

	#define FUN_CLASS_TYPE(s) bool is_##s() const { return fun_class == f_##s; }
	INTR_TYPE_FUN_CLASSES
	#undef FUN_CLASS_TYPE
	
	bool is_tuple_or_list() const
	{ return is_tuple() || is_value_list(); }
	bool is_indexable() const 
	{ return is_tuple() || is_value_list() || is_array(); }
	bool is_void() const;
	bool is_scalar() const
	{ return is_integer() || is_float() || is_rational() || is_binary() || is_boolean(); }
	bool is_integral() const
	{ return is_integer() || is_binary(); }
	bool is_fixed_number() const
	{ return is_integer() || is_float() || is_binary() ; }
	bool is_scope() const
	{ return is_module() || is_custom(); }
	
	static char const* name_of( fun_class_t f );
	
	/** 
		Distinct types create a new fundamental which is distinct from the one on which they are
		built. While there are conversions they are not the same type. Note that a distinct type
		doesn't actually need a name, this is to reinforce that this name need not be globally unique.
		Distinct types are identified by the address of the intr_type.
	*/
	bool distinct;
	boost::optional<std::string> distinct_name;
	bool is_distinct() const {
		return distinct;
	}
	//in which scope was this named type declared (may be null prior to scope addition)
	mutable scope * name_scope;
	
	//TODO: reliance on this id is going to be a problem and it must be removed. Multiple modules
	//must be able to determine the same distinct type and produce a unique name, which means
	//the "distinct_name" must truly be distinct (including scope name likely)
	intptr_t distinct_id() const {
		return reinterpret_cast<intptr_t>(this); 
	}
	
	/**
		A type may exist prior to being completed. This allows forward references and recursive
		types to be supported. Incomplete types can only be used in special circumstances.
	*/
	bool local_complete;
	bool is_complete() const;
	virtual bool is_complete_impl( util::validation_context & ) const
	{ return local_complete; }
	
	~intr_type();

	virtual shared_ptr<intr_type> clone() const = 0;
	
	/**
		Compares the types. This does a logical type comparison and will be equal only if the types 
		are exactly the same (this includes names).
		
		@param logical if true then non value-type flags are ignored, if false then a strict
			field by field comparison is done.
	*/
	int compare ( intr_type const & o, bool logical ) const;
	virtual int compare_impl( intr_type const & o, bool logical, util::validation_context & vctx ) const;
	
	bool operator<( intr_type const & o ) const
	{ return compare( o, false ) < 0; }
	bool operator==(intr_type const& o ) const
	{ return compare( o, false ) == 0; }
	bool operator!=(intr_type const& o ) const
	{ return compare( o, false ) != 0; }
	bool same_as(intr_type const& o) const
	{ return compare( o, true ) == 0; }
	
	virtual void validate() const;
	virtual void validate_impl( util::validation_context & vctx ) const;

	bool is_fixated() const;
	virtual bool is_fixated_impl( util::validation_context & vctx ) const = 0;
	
	///////////////////////////////////////////////////////
	// Simpilfied casting support
	template<typename T>
	T & cast() {
		PRE_CHECK( T::is_fun_class( fun_class ) );
		return static_cast<T&>(*this); 
	}
	template<typename T>
	T * if_cast() {
		if( T::is_fun_class( fun_class ) )
			return static_cast<T*>(this); 
		return nullptr;
	}
	template<typename T>
	T const & cast() const {
		PRE_CHECK( T::is_fun_class( fun_class ) );
		return static_cast<T const&>(*this); 
	}
	template<typename T>
	T const * if_cast() const {
		if( T::is_fun_class( fun_class ) )
			return static_cast<T const*>(this); 
		return nullptr;
	}

	virtual void write( serial_out & so ) const;
	virtual void read( serial_in & si );
	
	//intrusive values usable by the compiler
	mutable intptr_t cref;
protected:
	intr_type( fun_class_t fun_class );
private:
	fun_class_t fun_class;
};

/**
	An extrinsic type reference combines an itrinsic type along with extrinsic modifiers.
	The extrinsic modifiers are not considered to be part of the intrinsic type, and
	thus also not part of a named type (though could be par of an alias).
*/
struct extr_type_ref
{
	shared_ptr<intr_type const> const & get_fun() const
	{ 
		PRE_CHECK(fun);
		return fun; 
	}
	void set_fun( shared_ptr<intr_type const> nfun )
	{ fun = nfun; }
	bool has_fun() const
	{ return !!fun; }
	
	intr_type::fun_class_t get_fun_class() const
	{ return fun->fun_class; }
	
	/**
		Creates a clone of the fundemantal, resets this fun to that new one, and returns
		the new fundamental for modification.
	*/
	shared_ptr<intr_type> clone_fun() {
		auto t = fun->clone();
		fun = t;
		return t;
	}
	
	explicit extr_type_ref( shared_ptr<const intr_type> o ) 
		: fun( o )
	{ 
		lvalue = false;
		optional = false;
		cvt_optional = false;
		lossy = false;
		reference = r_value;
	}
	extr_type_ref() {
		lvalue = false;
		optional = false;
		cvt_optional = false;
		lossy = false;
		reference = r_value;
	}
	
	#define FUN_CLASS_TYPE(s) bool is_##s() const { return fun->fun_class == intr_type::f_##s; }
	INTR_TYPE_FUN_CLASSES
	#undef FUN_CLASS_TYPE
	
	bool is_tuple_or_list() const
	{ return is_tuple() || is_value_list(); }
	bool is_indexable() const 
	{ return is_tuple() || is_value_list() || is_array(); }
	bool is_void() const
	{ return fun->is_void(); }
	bool is_plain() const
	{ return !is_refer() && !is_optional(); }
	bool is_scalar() const
	{ return fun->is_scalar(); } //TODO: this one is correct, we should call intr_type's function to avoid mismatch
	bool is_integral() const
	{ return is_integer() || is_binary(); }
	bool is_fixed_number() const
	{ return is_integer() || is_float() || is_binary() ; }
	bool is_distinct() const {
		return fun->is_distinct(); 
	}
	bool is_value_ptr() const {
		return reference == r_value_ptr;
	}
	
	bool is_lossy() const { 
		return lossy;
	}
	void set_lossy( bool set ) {
		lossy = set;
	}
	
	void set_ref_shared() { 
		reference = r_shared; 
	}
	void set_ref_value_ptr() {
		reference = r_value_ptr;
	}
	void set_ref_value() {
		reference = r_value;
	}

	bool is_fixated() const {
		PRE_CHECK( fun );
		return fun->is_fixated();
	}
	bool is_fixated_impl( util::validation_context & vctx ) const {
		PRE_CHECK( fun );
		return fun->is_fixated_impl( vctx );
	}
	bool is_complete() const {
		return fun && fun->is_complete();
	}
	bool is_complete_impl( util::validation_context & vctx ) const {
		return fun && fun->is_complete_impl(vctx);
	}
	//this is an lvalue type
	bool is_lvalue() const
	{ return lvalue; }
	void set_lvalue( bool set )
	{ lvalue = set; }
	
	//a shared reference to this type
	bool is_shared() const
	{ return reference == r_shared; }
	
	bool is_refer() const
	{ return is_shared() || is_value_ptr(); }
	
	//the value is optionally set
	bool is_optional() const 
	{ return optional; }
	void set_optional( bool set )
	{ optional = set; }

	bool is_cvt_optional() const
	{ return cvt_optional; }
	void set_cvt_optional( bool set ) 
	{ cvt_optional = set; }
	
	///////////////////////////////////////////////////////
	// Simpilfied casting support
	template<typename T>
	T const & fun_cast() const {
		PRE_CHECK( fun );
		return fun->cast<T>();
	}
	template<typename T>
	T const * if_fun_cast() const {
		PRE_CHECK(fun);
		return fun->if_cast<T>();
	}
	
	void validate() const;
	void validate_impl(util::validation_context & vctx) const;
	
	int compare( extr_type_ref const& o, bool logical ) const;
	int compare_impl( extr_type_ref const & o, bool logical, util::validation_context & vctx ) const;
	bool operator<( extr_type_ref const & o ) const { 
		return compare( o, false ) < 0;
	}
	bool operator==( extr_type_ref const& o ) const { 	
		return compare( o, false ) == 0;
	}
	bool operator!=( extr_type_ref const & o ) const {
		return compare( o, false ) != 0;
	}
	bool same_as( extr_type_ref const & o ) const {
		return compare( o, true ) == 0;
	}
	
	void write( serial_out & so ) const;
	void read( serial_in & si );
private:
	//lvalue wraps optional
	bool lvalue;
	//optional wraps reference
	bool optional;
	
	enum reference_t : int8_t {
		//stores directly the fun type
		r_value = 0,
		//a shared reference to the value
		r_shared,
		//an unsafe pointer to a value
		r_value_ptr,
	} reference;
	
	shared_ptr<intr_type const> fun; //will be null if not yet defined (pre-typing)
	

	/**
		Lossy indicates the type is allowed to convert to another with a series of defined
		loss-of-precision/range rules.
	*/
	bool lossy;
	
	//the optional allows underlying conversion
	bool cvt_optional;
};

struct intr_type_unknown : public intr_type
{
	intr_type_unknown()
		: intr_type( f_unknown )
	{ }
	intr_type_CLONE( intr_type_unknown )
	
	bool is_fixated_impl(util::validation_context &) const {
		return true;
	}
};

struct intr_type_boolean : public intr_type
{
	intr_type_boolean()
		: intr_type( f_boolean )
	{ }
	intr_type_CLONE(intr_type_boolean)
	
	bool is_fixated_impl(util::validation_context &) const {
		return true;
	}
};

struct intr_type_char : public intr_type
{
	intr_type_char()
		: intr_type( f_char )
	{ }
	intr_type_CLONE(intr_type_char)
	
	bool is_fixated_impl(util::validation_context &) const {
		return true;
	}
};

struct intr_type_number : public intr_type
{
	static bool is_fun_class( fun_class_t val ) { 
		return val == f_rational || val == f_integer || val == f_binary || val == f_float; 
	}
	
protected:
	intr_type_number() = delete;
	intr_type_number( fun_class_t f )
		: intr_type( f )
	{ }
};

struct intr_type_rational : public intr_type_number
{
	typedef intr_type_number super;
	static bool is_fun_class( fun_class_t val ) { return val == f_rational; }
	
	virtual int compare_impl( intr_type const& o, bool logical, util::validation_context & vctx ) const;
	intr_type_rational()
		: intr_type_number( f_rational )
		, integer_signed( false )
	{}
	intr_type_CLONE(intr_type_rational)
	
	//if it is an integer it has a bitsize which can store that integer
	boost::optional<int> integer_bitsize;
	bool integer_signed;

	virtual void write( serial_out & so ) const;
	virtual void read( serial_in & si );
	
	bool is_fixated_impl(util::validation_context &) const {
		return false;
	}
};

struct intr_type_fixed_number : public intr_type_number
{
	typedef intr_type_number super;
	
	static bool is_fun_class( fun_class_t val ) { 
		return val == f_integer || val == f_float || val == f_binary; 
	}
	
	virtual int compare_impl( intr_type const & o, bool logical, util::validation_context & vctx ) const;
	virtual void validate_impl(util::validation_context & vctx) const;
	
	int data_bitsize;
	
	//will end up failing since copy ctor is protected, well, were it not for make_shared's access limitations!
	intr_type_CLONE(intr_type_fixed_number)
	
	bool is_fixated_impl(util::validation_context &) const {
		return true;
	}
	
	void write( serial_out & so ) const;
	void read( serial_in & si );
protected:
	intr_type_fixed_number( fun_class_t f )	
		: intr_type_number( f )
	{ }
};

struct intr_type_integer : public intr_type_fixed_number
{
	static bool is_fun_class( fun_class_t val ) { return val == f_integer; }

	intr_type_integer( int data_bitsize = platform::natural_integer_size )
		: intr_type_fixed_number( f_integer )
	{ 
		this->data_bitsize = data_bitsize;
	}
	intr_type_CLONE(intr_type_integer)
};

struct intr_type_binary : public intr_type_fixed_number
{
	static bool is_fun_class( fun_class_t val ) { return val == f_binary; }
	intr_type_binary( )
		: intr_type_fixed_number( f_binary )
	{ 
		data_bitsize = platform::natural_integer_size;
	}
	intr_type_CLONE(intr_type_binary)
};

struct intr_type_float : public intr_type_fixed_number
{
	static bool is_fun_class( fun_class_t val ) { return val == f_float; }
	intr_type_float( )
		: intr_type_fixed_number( f_float )
	{ 
		data_bitsize = platform::natural_float_size;
	}
	
	virtual void validate_impl(util::validation_context & vctx) const;
	intr_type_CLONE(intr_type_float)
};

struct intr_type_compound : public intr_type
{
	typedef intr_type super;
	
	static bool is_fun_class( fun_class_t val ) { 
		return val == f_tuple || val == f_value_list;
	}
	virtual int compare_impl( intr_type const & o, bool logical, util::validation_context & vctx ) const;
	virtual void validate_impl(util::validation_context & vctx) const;

	void write( serial_out & so ) const;
	void read( serial_in & si );

	struct sub_ref {
		std::string name;
		extr_type_ref type;
	};
	std::vector<sub_ref> sub;
	
	void push_sub( std::string name, extr_type_ref type ) {
		sub.push_back( sub_ref{ name, type } );
	}
	void push_sub( extr_type_ref type ) {
		sub.push_back( sub_ref{ "", type } );
	}
	
	boost::optional<size_t> find_name( std::string const & name ) const;
	
	bool is_fixated_impl(util::validation_context & vctx) const;
	bool is_complete_impl(util::validation_context & vctx) const;

	
	intr_type_CLONE( intr_type_compound )
protected:
	intr_type_compound( fun_class_t f )
		: intr_type( f )
	{ }
};

struct intr_type_tuple : public intr_type_compound
{
	typedef intr_type_compound super;
	
	static bool is_fun_class( fun_class_t val ) { return val == f_tuple; }

	intr_type_tuple()
		: intr_type_compound( f_tuple )
	{ 
		pack = p_normal;
		selectable = false;
	}
	
	enum pack_t : int8_t {
		p_normal,
		p_abi,
		p_byte,
	} pack;
	intr_type_CLONE(intr_type_tuple)
	
	//indicates that in type conversion this tuple can convert to any of types
	//of the elements of the tuple
	bool selectable;
	
	//complets an incomplete tuple with another definition
	void complete_with( intr_type_tuple const & o );
	
	void validate_impl(util::validation_context & vctx) const;
	int compare_impl( intr_type const & o, bool logical, util::validation_context & vctx ) const;

	void write( serial_out & so ) const;
	void read( serial_in & si );
};

struct intr_type_value_list : public intr_type_compound
{
	static bool is_fun_class( fun_class_t val ) { return val == f_value_list; }
	intr_type_value_list()
		: intr_type_compound( f_value_list )
	{ }
	
	void validate_impl(util::validation_context & vctx) const;
	intr_type_CLONE(intr_type_value_list)
};

struct intr_type_scope : public intr_type
{
	typedef intr_type super;
	
	static bool is_fun_class( fun_class_t val ) {
		return val == f_custom || val == f_module;
	}
	
	//the conext for this scope type
	shared_ptr<variable_context const> context;
	
	bool is_fixated_impl(util::validation_context &) const
	{ return true; } //TODO: assumed for now (would need to handle recursion)
	
	virtual int compare_impl( intr_type const & o, bool logical, util::validation_context & vctx ) const;
	virtual void validate_impl(util::validation_context & vctx) const;
	
	void write( serial_out & so ) const;
	void read( serial_in & si );
protected:
	intr_type_scope() = delete;
	intr_type_scope( fun_class_t f ) 
		: intr_type( f )
	{ }
};

struct intr_type_custom : public intr_type_scope
{
	typedef intr_type_scope super;
	
	static bool is_fun_class( fun_class_t val ) { return val == f_custom; }
	intr_type_custom()
		: intr_type_scope( f_custom )
	{ }
	
	//custom destroy code (called prior to builtin cleanup code)
	boost::optional<scope_ref> destroy;
	
	intr_type_CLONE(intr_type_custom)
	
	//TODO: serial?
};

struct intr_type_module : public intr_type_scope
{
	static bool is_fun_class( fun_class_t val ) { return val == f_module; }
	intr_type_module()
		: intr_type_scope( f_module )
	{ }
	
	intr_type_CLONE(intr_type_module)
};

struct intr_type_wrap : public intr_type
{
	typedef intr_type super;
	
	static bool is_fun_class( fun_class_t val ) { return val == f_array; }
	virtual int compare_impl( intr_type const & o, bool logical,util::validation_context & vctx ) const;
	virtual void validate_impl(util::validation_context & vctx) const;
	virtual void write( serial_out & so ) const;
	virtual void read( serial_in & si );
	extr_type_ref sub;

	bool is_fixated_impl(util::validation_context & vctx) const
	{ return sub.is_fixated_impl(vctx); }
	
	intr_type_CLONE(intr_type_wrap)
protected:
	intr_type_wrap( fun_class_t f )
		: intr_type( f )
	{ }
};

struct intr_type_array : public intr_type_wrap
{
	typedef intr_type_wrap super;
	
	static bool is_fun_class( fun_class_t val ) { return val == f_array; }
	intr_type_array()
		: intr_type_wrap( f_array ) { 
		raw = false;
	}
	
	//this is a raw array pointer as opposed to a managed/counted array
	bool raw;
	
	intr_type_CLONE(intr_type_array)
	virtual int compare_impl( intr_type const & o, bool logical, util::validation_context & vctx ) const;
	virtual void write( serial_out & so ) const;
	virtual void read( serial_in & si );
};

/**
	Note that since functions are immutable there is no distinction between the
	type of the function and a reference to a function (at this level at least, in
	the IR that changes).
*/
struct intr_type_function : public intr_type
{
	typedef intr_type super;
	
	static bool is_fun_class( fun_class_t val ) { return val == f_function; }
	intr_type_function()
		: intr_type( f_function )
		, access( a_bound )
	{ }
	
	shared_ptr<intr_type_value_list const> params, returns;
	//if non-null indicates the shared context type which is expected, if null then it is unique to this type
	shared_ptr<intr_type_custom const> context;
	
	enum access_t {
		//the type has a bound context
		a_bound,
		//the function has no context
		a_raw,
		//an explicit context must be passed to the function
		a_unbound,
	} access;
	
	bool has_context() const {
		return access != a_raw;
	}
	intr_type_CLONE(intr_type_function)
	
	bool is_bound() const 
	{ return access == a_bound; }
	bool is_raw() const
	{ return access == a_raw; }
	bool is_unbound() const
	{ return access == a_unbound; }
	
	bool is_fixated_impl(util::validation_context & vctx) const;
	virtual int compare_impl( intr_type const & o, bool logical,util::validation_context & vctx ) const;
	virtual void validate_impl(util::validation_context & vctx) const;
	
	virtual void write( serial_out & so ) const;
	virtual void read( serial_in & si );
};

#undef intr_type_CLONE

inline bool intr_type::is_void() const
{ return is_value_list() && this->cast<intr_type_value_list>().sub.empty(); }


/**
	A provisional intr_type specification which results from parsing. This contains details of what was parsed.
	It is meant to be converted to a intr_type during typing (where the additional information will be
	available to complete the type).
	
	If a spec includes a prototype then it becomes the dominant form. All other settings must be
	compliant otherwise the spec cannot determine a valid type.
*/
struct type_spec
{
	util::spec_optional<intr_type::fun_class_t> fun_class;
	shared_ptr<intr_type const> fun_prototype;
	
	util::spec_optional<intr_type_function::access_t> function_access;
	shared_ptr<intr_type_custom const> function_context;
	
	util::spec_optional<bool> signed_;
	util::spec_optional<int> data_bitsize;
	util::spec_optional<int> precision_variant;
	util::spec_optional<intr_type_tuple::pack_t> pack;
	util::spec_optional<bool> selectable, member, lossy;

	//extrinsic modifiers
	util::spec_optional<bool> shared, lvalue, optional, cvt_optional, value_ptr;
	
	struct sub_ref {
		std::string name;
		shared_ptr<type_spec> type;
	};
	std::vector<sub_ref> sub;	
	void push_sub( std::string name, shared_ptr<type_spec> type ) {
		sub_ref sr;
		sr.name = name;
		sr.type = type;
		sub.push_back( sr );
	}
	void push_sub( shared_ptr<type_spec> type ) {
		push_sub( "", type );
	}

	/** Is the spec free of any extrinsic modifiers (none are "set", default is okay) */
	bool extrinsic_free() const;
	
	//additional symbols added to the variable type
	std::vector<std::string> symbols;
	
	type_spec( intr_type::fun_class_t fun_class )
		: fun_class( fun_class )
	{ }
	//does a deep-copy
	type_spec( type_spec const & ) = default;
	type_spec() = default;
};

} //eon leaf


#endif