#ifndef SCOPE_HPP
#define SCOPE_HPP

#include "intr_type.hpp"
#include "scope_ref.hpp"
#include "fwd.hpp"
#include "variable_context.hpp"

#include <map>
#include <vector>

namespace leaf
{

class expression;

/**
	Encapsulates a logical scope in the code. This could be a function block, a
	structure, or an anonymous block.
*/
class scope : public std::enable_shared_from_this<scope>, public variable_context
{
	scope( scope_style_t style );
	scope() = delete;
	scope( scope const & o ) = delete;
public:
	template<typename ... T>
	static shared_ptr<scope> create( T&& ... all ) {
		return shared_ptr<scope>( new scope( std::forward<T>(all)... ) );
	}
	
	typedef boost::error_info<struct tag_error_symbol,std::string> error_symbol;
	
	struct capture_ref
	{
		scope_ref local, src;
	};
	typedef std::map<var_decl const *, capture_ref> capture_map_t;
	
	void set_parent( shared_ptr<scope> parent );
	
	/**
		This scope is logically a part of the "base" scope. "base" variables can be accessed
		as though local, but new local variables will exist only within this scope. The "base"
		needs to be mutable for captures -- since they must be captured in the base and not
		in this extended scope.
	*/
	void set_extends( shared_ptr<scope> base );
	
	/**
		Makes a separated context from the base but shares the same level.
	*/
	void set_continues( shared_ptr<scope> base );
	
	/**
		Makes this a namespace of the base scope.
	*/
	void set_namespace( shared_ptr<scope> base, std::string name );
	
	bool is_imperative() const { 
		return scope_style == scope_style_t::imperative; 
	}
	bool is_declarative() const { 
		return scope_style == scope_style_t::declarative; 
	}
	bool is_dynamic() const { 
		return scope_style == scope_style_t::dynamic; 
	}

	bool has_init_block() const {
		return !!init_block;
	}
	shared_ptr<statement_block> get_init_block() {
		PRE_CHECK( has_init_block() );
		return init_block;
	}
	
	/**
		Indicates this is a class scope and the context is explicitly represented by
		the given type.
	*/
	void mark_class_root( shared_ptr<intr_type_custom> context, shared_ptr<statement_block> init_block );
	bool is_class_root() const {
		return class_root;
	}
	shared_ptr<intr_type_custom> obtain_context() { 
		PRE_CHECK( context );
		return context; 
	}
	shared_ptr<intr_type_custom> get_class_scope_context() {
		if( class_root ) {
			PRE_CHECK( context );
			return context;
		}
		return get_parent()->get_class_scope_context();
	}
	
	shared_ptr<scope> get_parent() {
		PRE_CHECK( base_type == bt_parent );
		return base;
	}

	/**
		What is the parent to use if something is being defined (like a function). This is used to allow
		bypass functions to not function as a definition root, isntead delegating to the parent.
	*/
	shared_ptr<scope> get_defn_parent() {
		if( function_bypass ) {
			return get_parent();
		}
		return shared_from_this();
	}
	
	/**
		The depth away from the root scope.
	*/
	unsigned get_level() const;
	
	enum add_mode {
		add_in_order = 0x1,
	};
	/**
		Add a variable to this scope.
	*/
	scope_ref add_single( shared_ptr<var_decl> vs, unsigned flags = 0);

	/**
		Binds a variable in this scope. Binding allows the name to be attached
		to multiple variables.
	*/
	scope_ref add_multi( shared_ptr<var_decl> vs, unsigned flags = 0 );

	/**
		Brings a variable from a parent scope into the lookup of the current
		scope. This creates a scope-local binding for the variable: all references to this
		variable in this scope have this local binding. The captures table then contains
		a map between the local and source bindings.
		
		@param relative_to from where this variable was retrieved (default is from this)
		@return returns the new reference to the captured variable (it
			may be the same, for example if already local or global)
	*/
	scope_ref capture_var( scope_ref vs, scope const * relative_to = nullptr );
	
	/**
		Does the the reverse of capture_var:
					B = capture_var( A )
					->
					A = uncapture_var( B )
	*/
	scope_ref uncapture_var( scope_ref vs ) const;
	
	/**
		Imports a variable captured in another scope to this one... NOTE: see
		statement_block::import and note a lot of assumptions/limitations.
	*/
	scope_ref import( scope_ref sr );
	
	/**
		@throw exception if not found or no single variable exists
	*/
	scope_ref get_var( std::string const & name ) const;
	
	/**
		Obtain all variables bound to this name.
	*/
	var_vector_t get_vars( std::string const & name ) const;
	
	/**
		Obtain all variables bound to this name in the immediate scope only (no parent/extend)
	*/
	var_vector_t get_local_vars( std::string const & name ) const;
	
	/**
		Get all locally declared variables: those which first exist in this scope and are not params,
		returns, captures, or otherwise.
	*/
	var_vector_t get_local_decls( ) const;
	
	var_vector_t const & get_in_order() const
	{ return in_order; }
	
	capture_map_t const & get_captures() const
	{ return capture_map; }
	
	/**
		Obtain the value-list return type of all returns (in-order) for this scope.
	*/
	shared_ptr<intr_type_value_list> get_return_type() const;
	
	/**
		Introduces a defined type.
	*/
	void add_fun_type( shared_ptr<intr_type const> type );
	
	shared_ptr<intr_type const> if_get_fun_type( std::string const & name, unsigned get_flags = 0 ) const;
	
	/**
		Introduce a new type alias.
	*/
	void add_type_alias( std::string const & name, shared_ptr<type_spec const> spec );
	shared_ptr<type_spec const> get_type_alias( std::string const & name, unsigned get_flags = 0 );
	shared_ptr<type_spec const> if_get_type_alias( std::string const & name, unsigned get_flags = 0 );
	
	/**
		Mark this scope as the root of the scope for a function (it still has a parent,
		but this is where params/returns for the function are defined).
	*/
	void mark_function_root();
	/**
		This isn't a normal function. The parent of definitions here is actually the 
		part of this scope.
	*/
	void mark_function_bypass();
	
	/**
		Obtain the most recent function root for this scope.
		@throw error if not root found
	*/
	scope const & get_function_root() const;
	
	scope const * if_get_function_root() const;
	
	bool is_function_root() const { 
		return function_root; 
	}
	bool is_function_bypass() const {
		return function_bypass;
	}
	
	//intrusives usable by compiler
	mutable intptr_t fref;
	
	/**
		Qualifies a name to this scope by prepending namespace names.
	*/
	std::string qualify_name( std::string const & local, std::string const & join ) const;
private:
	typedef std::map<std::string,var_vector_t> var_map_t;
	var_map_t var_map;
	
	//captures are unique per linked variable statement
	capture_map_t capture_map;
	
	//all statements added with the order flag (used for parameter compilation)
	var_vector_t in_order;

	enum base_type_t {
		//is a proper child of base (start at 1 so 0 is undefined)
		bt_parent = 1, 
		//is a local extension of base
		bt_extends,
		//is a continuation/overlay of base
		bt_continues,
		//is a namespace of base
		bt_namespace,
	} base_type;
	shared_ptr<scope> base;
	std::string scope_name;
	
	scope_style_t scope_style;
	
	typedef std::map<std::string,shared_ptr<intr_type const>> fun_type_map_t;
	fun_type_map_t fun_type_map;
	
	typedef std::map<std::string,shared_ptr<type_spec const>> type_alias_map_t;
	type_alias_map_t type_alias_map;
	
	bool function_root, function_bypass;
	
	bool class_root;
	shared_ptr<intr_type_custom> context;

	//where initializations go in this declarative block
	shared_ptr<statement_block> init_block;

};

} //eon leaf
#endif