#ifndef LANG_SERIAL_HPP
#define LANG_SERIAL_HPP

#include "../core_types.hpp"
#include "intr_type.hpp"

#include <vector>
#include <set>
#include <map>
#include <type_traits>

namespace leaf {

/**
	Items in the stream are represented by serial_item's. This is mainly for loading.
*/
struct serial_item {
	enum type_t : int8_t {
		t_invalid = 0,
		t_end_marker = 1,
		t_extr_type = 2,
		t_intr_type = 3,
		t_var_decl = 4,
	} type;
	
	serial_item()
		: type( t_invalid )
	{ }
	
	extr_type_ref v_extr_type_ref;
	shared_ptr<intr_type> v_intr_type;
	shared_ptr<var_decl> v_var_decl;
};

/**
	A binary serialization processor for leaf code.
*/
class serial_out {
	std::ostream & out;
public:
	serial_out( std::ostream & out );
	
	///////////////////////////////////////////////////////
	// The high-level used to add objects to the output
	void add_extr_type( extr_type_ref t );
	void add_intr_type( intr_type const & t );
	void add_var_decl( var_decl const & vd );
	/**
		Should be called after all items have been written. This closes off the data stream
		and complete any internal processing.
	*/
	void done();

	///////////////////////////////////////////////////////
	// The low level interface used by the serializers
	template<typename T>
	void write( T const & t ) {
		static_assert( std::is_trivial<T>::value, "" );
		out.write( (char const*)&t, sizeof(t) );
	}
	template<typename T>
	void write( boost::optional<T> const & t ) {
		bool has = t;
		write(has);
		if( has )
			write(*t);
	}
	void write( std::string const & t );
	
	/**
		Writes a reference to the intr_type object. The object will also be queued for serialization
		if not already done.
	*/
	void write_intr_type_ref( intr_type const & t );
	void write_intr_type_ref( shared_ptr<intr_type const> t );
	
private:
	typedef std::vector<intr_type const*> pending_intr_types_t;
	pending_intr_types_t pending_intr_types;
	typedef std::set<intptr_t> known_intr_types_t;
	known_intr_types_t known_intr_types;
};

/**
	Reads the serialization written by serial_out.
*/
class serial_in {
	std::istream & in;
public:
	/**
		After construction the stream is no longer used (all items will have been loaded into
		memory).
	*/
	serial_in( std::istream & in );

	std::vector<serial_item> const & get_all_items() {
		return loaded_items;
	}
	
	/**
		Reads the primary assuming it is a intr_type.
	*/
	extr_type_ref read_primary_extr_type();
	
	template<typename T>
	void read( T & t ) {
		static_assert( std::is_trivial<T>::value, "" );
		in.read( (char*)&t, sizeof(t) );
	}
	template<typename T>
	void read( boost::optional<T> & t ) {
		bool has = t;
		read(has);
		if( !has ) {
			t = boost::optional<T>();
			return;
		}
			
		T tmp;
		read(tmp);
		t = tmp;
	}
	void read( std::string & t );
	
	shared_ptr<intr_type> read_intr_type_ref();
	
	template<typename T>
	shared_ptr<T> read_intr_type_ref_known() {
		auto x = read_intr_type_ref();
		STATE_CHECK( !x || T::is_fun_class( x->get_fun_class() ) )
		return std::static_pointer_cast<T>(x);
	}
private:
	typedef std::map<intptr_t, shared_ptr<intr_type>> intr_type_map_t;
	intr_type_map_t intr_type_known, intr_type_pending;
	
	std::vector<serial_item> loaded_items;
	void load_all_items();
	
	shared_ptr<intr_type> read_intr_type();
	shared_ptr<intr_type> create_intr_type(intr_type::fun_class_t cl, intptr_t id );
	
	shared_ptr<var_decl> read_var_decl();
	void complete();
};

} //eon leaf
#endif