#ifndef STATEMENT_HPP
#define STATEMENT_HPP

#include "../core_types.hpp"
#include "boost/optional.hpp"
#include "util/source_location.hpp"
#include "fwd.hpp"

namespace leaf
{

struct statement_block;

#define STATEMENT_FORMS \
	STATEMENT_FORM(assign) \
	STATEMENT_FORM(defer) \
	STATEMENT_FORM(do) \
	STATEMENT_FORM(expression) \
	STATEMENT_FORM(fail) \
	STATEMENT_FORM(for) \
	STATEMENT_FORM(init) \
	STATEMENT_FORM(import) \
	STATEMENT_FORM(loop_flow) \
	STATEMENT_FORM(multi) \
	STATEMENT_FORM(noop) \
	STATEMENT_FORM(return) \
	STATEMENT_FORM(type) \
	STATEMENT_FORM(var)
	
/**
	NOt a part of statement itself to avoid conflicts/confustion in naming. Things
	like expression have their own form, but derive from statement, thus an
	inner-name would conflict.
*/
enum statement_form_t {
	#define STATEMENT_FORM(form) sf_##form##_statement,
	STATEMENT_FORMS
	#undef STATEMENT_FORM
};

enum flow_type_t {
	flow_sequence = 1<<1, //the non-terminator type
	flow_return = 1<<2,
	flow_fail = 1<<3,
	flow_loop_break = 1<<4,
	flow_loop_next = 1<<5,
	flow_resume = 1<<6,
};

char const * name_of( statement_form_t f );

/**
	A base class for all types of statements.
*/
struct statement
{
	statement();
	virtual ~statement() { }
	
	virtual statement_form_t statement_form() const = 0;
	
	virtual bool is_declarative() const 
	{ return false; }
	
	/**
		A terminator ends a scope block. No code can exist past a terminator.
	*/
	bool is_terminator() const
	{ return (get_flow() & ~flow_sequence); }
	virtual unsigned get_flow() const
	{ return flow_sequence; }
	
	/**
		Checks the aspects of the expression which must be valid post-typing. This
		includes verifying all the types are correct/consistent.
	*/
	virtual void validate() const { }
	
	/**
		If set indicates the source location which best represents the location
		of this statement.
	*/
	boost::optional<util::source_location> src_loc;
	
	/**
		Used by the typer to track statements which are not yet complete.
	*/
	bool complete;
};

struct error_statement : public statement
{
	//on error handling
	fail_level on_match;
	shared_ptr<statement_block> on_error; //null if none set
};

//macros for doing checks on statements/expressions
#define STMT_CHECK( stmt_, cond_, ... )	\
	if( !(cond_) ) { auto x = error::create_exception<recoverable_error>( __VA_ARGS__ ); \
		x << error::tag_error_condition( #cond_ ); \
		if( (stmt_).src_loc ) x << util::source_location::error( *(stmt_).src_loc ); \
		BOOST_THROW_EXCEPTION( x ); }
#define STMT_FAIL( stmt_, ... )	{ \
	auto x = error::create_exception<recoverable_error>( __VA_ARGS__ ); \
	if( (stmt_).src_loc ) x << util::source_location::error( *(stmt_).src_loc ); \
	BOOST_THROW_EXCEPTION( x ); }

} //eon leaf

#endif