#ifndef LEAF_TYPE_CONVERTER_HPP
#define LEAF_TYPE_CONVERTER_HPP

#include "fwd.hpp"
#include "platform/platform.hpp"
#include "lang/cerr.hpp"
#include "lang/intr_type.hpp"

#include <boost/optional.hpp>

namespace leaf
{

struct cvt_expression
{
	//these conversions are listed in the order they are allowed to happen. The costing algorithm
	//will make use of this, and other places are allowed to assume this
	enum form_t
	{
		//nothing to convert at this level
		c_direct,
		c_field,
		//similar to c_field[0] but implies the input size as 1, also only for value lists
		c_collapse,
		//get the address of a variable
		c_addr_of,
		c_drop_lvalue,
		//deref an optional value (may trigger an error)
		c_deref_optional,
		//applies to shared or value_ptr
		c_deref,
		c_bind_raw_func,
		c_bind_func,
		//convert a rational to a sufficiently sized integer
		c_fixate,
		c_promote,
		c_loss,
		//wrap the sub-fields into a tuple
		c_pack,
		//wrap the sub-fields into a value-list
		c_pack_value,
		//create an optional value (set)
		c_optional,
		//creates a new unset optional value of target type
		c_unset_optional,
		//switches type to the defined "target_type" (current type is exact non-distinct match)
		c_set_distinct,
	} form;

	cvt_expression( form_t form )
		: form( form )
	{ }

	//if form == c_field
	size_t field;
	
	//if form == c_pack/c_pack_value
	struct sub_part {
		shared_ptr<cvt_expression> cvt;
		std::string name;
	};
	std::vector<sub_part> sub;
	
	//if form == c_pack
	intr_type_tuple::pack_t pack;
	//if form == c_promote, c_fixate (no from), c_loss
	enum fundamental_t {
		//in cost order
		p_integer,
		p_binary,
		p_float,
	} from_fun, to_fun;
	int to_bitsize;

	//NOTE: will only be set for c_unset_optional/c_set_distinct (TODO: perhaps it should be set for all?)
	//TODO: hmm, should this not just be a fundamental...?
	extr_type_ref target_type;
	
	inline bool no_cross_promote() const {
		return from_fun == to_fun;
	}
	
	//all forms
	shared_ptr<cvt_expression> then;
	
	bool chain( shared_ptr<cvt_expression> then ) {
		if( !then || then->is_none() )
			return false;
		this->then = then;
		return true;
	}
	shared_ptr<cvt_expression> extend( shared_ptr<cvt_expression> next ) {
		then = next;
		return next;
	}
	
	//true if there is no conversion operation to perform
	bool is_none() const {
		return form == c_direct && sub.empty();
	}
};

/**
	Tracks conversions between types and calculates the costs of conversions.
*/
class type_converter
{
	scope & sc;
public:
	type_converter( scope & sc );
	~type_converter();
	
	typedef cerr_impl<extr_type_ref> resolve_t;
	
	struct convert_t : public cerr_base
	{
		//the tree of conversions to perform, null if none/error
		shared_ptr<cvt_expression> conversions;
		
		int compare_cost( convert_t const & o ) const;
		
		convert_t() 
		{ }
		//providing a conversion assumes it is okay
		convert_t( shared_ptr<cvt_expression> cvt )
			: conversions( cvt )
		{ }
		//assumes there is no conversion
		convert_t( cerr result )
			: cerr_base( result )
		{ }
	};
	
	
	//Refers to any of the binary conversion functions
	typedef convert_t (type_converter::*binary_convert)( extr_type_ref const & /*to*/, 
		extr_type_ref const & /*from*/ );
	
	/**
		Calculates the cost of transforming one type to another.
	*/
	convert_t convert( extr_type_ref const & to, extr_type_ref const & from )
	{ return convert_impl( to, from, cf_allow_cvt ); }
	
	//to must be an lvalue, and then convers to non-lvalue type
	convert_t convert_to_lvalue( extr_type_ref const & to, extr_type_ref const & from )
	{ return convert_impl( to, from, cf_allow_cvt | cf_to_lvalue ); }
	//to must be a reference, and then convertes to non-shared type
	convert_t convert_to_refer( extr_type_ref const & to, extr_type_ref const & from )
	{ return convert_impl( to, from, cf_allow_cvt | cf_to_refer ); }

	//Refers to any of the unary conversion functions
	typedef convert_t (type_converter::*unary_convert)( extr_type_ref const & );
	
	/**
		Produces a "callable" type: a function or a selectable tuple. In the latter case it may
		not actually contain any functions, but the type itself is "callable".
	*/
	convert_t convert_any_callable( extr_type_ref const & from );
	convert_t convert_function_call( intr_type_value_list const & to_params,
		intr_type_value_list const & from_params );

	convert_t convert_any_refer( extr_type_ref const & from ) {
		return convert_any_common( from, cerr::requires_shared,
			acf_retain_refer | acf_require_refer );
	}
	convert_t convert_any_lvalue( extr_type_ref const & from ) {
		return convert_any_common( from, cerr::requires_lvalue,
			acf_retain_lvalue | acf_require_lvalue | acf_retain_optional | acf_retain_refer );
	}
	convert_t convert_nonlvalue_optional( extr_type_ref const & from ) {
		return convert_any_common( from, cerr::requires_optional ,
			acf_retain_optional | acf_require_optional | acf_retain_shared );
	}
	convert_t convert_any_nonlvalue( extr_type_ref const & from ) {
		return convert_any_common( from, cerr::mismatch_lvalue,
			acf_retain_optional | acf_retain_shared );
	}
	convert_t convert_any_scalar( extr_type_ref const & from ) {
		return convert_any_common( from,cerr::requires_scalar,
			acf_require_scalar );
	}
	convert_t convert_unqualified( extr_type_ref const & from ) {
		return convert_any_common( from, cerr::requires_unqualified, 0 );
	}
	convert_t convert_retain_lvalue( extr_type_ref const & from ) {
		return convert_any_common( from, cerr::requires_unqualified, acf_retain_lvalue );
	}
	convert_t convert_shared_or_unqualified( extr_type_ref const & from ) {
		return convert_any_common( from, cerr::requires_unqualified, acf_retain_shared );
	}
		
	/**
		Fixates all fixables types to their real forms (like rationals to integers/floats).
	*/
	resolve_t fixate( extr_type_ref const & from );
	
	/**
		Calculate the best type to which both given types can be converted. The best
		type is the one with the lowest cost of conversion.
	*/
	resolve_t unify( extr_type_ref const & a, extr_type_ref const & b );
	resolve_t unify( extr_type_ref const * begin, size_t len );
private:
	enum ci_flags
	{
		cf_allow_cvt = 1<<0,
		cf_to_lvalue = 1<<1,
		cf_to_refer = 1<<2,
	};
	convert_t convert_impl( extr_type_ref const & to, extr_type_ref const & from, unsigned flags );
		
	convert_t convert_list( intr_type_compound const& to,
		intr_type_compound const & from, unsigned sub_flags, cvt_expression::form_t form );
		
	convert_t convert_any_lvalue_impl( extr_type_ref const & from );
	convert_t convert_nonlvalue_optional_impl( extr_type_ref const & from );
	convert_t convert_any_scalar_impl( extr_type_ref const & from );
	convert_t convert_unqualified_impl( extr_type_ref const & from );
		
	enum any_common_flags {
		acf_require_lvalue = 1<<0,
		acf_retain_lvalue = 1<<1,
		acf_require_shared = 1<<2,
		acf_retain_shared = 1<<3,
		acf_require_optional = 1<<4,
		acf_retain_optional = 1<<5,
		acf_require_scalar = 1<<6,
		acf_retain_value_ptr = 1 << 7,
		acf_require_refer = 1 << 8,
		
		acf_retain_refer = acf_retain_shared | acf_retain_value_ptr,
	};
	convert_t convert_any_common( extr_type_ref const & from, cerr list_fail, unsigned flags );
		
	extr_type_ref unify_impl( extr_type_ref const & a, extr_type_ref const & b );
		
	extr_type_ref fixate_impl( extr_type_ref const & from );
	
	convert_t select_convert( extr_type_ref const & to, extr_type_ref const & from, unsigned flags );
};

} //eon leaf

#endif