#ifndef LEAF_TYPER_HPP
#define LEAF_TYPER_HPP

#include "../core_types.hpp"
#include "fwd.hpp"

#include "type_converter.hpp"
#include "type_identifier.hpp"
#include "statement.hpp"

namespace leaf
{

class typer
{
	shared_ptr<scope> global;
	shared_ptr<module_loader> mloader;
public:
	typer( shared_ptr<scope> global, 
		shared_ptr<module_loader> mloader = shared_ptr<module_loader>() );
	typer( ) { }
	
	typedef std::vector<shared_ptr<var_decl>> var_decls_t;

	enum type_expression_flags
	{
		//for full typing force the result to be converted to its inferred type
		tef_convert_inferred = 1<<0,
	};

	/**
		Adds a block to this module for typing.
	*/
	void full_add_block( shared_ptr<statement_block> block );
	/**
		Completes the typing of this modules.
	*/
	void full_complete_type();
	/** 
		Shortcut to add_block and complete_type
	*/
	void full_type( shared_ptr<statement_block> block );
	
	shared_ptr<expression> full_type( shared_ptr<expression> expr, unsigned flags = 0 );
	
	//bool value unused, just a placeholder for this tag
	typedef boost::error_info<struct tag_typer_error,bool> error_typer;
	//tracks additional information about errors
	typedef boost::error_info<struct cerr_source_type,std::string> tag_cerr_source_type;
	typedef boost::error_info<struct cerr_target_type,std::string> tag_cerr_target_type;
	
	/**
		Tracks the context in which typing is being done. This also serves as a builder class
		which provides convenience entry points and common funtions. 
		
		No state is maintained here, it merely wraps the scope and owner. It is important that
		this context can be recreated at any time, copied, etc.
	*/
	struct context {
		shared_ptr<scope> sc;
		
		typer * get_owner() const { 
			return owner; 
		}
		type_converter converter() const { 
			return type_converter(*sc); 
		}
		type_identifier identifier() const {
			return type_identifier(*sc); 
		}
		
		shared_ptr<expression> proc_expression( shared_ptr<expression> expr ) const;
		void add_pending_block( shared_ptr<statement_block> block ) const;
		bool proc_statement_block( shared_ptr<statement_block> block, context ctx ) const
		{ return owner->proc_statement_block( block, ctx ); }
		bool proc_statement_block( shared_ptr<statement_block> block ) const
		{ return owner->proc_statement_block( block ); }
		void scan_symbols( shared_ptr<statement_block> block ) const {
			owner->scan_symbols( block );
		}

		shared_ptr<expression> convert_expression( extr_type_ref const & target,
			shared_ptr<expression> source ) const;
		shared_ptr<expression> convert_expression( type_converter::unary_convert target,
			shared_ptr<expression> source ) const;
		shared_ptr<expression> convert_expression( type_converter::binary_convert how,
			shared_ptr<expression> target, shared_ptr<expression> source ) const;
		
		template<typename T>
		type_identifier::resolve_t determine_type( T const & source, 
			shared_ptr<type_spec> const & spec_type ) const {
			auto rt = identifier().determine( spec_type );
			STMT_CHECK( *source, rt.okay(), rt.explain_error() );
			return rt;
		}
		
		void add_export( shared_ptr<var_decl> vd ) {
			owner->var_exports.push_back( vd );
		}
		context( typer * owner, shared_ptr<scope> const & sc );
		context() = delete;
		context( context const & ) = default;
	private:	
		typer * owner;
	};
	
	struct job {
		virtual ~job() { }
		virtual bool process( typer * ty ) = 0;
	};
	
	friend class job_statement_block; //TODO: replace friend system with distinct open typer and closed high-level interface
	struct job_statement_block : public job {
		shared_ptr<statement_block> sb;
		bool process( typer * ty );
		job_statement_block( shared_ptr<statement_block> sb )
			: sb( sb) {
		}
	};
	
	var_decls_t const & get_var_exports() const {
		return var_exports;
	}
	std::vector<std::string> const & get_libraries() const {
		return libraries;
	}
private:
	typedef std::vector<shared_ptr<job>> job_vector_t;
	job_vector_t pending_jobs;
	void complete_type();
	size_t completed_statements;
	//no longer allow incomplete items and report errors instead
	bool no_incomplete;
	
	bool proc_statement_block( shared_ptr<statement_block> block, context ctx );
	bool proc_statement_block( shared_ptr<statement_block> block );
	
	shared_ptr<statement> proc_statement( shared_ptr<statement> s, context ctx );
	shared_ptr<statement> proc_error_statement( shared_ptr<error_statement> es, context ctx );

	shared_ptr<statement> proc_assign_statement( shared_ptr<assign_statement> as, context ctx );
	shared_ptr<statement> proc_defer_statement( shared_ptr<defer_statement> ds, context ctx );
	shared_ptr<statement> proc_do_statement( shared_ptr<do_statement> s, context ctx );
	shared_ptr<statement> proc_expression_statement( shared_ptr<expression_statement> es, context ctx );
	shared_ptr<statement> proc_fail_statement( shared_ptr<fail_statement> fs, context ctx );
	shared_ptr<statement> proc_for_statement( shared_ptr<for_statement> fs, context ctx );
	shared_ptr<statement> proc_import_statement( shared_ptr<import_statement> is, context ctx );
	shared_ptr<statement> proc_init_statement( shared_ptr<init_statement> is, context ctx );
	shared_ptr<statement> proc_loop_flow_statement( shared_ptr<loop_flow_statement> ls, context ctx );
	shared_ptr<statement> proc_return_statement( shared_ptr<return_statement> rs, context ctx );
	shared_ptr<statement> proc_type_statement( shared_ptr<type_statement> ts, context ctx );
	shared_ptr<statement> proc_var_statement( shared_ptr<var_statement> vs, context ctx );
	
	void scan_symbols( shared_ptr<statement_block> block );
	
	var_decls_t var_exports;
	std::vector<std::string> libraries;
};

} //eon leaf

#endif