#include "bin/expecter.hpp"
#include "ir/llvm/gen.hpp"
#include "parser/node_converter.hpp"
#include "util/file.hpp"
#include "lang/typer.hpp"
#include "parser/node_parser.hpp"

#include <boost/regex.hpp>
#include <boost/algorithm/string/predicate.hpp>

expecter::expecter( std::string const & src ) {
	this->src = src;
	omit = false;
	
	//check if there is a built-in test expect string
	boost::regex re_expect( "EXPECT\\(([^\\)]*)\\)" );
	boost::match_results<char const *> what;
	if( boost::regex_search( src.c_str(), src.c_str() + src.size(), what, re_expect ) ) {
		expect_output = what[1];
	}

	boost::regex re_omit( "OMIT:" );
	if( boost::regex_search( src.c_str(), src.c_str() + src.size(), what, re_omit ) ) {
		omit = true;
	}

	check_marker( "EXPECT-FAIL-TYPER", expect_fail_typer );
	check_marker( "EXPECT-FAIL-CONVERTER", expect_fail_converter );
	check_marker( "EXPECT-FAIL-PARSER", expect_fail_parser );
	check_marker( "EXPECT-FAIL-EXEC", expect_fail_exec );
	
	boost::regex re_src_line( "SRC-LINE\\(([0-9]+)\\)" );
	if( boost::regex_search( src.c_str(), src.c_str() + src.size(), what, re_src_line ) ){
		src_line = unsigned(atoi( what[1].str().c_str() ));
	}
	boost::regex re_rv( "EXPECT-RETURN\\(([0-9]+)\\)" );
	if( boost::regex_search( src.c_str(), src.c_str() + src.size(), what, re_rv ) ){
		expect_return_value = unsigned(atoi( what[1].str().c_str() ));
	}
}

void expecter::check_marker( std::string const & text, expect_fail & variable ) {
	boost::match_results<char const *> what;
	boost::regex re_mark( text + "(\\(([^)]*)\\))?" );
	if( boost::regex_search( src.c_str(), src.c_str() + src.size(), what, re_mark ) ) {
		variable.set = true;
		if( what[1].matched )
			variable.what = std::string( what[2].first, what[2].second );
	} else {
		variable.set = false;
	}
}

bool expecter::any_expect_fail() const { 
	return expect_fail_typer || expect_fail_converter || expect_fail_parser || expect_fail_exec; 
}

void expecter::check_result( std::string const & actual, int return_value ) {
	if( any_expect_fail() )
		throw recoverable_error( "Not expected to compile/run -- should have failed" );
		
	if( expect_output ) {
		if( *expect_output != actual )  {
			throw recoverable_error( "Unexpected results" )
				<< error_expect( *expect_output )
				<< error_actual( actual );
		}
	} else if( expect_return_value ) {
		if( *expect_return_value != return_value ) {
			char rv_expect[128], rv_actual[128];
			sprintf( rv_expect, "%d", *expect_return_value );
			sprintf( rv_actual, "%d", return_value );
			throw recoverable_error( "Unexpected return value" )
				<< error_expect( rv_expect )
				<< error_actual( rv_actual );
		}
	}
}

void expecter::catch_compile_exception( recoverable_error & re ) const {
	auto loc = boost::get_error_info<util::source_location::error>( re );
	auto reason = boost::get_error_info<error::tag_error_reason>( re );
	if( !reason ) {
		throw;
	}
	
	if( src_line ) {
		STATE_CHECK( loc, "missing error location" );
		if( loc->line() != *src_line ) {
			std::cout << "Expect Line: " << *src_line << std::endl;
			std::cout << "Actual Line: " << loc->line() << std::endl;
			STATE_FAIL( "line mismatch" );
		}
	}
	
	if( expect_fail_typer ) {
		if( boost::get_error_info<leaf::typer::error_typer>( re ) ) {
			if( expect_fail_typer.what.size()  
				&& !boost::starts_with( *reason, expect_fail_typer.what ) ) {
				std::cerr << "Expect Reason: " << expect_fail_typer.what <<std::endl;
				std::cerr << "Actual Reason: " << *reason << std::endl;
				STATE_FAIL( *reason + ": invalid reason" );
			}
			std::cout << "Caught expected typer error" << std::endl;
			return;
		}
	}

	if( expect_fail_converter ) {
		if( boost::get_error_info<leaf::node_converter::error_converter>( re ) )	{
			std::cout << "Caught expected converter error" << std::endl;
			return;
		}
	}
	
	if( expect_fail_parser ) {
		if( boost::get_error_info<leaf::node_parser::error_parser>( re ) ) {
			std::cout << "Caught expected parser error" << std::endl;
			return;
		}
	}
	
	throw;
}

bool expecter::catch_runtime_abort( std::string const & msg ) const {
	if( expect_fail_exec ) {
		if( expect_fail_exec.what.empty()
			|| ( msg.find( expect_fail_exec.what ) != std::string::npos ) ) {
			std::cout << "Caught expected exec error" << std::endl;
			return true;
		}
	}
	
	return false;
}
