#include "bin/expecter.hpp"

#include "lang/language.hpp"
#include "lang/typer.hpp"
#include "lang/dump.hpp"
#include "lang/serial.hpp"
#include "lang/module_loader.hpp"

#include "parser/node_converter.hpp"
#include "parser/node_parser.hpp"
#include "parser/dump.hpp"
#include "parser/source.hpp"

#include "ir/compiler_leaf.hpp"
#include "ir/dump.hpp"
#include "ir/module.hpp"
#include "ir/trace_functions.hpp"

#include "ir/llvm/gen.hpp"
#include "ir/llvm/runtime_test.hpp"

#include "util/file.hpp"

#include <boost/program_options.hpp>
namespace po = boost::program_options;
#include <boost/exception/diagnostic_information.hpp>
#include <boost/algorithm/string/predicate.hpp>
#include <boost/filesystem.hpp>
namespace fs = boost::filesystem;

#include <iostream>
#include <fstream>
#include <cstdlib>

#include <dlfcn.h>
#include <unistd.h>

#define SHOW_DEBUG 1
#include "util/debug.hpp"

/**
	Encapsulates the basic logic of compilation and execution (as required for tracing).
	
	This is not a settings class, but a stateful processor.
*/
struct runner {

	//debugging options
	bool dump_parse, dump_base, dump_pre_type, 
		dump_all_types, dump_ir, dump_ir_types, dump_post_type, dump_exports;
	std::string llvm_base_ll, dump_llvm;

	//all input files (reduced modes expect only 1 file)
	std::vector<std::string> input_files;
	//the name of the module being generated (if any)
	boost::optional<std::string> module_name;
	
	//dynamic libraries to open before execution
	std::vector<std::string> libraries;
	
	//optimizations
	ir_llvm::gen::options gen_opts;
	
	//for module resolution
	shared_ptr<leaf::module_loader> mloader;
	
	runner() {
		ut_capture_stdout = false;
		
		dump_ir_types = false;
		dump_pre_type = false;
		dump_base = false;
		dump_post_type = false;
		dump_exports = false;
		dump_all_types = false;
		dump_ir = false;
		dump_parse = false;
	}
	
	shared_ptr<leaf::statement_block> global;
	shared_ptr<leaf::scope> global_symbols;
	shared_ptr<leaf::typer> module_typer;
	
	void x_dump_parse( leaf::node const & n );
	void x_dump_post_type( leaf::statement_block const & sb );
	void x_dump_pre_type( leaf::statement_block const & sb );
	
	shared_ptr<leaf::expression> parse_expr( std::string const & expr );
	void typed_base_block( std::string const & filename );
		
	shared_ptr<leaf::statement_block> load_parse_block( std::string const & filename, 
		leaf::scope_style_t scope_tyle, bool dump = true );

	/**
		Execute the main function of the module.
		
		@return For a normal execution that cleanly completes this is the return value from the
			main entry point. For an expectation test that succeeds 0 is returned, an error is
			thrown on failure. Other errors will be thrown.
	*/
	int execute( shared_ptr<ir::module> module );
	
	enum gen_what_t {
		//static object file
		gw_object,
		//shared library
		gw_shared,
		//main executable
		gw_exe,
	};
	/**
		Generate an object file of the module.
		
		@param path_base the base path/name of the output (extensions will be added)
		@param gw what type of object file to generate
	*/
	void emit_object( shared_ptr<ir::module> module, std::string const & path_base, gen_what_t gw );
	
	/** 
		Prepares the expecter for unit_tests.
	*/
	void ut_create_expecter();
	shared_ptr<expecter> ut_expect;
	bool ut_capture_stdout;
	
	bool should_omit() const {
		return ut_expect && ut_expect->omit;
	}

	/**
		Build the source files down to the IR module.
	*/
	shared_ptr<ir::module> build_module();
private:
	shared_ptr<ir::module> get_expr();
	shared_ptr<ir::module> get_block();
	shared_ptr<ir::module> get_full();

};

int main( int argc, char ** argv )
{
	try {
		runner run;
		
		std::string expr, obj_file;
		bool exe_test = false, unit_test = false, skip_omit = false;
		po::options_description desc( "Options" );
		desc.add_options()
			( "help", "show help" )
			( "skip-omit", po::value( &skip_omit )->zero_tokens(), "skip checks marked as omit" )
			( "unit-test", po::value( &unit_test )->zero_tokens(), "the file is a Leaf unit test, enable test functions" )
			( "exe-test", po::value( &exe_test )->zero_tokens(), "variant of unit-test which captures stdout" )
			( "obj-file", po::value( &obj_file ), "produce object file (provide basename)" )
			( "input-file", po::value( &run.input_files ), "leaf input files" )
			( "shared", po::value( &obj_file ), "generate a shared library" )
			( "exe", po::value( &obj_file ), "generate an executable" )
			( "library", po::value( &run.libraries ), "libraries to load prior to execution" )
			;
		po::positional_options_description p;
		p.add( "input-file", -1 );

		po::options_description dopts( "Leaf Debugging" );
		dopts.add_options()
			( "dump-parse", po::value( &run.dump_parse )->zero_tokens(), "dump the parse tree" )
			( "dump-base", po::value( &run.dump_base )->zero_tokens(), "dump base content as well (must specify what to dump still)" )
			( "dump-pre-type", po::value( &run.dump_pre_type )->zero_tokens(), "dump the Leaf AST prior to typing" )
			( "dump-post-type", po::value( &run.dump_post_type )->zero_tokens(), "dump the Leaf AST post typing" )
			( "dump-all-types", po::value( &run.dump_all_types )->zero_tokens(), "dump all leaf AST tree types" )
			( "dump-ir", po::value( &run.dump_ir )->zero_tokens(), "dump the Leaf-IR" )
			( "dump-ir-types", po::value( &run.dump_ir_types )->zero_tokens(), "show all Leaf-IR types" )
			( "dump-llvm", po::value( &run.dump_llvm )->implicit_value("*"), "dump matching LLVM-IR symbols" )
			( "dump-exports", po::value( &run.dump_exports )->zero_tokens(), "dump exported symbols (full only)" )
			;
			
		po::options_description opts( "Optimizations" );
		opts.add_options()
			( "llvm-opt", po::value( &run.gen_opts.llvm_opt )->zero_tokens(), "run LLVM optimizers" )
			( "distinct-exception-flow", po::value( &run.gen_opts.distinct_exception_flow )->zero_tokens(), "Create separate exception paths" )
			( "reduce-branch", po::value( &run.gen_opts.reduce_branch )->implicit_value(true), "Use reduced branch flow" )
			( "reduce-flow", po::value( &run.gen_opts.reduce_flow )->implicit_value(true), "Use reduced blocks in flow" )
			;
			
		po::options_description all_opts;
		all_opts.add( desc ).add( opts ).add( dopts );
		
		po::variables_map vm;
		po::store( po::command_line_parser( argc, argv ).options( all_opts ).positional(p).run(), vm );
		po::notify( vm );

		if( vm.count( "help" ) ) {
			std::cerr << all_opts;
			return 0;
		}
		
		if( run.input_files.empty() ) {
			std::cerr << "requires input file" << std::endl;
			return 1;
		}
		
		//split obj_name
		fs::path pn(obj_file);
		fs::create_directories( pn.parent_path() );
		auto mod_name = pn.filename().string();

		//name is in obj-file
		runner::gen_what_t gw = runner::gw_object;
		if( vm.count( "shared" ) ) {
			gw = runner::gw_shared;
			run.gen_opts.pic = true;
			run.module_name = mod_name;
		} else if( vm.count( "exe" ) ) {
			gw = runner::gw_exe;
		}

		//configure module loader
		char * leaf_path_c = std::getenv( "LEAF_PATH" );
		STATE_CHECK( leaf_path_c, "LEAF_PATH" );
		std::string leaf_path( leaf_path_c );
		if( !boost::ends_with( leaf_path, "/" ) ) {
			leaf_path += '/';
		}
		run.mloader = make_shared<leaf::module_loader>( leaf_path );
		
		if( exe_test ) {
			run.ut_capture_stdout = true;
		}
	
		run.typed_base_block( unit_test ? leaf_path + "share/unit_test.leaf" : leaf_path + "share/base.leaf" );
		run.llvm_base_ll = leaf_path + "share/base.ll";
		
		//prepare expectations
		if( unit_test || exe_test ) {
			run.ut_create_expecter();
			if( skip_omit && run.should_omit() ) {
				return 2;
			}
		}

		auto module = run.build_module();
		if( !module ) {
			//happens only with expected failed compilation
			return 0;
		}

		if( obj_file.size() ) {
			run.emit_object( module, obj_file, gw );
			return 0;
		} else {
			return run.execute( module );	
		}
		
	} catch( recoverable_error & re ) {
		std::cerr << boost::current_exception_diagnostic_information();
		auto loc = boost::get_error_info<util::source_location::error>( re );
		if( loc )	{
			std::cout << "Source: " << loc->name()
				<< ":" << loc->line() << "," <<  loc->column()
				<< std::endl;
		}
		return 1;
	} catch(...) {
		std::cerr << boost::current_exception_diagnostic_information();
		return 1;
	}
}

shared_ptr<ir::module> runner::build_module() 
{
	shared_ptr<ir::module> m;
	try {
		STATE_CHECK( input_files.size() );
		auto & first_file = input_files[0];
		
		if( boost::ends_with( first_file, ".lfx" ) ) {
			STATE_CHECK( input_files.size() == 1, "only one lfx file supported as input" );
			m = get_expr();
		} else if( boost::ends_with( first_file, ".lfb" ) ) {
			STATE_CHECK( input_files.size() == 1, "only one lfb file supported as input" );
			m = get_block();
		} else if( boost::ends_with( first_file, ".leaf" ) ) {
			m = get_full();
		} else {
			STATE_FAIL( "unknown input-file type" );
		}
	} catch( recoverable_error & re ) {
		if( ut_expect ) {
			ut_expect->catch_compile_exception(re);
			return shared_ptr<ir::module>();
		}
		throw;
	}
	
	if( dump_ir ) {
		ir::dump d( std::cout );
		d.set_show_all_types( dump_ir_types );
		d.write( m.get() );
	}
	return m;
}

void runner::x_dump_parse( leaf::node const & n )
{
	if( !dump_parse )
		return;
	
	leaf::node_dump nd( std::cout );
	nd.header( "// -- Parse Tree --" );
	nd.dump( n );
	std::cout << std::endl;
}

void runner::x_dump_post_type( leaf::statement_block const & sb )
{
	if( !dump_post_type )
		return;
		
	std::cout << "// -- Post-Typing Leaf Tree --" << std::endl;
	leaf::dump d( std::cout );
	d.set_show_all_types( dump_all_types );
	d.write_statement_block( sb );
	std::cout << std::endl;
}

void runner::x_dump_pre_type( leaf::statement_block const & sb )
{
	if( !dump_pre_type )
		return;
		
	std::cout << "// -- Pre-Typing Leaf Tree --" << std::endl;
	leaf::dump d( std::cout );
	d.set_show_spec_type( true );
	d.set_validate( false );
	d.write_statement_block( sb );
	std::cout << std::endl;
}

shared_ptr<leaf::expression> runner::parse_expr( std::string const & expr )
{
	auto src = leaf::source::create( expr );

	shared_ptr<leaf::node> n = leaf::node_parser::parse_single( *src );
	x_dump_parse( *n );
	
	auto ex = leaf::node_converter::expression_convert( *n );
	if( dump_pre_type )
	{
		leaf::dump d( std::cout );
		d.set_show_spec_type( true );
		d.set_validate( false );
		d.write_expression( *ex );
		std::cout << std::endl;
	}
	
	//the expression must be converted to its inferred type to ensure it is representable as
	//an actual compiled type (such as removing rationals)
	ex = module_typer->full_type( ex, leaf::typer::tef_convert_inferred );
	
	if( dump_post_type ) {
		leaf::dump d( std::cout );
		d.set_show_all_types( dump_all_types );
		d.write_expression( *ex );
		std::cout << std::endl;
	}

	return ex;
}

shared_ptr<leaf::statement_block> runner::load_parse_block( std::string const & filename, 
	leaf::scope_style_t scope_style, bool dump )
{
	auto s = leaf::source::create( util::load_file( filename ) );
	s->set_name( filename );
	
	auto n = leaf::node_parser::parse_as_block( *s );
	if( dump )
		x_dump_parse( *n );
	
	auto block = leaf::node_converter::convert_block( *n, scope_style );
	if( dump )
		x_dump_pre_type( *block );
	
	return block;
}

void runner::typed_base_block( std::string const & filename )
{
	PRE_CHECK(!global);
	auto block = load_parse_block( filename, leaf::scope_style_t::declarative, false );
	leaf::typer t;
	t.full_type( block );
	if( dump_base ) {
		x_dump_post_type( *block );
	}
	global = block;
	global_symbols = global->symbols;
	if( module_name ) {
		auto module_space = leaf::scope::create( leaf::scope_style_t::declarative );
		module_space->set_namespace( global_symbols, *module_name );
		global_symbols = module_space;
	}
	module_typer = make_shared<leaf::typer>( global_symbols, mloader );
}

shared_ptr<ir::module> runner::get_expr() 
{
	shared_ptr<leaf::expression> ex = parse_expr( util::load_file( input_files.at(0) ) );
	ir::compiler_leaf c( global_symbols );
	
	shared_ptr<ir::module> m = c.compile_expr( *ex );
	auto f = m->if_get_function( "eval_expr" );
	PRE_CHECK(f);
	main_trace_function( c, f );

	return m;
}

shared_ptr<ir::module> runner::get_block()
{
	shared_ptr<leaf::statement_block> block = load_parse_block( input_files.at(0), 
		leaf::scope_style_t::dynamic );
	block->symbols->set_continues( global_symbols );
	
	module_typer->full_type( block );
	x_dump_post_type( *block );
	
	ir::compiler_leaf c( global_symbols );

	return c.compile_block( *block );
}

shared_ptr<ir::module> runner::get_full()
{
	std::vector<leaf::statement_block const*> raw_blocks;
	std::vector<shared_ptr<leaf::statement_block>> blocks;
	
	auto single_scope = leaf::scope::create( leaf::scope_style_t::declarative );
	single_scope->set_continues( global_symbols );
	
	for( auto & file : input_files ) {
		auto block = load_parse_block( file, leaf::scope_style_t::declarative );
		block->symbols = single_scope; //hmmm....
		
		module_typer->full_add_block( block );
		blocks.push_back( block );
		raw_blocks.push_back( block.get() );
	}
	
	module_typer->full_complete_type();
	for( auto b : raw_blocks ) {
		x_dump_post_type( *b );
	}
	
	
	if( dump_exports ) {
		std::cout << "[Exports]" << std::endl;
		leaf::dump d( std::cout );
		for( auto & ex : module_typer->get_var_exports() ) {
			d.write_var_decl( *ex );
			d.line_break();
		}
		std::cout << std::endl;
	}

	ir::compiler_leaf c( global_symbols );

	return c.compile_full( &raw_blocks[0], raw_blocks.size() );
}

void runner::ut_create_expecter() {
	//one of the files contains the test conditions
	std::string xcode;
	if( input_files.size() == 1 ) {
		xcode = util::load_file( input_files[0] );
	} else {
		for( auto & file : input_files ) {
			//only the full mode can have multiple files now, so this check is specific to that
			if( boost::ends_with( file, "main.leaf" ) ) {
				xcode = util::load_file( file );
			}
		}
	}
	
	ut_expect = make_shared<expecter>(xcode);
	STATE_CHECK( ut_expect->any_expect(), "no test conditions found" );
}

int runner::execute( shared_ptr<ir::module> m )
{
	//load explicit libraries (mainly for testing)
	std::vector<void*> lib_handles;
	auto load_lib = [&]( std::string const & path ) {
		auto handle = dlopen( path.c_str(), RTLD_NOW | RTLD_GLOBAL );
		STATE_CHECK( handle, std::string( "failed loading library: " )
			+ path + ": " + dlerror() );
		lib_handles.push_back( handle );
	};
	
	for( auto & lib : libraries ) {
		load_lib( lib );
	}
	//load imported libraries
	for( auto & lib : module_typer->get_libraries() ) {
		load_lib( mloader->resolve_dynload(lib) );
	}

	//setup capturing of stdout
	volatile int old_stdout = 0; //GCC: complains it will get clobbered by longjmp without volatile
	char out_filename[128] = "/tmp/output_XXXXXX";
	if( ut_capture_stdout ) {
		old_stdout = dup(STDOUT_FILENO);
 		int out_file = mkstemp(out_filename);
 		fflush( stdout );
 		
 		dup2( out_file, STDOUT_FILENO );
 	}
 	
 	auto end_redirect = [&]() {
		if( !old_stdout ) {
			return;
		}
		
 		fflush( stdout );
 		dup2(old_stdout, STDOUT_FILENO);
 	};
	auto clean_redirect = [&]() {
 		unlink(out_filename);
	};
	
	int return_value = 0;
	
	try {
		jmp_buf abort_point;
		int i = setjmp(abort_point);
		if( i ) {
			end_redirect();
			if( !ut_expect || !ut_expect->catch_runtime_abort( runtime_test::get_error() ) ) {
				throw recoverable_error( std::string( "abort: " ) + runtime_test::get_error() );
			}
		} else {
			runtime_test::prepare( &abort_point );
			ir_llvm::gen lg( llvm_base_ll, gen_opts );
			int rv = lg.execute( *m, dump_llvm );
			end_redirect();
			
			runtime_test::check_state();
			if( ut_expect ) {
				std::string output = runtime_test::get_trace();
				if( old_stdout ) {
					output = util::load_file( out_filename );
				}
				//DEBUG( "OUTPUT:", output );
				ut_expect->check_result( output, rv );
			} else {
				return_value = rv;
			}
		}
		
		clean_redirect();
	} catch( ... ) {
		end_redirect();
		clean_redirect();
		throw;
	}
	
	return return_value;
}

void runner::emit_object( shared_ptr<ir::module> m, std::string const & path_base, gen_what_t gw )
{
	ir_llvm::gen lg( llvm_base_ll, gen_opts );
	lg.emit_object( *m, path_base, dump_llvm );
	
	//create leaf object file
	std::ofstream lo( path_base + ".lo" );
	leaf::serial_out sout( lo );
	for( auto & ex : module_typer->get_var_exports() ) {
		sout.add_var_decl( *ex );
	}
	sout.done();
	
	if( gw == gw_shared ) {
		std::string cmd( "ld --eh-frame-hdr --build-id --export-dynamic -shared -o " );
		cmd += path_base + ".so -init "
			+ *module_name + platform::export_sep + "_init_module -fini "
			+ *module_name + platform::export_sep + "_final_module"
			//+ " " + mloader->resolve_dyn_core( "runtime" )
			+ " " + path_base + ".o";
		int ret = std::system( cmd.c_str() );
		STATE_CHECK( ret == 0, cmd );
	} else if( gw == gw_exe ) {
		//NOTE: switch to g++ for rare debugging needs with iostream
		std::string cmd( "gcc -Wl,--eh-frame-hdr,--build-id,--export-dynamic -o " );
		cmd += path_base + " " + path_base + ".o"
			//+ " " + mloader->resolve_static_core( "runtime_exception" )
			//+ " " + mloader->resolve_static_core( "runtime_support" )
			+ " " + mloader->resolve_dyn_core( "runtime" )
			;
			
		for( auto & lib : module_typer->get_libraries() ) {
			cmd = cmd + " " + mloader->resolve_dynlink( lib );
		}
		
		int ret = std::system( cmd.c_str() );
		STATE_CHECK( ret == 0, cmd );
	}
}
