#!/usr/bin/python
"""
	Execute a group of tests.
"""
import glob, os, sys, argparse
from shelljob import job, fs

parser = argparse.ArgumentParser( description = 'Run Leaf tests' )
parser.add_argument( '--all', action='store_const', const=True )
parser.add_argument( '--unit', action='store_const', const=True )
parser.add_argument( '--mod', action='store_const', const=True )
parser.add_argument( '--exe', action='store_const', const=True )
parser.add_argument( '--exe-jit', action='store_const', const=True )

parser.add_argument( '--leaf-opts', action='store_const', const=True )
parser.add_argument( '--no-opts', action='store_const', const=True )
parser.add_argument( '--libs', action='store_const', const=True )

args = parser.parse_args()

# Collect build options
leaf_opts = ''
if args.leaf_opts:
	# TODO: not supported now, should add back
	leaf_opts = '--distinct-exception-flow'
if args.no_opts:
	# TODO: reduce-flow not support, reduce-branch no longer relevant
	leaf_opts = '--reduce-branch=0 --reduce-flow=0'

 
class TestMonitor(job.FileMonitor):
	def __init__(self):
		super(TestMonitor,self).__init__()
		self.passed = 0
		self.skipped = 0
		self.failed = 0
		self.okay = True
		
	def job_finished(self,jobj):
		super(TestMonitor,self).job_finished(jobj)
		if jobj.exit_code == 0:
			self.passed += 1
			msg = "\x1b[0;32mPASS\x1b[m"
		elif jobj.exit_code == 2:
			self.skipped += 1
			msg = "\x1b[0;33mSKIP\x1b[m"
		else:
			self.failed += 1
			msg = "\x1b[0;31mFAIL\x1b[m"
			
		print( "{} {}".format( jobj.name, msg ) )
		
	def gen_feedback(self):
		pass
		
	def report(self):
		sys.stdout.write( '{} of {} Passed'.format( self.passed, self.passed + self.failed ) )
		if self.skipped != 0:
			sys.stdout.write( " (\x1b[0;33m{} Skipped\x1b[m)".format( self.skipped ) )
		if self.failed != 0:
			sys.stdout.write( " (\x1b[0;31m{} Failed\x1b[m)".format( self.failed ) )
			self.okay = False
		print( "")

def common_job( test, line ):
	jobj = job.Job('build/leaf {} --skip-omit {}'.format( leaf_opts, line ))
	jobj.name = test
	return jobj
	
def build_job( test ):
	return common_job( test, '--unit-test ' + test )

def build_dir_job( test ):
	files = fs.find( test, include_dirs = False, name_regex = '.*\\.leaf', limit_depth = 0 )
	files_str = " ".join(files)
	if len(files_str) == 0:
		return None
	return common_job( test, '--unit-test ' + files_str )

def build_exe_jit_job( test ):
	files = fs.find( test, include_dirs = False, name_regex = '.*\\.leaf' )
	return common_job( test, '--exe-test ' + " ".join(files) )

def build_lib_job( test ):
	files = fs.find( test, include_dirs = False, name_regex = '.*\\.leaf' )
	return common_job( test, '--shared build/libs/{} {}'.format( 
		os.path.basename(test), " ".join(files)  ) )

def build_exe_job( test ):
	jobj = job.Job('src/bin/exe_test.py {}'.format( test ) )
	jobj.name = test
	return jobj

def build_libs( libs ):
	cmds = [ build_lib_job(test) for test in sorted(libs) ]
	# allows buidler to return None on some inputs
	cmds = filter(None,cmds)
	print( "Building libraries ({})".format( leaf_opts ) )
	libs_run = TestMonitor()
	libs_run.run( cmds )
	libs_run.report()
	if not libs_run.okay:
		sys.exit(1)

def run( name, tests, builder ):
	# run tests
	if not args.libs:
		cmds = [ builder(test) for test in sorted(tests) ]
		# allows buidler to return None on some inputs
		cmds = filter(None,cmds)

		print( "Running {} tests ({})".format( name, leaf_opts ) )
		tests_run = TestMonitor()
		tests_run.run( cmds )
		tests_run.report()
		if not tests_run.okay:
			sys.exit(1)

# Run selected tests
if args.unit or args.all:
	tests = fs.find( 'share/unit_tests', include_dirs = False, name_regex = '.*\\.(leaf|lfx|lfb)' )
	run( "Unit", tests, build_job )

if args.mod or args.all:
	libs = fs.find( 'share/mod_tests/_libs', include_files = False, exclude_root = True )
	build_libs( libs )
	dirs = fs.find( 'share/mod_tests', include_files = False, not_whole_name_regex = '.*/_.*', exclude_root = True )
	run( "Modules", dirs, build_dir_job )

if args.exe or args.exe_jit or args.all:
	libs = fs.find( 'share/exe_tests/_libs', include_files = False, exclude_root = True )
	build_libs( libs )
	dirs = list(fs.find( 'share/exe_tests', include_files = False, 
		not_whole_name_regex = '.*/_.*', exclude_root = True) )
	if args.exe or args.all:
		run( "Standalone", dirs, build_exe_job )
	if args.exe_jit or args.all:
		# don't do those with tests.yaml since they can't be tested in JIT now
		run( "JIT", [ d for d in dirs if not os.path.exists( d+ '/tests.yaml' ) ], build_exe_jit_job )

else:
	print( "Missing test suite" )
	sys.exit(1)
