#include "ir/compiler_leaf.hpp"
#include "ir/builder.hpp"
#include "ir/type_manager_leaf.hpp"
#include "ir/std_functions.hpp"

#include "lang/language.hpp"

#define SHOW_DEBUG 1
#include "util/debug.hpp"
#include "lang/dump.hpp"

namespace ir
{

namespace {
	std::string export_name_of( leaf::var_decl const & decl, type const & t ) {
		std::string use_name = decl.var_scope->qualify_name( decl.symbol, platform::export_sep);
		//multi's must have a distinct name, but predictable as export symbol
		if( decl.form == leaf::var_decl::f_multi ) {
			use_name += platform::export_sep;
			//each override must have a unique type, so this works
			use_name += t.get_export_name();
		}
		
		DEBUG( "EXPORT", use_name );
		return use_name;
	}
	
	std::string import_name_of( leaf::var_decl const & decl, type const & t ) {
		std::string use_name;
		switch( decl.name_form ) {
			case leaf::var_decl::nf_bare:
				use_name = decl.import_name;
				break;
				
			case leaf::var_decl::nf_leaf:
				if( decl.import_module.size() ) {
					use_name = decl.import_module;
					use_name += platform::export_sep;
				}
				use_name += decl.import_name;
				
				if( decl.form == leaf::var_decl::f_multi ) {
					use_name += platform::export_sep;
					use_name += t.get_export_name();
				}
				break;
		}
		
		DEBUG( "IMPORT", use_name );
		return use_name;
	}
}

compiler_leaf::compiler_leaf( shared_ptr<leaf::scope> globals )
	: globals( globals ) 
	, types( new type_manager_leaf(*this) ) {
}

/*virtual*/ type_manager & compiler_leaf::get_types() {
	return *types;
}

shared_ptr<ir::module> compiler_leaf::compile_expr( leaf::expression const & expr )
{
	expr.validate();
	init_module( "single_expr" );
	
	module_builder mb( *this );
	auto fb = mb.create_function( "eval_expr" );
	//no params
	auto rv = fb.define_return( types->get( expr.type ), "value" );
	
	builder bb( fb );
	bb.add_store_inst( bb.access_raw_pointer(rv), eval_single_expr( expr, bb ) );
	return cur_module;
}

shared_ptr<ir::module> compiler_leaf::compile_block( leaf::statement_block const & bk )
{
	bk.validate();
	init_module( "single_block" );

	module_builder mb( *this );
	auto fb = mb.create_function( "_entry" );
	//no params
	//no returns

	builder bb(fb);
	build_block( bk, bb );
	return cur_module;
}

shared_ptr<ir::module> compiler_leaf::compile_full( leaf::statement_block const ** bks, size_t len )
{
	init_module( "program" );


	module_builder mb( *this );
	auto initf = mb.create_function( globals->qualify_name("_init_module",platform::export_sep) );
	initf.owner_function->is_export = true;
	auto finalf = mb.create_function( globals->qualify_name("_final_module",platform::export_sep) );
	finalf.owner_function->is_export = true;
	
	builder bb(initf);
	bb.split_defer = finalf.owner_function->entry_block.get();
	for( size_t i=0; i < len; ++i ) {
		bks[i]->validate();
		build_block( *bks[i], bb );
	}
	
	auto main = cur_module->if_get_global_variable( "main" );
	if( main ) {
		auto entryf = mb.create_function( "_entry" );
		builder eb( entryf );
		
		eb.add_funccall_inst( initf.get_value_defn() );
		
		auto & ftype = *main->direct_type;
		STATE_CHECK( ftype.is_callable() );
		auto fc = eb.add_funccall_inst( value_defn(main) );
		
		//is there a return value
		if( ftype.sub.size() ) {
			auto func_ret = entryf.define_return( ftype.sub_type(0), "rv" );
			auto rv = fc.define_return( ftype.sub_type(0) );
			assign( eb.access_raw_pointer(func_ret), rv, eb, eb );
		}
		
		eb.add_funccall_inst( finalf.get_value_defn() );
	}
	
	return cur_module;
}

/**
	This evaluates an expression where a value-list is allowed. This is done to avoid creating
	tuples where they are not actually in the source-code -- which involves potentially 
	needless copying	and a loss of clarity in the output IR.
	
	Compoung value-lists are flattened here. As it is assumed the input leaf structure is valid
	this shouldn't introduce any problems.
*/
compiler_leaf::value_ref_vector_t compiler_leaf::eval_list_expr( leaf::expression const & expr, builder & where, 
	unsigned flags )
{
	value_ref_vector_t out;
	eval_list_expr( out, expr, where, flags );
	return out;
}

void compiler_leaf::eval_list_expr( value_ref_vector_t & out, leaf::expression const & expr, builder & where,
	unsigned flags )
{
	switch( expr.form )
	{
		case leaf::expression::f_value_list:
		{
			auto & evl = static_cast<leaf::expr_value_list const&>(expr);
			for( size_t i=0; i < evl.sub.size(); ++i )
				eval_list_expr( out, *evl.sub[i], where, flags );
			break;
		}
			
		case leaf::expression::f_funccall:
			eval_funccall_expr( out, static_cast<leaf::expr_funccall const&>(expr), where, flags );
			break;

		default: 
			out.push_back( eval_single_expr( expr, where, flags ) );
			break;
	}
}

void compiler_leaf::eval_funccall_expr( value_ref_vector_t & out, leaf::expr_funccall const & expr, builder & where,
	unsigned /*flags*/ )
{
	auto fi = where.create_funccall_inst( eval_single_expr( *expr.func, where, ef_calling ) );
	
	auto & ft = expr.func->type.fun_cast<leaf::intr_type_function>();
	auto ptypes = ft.params;
	
	STATE_CHECK( expr.params->form == leaf::expression::f_value_list );
	auto & ep = static_cast<leaf::expr_value_list&>(*expr.params);
	STATE_CHECK( ep.sub.size() == ptypes->sub.size() );
	for( size_t i=0; i < ep.sub.size(); ++i )
	{
		unsigned flags = 0;
			
		value_ref ex = eval_single_expr( *ep.sub[i], where, flags );
		fi.add_param( assign_param( ex, where ) );
	}
		
	if( expr.context ) {
		value_ref ex = eval_single_expr( *expr.context, where );
		//TODO: this access_raw_pointer seems highly suspect!
		if( ex.result_type()->is_shared_pointer() ) {
			fi.set_context( where.access_raw_shared(ex) );
		} else {
			fi.set_context( where.access_raw_pointer(ex) );
		}
	}
	fi.add();
	
	leaf::intr_type const & vt = *ft.returns;
	auto & vl = vt.cast<leaf::intr_type_value_list>();
	for( size_t i=0; i < vl.sub.size(); ++i )
	{
		auto ref = fi.define_return( types->get( vl.sub[i].type ) );
		out.push_back( ref );
		//returns will be acquired in target function, so release them now
		//TODO: should release at end of current statement (not end of block)
		release( ref, where, block_flow::defer );
	}
	
}

value_ref compiler_leaf::eval_single_expr( leaf::expression const & expr, builder & where, unsigned /*flags*/ )
{
	value_ref vr;
	
	switch( expr.form )
	{
		case leaf::expression::f_block:
			//TODO:
			STATE_CHECK( false );

		case leaf::expression::f_symbol:
		{
			auto es = static_cast<leaf::expr_symbol const&>(expr);
			
			STATE_CHECK( es.sref.decl );
			value_defn sd;
			switch( es.sref.decl->import_type ) {
				case leaf::var_decl::it_none: {
					sd = obtain_symbol_value_defn( es.sref );
					break;
				}
				case leaf::var_decl::it_abi_code: {
					auto t = types->get( es.sref.decl->type );
					sd = where.define_value( value::f_import_code, t, import_name_of( *es.sref.decl, *t ) );
					break;
				}
				case leaf::var_decl::it_abi_data: {
					auto t = types->get( es.sref.decl->type );
					sd = where.define_value( value::f_import_data, t, import_name_of( *es.sref.decl, *t ) );
					break;
				}
			}
			if( es.type.is_lvalue() )
				vr = value_ref::as_raw_pointer( sd );
			else
				vr = value_ref::as_value( sd );
			return vr;
		}
			
		case leaf::expression::f_funccall:
		{
			value_ref_vector_t rets;
			eval_funccall_expr( rets, static_cast<leaf::expr_funccall const&>(expr), where );
			if( rets.size() != 1 ) {
				auto bt = where.create_pack_tuple_inst( types->get( expr.type ) );
				for( size_t i=0; i < rets.size(); ++i )
					bt->sub.push_back( rets[i] );
				where.add( bt );
				vr = bt->result;
			} else {
				vr = rets[0];
			}
			return vr;
		}

		//NOTE: I'm trying to prevent needing this, I'd prefer the value_list's don't exist at this
		//point as actual values. Certain conversions may require this to be reintroduced (just
		//remove the check and allow it to work as a tuple)
		case leaf::expression::f_value_list:
			STATE_CHECK(false);
			
		case leaf::expression::f_tuple:
		{
			auto & et = static_cast<leaf::expr_with_sub const&>(expr);
			auto bt = where.create_pack_tuple_inst( types->get( expr.type ) );
			for( size_t i=0; i < et.sub.size(); ++i )
				bt->sub.push_back( eval_single_expr( *et.sub[i], where ) );
			where.add( bt );
			vr = bt->result;
			return vr;
		}
		
		case leaf::expression::f_const_number:
		{
			auto ecn = static_cast<leaf::expr_const_number const&>(expr);
			switch( ecn.src_form )
			{
				case leaf::expr_const_number::sf_rational:
					STATE_FAIL( "rational no longer valid" );
				case leaf::expr_const_number::sf_integer:
					vr = where.const_integer( types->get( expr.type ), ecn.integer_value() );
					break;
				case leaf::expr_const_number::sf_binary:
					vr = where.const_binary( types->get( expr.type ), ecn.integer_value() ); //TODO: requires unsigned conversion
					break;
				case leaf::expr_const_number::sf_float:
					vr = where.const_float( types->get( expr.type ), double(ecn.float_value()) );
					break;
				case leaf::expr_const_number::sf_boolean:
					vr = where.const_boolean( ecn.boolean_value() );
					break;
			}
			return vr;
		}
			
		case leaf::expression::f_switch:
			return eval_switch( static_cast<leaf::expr_switch const&>(expr), where );
		
		case leaf::expression::f_funcdefn:
			return build_function( static_cast<leaf::expr_funcdefn const&>(expr), where );
			
		case leaf::expression::f_subscript:
			return eval_subscript( static_cast<leaf::expr_subscript const&>(expr), where );
			
		case leaf::expression::f_field:
			return eval_field( static_cast<leaf::expr_field const&>(expr), where );
			
		case leaf::expression::f_special_field:
			return eval_special_field( static_cast<leaf::expr_special_field const&>(expr), where );

		case leaf::expression::f_transform:
			return eval_transform( static_cast<leaf::expr_transform const&>(expr), where );
			
		case leaf::expression::f_binary:
			return eval_binary( static_cast<leaf::expr_binary const&>(expr), where );
			
		case leaf::expression::f_unary:
			return eval_unary( static_cast<leaf::expr_unary const&>(expr), where );
		
		case leaf::expression::f_new_func:
		{
			auto & nf = static_cast<leaf::expr_new_func const&>(expr);
			vr = get_new_func( *this, types->get( nf.new_type ) );
			return vr;
		}
		
		case leaf::expression::f_default:
			return gen_default( types->get(expr.type), where );
			
		case leaf::expression::f_conditional:
			return eval_conditional( static_cast<leaf::expr_conditional const&>(expr), where );
			
		case leaf::expression::f_text:
			return eval_text( static_cast<leaf::expr_text const&>(expr), where );
	}

	STATE_FAIL("Unsupported" );
}

/**
	To use a lazy-switch in evaluation requires a local variable of the result. This is required
	since the actual result will be elevated out of the block in which is created, which is
	not allowed. So instead we store in a local variable. This also allows the deferred destruction
	to happen in the correct scope.
*/
value_ref compiler_leaf::eval_switch( leaf::expr_switch const & es, builder & where )
{
	auto sb = where.create_switch( types->get(es.type) );
	for(size_t i=0; i < es.expr_list.size(); ++i ) {
		sb.add_case();
		if( es.expr_list[i].cond ) {
			sb.set_cond_value( eval_single_expr( *es.expr_list[i].cond, sb.get_cond_block() ));
		}
			
		sb.set_expr_value( eval_single_expr( *es.expr_list[i].value, sb.get_expr_block() ) );
	}
	sb.add();
	return sb.get_result();
#if 0
	//this form could be used if we can prove there is no defer in any of the contained blocks
	auto sb = where.create_switch( types->get(es.type) );
	for(size_t i=0; i < es.expr_list.size(); ++i ) {
		sb.add_case();
		if( es.expr_list[i].cond )
			sb.set_cond_value( eval_single_expr( *es.expr_list[i].cond, sb.get_cond_block() ));
		sb.set_expr_value( eval_single_expr( *es.expr_list[i].value, sb.get_expr_block() ));
	}
	vr = sb.get_result();
	sb.add();
	return vr;
#endif
}

value_ref compiler_leaf::eval_transform( leaf::expr_transform const & et, builder & where, unsigned /*eval_flags*/ )
{
	auto what_pl = []( leaf::intr_type const & vt ) {
		if( vt.is_integer() )
			return promote_loss_inst::w_integer;
		if( vt.is_binary() )
			return promote_loss_inst::w_binary;
		if( vt.is_float() )
			return promote_loss_inst::w_float;
		STATE_FAIL( "unsupported promotion" );
	};
	
	switch( et.transform )
	{
		case leaf::expr_transform::t_deref: {
			bool retain_lvalue = et.param->type.is_lvalue();
			value_ref vr = eval_single_expr( *et.param, where );
			if( et.param->type.is_shared() ) {
				if( retain_lvalue ) {
					return where.access_raw_shared( where.access_deref( vr ) );
				}
				return where.access_deref_shared( vr );
			} else {
				STATE_CHECK( et.param->type.is_value_ptr() );
//  				if( retain_lvalue ) {
//  					return vr;
//  				}
				return where.access_deref( vr );
			}
		}
	
		case leaf::expr_transform::t_drop_lvalue: {
			value_ref vr = eval_single_expr( *et.param, where );
			return where.access_deref( vr );
		}
		
		case leaf::expr_transform::t_addr_of: {
			value_ref vr = eval_single_expr( *et.param, where );
			if( et.param->type.is_array() ) {
				vr = where.access_raw_array( vr );
			} else if( et.param->type.is_lvalue() ) {
				//already a pointer
			} else {
				STATE_CHECK( "expecting-address" );
			}
			return vr;
		}
			
		case leaf::expr_transform::t_shared_expr: {
			if( et.cref ) {
				return value_ref::as_value( reinterpret_cast<symbol_cref*>(et.cref)->ref );
			} else {
				value_ref vr = eval_single_expr( *et.param, where );
				auto vd = where.resolve( vr );
				
				symbol_deque.resize( symbol_deque.size() + 1 );
				symbol_cref & sc = symbol_deque.back();
				sc.ref = vd.origin;
				et.cref = reinterpret_cast<intptr_t>(&sc);
				return vd;
			}
		}
			
		case leaf::expr_transform::t_promote: {
			return where.add_promote_loss( false, 
				eval_single_expr( *et.param, where ), 
				types->get( et.type ), what_pl( *et.param->type.get_fun() ), what_pl( *et.type.get_fun() ) );
		}
		case leaf::expr_transform::t_loss: {
			return where.add_promote_loss( true,
				eval_single_expr( *et.param, where ), 
				types->get( et.type ), what_pl( *et.param->type.get_fun() ), what_pl( *et.type.get_fun() ) );
		}
		
		case leaf::expr_transform::t_code_val: {
			return where.add_cast( eval_single_expr( *et.param, where ), types->get( et.type) );
		}
		case leaf::expr_transform::t_char_val: {
			return where.add_cast( eval_single_expr( *et.param, where ), types->get( et.type ) );
		}
		
		case leaf::expr_transform::t_make_optional: {
			return where.create_optional( 
				eval_single_expr( *et.param, where ),
				types->get( et.type )
			);
		}
		
		case leaf::expr_transform::t_deref_optional: {
			auto sb = where.create_switch();
			
			sb.add_case();
			auto base = eval_single_expr( *et.param, sb.get_cond_block() );
			sb.set_cond_value( sb.get_cond_block().access_optional_set( base ) );
			
			//TODO: rather than else a "not" on the cond
			sb.add_else();
			auto fail = make_shared<type>( type::a_raw );
			auto v = where.define_value( value::f_import_code, fail, "access_unset_optional" );
			sb.get_expr_block().add_funccall_inst( v );
			
			sb.add();
			return where.access_optional_value( base );
		}
		
		case leaf::expr_transform::t_bind_raw_func: {
			value_ref vr = eval_single_expr( *et.param, where );
			auto wf = where.add_bind_raw_func_inst( vr );
			release( wf->result, where, block_flow::defer );
			return wf->result;
		}
		
		case leaf::expr_transform::t_bind_func:
			return eval_transform_bind_func( et, where );
			
	}
	
	STATE_FAIL("invalid transform");
}

value_ref compiler_leaf::eval_conditional( leaf::expr_conditional const & ec, builder & where, unsigned /*eval_flags*/ )
{
	switch( ec.op ) {
		case leaf::expr_conditional::o_optional_or: {
			auto sb = where.create_switch( types->get(ec.type) );
			
			sb.add_case();
			auto base = eval_single_expr( *ec.cond, sb.get_cond_block() );
			sb.set_cond_value( sb.get_cond_block().access_optional_set( base ) );
			auto value = eval_single_expr( *ec.a, sb.get_expr_block() );
			sb.set_expr_value( value );
						
			sb.add_else();
			value = eval_single_expr( *ec.b, sb.get_expr_block() );
			sb.set_expr_value(value);
			
			sb.add();
			return sb.get_result();
		}
		
		case leaf::expr_conditional::o_optional_cond: {
			auto sb = where.create_switch( types->get(ec.type) );
			
			sb.add_case();
			sb.set_cond_value( eval_single_expr( *ec.cond, sb.get_cond_block() ) );
			auto value = eval_single_expr( *ec.a, sb.get_expr_block() );
			value = sb.get_expr_block().create_optional( value, types->get(ec.type) );
			sb.set_expr_value( value);
			
			sb.add_else();
			//type is optional, so default is an unset optional
			value = gen_default( types->get(ec.type), sb.get_expr_block() );
			sb.set_expr_value( value );
			
			sb.add();
			return sb.get_result();
		}
	}
	
	STATE_FAIL("invalid conditional");
}

value_ref compiler_leaf::eval_text( leaf::expr_text const & expr, builder & where )
{
	switch( expr.what) {
		case leaf::expr_text::w_char: {
			return where.const_binary( types->get(expr.type), uintmax_t(expr.text[0]) );
		}
		
		case leaf::expr_text::w_string: {
			auto at = types->get(expr.type);
			auto ctype = at->sub[0];
			auto ar = where.add_alloc_array_inst( at, ctype, 
				where.const_integer_normal( int(expr.text.size())) );
				
			auto local = where.define_local( at, "const_string" );
			auto lptr = where.access_raw_pointer( local );
			where.add_store_inst( lptr, ar->result );
			auto aptr = lptr;
			for( size_t i=0; i < expr.text.size(); ++i ) {
				auto off = where.add_array_offset( aptr, 
					where.const_integer_normal(int(i)) );
				auto val = where.const_binary( ctype, uintmax_t(expr.text[i]) );
				assign( off, val, where, where );
			}
			
			release( ar->result, where, block_flow::defer );
			return ar->result;
		}
	}
	
	STATE_FAIL("invalid text" );
}

/**
	Convert an unbound function into a bound one by creating the context and assigning
	the variables.
*/
value_ref compiler_leaf::eval_transform_bind_func( leaf::expr_transform const & et, builder & where )
{
	STATE_CHECK( et.param->form == leaf::expression::f_funcdefn );
	auto & func_defn = static_cast<leaf::expr_funcdefn const&>(*et.param);
	
	auto & func_type = et.param->type.fun_cast<leaf::intr_type_function>();
	
	builder assign_block( *where.insert_block() );
	
	shared_ptr<bind_func_inst> ci;
	value_ref func = eval_single_expr( *et.param, where );
	ci = assign_block.create_bind_func_inst( func, 
		types->get( leaf::extr_type_ref(func_type.context) ) );
	assign_block.add( ci );
	
	auto & q = func_defn.block->symbols->get_captures();
	unsigned index = 0;
	for( auto cit = q.begin(), end = q.end(); cit != end; ++cit, ++index ) {
		auto & local_decl = *cit->second.local.decl;
		auto local_type = types->get( local_decl.type );
		
		auto sp = make_shared<struct_pointer_inst>();
		sp->struct_ptr = ci->ctx_data.result;
		STATE_CHECK( local_decl.xref );
		sp->index = local_decl.xref - 1;
		sp->result = where.make_temp( type::get_raw_pointer( local_type ) );
		assign_block.add( sp );
		
		auto & src_decl = *cit->second.src.decl;
		symbol_cref  & sc = *reinterpret_cast<symbol_cref*>(src_decl.cref);
		auto ci_ref = value_ref::as_value( sc.ref );
		
		assign( sp->result, ci_ref, assign_block, assign_block, af_no_release ); //don't need distinct store_block
	}
	
	configure_destructor( ci->ctx_data, where );
	//release at end of construction block
	release( ci->result, where, block_flow::defer );
	
	return ci->result;	
}

value_ref compiler_leaf::eval_binary( leaf::expr_binary const & et, builder & where, unsigned /*flags*/ )
{
	auto a = eval_single_expr( *et.a, where );
	auto b = eval_single_expr( *et.b, where );
	
	binary_inst::op_t op = binary_inst::o_add; //keep GCC happy
	switch( et.op ) {
		#define MAP(t) case leaf::expr_binary::o_##t: op = binary_inst::o_##t; break;
		MAP(add)
		MAP(sub)
		MAP(mul)
		MAP(div)
		
		MAP(mod)
		
		MAP(bit_shl)
		MAP(bit_shr)
		MAP(bit_or)
		MAP(bit_and)
		MAP(bit_xor)
		
		MAP(eq)
		MAP(neq)
		MAP(lt)
		MAP(lte)
		MAP(gt)
		MAP(gte)
		#undef MAP
	}
	
	return where.add_binary( op, a, b, types->get( et.type ) );
}

value_ref compiler_leaf::eval_unary( leaf::expr_unary const & et, builder & where, unsigned /*flags*/ )
{
	auto param = eval_single_expr( *et.param, where );
	
	unary_inst::op_t op = unary_inst::o_neg; //keep GCC happy;
	switch( et.op ) {
		#define MAP(t) case leaf::expr_unary::o_##t: op = unary_inst::o_##t; break;
		MAP(neg)
		MAP(abs)
		
		MAP(bit_cpl)
		#undef MAP
	}
	
	return where.add_unary( op, param, types->get( et.type ) );
}

value_ref compiler_leaf::eval_subscript( leaf::expr_subscript const & ex, builder & where, unsigned /*flags*/ )
{
	auto & arr = ex.base->type.fun_cast<leaf::intr_type_array>();
	
	 if( !arr.raw ) {
		return where.add_array_offset(
			eval_single_expr( *ex.base, where ),
			eval_single_expr( *ex.offset, where ) );
	} else {
		return where.add_array_ptr_offset(
			eval_single_expr( *ex.base, where ),
			eval_single_expr( *ex.offset, where ),
			ex.base->type.is_lvalue() );
	}
}

value_ref compiler_leaf::eval_field( leaf::expr_field const & expr, builder & where, unsigned /*eval_flags*/ )
{
	STATE_CHECK( expr.base->type.is_tuple_or_list() || expr.base->type.is_custom() );

	value_ref base = eval_single_expr( *expr.base, where );
	value_ref result = base;
	
	size_t index;
	if( expr.sref ) {
		//resolving the base and its type above should have fixated all fields
		STATE_CHECK( expr.sref->decl->xref );
		//scope_ref's must be resolved to an index
		index = expr.sref->decl->xref - 1;
	} else {
		index = *expr.field_index;
	}
	
	if( expr.base->type.is_lvalue() ) {
		if( expr.base->type.is_shared() ) {
			result = where.access_deref( result ); //drop lvalue on raw_ptr to get shared
			result = where.access_raw_shared( result ); //get pointer to shared value
		} else if( expr.base->type.is_value_ptr() ) {
			//TODO: already a  pointer...?
		}
	} else {
		if( expr.base->type.is_shared() ) {
			result = where.access_deref_shared( result );
		} else if( expr.base->type.is_value_ptr() ) {
			result = where.access_deref( result ); //TODO: untested
		}
	}
	
	auto vr = where.access_post_field( result, index );
	
	if( expr.bind ) {	
		//TODO: like in the funccall this raw_pointer is suspect here (also needs shared)
		value_ref bvr = base;
		//if( !expr.base->type.is_value_ptr() )	
			bvr = where.access_raw_pointer(base);
			
		auto ex = where.add_bind_func_explicit_inst( vr, bvr );
		vr = ex->result;
		release( vr, where, block_flow::defer );
	}
	
	return vr;
}

value_ref compiler_leaf::eval_special_field( leaf::expr_special_field const & expr, builder & where, unsigned eval_flags )
{
	value_ref base = eval_single_expr( *expr.base, where, eval_flags );
	
	switch( expr.field )
	{
		case leaf::expr_special_field::f_array_size: {
			auto sf = make_shared<special_field_inst>();
			sf->object = base;
			sf->field = special_field_inst::f_array_len;
			sf->result = where.make_temp( types->get( expr.type ) );
			where.add( sf );
			return sf->result;
		}
		case leaf::expr_special_field::f_optional_set: {
			return where.access_optional_set( base );
		}
	}
	
	STATE_FAIL( "invalid field" );
}

builder compiler_leaf::maybe_build_on_error( leaf::error_statement const & es, builder & where )
{
	if( !es.on_error )
		return where;
		
	auto db = where.insert_block();
	db->error_block = where.create_block();
	db->error_level = int(es.on_match);
		
	builder eb( *db->error_block );
	build_block( *es.on_error, eb );

	return builder( *db );
}

void compiler_leaf::build_do_statement( leaf::do_statement const & ds, builder & where_ )
{
	if( !ds.expr_list.size() )
		return;

	builder where = maybe_build_on_error( ds, where_ );

	//a local must be defined otherwise the condition value would elevate out of the containing
	//block. This is problematic where there is a defere, since two blocks merge and only one
	//has set the actual value. Refer to the test <unit_tests/block/do/then_defer.lfb>

	if( ds.loop ) {
		STATE_CHECK( ds.expr_list.size() == 1, "unsupported-looping-construct" );
		auto loop_inst = where.add_loop();
		
		auto & item = ds.expr_list[0];
		if( item.cond ) {
			builder cond_block( *loop_inst->cond_block );
			auto result = where.define_local( types->get_boolean(), "cond_result" );
			auto lres = eval_single_expr( *item.cond, cond_block );
			assign( value_ref::as_raw_pointer( result ), lres, cond_block, cond_block );
			loop_inst->cond_value = result;
		} else {
			loop_inst->cond_value = where.const_boolean( true );
		}
		
		builder body_block( *loop_inst->body_block );
		build_block( *item.block, body_block );
	} else {

		auto result = where.define_local( types->get_boolean(), "cond_result" );
		auto sb = where.create_switch();
		for( size_t i=0; i < ds.expr_list.size(); ++i ) {
			auto & item = ds.expr_list[i];
			sb.add_case();
			if( item.cond ) {
				auto lres = eval_single_expr( *item.cond, sb.get_cond_block() );
				assign( value_ref::as_raw_pointer( result ), lres, sb.get_cond_block(), sb.get_cond_block() );
				sb.set_cond_value( result );
			}
				
			build_block( *item.block, sb.get_expr_block() );
		}
		sb.add();
	}
}

void compiler_leaf::build_type_statement( leaf::type_statement const & ts, builder & where )
{
	if( ts.init_block ) {
		build_multi_statement( *ts.init_block, where );
	}
}

void compiler_leaf::build_defer_statement( leaf::defer_statement const & ds, builder & where )
{
	auto db = make_shared<block>( *where.owner_function );
	builder bb( *db );
	build_block( *ds.block, bb );
	if( ds.error_only ) {
		where.defer_error( db );
	} else {
		where.defer( db );
	}
}

void compiler_leaf::build_return_statement( leaf::return_statement const & rs, builder & where )
{
	builder bb(where);
	if( rs.value ) {
		builder assign_block( *where.insert_block() );
		
		value_ref_vector_t out;
		for( size_t i=0; i < where.owner_function->returns.size(); ++i )
			out.push_back( value_ref::as_raw_pointer( where.owner_function->returns[i] ) );
			
		auto right = eval_list_expr( *rs.value, assign_block );
		assign( out, right, assign_block );
	}
	
	auto ri = make_shared<return_inst>();
	bb.add( ri );
}

void compiler_leaf::build_fail_statement( leaf::fail_statement const & fs, builder & where )
{
	if( fs.level == leaf::fail_level::resume ) {
		auto ri = make_shared<resume_inst>();
		where.add( ri );
	} else {
		auto fi = make_shared<fail_inst>();
		fi->level = int(fs.level);
		where.add( fi );
	}
}

void compiler_leaf::build_loop_flow_statement( leaf::loop_flow_statement const & ls, builder & where )
{
	auto lfi = make_shared<loop_flow_inst>();
	switch( ls.flow ) {
		case leaf::loop_flow_statement::f_break: 
			lfi->flow = loop_flow_inst::f_break; 
			break;
		case leaf::loop_flow_statement::f_next: 
			lfi->flow = loop_flow_inst::f_next; 
			break;
	}
	where.add(lfi);
}

void compiler_leaf::build_expression_statement( leaf::expression_statement const & es, builder & where )
{ 
	builder eb = maybe_build_on_error( es, where );
	eval_list_expr( *es.expr, eb ); 
}

void compiler_leaf::build_block( leaf::statement_block const & bk, builder & where )
{
	for( size_t i=0; i < bk.list.size(); ++i )
		build_statement( *bk.list[i], where );
}

void compiler_leaf::build_multi_statement( leaf::multi_statement const & ms, builder & where )
{
	for( size_t i=0; i < ms.list.size(); ++i )
		build_statement( *ms.list[i], where );
}

void compiler_leaf::build_statement( leaf::statement const & st, builder & where )
{
	switch( st.statement_form() )
	{
		#define BUILDS(type) \
			case leaf::sf_##type: \
				DEBUG("Statement: ", #type); \
				build_##type( static_cast<leaf::type const&>(st), where ); \
				break;
		BUILDS(assign_statement)
		BUILDS(defer_statement)
		BUILDS(do_statement)
		BUILDS(expression_statement)
		BUILDS(fail_statement)
		BUILDS(init_statement)
		BUILDS(loop_flow_statement)
		BUILDS(multi_statement)
		BUILDS(return_statement)
		BUILDS(type_statement)
		#undef BUILDS
		
		case leaf::sf_var_statement:
			//TODO: hmm, adding symbol scanner leaves some around (likely pushed as incomplete statements)
			//STATE_FAIL("unexpected"); //shouldn't be any left
			break;
		case leaf::sf_noop_statement:
		case leaf::sf_import_statement:
		case leaf::sf_for_statement:
			break;
	}
}
	
/**
	At this level it isn't easy (and thus not done) to determine if a variable may be accessed
	prior to the "init" for the variable. The LLVM lowering will however catch it.
*/
value_defn compiler_leaf::obtain_symbol_value_defn( leaf::scope_ref sr )
{
	auto sval = get_symbol_cref( sr.decl->cref );
	if( !sval->direct_type ) {
		auto & decl = *sr.decl;
		sval->direct_type = types->get( decl.type );
		
		if( decl.form == leaf::var_decl::f_multi || decl.form == leaf::var_decl::f_single ) {
			if( sr.root_depth == 0 ) {
				sval->form = value::f_global;
			} else {
				sval->form = value::f_local;
			}
		} else if( decl.form == leaf::var_decl::f_capture ) {
			sval->form = value::f_capture_ref;
		} else {
			STATE_FAIL( decl.symbol );
		}
		
		sval->name = export_name_of( decl, *sval->direct_type );
	}
	return sval;
}

void compiler_leaf::build_init_statement( leaf::init_statement const & vs, builder & outer_where )
{
	auto & decl = *vs.sref.decl;
	if( decl.import_type != leaf::var_decl::it_none )
		return;

	builder where( *outer_where.insert_block() );
	
	value_defn vd = obtain_symbol_value_defn( vs.sref );
	if( vd.origin->form == value::f_global ) {
		if( decl.is_export )
			vd.origin->is_export = true;
		cur_module->register_global( vd.origin );
	} else if( vd.origin->form == value::f_local ) {
		where.owner_function->locals.push_back( vd.origin );
	}
	
	value_ref nv = value_ref::as_raw_pointer( vd );

	unsigned flags = 0;
	if( vs.init_refer_value )
		flags |= if_init_refer_value;
	if( vs.undefined )
		flags |= if_undefined_value;
	initialize_leaf( value_ref::as_raw_pointer(vd), vs.sref.decl->type,  vs.value, flags, where );

	//initialization comes before release, this is why there is a "code" block
	if( vs.code ) {
		build_statement( *vs.code, where );
	}
	
	//by default release at end of block, if "no_release" specified it means only normal flow doesn't
	//need release but error flow still does
	//TODO: this probably needs to happen prior to the value assignment,
	//such that for shared pointers the item can be released if there is an error in the assignment
	release( vd, outer_where, 
		vs.no_release ? block_flow::defer_error : block_flow::defer, 
		rf_cur_value );
}

void compiler_leaf::initialize_leaf( value_ref ptr, leaf::extr_type_ref const & init_type_, 
	shared_ptr<leaf::expression> value, unsigned flags, builder & where )
{
	auto init_type = types->get( init_type_ );
	if( value ) {
		if( flags & if_init_refer_value )
			ptr = initialize( ptr, init_type, flags | if_undefined_value, where );
		
		auto right = eval_single_expr( *value, where );
		assign( ptr, right, where, where, af_no_release );
	} else {
		initialize( ptr, init_type, flags, where );
	}
}

void compiler_leaf::build_assign_statement( leaf::assign_statement const& as, builder & where )
{
	//context of implicit refer/relese is this statement itself.
	builder assign_block( *where.insert_block() );
	
	bool left_refer = as.left->type.is_refer();
	bool right_refer = as.right->type.is_refer();
	
	unsigned flags = 0;
	
	auto left = eval_list_expr( *as.left, assign_block ); 
	
	switch( as.how ) {
		case leaf::assign_statement::h_value:
			STATE_CHECK( !left_refer || right_refer );
			break;
		case leaf::assign_statement::h_init_refer:
			//STATE_CHECK( left_shared ); //could also be a value-list
			flags |= af_init_refer;
			break;
		case leaf::assign_statement::h_default: {
			STATE_CHECK( left.size() == 1 ); //TODO: shoudl be safe to just default each part
			
			auto sb = where.create_switch();
			//TODO: a "not" expression is really needed now...
			sb.add_case();
			sb.set_cond_value( sb.get_cond_block().access_optional_set( 
				sb.get_cond_block().access_deref( left[0] ) ) );
			
			sb.add_else();
			//further evaluation happens inside this conditional block
			assign_block = sb.get_expr_block();
			
			sb.add();
			break;
		}
	}
	
	auto right = eval_list_expr( *as.right, assign_block );
	assign( left, right, assign_block, flags );
}

value_ref compiler_leaf::build_function( leaf::expr_funcdefn const & expr, builder & where )
{
	//generate a suitable name
	static unsigned count = 0;
	char buf[128];
	std::string use_name = where.owner_function->defn_value->name;
	if( use_name == "_entry" ) //kind of cheap, but good enough for debugging
		use_name = "";
		
	if( expr.sug_name.size() ) {
		sprintf( buf, "_%d", ++count );
		use_name += '_';
		use_name += expr.sug_name;
		use_name += buf;
	} else {
		sprintf( buf, "_anon_%d", ++count );
		use_name += buf;
	}

	auto & func_type = expr.type.fun_cast<leaf::intr_type_function>();
	auto result = get_symbol_cref( expr.cref );
	auto fbb = where.create_function( use_name, types->get(expr.type), result );
	builder fbody( fbb );
	
	//track current scope function for later reference
	STATE_CHECK( expr.block->symbols->is_function_root(), use_name );
	STATE_CHECK( !expr.block->symbols->fref );
	expr.block->symbols->fref = reinterpret_cast<intptr_t>( fbb.owner_function );
	
	leaf::scope::var_vector_t const & srs = expr.block->symbols->get_in_order();
	for( size_t i=0; i < srs.size(); ++i )
	{
		leaf::var_decl const & decl = *srs[i].decl;
		if( decl.form == leaf::var_decl::f_param ) 	{
			fbb.define_param( types->get( decl.type ), decl.symbol, get_symbol_cref( decl.cref ) );
			
		} else if( decl.form == leaf::var_decl::f_return ) {
			fbb.define_return( types->get( decl.type ), decl.symbol, 
				get_symbol_cref( decl.cref ) );
			//an init statement is expected for returns (it is not done automatically here)
		}
	}
 
	if( func_type.access != leaf::intr_type_function::a_raw ) {
		STATE_CHECK( func_type.access == leaf::intr_type_function::a_unbound );
		
		auto & q = expr.block->symbols->get_captures();
		for( auto cit = q.begin(), end = q.end(); cit != end; ++cit )
		{
			auto & local_decl = *cit->second.local.decl;
			auto local_type = types->get( local_decl.type );
			//DEBUG( "CAPTURE", &local_decl.cref, local_decl.symbol, local_decl.xref );
			auto v = fbb.define_capture( local_type, local_decl.symbol, 
				get_symbol_cref( local_decl.cref ) );
			

			//typing of the function will resolve the "xref" of the type. If there local decl has an xref
			//we use that (happens for distinct closure context) otherwise the source must have
			//an xref (happens for class members)
			if( local_decl.xref ) {
				v.origin->capture_index = local_decl.xref - 1;
			} else {
				STATE_CHECK( cit->second.src.decl->xref );
				v.origin->capture_index = cit->second.src.decl->xref -1;
			}
		}
	}
			
	build_block( *expr.block, fbody );

	return value_ref::as_value(result);
}

} //eon ir