#include "ir/instruction.hpp"
#include "ir/block.hpp"
#include "ir/type.hpp"
#include "ir/dump.hpp"

namespace ir
{

switch_inst::conditional_block::conditional_block( block & parent )
{
	cond_block.reset( new block( parent.owner_function ) );
	expr_block.reset( new block( parent.owner_function ) );
}

void store_inst::validate() const
{
	left.validate();
	right.validate();

	auto t = left.result_type();
	if( t->is_raw_pointer() ) {
		auto pright = type::get_raw_pointer( right.result_type() );
		STATE_CHECK( *left.result_type() == *pright,
			dump::get_type( left.result_type().get() ) + " == " + dump::get_type(pright.get()) );
	} else if( t->is_shared_pointer() ) {
		STATE_CHECK( *left.result_type() == *type::get_shared_pointer( right.result_type() ) );
	} else {
		STATE_FAIL( "requires raw/shared pointer for storage" );
	}
}

void funccall_inst::validate() const
{
	call.validate();
	auto ct = call.result_type();
	STATE_CHECK( ct->is_callable() );
	
	STATE_CHECK( returns.size() == ct->sub.size() );
	STATE_CHECK( params.size() == ct->params.size() );
	
	for( size_t i=0; i < returns.size(); ++i )
		STATE_CHECK( *returns[i].result_type() == *ct->sub[i] );
	for( size_t i=0; i < params.size(); ++i )
		STATE_CHECK( *params[i].result_type() == *ct->params[i] );
}

void pack_tuple_inst::validate() const
{
	result.validate();
	auto pt = result.result_type();
	STATE_CHECK( pt->sub.size() == sub.size() );
	for( size_t i=0; i < sub.size(); ++i )
		STATE_CHECK( *sub[i].result_type() == *pt->sub[i] );
}

void switch_inst::validate() const
{
	if( with_result )
		result.validate();

	for( auto & cb : blocks )
	{
		cb->cond_block->validate();
		cb->expr_block->validate();
		if( cb->cond_value ) {
			cb->cond_value->validate();
			STATE_CHECK( cb->cond_value->result_type()->is_boolean() );
		}
		
		if( with_result )
		{
			cb->expr_value.validate();
			STATE_CHECK( *result.result_type() == *cb->expr_value.result_type() );
		}
	}
}

void resolve_inst::validate() const
{
	source.validate();
	result.validate();
	STATE_CHECK( *result.result_type() == *source.result_type() );
}

void malloc_inst::validate() const
{
	STATE_CHECK( direct_type );
	direct_type->validate();
	result.validate();
	STATE_CHECK( *result.result_type() == *type::get_raw_pointer( direct_type ) );
}

void alloc_shared_inst::validate() const
{
	sod.validate();
	STATE_CHECK( *sod.result.result_type() == *type::get_shared_pointer( sod.direct_type ) );
}

void bind_raw_func_inst::validate() const
{
	raw_func.validate();
	STATE_CHECK( raw_func.result_type()->fundamental == type::f_function_ptr );
	STATE_CHECK( raw_func.result_type()->access == type::a_raw );
	result.validate();
	STATE_CHECK( result.result_type()->fundamental == type::f_function_ptr );
	STATE_CHECK( result.result_type()->access == type::a_bound );
	STATE_CHECK( result.result_type()->sub_compare( *raw_func.result_type() ) == 0 );
}

void bind_func_inst::validate() const
{
	ctx_data.validate();
	STATE_CHECK( *ctx_data.result.result_type() == *type::get_raw_pointer( ctx_data.direct_type ) );
	
	ctx_func.validate();
	STATE_CHECK( ctx_func.result_type()->fundamental == type::f_function_ptr );
	STATE_CHECK( ctx_func.result_type()->access == type::a_unbound );
	
	result.validate();
	STATE_CHECK( result.result_type()->fundamental == type::f_function_ptr );
	STATE_CHECK( result.result_type()->access == type::a_bound );
}

void bind_func_explicit_inst::validate() const
{
	ctx_data.validate();
	
	ctx_func.validate();
	STATE_CHECK( ctx_func.result_type()->fundamental == type::f_function_ptr );
	STATE_CHECK( ctx_func.result_type()->access == type::a_unbound );
	
	result.validate();
	STATE_CHECK( result.result_type()->fundamental == type::f_function_ptr );
	STATE_CHECK( result.result_type()->access == type::a_bound );
}

void shared_object_defn::validate() const
{
	STATE_CHECK( direct_type );
	direct_type->validate();
	result.validate();
	if( !static_destructor )
		destructor.validate();
}

void release_shared_inst::validate() const
{
	object.validate();
	STATE_CHECK( object.result_type()->is_shared_object_pointer() );
	
	switch( destroy_how )
	{
		case dh_trivial:
		case dh_dynamic:
			break;
	}
}

void acquire_shared_inst::validate() const
{
	object.validate();
	STATE_CHECK( object.result_type()->is_shared_object_pointer() );
}

void alloc_array_inst::validate() const
{
	result.validate();
	STATE_CHECK( item_type );
	item_type->validate();
	length.validate();
	STATE_CHECK( length.result_type()->is_integer() );
}

void free_array_inst::validate() const
{
	array.validate();
}

void array_offset_inst::validate() const
{
	array.validate();
	offset.validate();
	result.validate();
	auto rt = array.result_type();
	STATE_CHECK( rt->is_array() 
		|| (rt->is_raw_pointer() && rt->sub[0]->is_array()) );
	STATE_CHECK( offset.result_type()->is_integer() );
}

void array_ptr_offset_inst::validate() const
{
	array.validate();
	offset.validate();
	result.validate();
	auto rt = array.result_type();
	STATE_CHECK( rt->is_raw_pointer() );
	STATE_CHECK( offset.result_type()->is_integer() );
}

void special_field_inst::validate() const
{
	object.validate();
	result.validate();
	
	switch( field )
	{
		case f_array_len:
			STATE_CHECK( object.result_type()->fundamental == type::f_array );
			return;
	}
	
	STATE_CHECK( false );
}

loop_inst::loop_inst( block & parent )
	: instruction( t_loop )
{
	cond_block = make_shared<block>( parent.owner_function );
	body_block = make_shared<block>( parent.owner_function );
}

void loop_inst::validate() const
{
	STATE_CHECK(cond_block);
	cond_block->validate();
	STATE_CHECK(body_block);
	body_block->validate();
	cond_value.validate();
	STATE_CHECK( cond_value.result_type()->is_boolean() );
}

void promote_loss_inst::validate() const
{
	source.validate();
	result.validate();
	STATE_CHECK( to_what >= w_integer && to_what <= w_binary );
	STATE_CHECK( from_what >= w_integer && from_what <= w_binary );
}

void loop_flow_inst::validate() const
{
	STATE_CHECK( flow == f_break || flow == f_next );
}

void binary_inst::validate() const
{
	a.validate();
	b.validate();
	result.validate();
	STATE_CHECK( *a.result_type() == *b.result_type() );
	if( is_boolean() ){
		STATE_CHECK( result.result_type()->is_boolean() );
	} else {	
		STATE_CHECK( *result.result_type() == *a.result_type() );
		STATE_CHECK( a.result_type()->is_scalar() );
	}
	STATE_CHECK( op >= min_op_ && op <= max_op_ );
}

void unary_inst::validate() const
{
	param.validate();
	result.validate();
	STATE_CHECK( *param.result_type() == *result.result_type() );
	STATE_CHECK( param.result_type()->is_scalar() );
	STATE_CHECK( op >= min_op_ && op <= max_op_ );
}

void clone_inst::validate() const
{
	source.validate();
	result.validate();
	if( source_value ) {
		STATE_CHECK( result.result_type()->is_raw_pointer() );
		STATE_CHECK( *result.result_type()->sub[0] == *source.result_type() );
	} else {
		STATE_CHECK( source.result_type()->is_raw_pointer() );
		STATE_CHECK( *result.result_type() == *source.result_type() );
	}
}

void cast_inst::validate() const
{
	source.validate();
	result.validate();
}

} //eon ir
