#include "ir/llvm/builder.hpp"

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#pragma GCC diagnostic ignored "-Wunused-parameter"
	#include "llvm/IR/Module.h"
	#include "llvm/IR/Function.h"
#pragma GCC diagnostic pop

#define SHOW_DEBUG 1
#include "util/debug.hpp"
#include <fstream>

namespace ir_llvm 
{

block_ptr block_data::get_flow( block_flow bf ) const
{
	if( bf == bf_seq && block_seq ) {
		return block_seq;
	}
	return context->get_flow(bf);
}

block_ptr block_context::get_flow( block_flow bf ) const 
{
	if( flow[bf] )
		return flow[bf];
	if( parent )
		return parent->get_flow(bf);
	return block_ptr();
}

block_ptr block_context::get_error_land() const
{
	if( !error_land && parent ) {
		return parent->get_error_land();
	}
	return error_land;
}

int block_context::get_max_error_level() const
{
	if( !parent )
		return error_level;
	return std::max( error_level, parent->error_level );
}

/////////////////////////////////////////////////////////////
// function_builder
function_builder::function_builder( llvm::Module & module, llvm::Function & function, 
	type_manager & types, gen::options const & gopts ) 
	: lbuilder( function.getContext() )
	, module( module )
	, llvm_context( function.getContext() )
	, function( function )
	, types( types )
	, gopts( gopts )
{ 
	entry_block = cur_block = block_data::create( 
		make_shared<block_context>(),
		llvm::BasicBlock::Create( llvm_context, "entry", &function ) );
	all_blocks.push_back( cur_block );
	lbuilder.SetInsertPoint( cur_block->ll_block );
	
	returnPath = nullptr;
	landingPadResult = nullptr;
	label_count = 0;
}

block_ptr function_builder::get_resume_block()
{
	if( resume_block )
		return resume_block;
			
	auto old_block = cur_block;
	resume_block = create_child_block( "resume" ); //TODO: should be child of function, not current
	set_insert_point( resume_block );
	auto lr = module.getFunction( "_Unwind_Resume" );
	direct_call( lr, CreateExtractValue( CreateLoad( landingPadResult, "lpad_result" ), 0, "except" ) );
	unreachable();
	
	set_insert_point( old_block );
	return resume_block;
}

block_ptr function_builder::get_rethrow_block()
{
	if( rethrow_block )
		return rethrow_block;
	
	auto old_block = cur_block;
	rethrow_block = create_child_block( "rethrow" );
	set_insert_point( rethrow_block );
	auto ur = module.getFunction( "leaf_rethrow" );
	direct_call( ur, CreateExtractValue( CreateLoad( landingPadResult, "lpad_result" ), 0, "except" ) );
	unreachable();
	set_insert_point( old_block );
	return rethrow_block;
}

void function_builder::complete_blocks()
{
	propagate_flow( entry_block, block_flow_flag(bf_seq) );

#if 1
	//debugging dump
	auto flow_str = []( unsigned flow ) {
		std::string result;
		#define BFN(p) if( flow & block_flow_flag(bf_##p) ) { result += #p " "; }
		BFN(seq)
		BFN(return)
		BFN(loop_break)
		BFN(loop_next)
		BFN(error)
		#undef BFN
		return result;
	};
	
	std::ofstream pd( (std::string("/tmp/dump_") + 
		function.getName() + ".txt").str().c_str() );
		
	auto dblock = [&]( std::string const & label, block_ptr bk ) {
		pd << "\t" << label << ": "
			<< (bk ? bk->ll_block->getName().str() : "MISSING" ) 
			<< std::endl;
	};
	
	for( auto & mb : all_blocks ) {
		pd << mb->ll_block->getName().str();
		if( mb->complete )
			pd << " complete";
		if( mb.get() == entry_block.get() )
			pd << " entry";
		if( mb->invoke_inst )
			pd << " invoke";
		pd << std::endl;
		
		pd << "\tflow_in: "  << flow_str( mb->flow_in ) << std::endl;
		pd << "\tflow_out: " << flow_str( mb->flow_out ) << std::endl;
		if( mb->block_undivert ) {
			pd << "\tundivert: " << mb->block_undivert->ll_block->getName().str() << std::endl;
		}
		if( mb->undiverts ) {
			pd << "\tundiverts" << std::endl;
		};
		for( int i =0; i < mb->cvt_size; ++i ) {
			pd << "\tconvert: " << flow_str(block_flow_flag(mb->cvt_in[i]))
				<< " -> " << flow_str(block_flow_flag(mb->cvt_out[i])) << std::endl;
		}
		
		for( int8_t i = 0; i < block_flow_size; ++i ) {
			if( !(mb->flow_out & block_flow_flag(block_flow(i)) ) ) {
				continue;
			}			
			auto db = mb->get_flow(block_flow(i));
			dblock( flow_str(block_flow_flag(block_flow(i))), db );
		}
		if( mb->cond_value ) {
			dblock( "block_cond_true", mb->block_cond_true );
		}
		if( mb->block_undivert ) {
			dblock( "block_undivert", mb->block_undivert );
		}
	}
	pd.close();
#endif

	//DEBUG( "------", GetInsertBlock().getFunction()->getName().str() );

	//using indexing since "all_blocks" can be extended within complete_block (iterators would invalidate)
	for( size_t i=0; i < all_blocks.size(); ++i ) {
		complete_block( all_blocks[i] );
	}
}

void function_builder::propagate_flow( block_ptr block, unsigned flow )
{
	unsigned c = block->flow_in | flow;
	//already handled
	if( c == block->flow_in ) {
		return;
	}
	block->flow_in = c;

	//converted flow (these are applied in order since they could be chained
	//for example, a error_post block which then resumes)
	for( int i=0; i < block->cvt_size; ++i ) {
		if( c & block_flow_flag(block->cvt_in[i]) ) {
			c &= ~block_flow_flag(block->cvt_in[i]);
			c |= block_flow_flag(block->cvt_out[i]);
		}
	}

	// all diverted flow becomes seq for this block
	if( block->block_undivert ) {
		propagate_flow( block->block_undivert, c );
		c = block_flow_flag(bf_seq);
	}
	
	block->flow_out |= c;
	if( block->invoke_inst ) {
		STATE_CHECK( block->context->get_error_land()->ll_block == block->invoke_inst->getUnwindDest() );
		propagate_flow( block->context->get_error_land(), block_flow_flag(bf_error) );
		//don't add bf_error to flow_out, since it has the above special propagation
	}
	
	unsigned pflow = block->flow_out;
	//undiverts doesn't propagate seq flow, since previous block_undivert did proper flow
	if( block->undiverts ) {
		pflow &= ~block_flow_flag(bf_seq);
	}
	
	//conditionals don't interrupt flow
	if( block->cond_value ) {
		STATE_CHECK( block->block_cond_true );
		STATE_CHECK( block->block_seq );
		propagate_flow( block->block_cond_true, pflow );
		propagate_flow( block->block_seq, pflow );
	}
	
	//branched flow
	for( int8_t i=0; i < block_flow_size; ++i ) {
		auto bf = block_flow(i);
		if( pflow & block_flow_flag(bf) ) {
			auto f = block->get_flow(bf);
			//STATE_CHECK(f) //should make resume another block
			if( f ) {
				propagate_flow( f, block_flow_flag(bf) );
			}
		}
	}
}
		
void function_builder::complete_block( block_ptr mb )
{
	if( mb->complete ) {
		return;
	}
	
	//check for unneeded blocks
#if 0
	if( !mb->flow_in && !mb->flow_out ) {
		mb->complete = true;
		mb->ll_block->removeFromParent();
		return;
	}
#endif
	
	STATE_CHECK( mb );
	STATE_CHECK( mb->ll_block );
	set_insert_point( mb );
	
	if( mb->cond_value ) {
		//STATE_CHECK( mb->flow_out == block_flow_flag(bf_seq),
		//	std::string( "invalid-block-out-flow:" ) + mb->ll_block->getName().str() );
		STATE_CHECK( mb->block_cond_true );
		STATE_CHECK( mb->block_seq );
		lbuilder.CreateCondBr( mb->cond_value, 
			mb->block_cond_true->ll_block, mb->block_seq->ll_block );
		mb->complete = true;
		return;
	}
	
	struct case_pair {
		llvm::ConstantInt * value;
		llvm::BasicBlock * dest;
	};
	std::vector<case_pair> case_pairs;
	for( int8_t bfi=0; bfi < block_flow_size; ++bfi ) {
		auto bf = block_flow(bfi);
		if( mb->flow_out & block_flow_flag(block_flow(bf)) ) {
			auto sb = mb->get_flow(bf);
			if( !sb && bf == bf_error ) {
				sb = get_rethrow_block();
			}
			STATE_CHECK( sb );
			case_pair cp;
			cp.value = ConstInt8(bfi);
			cp.dest = sb->ll_block;
			case_pairs.push_back( cp );
		}
	};
	
	if( case_pairs.size() == 0 ) {
		//STATE_CHECK( !case_pairs.empty() );
		lbuilder.CreateUnreachable();
	} else if( case_pairs.size() == 1 ) {
		lbuilder.CreateBr( case_pairs[0].dest );
	} else {
		auto sw = lbuilder.CreateSwitch( 
			CreateLoad( returnPath, "return_path" ),
			case_pairs[0].dest, 
			int(case_pairs.size()) );
		for( auto & cp : case_pairs )
			sw->addCase( cp.value, cp.dest );
	}
	
	mb->complete = true;
}

#if 0
void function_builder::complete_block( block_ptr mb )
{
	if( mb.is_complete() )
		return;
	
	SetInsertPoint( mb );
	STATE_CHECK( mb.seq_block() );
	STATE_CHECK( mb.return_block(), mb.get_name() );
	
	basic_block_ptr use_break, use_next;
	
	auto use_cleanup = [&]() { 
		return mb.cleanup_blockCont() ? mb.cleanup_blockCont() : get_resume_block(); 
	};
	auto use_resume = [&]() {
		return mb.cleanup_blockCont() ? mb.cleanup_blockCont() : get_rethrow_block();
	};
	
	if( mb.loop_break_block() ) {
		use_break = mb.loop_break_block();
	} else {
		use_break = mb.return_block();
	}
	
	if( mb.loop_next_block() ) {
		use_next = mb.loop_next_block();
	} else {
		use_next = mb.return_block();
	}
		
	struct case_pair {
		llvm::ConstantInt * value;
		llvm::BasicBlock * dest;
	};
	std::vector<case_pair> case_pairs;
	auto add_case_pair = [&]( int8_t value, basic_block_ptr ptr ) {
		case_pair cp;
		cp.value = ConstInt8( value );
		cp.dest = ptr.complete_case_block();
		case_pairs.push_back( cp );
	};
	if( !gopts.reduce_branch || mb.has_flow_in( basic_block_ptr::ff_return ) ) 
		add_case_pair( function_builder::rpv_return, mb.return_block() );
	if( !gopts.reduce_branch || mb.has_flow_in( basic_block_ptr::ff_resume ))
		add_case_pair( function_builder::rpv_resume, use_resume() );
	if( !gopts.reduce_branch || mb.has_flow_in( basic_block_ptr::ff_cleanup ))
		add_case_pair( function_builder::rpv_cleanup, use_cleanup()  );
	if( !gopts.reduce_branch || mb.has_flow_in( basic_block_ptr::ff_loop_break ))
		add_case_pair( function_builder::rpv_loop_break, use_break );
	if( !gopts.reduce_branch || mb.has_flow_in( basic_block_ptr::ff_loop_next))
		add_case_pair( function_builder::rpv_loop_next, use_next );
	
	if( case_pairs.size() == 0 ) {
		branch_sequence( mb.seq_block() );
	} else {
		auto sw = branch_switch( CreateLoad( returnPath, "return_path" ), mb.seq_block(), int(case_pairs.size()) );
		//TODO: these completions don't properly complete our block graph -- not used
		//right now, but perhaps we might want it for debugging later
		for( auto & cp : case_pairs )
			sw->addCase( cp.value, cp.dest );
	}
	
	mb.set_complete();
}
#endif

//NOTE: I'm unsure if this should actually be in builder at all and not just in gen
llvm::Value * function_builder::create_abs( llvm::Value * param )
{
	auto in_block = cur_block;
	auto cond = create_lt( param, const_zero( param->getType() ) );
	
	auto neg_block = create_child_block();
	auto post_block = create_child_block();
	branch_conditional( cond, neg_block, post_block );
	
	set_insert_point( neg_block );
	auto abs = create_neg( param );
	branch_sequence( post_block );
	
	set_insert_point( post_block );
	auto phi = create_phi( param->getType(), 2 );
	phi->addIncoming( param, in_block->ll_block );
	phi->addIncoming( abs, neg_block->ll_block );
	return phi;
}

} //eon ir_llvm