#include "ir/language.hpp"

#include "ir/llvm/gen.hpp"
#include "ir/llvm/type_manager.hpp"

#define SHOW_DEBUG 0
#include "util/debug.hpp"

#include "ir/llvm/builder.hpp"

#include "platform/platform.hpp"

//NOTE: I'm not entirely clear on these defines
#define __STDC_LIMIT_MACROS
#define __STDC_CONSTANT_MACROS 

// LLVM: doesn't seem to care about conversion warnings
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#pragma GCC diagnostic ignored "-Wunused-parameter"
	#include "llvm/Analysis/Verifier.h"
	#include "llvm/Assembly/AssemblyAnnotationWriter.h"
	#include "llvm/Assembly/PrintModulePass.h"
	#include "llvm/IR/Constants.h"
	#include "llvm/IR/DataLayout.h"
	#include "llvm/IR/DerivedTypes.h"
	#include "llvm/ExecutionEngine/GenericValue.h"
	#include "llvm/ExecutionEngine/Interpreter.h"
	#include "llvm/ExecutionEngine/JIT.h"
	#include "llvm/LinkAllPasses.h"
	#include "llvm/Linker.h"
	#include "llvm/IR/Module.h"
	#include "llvm/PassManager.h"
	#include "llvm/Support/FormattedStream.h"
	#include "llvm/IRReader/IRReader.h"
	#include "llvm/Support/NoFolder.h"
	#include "llvm/Support/raw_ostream.h"
	#include "llvm/Support/TargetSelect.h"
	#include "llvm/Support/SourceMgr.h"
	//added for object files
	#include "llvm/MC/MCAsmBackend.h"
	#include "llvm/MC/MCAsmInfo.h"
	#include "llvm/MC/MCCodeEmitter.h"
	#include "llvm/MC/MCContext.h"
	#include "llvm/MC/MCInstPrinter.h"
	#include "llvm/MC/MCInstrInfo.h"
	#include "llvm/MC/MCObjectFileInfo.h"
	#include "llvm/MC/MCRegisterInfo.h"
	#include "llvm/MC/MCSectionMachO.h"
	#include "llvm/MC/MCStreamer.h"
	#include "llvm/MC/MCSubtargetInfo.h"
	#include "llvm/MC/SubtargetFeature.h"
	#include "llvm/Support/Host.h"
	#include "llvm/Support/TargetRegistry.h"
	#include "llvm/Support/ToolOutputFile.h"
	#include "llvm/Support/raw_os_ostream.h"
#pragma GCC diagnostic pop

#include <fstream>

namespace
{
	template<typename T>
	void dump( T & t )
	{
		std::string out;
		llvm::raw_string_ostream outs( out );
		t.print( outs, nullptr );
		std::cout << out << std::endl;
	}
	
	template<typename T>
	void dumps( T & t )
	{
		std::string out;
		llvm::raw_string_ostream outs( out );
		t.print( outs );
		std::cout << out << std::endl;
	}
}

namespace ir_llvm
{

gen::gen( std::string const & base_ll, options gopts )
	: gopts( gopts )
{
	llvm::InitializeNativeTarget();
	llvm_context.reset( new llvm::LLVMContext );
	module.reset( new llvm::Module( "test", *llvm_context ) );
	data_layout.reset( new llvm::DataLayout( module.get() ) );
	types.reset( new type_manager( *llvm_context, *data_layout ) );

	except_personality = module->getOrInsertFunction( "leaf_exception_personality",
		types->f_natural_int(), 
		types->f_natural_int(), types->f_intx(32), types->f_intx(64), types->generic_ptr(), types->generic_ptr(),
		(llvm::Type*)0 );
	llvm::Type *caughtFields[] = { types->generic_ptr(), types->f_natural_int() };
	except_type = llvm::StructType::get( *llvm_context, llvm::ArrayRef<llvm::Type*>(caughtFields) );
	
	create_std_functions();
	
	llvm::SMDiagnostic err;	
	base_module.reset( llvm::ParseIRFile( base_ll, err, *llvm_context ) );
	if( !base_module )
		STATE_FAIL( err.getMessage().str() + ": " + err.getLineContents().str() );
	STATE_CHECK( llvm::verifyModule( *base_module ) == 0 );
}

gen::~gen()
{
}

/**
	This wraps the Leaf-IR entry point with a final exception handling clause. Though perhaps
	we shouldn't do this and instead let Leaf terminate (since the debugger then knows where
	it is).
*/
void gen::build_wrapped_entry()
{
	//if there is no entry don't create main (see compiler_leaf::compile_full)
	if( !module->getFunction("_entry" ) ) {
		return;
	}
	auto main = llvm::cast<llvm::Function>(module->getFunction( "_entry" ));
	auto uncaught = module->getFunction( "_uncaught_exception" );
	STATE_CHECK( uncaught );

	STATE_CHECK( !module->getFunction("main" ) );
	auto func = llvm::cast<llvm::Function>(module->getOrInsertFunction( "main",
		types->f_natural_int(), //return
		types->main_argc(), types->main_argv(), //params
		(llvm::Type*)0 ) );
	func->setHasUWTable();
	func->setLinkage(llvm::GlobalValue::ExternalLinkage);
	STATE_CHECK( func );
	
	function_builder builder( *module, *func, *types, gopts );
	
	//store argc/argv in globals
	auto params_at = func->arg_begin();
	auto argc = new llvm::GlobalVariable( *module, types->main_argc(), /*constant*/false,
		llvm::GlobalValue::ExternalLinkage, /*init*/0, "__leaf_argc" );
	argc->setInitializer( llvm::UndefValue::get( types->main_argc() ) );
	builder.CreateStore( params_at++, argc );
	
	auto argv = new llvm::GlobalVariable( *module, types->main_argv(), /*constant*/false,
		llvm::GlobalValue::ExternalLinkage, /*init*/0, "__leaf_argv" );
	argv->setInitializer( llvm::UndefValue::get( types->main_argv() ) );
	builder.CreateStore( params_at++, argv );
	
	auto cblock = builder.create_child_block( "catch" );
	auto eblock = builder.create_child_block( "end" );
	auto result = builder.add_invoke( main, eblock, cblock );
	
	builder.set_insert_point( eblock );
	if( main->getReturnType()->isVoidTy() ) {
		builder.func_return( builder.ConstNaturalInt( 0 ) );
	} else {
		builder.func_return( result );
	}
	
	//only global variables can be used for the clauses
	builder.set_insert_point( cblock );
	auto lp = builder.CreateLandingPad( except_type, except_personality );
	//catch_all and report the error
	auto g = new llvm::GlobalVariable( *module, types->f_natural_int(), /*constant*/true, 
		llvm::GlobalValue::LinkOnceODRLinkage, 
		llvm::ConstantInt::get( types->f_natural_int(), 0 ), "levelCatchAll" );
	g->setUnnamedAddr( true );
	lp->addClause( builder.CreateBitCast( g, types->generic_ptr() ) );
	builder.Call( uncaught, builder.CreateExtractValue(lp, 0 ) );
	builder.func_return( builder.ConstNaturalInt( 255 ) );
}

void gen::show_code( std::string const & sc )
{
	if( sc == "*" ) {
		dump( *module ); 
	} else if( !sc.empty() ) {
		for( auto p =  func_map.equal_range( sc ); p.first != p.second; ++p.first ) {
			if( p.first->first == sc ) {
				dump( *p.first->second );
			}
		}
	}
}


int gen::execute( ir::module const & m, std::string const & sc ) //main
{
	m.validate();
	build_module( m );
	build_wrapped_entry();
	
	//print out the IR before adding imported functions (thus showing only interesting bits)
	show_code( sc );

	link();
	auto ee = create_engine();
	auto main = linker->getModule()->getFunction( "main" );
	
	std::vector<llvm::GenericValue> args(2);
	args[0].IntVal = llvm::APInt( platform::abi_int_size, 0 );
	args[1].PointerVal = nullptr;
	llvm::GenericValue gv = ee->runFunction( main, args );
	return int(gv.IntVal.getSExtValue());
}

void gen::create_std_functions( )
{
	module->getOrInsertFunction(
		"count_malloc", types->generic_ptr(), types->f_natural_int(), (llvm::Type*)0 );
	module->getOrInsertFunction(
		"count_free", types->type_void(), types->generic_ptr(), (llvm::Type*)0 );
	module->getOrInsertFunction(
		"_void_destructor", types->type_void(), types->generic_ptr(), (llvm::Type*)0 );
	module->getOrInsertFunction(
		"_duplicate_release_shared", types->type_void(), (llvm::Type*)0 );
	module->getOrInsertFunction(
		"_uncaught_exception", types->type_void(), types->generic_ptr(), (llvm::Type*)0 );
	module->getOrInsertFunction(
		"_raise_exception", types->type_void(), types->f_natural_int(), (llvm::Type*)0 );

	//TODO: mark these as "noreturn" ?
	module->getOrInsertFunction(
		"_Unwind_Resume", types->type_void(), types->generic_ptr(), (llvm::Type*)0);
	module->getOrInsertFunction(
		"leaf_rethrow", types->type_void(), types->generic_ptr(), (llvm::Type*)0);
}

llvm::Value * gen::gen_malloc( llvm::Value * size, function_builder & bb )
{
	//TODO: called function should throw if cannot allocate
	auto func = llvm::cast<llvm::Function>( module->getFunction( "count_malloc" ) );
	STATE_CHECK( func );
	return bb.Call( func, size ); 
}

llvm::Value * gen::gen_malloc( size_t size, function_builder & bb ) 
{
	return gen_malloc( bb.ConstNaturalInt(int64_t(size)), bb );
}

llvm::Value * gen::heap_alloc_type( llvm::Type * t, function_builder & bb )
{
	auto raw = gen_malloc( types->alloc_size( t ), bb );
	return bb.CreateBitCast( raw, types->ptr_to(t) );
}

void gen::gen_free( llvm::Value * ptr, function_builder & bb )
{
	auto func = llvm::cast<llvm::Function>( module->getFunction( "count_free" ) );
	STATE_CHECK( func );
	
	std::vector<llvm::Value*> args;
	args.push_back( bb.CreateBitCast( ptr, types->generic_ptr() ) );
	bb.Call( func, args );
}

void gen::build_module( ir::module const & m )
{
	for( auto git = m.global_variable_map.begin(), end = m.global_variable_map.end(); git != end; ++git )
		build_global_variable( *git->second );
		
	for( auto fit = m.function_map.begin(), end = m.function_map.end(); fit != end; ++fit )
		setup_function( *fit->second );
	for( auto fit = m.function_map.begin(), end = m.function_map.end(); fit != end; ++fit )
		build_function( *fit->second );
}

void gen::create_symbol_cref( intptr_t & track, llvm::Value * v, symbol_cref::ref_type_t rt )
{
	//TODO: This was a nice safety check, it's disabled since exceptions regenerate "defer" code
	//STATE_CHECK( !track );
	
	symbol_deque.resize( symbol_deque.size() + 1 );
	symbol_cref & sc = symbol_deque.back();
	sc.ref_type = rt;
	sc.value = v;
	
	track = reinterpret_cast<intptr_t>(&sc);
}

void gen::build_global_variable( ir::value const & gv )
{
	auto gtype = types->value( *gv.direct_type );
	
	auto use_name = gv.name;
	if( use_name == "main" ) {
		use_name = "_main";
	}
	auto g = new llvm::GlobalVariable( *module, gtype.ll_type, /*constant*/false, 
		gv.is_export ? llvm::GlobalValue::ExternalLinkage : llvm::GlobalValue::InternalLinkage, 
		/*init*/0, use_name );
	g->setInitializer( llvm::UndefValue::get( gtype.ll_type ) );
		
	create_symbol_cref( gv.cref, g, symbol_cref::rt_global );
}

void gen::setup_function( ir::function const & f )
{
	DEBUG( "SETUP-FUNCTION: ", f.defn_value->name );
	
	auto func_type = types->true_function( *f.defn_value->direct_type );
	
	auto use_name = f.defn_value->name;
	
	auto func = llvm::cast<llvm::Function>(module->getOrInsertFunction( use_name, 
		func_type.ll_func_type ) );
	func->setHasUWTable();
	func->setLinkage( f.is_export ? llvm::GlobalValue::ExternalLinkage : llvm::GlobalValue::InternalLinkage);
	create_symbol_cref( f.defn_value->cref, func, symbol_cref::rt_value );
	
	func_map.insert( std::make_pair( f.defn_value->name, func ) );
}

void gen::build_function( ir::function const & f )
{
	DEBUG( "BUILD-FUNCTION: ", f.defn_value->name );
	
	auto sr = reinterpret_cast<symbol_cref*>( f.defn_value->cref );
	auto func = llvm::cast<llvm::Function>(sr->value);
	auto func_type = types->true_function( *f.defn_value->direct_type );
	
	function_builder builder( *module, *func, *types, gopts );
	
	//track whcih return path is being used, 0=normal, 1=shortened
	builder.returnPath = builder.CreateAlloca( types->f_intx(8), 0, "_return_path" );
	builder.CreateStore( builder.ConstInt8(function_builder::rpv_sequential), builder.returnPath );
	builder.landingPadResult = builder.CreateAlloca( except_type, 0, "_exception" );
	
	//connect params/return
	auto params_at = func->arg_begin();
	//either we create a local return allocation or use the provided one (first parameter)
	llvm::Value * ret_value = nullptr;
	if( f.returns.size() )
	{
		if( func_type.is_direct_return )
			ret_value = builder.CreateAlloca( func->getReturnType(), 0, "_return_val" );
		else
			ret_value = params_at++;
	}
	
	if( func_type.is_context )
	{
		llvm::Type * ctx_type = types->value( *f.defn_value->direct_type->context ).ll_type;
		auto closure_ptr = builder.CreateBitCast( params_at++,  types->ptr_to( ctx_type ) );
		
		//DEBUG( "CAPTURES" );
		for( size_t i=0; i < f.captures.size(); ++i )
		{
			auto & c = f.captures[i];
			size_t index = c.origin->capture_index;
			//DEBUG( "CAPTURE", i, index, c.origin->name, reinterpret_cast<intptr_t>(&c.origin->cref) );
				
			auto left = builder.CreateStructGEP( closure_ptr, unsigned(index), c.origin->name.c_str() );
			create_symbol_cref( c.origin->cref, left, symbol_cref::rt_pointer );
		}
	}
	
	for( size_t i=0; i < f.params.size(); ++i )
	{
		//parameters must be copied since LLVM doesn't provide a pointer. Ultimately we should
		//optimize this to only do those which are needed (though LLVM seems to do a fine job
		//of identifying and removing those in its own)
		STATE_CHECK( params_at != func->arg_end() );
		auto local = builder.CreateAlloca( params_at->getType() );
		builder.CreateStore( params_at++, local );
		create_symbol_cref( f.params[i].origin->cref, local, symbol_cref::rt_pointer );
		
		//if poiner not needed
		//create_symbol_cref( f.params[i].origin->cref, params_at++, symbol_cref::rt_value );
	}
	
	unsigned ret_index = 0;
	for( size_t i=0; i < f.returns.size(); ++i )
	{
		llvm::Value * v;
		if( f.returns.size() == 1 )
			v = ret_value;
		else
			v = builder.CreateStructGEP( ret_value, ret_index++, f.returns[i].origin->name.c_str() );
			
		create_symbol_cref( f.returns[i].origin->cref, v, symbol_cref::rt_pointer );
	}

	//build locals
	for( size_t i=0; i < f.locals.size(); ++i )
	{
		auto & local = *f.locals[i];
		auto v = builder.CreateAlloca( types->value( *local.direct_type ).ll_type, 0, local.name );
		create_symbol_cref( local.cref, v, symbol_cref::rt_pointer );
	}

	auto context = builder.get_context();
	auto end_block = builder.create_child_block( "func_ret" );
	context->flow[bf_seq] = end_block;
	context->flow[bf_return] = end_block;
	
	build_block( *f.entry_block, builder );
	
	builder.set_insert_point( end_block );
	//with no returns, or returns as a pointer, we return void
	if( f.returns.size() == 0 || !func_type.is_direct_return )
		builder.func_return_void();
	else	
		builder.func_return( builder.CreateLoad( ret_value ) );
		
	builder.complete_blocks();
}

void gen::build_block_part( ir::block_seq_item const & bk, function_builder & bb )
{
	if( bk.an_instruction ) {
		build_instruction( *bk.an_instruction, bb );
	}
	if( bk.a_block ) {
		build_block( *bk.a_block, bb );
	}
}

void gen::build_block( ir::block const & bk, function_builder & bb )
{
	//needed in case the context changes to go back to original context
	block_ptr seq_block;
	
	if( bk.error_block ) {
		auto end_block = bb.create_child_block( "error_end" );
		auto error_cont = bb.create_child_block( "error_cont" );
		
		auto error_body = bb.create_child_context( "error_body" );
		error_body->context->flow[bf_seq] = end_block;
		error_body->context->flow[bf_error] = error_cont;
		bb.branch_sequence( error_body );
		
		auto error_land = bb.create_child_block( error_body, "error_land" );
		error_body->context->error_land = error_land;
		auto catch_block = create_landing_block( error_land, error_cont, bb, bk.error_level );
		
		bb.set_insert_point( catch_block );
		build_block( *bk.error_block, bb );
		bb.branch_sequence( end_block );
		
		bb.set_insert_point( error_body );
		seq_block = end_block;
	}
	
	for( size_t i=0; i < bk.items.size(); ++i ) {
		auto & si = bk.items[i];
		
		switch( si.flow ) {
			case ir::block_flow::defer: {
				auto end_block = bb.create_child_block( "defer_end" );
				auto cont_block = bb.create_child_block( "defer_cont" );
				auto defer_block = bb.create_child_block( "defer_block" );
				
				auto defer_body = bb.create_child_context( "defer_body" );
				defer_body->context->flow[bf_seq] = defer_block;
				defer_body->context->flow[bf_loop_break] = defer_block;
				defer_body->context->flow[bf_loop_next] = defer_block;
				defer_body->context->flow[bf_return] = defer_block;
				defer_body->context->flow[bf_error] = defer_block;
				bb.branch_sequence(defer_body);
				
				auto defer_land = bb.create_child_block( defer_body, "defer_land" );
				defer_body->context->error_land = defer_land;
				create_landing_block( defer_land, defer_block, bb, lt_cleanup );
				
				bb.set_insert_point( defer_block );
				auto old_return_path = bb.flow_divert(cont_block);
				build_block_part( si, bb );
				bb.flow_undivert( cont_block, old_return_path );
				
				//cont_block handles cases where there is a conditional branch in the defer_end block,
				//which can't be handled at the moment, so this split is done
				cont_block->block_seq = end_block;
				
				if( !seq_block )
					seq_block = end_block;
				
				bb.set_insert_point( defer_body );
				break;
			}
				
			case ir::block_flow::defer_error: {
				auto end_block = bb.create_child_block( "defer_error_end" );
				auto defer_block = bb.create_child_block( "defer_error_block" );
				
				auto defer_body = bb.create_child_context( "defer_error_body" );
				defer_body->context->flow[bf_error] = defer_block;
				bb.branch_sequence( defer_body );
				
				auto defer_land = bb.create_child_block( defer_body, "defer_error_land" );
				defer_body->context->error_land = defer_land;
				create_landing_block( defer_land, end_block, bb, lt_cleanup );
				
				bb.set_insert_point( defer_block );
				auto old_return_path = bb.flow_divert( end_block );
				build_block_part( si, bb );
				bb.flow_undivert( end_block, old_return_path );
				
				bb.set_insert_point( defer_body );
				break;
			}
			
			case ir::block_flow::sequential: {
				build_block_part( si, bb );
				break;
			}
		} //end flow-switch
	} //end items loop

	if( seq_block )
		bb.set_insert_point( seq_block );
}

block_ptr gen::create_landing_block( block_ptr block, block_ptr seq, 
	function_builder & bb, int landing_type )
{
	bb.set_insert_point( block );
	auto lp = bb.CreateLandingPad( except_type, except_personality );
	if( landing_type == lt_cleanup ) {
		lp->setCleanup(true);
		bb.CreateStore( lp, bb.landingPadResult );
		bb.CreateStore( bb.ConstInt8(bf_error), bb.returnPath );
		return seq;
	} else {
		STATE_CHECK( landing_type ); //no catch-all's except cleanup (if needed,just don't call addClause)
		block->context->error_level = landing_type;
		int max_level = block->context->get_max_error_level();
		
		auto lg = clause_level_globals[max_level];
		if( !lg ) {
			char buf[128];
			sprintf( buf, "level%d", max_level );
			
			std::vector<llvm::Constant*> vals;
			vals.push_back( llvm::ConstantInt::get( types->f_natural_int(), landing_type ) );
			auto g = new llvm::GlobalVariable( *module, types->f_natural_int(), /*constant*/true, 
				llvm::GlobalValue::LinkOnceODRLinkage, 
				llvm::ConstantInt::get( types->f_natural_int(), max_level ), buf );
			//g->setUnnamedAddr( true );
			clause_level_globals[max_level] = g;
			lg = g;
		}
		lp->addClause( bb.CreateBitCast( lg, types->generic_ptr() ) );
		bb.CreateStore( lp, bb.landingPadResult );
		
		bb.set_insert_point(seq);
		//LLVM/ABI: landing pads may be combined, as well as lacking intermediate catch points
		//means we must chain the error handling and check for matches here
		auto cnd = bb.CreateICmpSLT( 
			bb.ConstNaturalInt(landing_type),
			bb.CreateExtractValue( bb.CreateLoad( bb.landingPadResult ), 1 )
			);
			
		auto catch_block = bb.create_child_block( "catch" );
		//TODO: cleanup this handling, also see in builder.cpp
		auto cont_error = seq->context->get_flow(bf_error);
		if( !cont_error )
			cont_error = bb.get_resume_block();
		bb.branch_conditional( cnd, cont_error, catch_block );
		
		//error is handled so revert to sequential flow (it is still handled even if part of a continuation)
		bb.set_insert_point( catch_block );
		bb.CreateStore( bb.ConstInt8(bf_seq), bb.returnPath );
		seq->add_cvt( bf_error, bf_seq );
		return catch_block;
	}
}

void gen::build_instruction( ir::instruction const & i, function_builder & bb )
{
	DEBUG( "INSTRUCTION: ", typeid( i ).name() );
	switch( i.type )
	{
		#define BINST(type) \
			case ir::instruction::t_##type: \
				return build_##type##_inst( static_cast<ir::type##_inst const &>(i), bb );
			LLVM_GEN_BUILD_INSTRUCTIONS
		#undef BINST
	}
	
	STATE_CHECK( false );
}

void gen::build_store_inst( ir::store_inst const & si, function_builder & bb )
{
	auto left = obtain_value( si.left, bb );
	auto right = obtain_value( si.right, bb );
#if 0
	dump( *module );
	dumps( *left );
	dumps( *right );
#endif
	
	//assigning to a shared pointer assigns to its value
	if( si.left.result_type()->is_shared_pointer() )
		left = bb.CreateStructGEP( left, sf_data );
		
	bb.CreateStore( right, left );
}

void gen::build_funccall_inst( ir::funccall_inst const & fi, function_builder & bb )
{
	auto func_type = types->true_function( *fi.call.result_type() );
	auto call = obtain_value( fi.call, bb );
	
	std::vector<llvm::Value*> args;
	
	//some returns are done as the first parmater
	llvm::Value * ret_ptr = nullptr;
	if( !func_type.is_direct_return )
	{
		ret_ptr = bb.CreateAlloca( func_type.ll_ret_type );
		args.push_back( ret_ptr );
	}

	//deal with the context-style call (after return param)
	if( func_type.is_context ) {
		if( fi.context ) {
			//explicit context provided
			auto val = obtain_value( *fi.context, bb ); //TODO: must be a ptr!
			args.push_back( bb.CreateBitCast( val, types->generic_ptr() ) );
		} else {
			auto gptr = bb.CreateStructLoad( call, cf_ctx_ptr, "ctx_gptr" );
			args.push_back( bb.CreateBitCast( gptr, types->generic_ptr(), "ctx_ptr" ) );
			call = bb.CreateStructLoad( call, cf_func_ptr, "func_ptr" );
		} 
	} else {
		STATE_CHECK( !fi.context );
	}

	//then add params
	for( size_t i=0; i < fi.params.size(); ++i ) {
		auto p = obtain_value( fi.params[i], bb );
		args.push_back( p );
	}
	
	llvm::Value * rv = bb.Call( call, args );
	if( ret_ptr )
		rv = bb.CreateLoad( ret_ptr );
		
	//map return values
	for( size_t i=0; i < fi.returns.size(); ++i )
	{	
		auto & r = fi.returns[i];
		
		llvm::Value * v;
		if( fi.returns.size() == 1 )
			v = rv;
		else
			v = bb.CreateExtractValue( rv, i );
		
		create_symbol_cref( r.origin->cref, v, symbol_cref::rt_value );
	}
}

void gen::build_pack_tuple_inst( ir::pack_tuple_inst const & bt, function_builder & bb )
{
	llvm::Value * cur = llvm::UndefValue::get( types->value( *bt.result.origin->direct_type ).ll_type );
	
	for( size_t i=0; i < bt.sub.size(); ++i )
	{
		//DEBUG( "PACK", i );
		auto part = obtain_value( bt.sub[i], bb );
		//dump( *module );
		cur = bb.CreateInsertValue( cur, part, i );
	}
	
	create_symbol_cref( bt.result.origin->cref, cur, symbol_cref::rt_value );
}

void gen::build_resolve_inst( ir::resolve_inst const & ei, function_builder & bb )
{
	auto source = obtain_value( ei.source, bb );
	create_symbol_cref( ei.result.origin->cref, source, symbol_cref::rt_value );	
}

void gen::build_struct_pointer_inst( ir::struct_pointer_inst const & sp, function_builder & bb )
{
	auto source = obtain_value( sp.struct_ptr, bb );
	auto sub = bb.CreateStructGEP( source, unsigned(sp.index) );
	create_symbol_cref( sp.result.origin->cref, sub, symbol_cref::rt_value );
}

void gen::build_malloc_inst( ir::malloc_inst const & mi, function_builder & bb )
{
	auto obj_type = types->value( *mi.direct_type );
	auto cast_ptr = heap_alloc_type( obj_type.ll_type, bb );

	create_symbol_cref( mi.result.origin->cref, cast_ptr, symbol_cref::rt_value );
}

/**
	There are two types of wrapper functions:
	1. The value being wrapped is a constant function value. In this case a specific wrapper is
	built which calls the target function.
	2. The value is a runtime variable. This pointer is used as the context and a generic runtime
	function is created which routes to the target function.
*/
void gen::build_bind_raw_func_inst( ir::bind_raw_func_inst const & wi, function_builder & bb )
{
	auto raw = obtain_value( wi.raw_func, bb );
	//auto raw_type = raw->getType();
	auto ctx_fp_type = llvm::cast<llvm::StructType>(
		types->true_ctx_function( *wi.result.origin->direct_type ).ll_type );
	auto ctx_func_type = types->true_function( *wi.result.origin->direct_type ).ll_func_type;

	
	auto lf = dynamic_cast<llvm::Function*>( raw );
	llvm::Function * wrap_func;
	
	static int name_count = 0;
	char buf_name[32];
	sprintf( buf_name, "_%d", name_count++ );
	std::string name = "_ctx_wrap_" + raw->getName().str() + buf_name;
	
	bool need_func = true; //I was hoping for runtime we could reuse, but it is still signature specific!
	wrap_func = llvm::cast<llvm::Function>( module->getOrInsertFunction( 
		name, ctx_func_type ) );
	
	if( need_func ) {
		wrap_func->setHasUWTable();
		if( !wrap_func->size() )
		{
			function_builder builder( *module, *wrap_func, *types, gopts );
	
			auto in_params = wrap_func->arg_begin();
			auto ctx = in_params++; //TODO: what happens if there is a ptr return value?
			std::vector<llvm::Value*> args;
			while( in_params != wrap_func->arg_end() )
				args.push_back( in_params++ );
			
			llvm::Value * result;
			if( lf ) {
				result = builder.direct_call( raw, args );
			} else {
				//the context ptr is actually the function ptr
				auto pptr = builder.CreateBitCast( ctx, raw->getType() );
				result = builder.direct_call( pptr, args );
			}
			if( result->getType()->isVoidTy() )
				builder.func_return_void();
			else
				builder.func_return( result );
		}
	}
	
	//build ctx func ptr
	auto fp = heap_alloc_type( ctx_fp_type, bb );
	bb.CreateStructStore( fp, bb.ConstNaturalInt(1), cf_ref_count, "ref_count" );
	bb.CreateStructStore( fp, wrap_func, cf_func_ptr, "func_ptr" );
	if( !lf ) {
		//use function pointer as the context
		bb.CreateStructStore( fp, bb.CreateBitCast( raw, types->generic_ptr() ), cf_ctx_ptr, "ctx_func" );
	}
	auto dtor =  module->getFunction( "_void_destructor" );
	STATE_CHECK( dtor );
	bb.CreateStructStore( fp, dtor, cf_destructor, "dtor" );
	
	create_symbol_cref( wi.result.origin->cref, fp, symbol_cref::rt_value );
}

void gen::build_bind_func_explicit_inst( ir::bind_func_explicit_inst const & ci, function_builder & bb )
{
	auto ctx_func = obtain_value( ci.ctx_func, bb );
	//TODO: what a misleading function name
	auto ctx_fp_type = types->true_ctx_function( *ci.result.origin->direct_type ).ll_type;
	
	auto ctx_data = obtain_value( ci.ctx_data, bb );
	auto fp = heap_alloc_type( ctx_fp_type, bb ); //TODO: huh, why this type?!
	bb.CreateStructStore( fp, bb.ConstNaturalInt(1), cf_ref_count, "ref_count" );
	bb.CreateStructStore( fp, ctx_func, cf_func_ptr, "func_ptr" );
	bb.CreateStructStore( fp, bb.CreateBitCast( ctx_data, types->generic_ptr() ), cf_ctx_ptr, "ctx_data" );
	auto dtor =  module->getFunction( "_void_destructor" );
	STATE_CHECK( dtor );
	bb.CreateStructStore( fp, dtor, cf_destructor, "dtor" );
	
	create_symbol_cref( ci.result.origin->cref, fp, symbol_cref::rt_value );
	
}

/**
	The standard context builder puts the context data at the end of the same structure. There is
	still a pointer to this data inside the structure since this other style is also required (to bind
	existing contexts to functions).
*/
void gen::build_bind_func_inst( ir::bind_func_inst const & ci, function_builder & bb )
{
	auto ctx_func = obtain_value( ci.ctx_func, bb );
	auto ctx_fp_type = types->true_ctx_function( *ci.result.origin->direct_type ).ll_type;
	auto fp_base_size = types->alloc_size( ctx_fp_type );
	
	auto ctx_type = types->value( *ci.ctx_data.direct_type ).ll_type; //TODO: use name again
	auto ctx_type_size = types->alloc_size(ctx_type);
	
	auto fp_raw = gen_malloc( fp_base_size + ctx_type_size, bb );
	auto fp = bb.CreateBitCast( fp_raw, types->ptr_to( ctx_fp_type ) );
	
	bb.CreateStructStore( fp, bb.ConstNaturalInt( 1 ), cf_ref_count );
	bb.CreateStructStore( fp, ctx_func, cf_func_ptr );
	
	if( ci.ctx_data.static_destructor )
	{
		//TODO: should actually be a pure-throwing function this should never be called (at least in debug builds)
		bb.CreateStructStore( fp, module->getFunction( "_void_destructor" ), cf_destructor );
	}
	else
	{
		auto dtor = obtain_value( ci.ctx_data.destructor, bb );
		dtor = bb.CreateBitCast( dtor, types->ptr_to( types->destructor() ) );
		bb.CreateStructStore( fp, dtor, cf_destructor );
	}
	
	auto ctx_raw_ptr = bb.CreateStructGEP( fp, cf_ctx_offset );
	auto ctx_ptr = bb.CreateBitCast( ctx_raw_ptr, types->ptr_to( ctx_type ) );
	bb.CreateStructStore( fp, bb.CreateBitCast( ctx_ptr, types->generic_ptr() ), cf_ctx_ptr );
	
	create_symbol_cref( ci.ctx_data.result.origin->cref, ctx_ptr, symbol_cref::rt_value );
	create_symbol_cref( ci.result.origin->cref, fp, symbol_cref::rt_value );
}

void gen::build_void_context_inst( ir::void_context_inst const & vc, function_builder & bb )
{
	auto ctx_fp_type = llvm::cast<llvm::StructType>(
		types->true_ctx_function( *vc.result.origin->direct_type ).ll_type );
	
	auto fp = heap_alloc_type( ctx_fp_type, bb );
	bb.CreateStructStore( fp, bb.ConstNaturalInt( 1 ), cf_ref_count );
	bb.CreateStructStore( fp, 
		llvm::ConstantPointerNull::get( llvm::cast<llvm::PointerType>(ctx_fp_type->getElementType( cf_func_ptr ) )),
		cf_func_ptr ); //TODO: A function which throws
	bb.CreateStructStore( fp, 
		llvm::ConstantPointerNull::get( llvm::cast<llvm::PointerType>(ctx_fp_type->getElementType( cf_ctx_offset ) ) ),
		cf_ctx_offset );
	auto dtor = module->getFunction( "_void_destructor" );
	STATE_CHECK( dtor );
	bb.CreateStructStore( fp, dtor, cf_destructor );
	
	create_symbol_cref( vc.result.origin->cref, fp, symbol_cref::rt_value );
}

void gen::build_alloc_shared_inst( ir::alloc_shared_inst const & as, function_builder & bb )
{
	//auto direct = types->value( *as.direct_type ).ll_type;
	auto wrap = types->true_shared( *as.sod.result.origin->direct_type ).ll_type;
	
	auto fp = heap_alloc_type( wrap, bb );
	bb.CreateStructStore( fp, bb.ConstNaturalInt( 1 ), sf_ref_count );
	
	if( as.sod.static_destructor )
	{
		//TODO: should actually be a pure-throwing function this should never be called
		bb.CreateStructStore( fp, module->getFunction( "_void_destructor" ), sf_destructor );
	}
	else
	{
		auto dtor = obtain_value( as.sod.destructor, bb );
		dtor = bb.CreateBitCast( dtor, types->ptr_to( types->destructor() ) );
		bb.CreateStructStore( fp, dtor, sf_destructor );
	}
	
	create_symbol_cref( as.sod.result.origin->cref, fp, symbol_cref::rt_value );
}

void gen::build_release_shared_inst( ir::release_shared_inst const & rs, function_builder & bb )
{
	//dump( *module );
	auto object = obtain_value( rs.object, bb );
	auto rc = bb.CreateStructGEP( object, cf_ref_count, "ref_count" );
	auto old_count = bb.CreateAtomicRMW( llvm::AtomicRMWInst::Sub, rc, 
		bb.ConstNaturalInt(1), llvm::Acquire );
		
	//debugging check for duplicate release of shared
	{
		auto cnd = bb.CreateICmpSLT( old_count, bb.ConstNaturalInt(1) );
		auto fail_block = bb.create_child_block(); 
		auto post_block = bb.create_child_block();
		bb.branch_conditional( cnd, fail_block, post_block );
		
		bb.set_insert_point( fail_block );
		auto func = llvm::cast<llvm::Function>( module->getFunction( "_duplicate_release_shared" ) );
		STATE_CHECK( func );
		bb.direct_call( func );
		bb.branch_sequence( post_block );
		
		bb.set_insert_point( post_block );
	}
	auto release_block = bb.create_child_block();
	auto post_block = bb.create_child_block();
	auto cnd = bb.CreateICmpEQ( old_count, bb.ConstNaturalInt( 1 ) );
	bb.branch_conditional( cnd, release_block, post_block );
	
	bb.set_insert_point( release_block );
	
	if( rs.destroy_how == ir::release_shared_inst::dh_dynamic )
	{
		unsigned field = 0;
		auto &rs_type = *rs.object.result_type();
		if( rs_type.is_shared_pointer() )
			field = sf_data;
		else if( rs_type.fundamental == ir::type::f_function_ptr 
			&& rs_type.access == ir::type::a_bound)
			field = cf_ctx_offset;
		STATE_CHECK(field);
		
		std::vector<llvm::Value*> args;
		args.push_back( bb.CreateBitCast( bb.CreateStructGEP( object, field, "object" ), types->generic_ptr() ) );
		auto destructor = bb.CreateStructLoad( object, sf_destructor, "dtor" );
		bb.Call( destructor, args );
	}
	
	gen_free( object, bb );
	bb.branch_sequence( post_block );
	
	bb.set_insert_point( post_block );
}

void gen::build_acquire_shared_inst( ir::acquire_shared_inst const & as, function_builder & bb )
{
	auto object = obtain_value( as.object, bb );
	auto rc = bb.CreateStructGEP( object, cf_ref_count, "ref_count" );
	bb.CreateAtomicRMW( llvm::AtomicRMWInst::Add, rc, bb.ConstNaturalInt(1), llvm::Release );
}

void gen::build_switch_inst( ir::switch_inst const & si, function_builder & bb )
{
	auto start_block = bb.get_insert_block();
	
	//setup final Phi
	auto end_block = bb.create_child_block( "switch_end" );
	llvm::PHINode * phi = nullptr;
	if( si.with_result )
	{
		bb.set_insert_point(end_block);
		auto t = types->value(*si.result.origin->direct_type).ll_type;
		phi = bb.create_phi( t, unsigned(si.blocks.size()) );
		bb.set_insert_point(start_block);
	}
	
	auto cond_block = start_block;
	for( size_t i=0; i < si.blocks.size(); ++i )
	{
		bool last = (i == (si.blocks.size()-1));
		
		auto & cb = *si.blocks[i];
		STATE_CHECK( cb.cond_value || last );
		STATE_CHECK( !last || !si.with_result || !cb.cond_value );
		
		bb.set_insert_point( cond_block );
		auto then_block = bb.create_child_block( "switch_then" );
		auto else_block = last ? end_block : bb.create_child_block( "switch_else" );
		
		build_block( *cb.cond_block, bb );
		
		if( !cb.cond_value ) {
			bb.branch_sequence( then_block );
		} else {
			auto cond = obtain_value( *cb.cond_value, bb );
			bb.branch_conditional( cond, then_block, else_block );
		}
	
		bb.set_insert_point( then_block );
		build_block( *cb.expr_block, bb );
		auto phi_block = bb.get_insert_block();
		
		llvm::Value * then_val = nullptr;
		if( si.with_result )
			then_val = obtain_value( cb.expr_value, bb );
		bb.branch_sequence( end_block );
		
		if( si.with_result )
			phi->addIncoming( then_val, phi_block->ll_block );
		cond_block = else_block;
	}
	
	if( si.with_result )
		create_symbol_cref( si.result.origin->cref, phi, symbol_cref::rt_value );

	bb.set_insert_point( end_block );
}

void gen::build_loop_inst( ir::loop_inst const & lo, function_builder & bb )
{
	//in current context
	auto end_block = bb.create_child_block( "loop_end" );
	//in new context
	auto cond_block = bb.create_child_context( "loop_cond" );
	bb.branch_sequence( cond_block );
	bb.set_insert_point( cond_block );
	bb.get_context()->flow[bf_seq] = end_block;
	build_block( *lo.cond_block, bb );
	
	auto body_block = bb.create_child_block( "loop_body" );
	auto cond = obtain_value( lo.cond_value, bb );
	bb.branch_conditional( cond, body_block, end_block );
	
	auto loop_next_block = bb.create_child_block( "loop_next" );
	bb.get_context()->flow[bf_loop_next] = loop_next_block;
	loop_next_block->add_cvt( bf_loop_next, bf_seq );
	loop_next_block->block_seq = cond_block;
		
	auto loop_break_block = bb.create_child_block( "loop_break" );
	bb.get_context()->flow[bf_loop_break] = loop_break_block;
	loop_break_block->add_cvt( bf_loop_break, bf_seq );
	loop_break_block->block_seq = end_block;
	
	bb.set_insert_point( body_block );
	build_block( *lo.body_block, bb );
	bb.branch_sequence( cond_block );
		
	bb.set_insert_point( end_block );
}

void gen::build_return_inst( ir::return_inst const & /*r*/, function_builder & bb )
{
	bb.flow_return();
}

void gen::build_alloc_array_inst( ir::alloc_array_inst const & aa, function_builder & bb )
{
	auto len_type = types->f_natural_int();
	auto length_size = types->alloc_size( len_type );
	
	auto item_type = types->value( *aa.item_type ).ll_type;
	auto one_item_size = types->alloc_size( item_type );
	
	auto align = types->max_align();
	STATE_CHECK( align >= length_size );
	
	auto length = bb.create_sign_ext( obtain_value( aa.length, bb ), len_type );
	
	auto items_size = bb.create_mul( llvm::ConstantInt::get( len_type, one_item_size ),
		length );
	auto full_size = bb.create_add( llvm::ConstantInt::get( len_type, align ), items_size );
		
	auto memory = gen_malloc( full_size, bb );
	
	auto length_ptr = bb.CreateBitCast( memory, types->ptr_to( len_type ));
	bb.CreateStore( length, length_ptr );
	
	create_symbol_cref( aa.result.origin->cref, memory, symbol_cref::rt_value );
}

void gen::build_fail_inst( ir::fail_inst const & fi, function_builder & bb )
{
	auto fail = module->getFunction( "_raise_exception" );
	bb.Call( fail, bb.ConstNaturalInt( fi.level ) );
}

void gen::build_free_array_inst( ir::free_array_inst const & fa, function_builder & bb )
{
	gen_free( obtain_value( fa.array, bb ), bb );
}

void gen::build_special_field_inst( ir::special_field_inst const & sf, function_builder & bb )
{
	switch( sf.field )
	{	
		case ir::special_field_inst::f_array_len:
		{
			auto array = obtain_value( sf.object, bb );
			auto length_ptr = bb.CreateBitCast( array, types->ptr_to( types->f_natural_int() ));
			auto length = bb.CreateLoad( length_ptr );
			create_symbol_cref( sf.result.origin->cref, length, symbol_cref::rt_value );
			return;
		}
	}
	
	STATE_FAIL( "invalid field" );
}

void gen::build_array_offset_inst( ir::array_offset_inst const & ao, function_builder & bb )
{
	auto offset = obtain_value( ao.offset, bb );
	auto base = obtain_value( ao.array, bb );
	
	auto array_type = ao.array.result_type();
	bool is_pointer = false;
	if( array_type->is_raw_pointer() ) {
		array_type = array_type->sub[0];
		is_pointer = true;
		base = bb.CreateLoad( base );
	}
	
	//TODO: seems quite similar to the bits in build_alloc_array_inst
	auto len_type = types->f_natural_int();
	auto item_type = types->value( *array_type->sub[0] ).ll_type;
	auto one_item_size = types->alloc_size( item_type );
	auto align = types->max_align();
	
	auto item_offset = bb.create_mul( llvm::ConstantInt::get( len_type, one_item_size ), 
		bb.create_sign_ext( offset, len_type ) );
	//data starts after the size field (of align size)
	item_offset = bb.create_add( llvm::ConstantInt::get( len_type, align ), item_offset );
	
	auto raw_ptr = bb.create_inbounds_gep( base, item_offset );
	auto item_ptr = bb.CreateBitCast( raw_ptr, types->ptr_to( item_type ) );
	
	if( is_pointer ) {
		create_symbol_cref( ao.result.origin->cref, item_ptr, symbol_cref::rt_value );
	} else {
		create_symbol_cref( ao.result.origin->cref, bb.CreateLoad( item_ptr ), symbol_cref::rt_value );
	}
}

void gen::build_array_ptr_offset_inst( ir::array_ptr_offset_inst const & ao, function_builder & bb )
{
	auto offset = obtain_value( ao.offset, bb );
	auto base = obtain_value( ao.array, bb );
	
	auto array_type = ao.array.result_type();
	if( ao.lvalue ) {
		array_type = array_type->sub[0];
		base = bb.CreateLoad(base);
	}
	
	auto len_type = types->f_natural_int();
	auto item_type = types->value( *array_type->sub[0] ).ll_type;
	auto one_item_size = types->alloc_size( item_type );
	
	auto item_offset = bb.create_mul( llvm::ConstantInt::get( len_type, one_item_size ), 
		bb.create_sign_ext( offset, len_type ) );
	
	//since we calculate offset the base must be a generic pointer
	base = bb.CreateBitCast( base, types->generic_ptr() );
	auto raw_ptr = bb.create_inbounds_gep( base, item_offset );
	auto item_ptr = bb.CreateBitCast( raw_ptr, types->ptr_to( item_type ) );
	
	if( ao.lvalue ) {
		create_symbol_cref( ao.result.origin->cref, item_ptr, symbol_cref::rt_value );
	} else {
		create_symbol_cref( ao.result.origin->cref, bb.CreateLoad( item_ptr ), symbol_cref::rt_value );
	}
}

void gen::build_promote_inst( ir::promote_inst const & pi, function_builder & bb )
{
	auto source = obtain_value( pi.source, bb );
	
	#define TO_FROM( t, f ) (pi.to_what == ir::promote_inst::t && pi.from_what == ir::promote_inst::f)
	llvm::Value * promote;
	if( TO_FROM( w_integer, w_integer ) ) {
		auto target_type = types->f_intx( pi.to_bitsize );
		promote = bb.create_sign_ext( source, target_type );
	} else if( TO_FROM( w_binary, w_binary ) 
		|| TO_FROM( w_integer, w_binary ) ) {
		auto target_type = types->f_intx( pi.to_bitsize );
		promote = bb.create_zero_ext( source, target_type );
	} else if( TO_FROM( w_float, w_float ) ) {
		auto target_type = types->f_float( pi.to_bitsize );
		promote = bb.create_fp_ext( source, target_type );
	} else if( TO_FROM( w_float, w_integer ) ) {
		auto target_type = types->f_float( pi.to_bitsize );
		promote = bb.create_si_to_fp( source, target_type );
	} else if( TO_FROM( w_float, w_binary ) ) {
		auto target_type = types->f_float( pi.to_bitsize );
		promote = bb.create_ui_to_fp( source, target_type );
	} else {
		STATE_FAIL();
	}
	#undef TO_FROM	
	create_symbol_cref( pi.result.origin->cref, promote, symbol_cref::rt_value );
}

void gen::build_loss_inst( ir::loss_inst const & li, function_builder & bb )
{
	auto source = obtain_value( li.source, bb );
	
	#define TO_FROM( t, f ) (li.to_what == ir::promote_inst::t && li.from_what == ir::promote_inst::f)
	llvm::Value * loss;
	if( TO_FROM( w_integer, w_integer ) ||
		TO_FROM( w_binary, w_integer ) || 
		TO_FROM( w_integer, w_binary ) ) {
		auto target_type = types->f_intx( li.to_bitsize );
		loss = bb.create_sext_or_trunc( source, target_type );
	} else if( TO_FROM( w_integer, w_float ) ) {
		auto target_type = types->f_intx( li.to_bitsize );
		loss = bb.create_fp_to_si( source, target_type );
	} else if( TO_FROM( w_float, w_float ) ) {
		auto target_type = types->f_float( li.to_bitsize );
		loss = bb.create_fp_trunc( source, target_type );
	} else if( TO_FROM( w_float, w_integer ) ) {
		auto target_type = types->f_float( li.to_bitsize );
		loss = bb.create_si_to_fp( source, target_type );
	} else if( TO_FROM( w_float, w_binary ) ) {
		auto target_type = types->f_float( li.to_bitsize );
		loss = bb.create_ui_to_fp( source, target_type );
	} else if( TO_FROM( w_binary, w_float ) ) {
		auto target_type = types->f_intx( li.to_bitsize );
		loss = bb.create_fp_to_ui( source, target_type );
	} else {
		STATE_FAIL();
	}
	#undef TO_FROM
	create_symbol_cref( li.result.origin->cref, loss, symbol_cref::rt_value );
}

void gen::build_cast_inst( ir::cast_inst const & ci, function_builder & bb )
{
	auto source = obtain_value( ci.source, bb );
	auto ret = bb.CreateBitCast( source, types->value( *ci.result.result_type() ).ll_type );
	create_symbol_cref( ci.result.origin->cref, ret, symbol_cref::rt_value );
}

void gen::build_resume_inst( ir::resume_inst const & /*ri*/, function_builder & bb )
{
	bb.flow_resume();
}

void gen::build_loop_flow_inst( ir::loop_flow_inst const & lfi, function_builder & bb )
{
	switch( lfi.flow ) {
		case ir::loop_flow_inst::f_break: 
			bb.flow_loop_break();
			break;
		case ir::loop_flow_inst::f_next:
			bb.flow_loop_next();
			break;
	}
}

void gen::build_binary_inst( ir::binary_inst const & bi, function_builder & bb )
{
	auto a = obtain_value( bi.a, bb );
	auto b = obtain_value( bi.b, bb );
	
	auto rt = bi.a.result_type();
	bool sign = !rt->is_binary() && !rt->is_boolean();
	
	llvm::Value * result = nullptr;
	switch( bi.op ) {
		#define SIMPLE(t) \
			case ir::binary_inst::o_##t: \
				result = bb.create_##t( a, b, sign ); \
				break;
		SIMPLE(add)
		SIMPLE(sub)
		SIMPLE(mul)
		SIMPLE(div)
		
		SIMPLE(mod)
		
		SIMPLE(bit_shl)
		SIMPLE(bit_shr)
		SIMPLE(bit_or)
		SIMPLE(bit_and)
		SIMPLE(bit_xor)
		
		SIMPLE(eq)
		SIMPLE(neq)
		SIMPLE(lt)
		SIMPLE(lte)
		SIMPLE(gt)
		SIMPLE(gte)
		#undef SIMPLE
	}
	STATE_CHECK(result);
	
	create_symbol_cref( bi.result.origin->cref, result, symbol_cref::rt_value );
}

void gen::build_comment_inst( ir::comment_inst const &, function_builder & )
{
	//no effect
}

void gen::build_clone_inst( ir::clone_inst const & ci, function_builder & bb )
{
	auto source = obtain_value( ci.source, bb );
	auto temp = bb.CreateAlloca( types->value( *ci.result.result_type()->sub[0] ).ll_type );
	bb.CreateStore( ci.source_value ? source : bb.CreateLoad(source), temp );
	create_symbol_cref( ci.result.origin->cref, temp, symbol_cref::rt_value ); //not a pointer since the result is the value of the pointer
}

void gen::build_unary_inst( ir::unary_inst const & un, function_builder & bb )
{
	auto param = obtain_value( un.param, bb );
	
	llvm::Value * result = nullptr;
	switch( un.op ) {
		#define SIMPLE(t) \
			case ir::unary_inst::o_##t: \
				result = bb.create_##t( param ); \
				break;
		SIMPLE(neg)
		SIMPLE(abs)
		
		SIMPLE(bit_cpl)
		#undef SIMPLE
	}
	
	STATE_CHECK(result);
	create_symbol_cref( un.result.origin->cref, result, symbol_cref::rt_value );
}

llvm::Value* gen::obtain_value( ir::value_ref const & vr_, function_builder & bb )
{
	typedef boost::error_info<struct value_name,std::string> ex_value_name;
	ir::value_ref vr = vr_;
	ir::value const * v = vr.origin.get();
	
	llvm::Value * ret = nullptr;
	auto std_proc = [&]( llvm::Value * raw, symbol_cref::ref_type_t rt ) {
		ret = raw;
		auto ret_type = v->direct_type;
		
		bool is_ptr = false;
		switch( rt )
		{
			case symbol_cref::rt_value:
				break;
				
			case symbol_cref::rt_pointer:
				if( vr.how != vr.h_raw_pointer ) {
					ret = bb.CreateLoad( ret );
				} else {
					is_ptr = true;
				}
				break;
				
			case symbol_cref::rt_global:
				ret = bb.create_gep( ret, 0 );
				if( vr.how != vr.h_raw_pointer ) {
					ret = bb.CreateLoad( ret );
				} else {
					is_ptr = true;
				}
				break;
		}
		
		//resolve pre_field
		if( vr.pre_field ) 	{
			if( ret_type->is_raw_pointer() || is_ptr ) {
				ret = bb.CreateStructGEP( ret, unsigned(*vr.pre_field) );
			} else {
				ret  = bb.CreateExtractValue( ret, *vr.pre_field );
			}
		}
		
		if( vr.how == vr.h_raw_pointer ) {
			STATE_CHECK( is_ptr, "cannot get pointer to value" );
		} else if( vr.how == vr.h_deref ) {
			ret = bb.CreateLoad( ret );
		} else if( vr.how == vr.h_deref_shared ) {
			ret = bb.CreateStructLoad( ret, sf_data );
		} else if( vr.how == vr.h_raw_shared ) {
			ret = bb.CreateStructGEP( ret, sf_data );
		} else if( vr.how == vr.h_raw_array ) {
			STATE_CHECK( ret_type->is_array() );
			auto base = bb.create_inbounds_gep( ret,
				llvm::ConstantInt::get( types->f_natural_int(), types->max_align() ) );
			ret = bb.CreateBitCast( base, types->ptr_to(	
				types->value(*ret_type->item_type()).ll_type ) );
			dumps( *ret->getType() );
		} else /*h_value*/ {
		}
	};
	
	try {
		switch( v->form ) {
			case ir::value::f_const_boolean:
				ret = llvm::ConstantInt::get( llvm::Type::getInt1Ty( *llvm_context ), v->boolean_value );
				break;
			case ir::value::f_const_integer:
				//TODO: > int64 constant value
				ret = llvm::ConstantInt::get( types->value( *v->direct_type ).ll_type, v->int_value, true );
				break;
			case ir::value::f_const_binary:
				//TODO: > uint64 constant value
				ret = llvm::ConstantInt::get( types->value( *v->direct_type ).ll_type, v->uint_value, false );
				break;
			case ir::value::f_const_float:
				//TODO: What about 80/128 bit constants?!
				ret = llvm::ConstantFP::get( types->value( *v->direct_type ).ll_type, double(v->float_value) );
				break;
			case ir::value::f_nullptr:
				ret = llvm::ConstantPointerNull::get( llvm::cast<llvm::PointerType>(
					types->value( *v->direct_type ).ll_type) );
				break;
			case ir::value::f_undefined:
				ret = llvm::UndefValue::get( types->value( *v->direct_type ).ll_type );
				break;
				
			case ir::value::f_global:
			case ir::value::f_local:
			case ir::value::f_temp:
			case ir::value::f_return_ref:
			case ir::value::f_param_ref:
			case ir::value::f_capture_ref:
			{
				DEBUG( "VALUE", reinterpret_cast<intptr_t>(&v->cref) );
				STATE_CHECK( v->cref, std::string("uninitialized-value: ") + v->name );
				auto sc = reinterpret_cast<symbol_cref*>( v->cref );
				std_proc( sc->value, sc->ref_type );
				break;
			}
				
			case ir::value::f_import_code: {
				auto type = types->true_function(*v->direct_type ).ll_func_type;
				auto func = llvm::cast<llvm::Function>( module->getOrInsertFunction( v->name,
					type ) );
				std_proc( func, symbol_cref::rt_value );
				break;
			}
			
			case ir::value::f_import_data: {
				auto gv = module->getGlobalVariable( v->name );
				if( !gv ) {
					auto type = types->value( *v->direct_type ).ll_type;
					gv = new llvm::GlobalVariable( *module, type, /*constant*/false,
						llvm::GlobalValue::ExternalLinkage, /*init*/0, v->name );
				}
				std_proc( gv, symbol_cref::rt_global );
				break;
			}
		}
		
		STATE_CHECK( ret );
		return ret;
	} catch( boost::exception & be ) {
		be << ex_value_name( v->name );
		throw;
	}
}

llvm::ExecutionEngine* gen::create_engine()
{
	llvm::TargetOptions topts;
	topts.JITExceptionHandling = 1;
	topts.JITEmitDebugInfo = 1;
	
	std::string errStr;
	llvm::ExecutionEngine * ee = llvm::EngineBuilder( linker->getModule() )
		.setErrorStr( &errStr )
		.setEngineKind( llvm::EngineKind::JIT )
		.setTargetOptions( topts )
		.create();
		
	STATE_CHECK( ee );

	if( gopts.llvm_opt )
	{
		llvm::PassManager fpm;
		fpm.add( new llvm::DataLayout( *ee->getDataLayout() ) );
		fpm.add( llvm::createBasicAliasAnalysisPass() );
		fpm.add( llvm::createPromoteMemoryToRegisterPass());
		fpm.add( llvm::createInstructionCombiningPass());
		fpm.add( llvm::createAAEvalPass() );
		
		fpm.add( llvm::createReassociatePass());
		fpm.add( llvm::createGVNPass());
		fpm.add( llvm::createConstantPropagationPass() );
		fpm.add( llvm::createFunctionInliningPass());
		fpm.add( llvm::createPartialInliningPass());
		fpm.add( llvm::createInstructionCombiningPass());
		fpm.add( llvm::createPromoteMemoryToRegisterPass());
		fpm.add( llvm::createTailCallEliminationPass());
		fpm.add( llvm::createCFGSimplificationPass());
		
		fpm.add( llvm::createDeadStoreEliminationPass() );
		fpm.add( llvm::createDeadCodeEliminationPass() );
		fpm.add( llvm::createUnreachableBlockEliminationPass() );
		
		fpm.run( *linker->getModule() );
		
		//print out the IR
		dump( *linker->getModule() );
	}
	
	return ee;
}

void gen::link()
{
	//import functions
	auto & fl = base_module->getFunctionList();
	for( auto it = fl.begin(); it != fl.end(); ++it )
		module->getOrInsertFunction( it->getName(), it->getFunctionType(), it->getAttributes() );

	//LLVM3.3
	linker.reset( new llvm::Linker( base_module.get() ) );
	std::string err_msg;
	bool err = linker->linkInModule( module.get(), &err_msg );
	STATE_CHECK( !err, err_msg );
	
	STATE_CHECK( llvm::verifyModule( *linker->getModule() ) == 0 );
}

void gen::emit_object( ir::module const & m, std::string const & base_path,
	std::string const & sc )
{
	//LLVM: Not clear why I need this, especially since InitializeNativeTarget was called already
	llvm::InitializeNativeTargetAsmPrinter();

	m.validate();
	build_module( m );
	build_wrapped_entry();
	show_code(sc);
	link();

	std::string err_str;
	
	//emit LL for reference (for debugging)
	{
		std::ofstream out( (base_path + ".ll").c_str() );
		llvm::raw_os_ostream rout( out );
		linker->getModule()->print( rout, nullptr );
	}
	
	std::string triple_name = llvm::sys::getDefaultTargetTriple();
	//DEBUG( "Triple", triple_name );
	llvm::Triple triple(llvm::Triple::normalize(triple_name));
	llvm::Target const *  target = llvm::TargetRegistry::lookupTarget( "x86-64", triple, err_str );
	STATE_CHECK( target, err_str );
	
	llvm::TargetOptions topts;
	llvm::TargetMachine * machine = target->createTargetMachine( triple.getTriple(),
		"" /*cpu*/, "" /*featurestr*/, topts, 
		gopts.pic ? llvm::Reloc::PIC_ : llvm::Reloc::Default, 
		llvm::CodeModel::Default,
		llvm::CodeGenOpt::Default );
	//machine->setMCUseLoc( true );
	//machine->setMCUseCFI( true );
	//machine->setMCUseDwarfDirectory( true );
	
	auto mcasm = machine->getMCAsmInfo();
	STATE_CHECK( mcasm->doesSupportExceptionHandling() );
	STATE_CHECK( mcasm->isExceptionHandlingDwarf() );
	
	llvm::tool_output_file out( (base_path + ".o").c_str(), err_str, llvm::raw_fd_ostream::F_Binary );
	STATE_CHECK( err_str.empty(), err_str );
	llvm::formatted_raw_ostream fout( out.os() );
	
	llvm::PassManager fpm;
	fpm.add( new llvm::DataLayout( *data_layout ) );
	bool ret = machine->addPassesToEmitFile( fpm, fout, llvm::TargetMachine::CGFT_ObjectFile );
	STATE_CHECK( !ret );
	fpm.run( *linker->getModule() );
	
	out.keep();
}

} //eon ir

