#include "ir/type.hpp"
#include "ir/function.hpp"

#include "platform/platform.hpp"

#include "ir/llvm/type_manager.hpp"

//NOTE: I'm not entirely clear on these defines
#define __STDC_LIMIT_MACROS
#define __STDC_CONSTANT_MACROS 

// LLVM: doesn't seem to care about conversion warnings
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#pragma GCC diagnostic ignored "-Wunused-parameter"
	#include "llvm/IR/DerivedTypes.h"
	#include "llvm/IR/Constants.h"
	#include "llvm/Support/TargetSelect.h"
	#include "llvm/IR/DataLayout.h"
#pragma GCC diagnostic pop

#define SHOW_DEBUG 0
#include "util/debug.hpp"
#include "ir/dump.hpp"

namespace ir_llvm
{

type_manager::type_manager( llvm::LLVMContext & context, llvm::DataLayout & data_layout )
	: context( context )
	, data_layout( data_layout )
{
	generic_ptr_ = llvm::PointerType::get( llvm::Type::getInt8Ty(context), 0 );
	void_ = llvm::Type::getVoidTy(context);
	
	//construct destructor function type
	std::vector<llvm::Type*> params;
	params.push_back( generic_ptr_ );
	destructor_ = llvm::FunctionType::get( void_, params, false/*vararg*/ );
	
	//extype_type
	extype_type_ = llvm::StructType::create( context, "_extype_type" );
	std::vector<llvm::Type*> fields;
	fields.push_back( f_natural_int() );
	extype_type_->setBody( fields );
}

bool type_manager::type_ref::operator<( type_ref const & o ) const
{
	if( ctx != o.ctx )
		return ctx < o.ctx;
	if( shared != o.shared )
		return shared < o.shared;
		
	return *ir_type < *o.ir_type;
}

llvm::IntegerType * type_manager::f_intx( unsigned bits )
{
	return llvm::IntegerType::get(context, bits);
}

llvm::IntegerType * type_manager::f_abi_char()
{
	return llvm::IntegerType::get(context, platform::abi_char_size );
}

llvm::IntegerType *  type_manager::main_argc() {
	return f_intx( platform::abi_int_size );
}

llvm::PointerType * type_manager::main_argv() {
	return ptr_to( ptr_to( f_abi_char() ) );
}

llvm::Type * type_manager::f_float( unsigned bits )
{
	//TODO: float/double should be matched to platform sizes (though the same on all known platforms)
	switch( bits ) {
		case 32: return llvm::Type::getFloatTy(context);
		case 64: return llvm::Type::getDoubleTy(context);
		case 80: return llvm::Type::getX86_FP80Ty(context);
		case 128: return llvm::Type::getFP128Ty(context);
	}
	STATE_FAIL( "unsupported float bitsize" );
}

llvm::IntegerType * type_manager::f_natural_int()
{
	return f_intx( platform::natural_integer_size );
}

llvm::PointerType * type_manager::ptr_to( llvm::Type * t ) 
{
	return llvm::PointerType::get( t, 0 ); 
}

size_t type_manager::alloc_size( llvm::Type * t )
{ 
	return data_layout.getTypeAllocSize( t ); 
}

size_t type_manager::max_align()
{
	return platform::max_align;
}

type_manager::type_entry type_manager::value( ir::type const & t )
{
	auto tmh = type_map_helper( t );
	if( tmh.has() )
		return tmh.get();
		
	type_entry ret;
	
	switch( t.fundamental )
	{
		case ir::type::f_integer:
		case ir::type::f_binary: //LLVM: has no distinction between signed/unsigned integers (for storage, for ops it does)
			ret.ll_type = f_intx( t.trivial_bitsize );
			ret.direct_return = true; //TODO: what about very large integers?
			break;
			
		case ir::type::f_float:
			ret.ll_type = f_float( t.trivial_bitsize );
			ret.direct_return = true;
			break;
			
		case ir::type::f_boolean:
			ret.ll_type = llvm::Type::getInt1Ty( context );
			ret.direct_return = true;
			break;
			
		case ir::type::f_tuple:
		{
			auto st = llvm::StructType::create( context, t.name.c_str() );
			ret.ll_type = st;
			//do now to allow recursion
			tmh.set(ret);
			
			std::vector<llvm::Type*> fields;
			for( unsigned i=0; i < t.sub.size(); ++i )
				fields.push_back( value( *t.sub[i] ).ll_type );
			bool packed = false;
			switch( t.pack ) {
				case ir::type::p_normal: break;
				case ir::type::p_abi: break;
				case ir::type::p_byte: packed = true; break;
			}
			st->setBody( fields, packed );
			break;
		}
		
		case ir::type::f_optional:
		{
			STATE_CHECK( t.sub.size() == 2 );
			auto st = llvm::StructType::create( context, t.name.c_str() );
			ret.ll_type = st;
			tmh.set(ret);
			
			std::vector<llvm::Type*> fields;
			fields.push_back( value( *t.sub[0] ).ll_type );
			fields.push_back( value( *t.sub[1] ).ll_type );
			st->setBody( fields );
			break;
		}

		case ir::type::f_function_ptr: {
			ret.direct_return = true;
			switch( t.access ) {
				case ir::type::a_raw: {
					auto fr = true_function( t );
					ret.ll_type = llvm::PointerType::get( fr.ll_func_type, 0 );
					break;
				}
				case ir::type::a_unbound: {
					auto fr = true_function( t );
					ret.ll_type = llvm::PointerType::get( fr.ll_func_type, 0 );
					break;
				}
				case ir::type::a_bound: {
					ret.ll_type = llvm::PointerType::get( true_ctx_function( t ).ll_type, 0 );
					break;
				}
			}
			break;
		}
		
		case ir::type::f_raw_pointer:
		{
			STATE_CHECK( t.sub.size() == 1 );
			auto pt = value( *t.sub[0] ).ll_type;
			ret.ll_type = llvm::PointerType::get( pt, 0 );
			ret.direct_return = true;
			break;
		}
		
		case ir::type::f_shared_pointer:
		{
			//LLVM: can't create an empty body pointer like for Struct, but it appears it doesn't care about
			//multiple equivalent pointers (or it manages them itself)
			auto shared = true_shared( t ).ll_type;
			ret.ll_type = llvm::PointerType::get( shared, 0 );
			ret.direct_return = true;
			break;
		}
			
		case ir::type::f_array:
			ret.direct_return = true;
			ret.ll_type = generic_ptr();
			break;
			
		default:
			STATE_CHECK( false );
	}
	
	STATE_CHECK( ret.ll_type );
	tmh.set(ret);
	return ret;
}

type_manager::type_entry type_manager::true_shared( ir::type const & t )
{
	auto tmh = type_map_helper( t );
	tmh.tr.shared = true;
	if( tmh.has() )
		return tmh.get();
		
	auto st = llvm::StructType::create( context );

	type_entry te;
	te.ll_type = st;
	te.direct_return = false;
	tmh.set(te);

	STATE_CHECK( t.sub.size() == 1 );
	STATE_CHECK( t.fundamental == ir::type::f_shared_pointer );
	std::vector<llvm::Type*> fields(sf_size);
	fields[sf_ref_count] = f_natural_int();
	fields[sf_destructor] = ptr_to( destructor() );
	fields[sf_data] = value( *t.sub[0] ).ll_type;
	st->setBody( fields );
	
	return te;
}

type_manager::func_type_entry type_manager::true_function( ir::type const & t )
{
	STATE_CHECK( t.fundamental == ir::type::f_function_ptr );

	//DEBUG( "TRUE_FUNCTION", ir::dump::get_type( &t ), "ACCESS", t.access );
	auto tmh = func_type_map_helper( t );
	if( tmh.has() )
		return tmh.get();
	//LLVM: offers no empty-body FunctionType, we have to hope it accepts duplicates
	
	func_type_entry ret;
	
	//collapse returns as possible
	if( t.sub.size() == 0 ) {
		ret.ll_ret_type = type_void();
		ret.is_direct_return = true;
	} else if( t.sub.size() == 1 ) {
		auto rt = value( *t.sub[0] );
		ret.ll_ret_type = rt.ll_type;
		ret.is_direct_return = rt.direct_return;
	} else {
		std::vector<llvm::Type*> fields;
		for( size_t i=0; i < t.sub.size(); ++i )
			fields.push_back( value( *t.sub[i] ).ll_type );
			
		ret.ll_ret_type = llvm::StructType::create( context, fields );
		ret.is_direct_return = false; //LLVM:FEATURE: there are actually some structures which can be directly returned (platform/ABI dependent)
	}

	std::vector<llvm::Type*> params;
	
	//if we can't return the return type we put a pointer to it as the first parameter
	llvm::Type * use_ret_type = ret.ll_ret_type;
	if( !ret.is_direct_return ) {
		params.push_back( llvm::PointerType::get( ret.ll_ret_type, 0 ) );
		use_ret_type = type_void();
	}
	
	//do we need a context parameter (comes after returns value)
	if( t.access != ir::type::a_raw ) {
		params.push_back( generic_ptr() );
		ret.is_context = true;
	} else {
		ret.is_context = false;
	}
		
	for( size_t i=0; i < t.params.size(); ++i )
		params.push_back( value( *t.params[i] ).ll_type );
		
	ret.ll_func_type = llvm::FunctionType::get( use_ret_type, params, false/*vararg*/ );
	tmh.set(ret);
	return ret;
}

/**
	A context function is a shared object with this format which contains a ref_count and
	a function pointer. It also includes the capture data as part of the structure, stored
	at a common offset 'cf_ctx_offset' from the base of the structure.
	
	This makes this type an interface type since it will be the same for any equivalent
	function signature irrespective of the context data.
*/
type_manager::type_entry type_manager::true_ctx_function( ir::type const & t )
{
	auto tmh = type_map_helper( t );
	tmh.tr.ctx = true;
	if( tmh.has() )
		return tmh.get();
	
	auto st = llvm::StructType::create( context );

	type_entry te;
	te.ll_type = st;
	te.direct_return = false;
	tmh.set( te );
	
	std::vector<llvm::Type*> fields(cf_size);
	//for the shared object count
	fields[cf_ref_count] = f_natural_int();
	fields[cf_destructor] = ptr_to( destructor() );
	auto fr = true_function( t );
	fields[cf_func_ptr] = ptr_to( fr.ll_func_type );
	fields[cf_ctx_ptr] = generic_ptr();
	//TODO: This shouldn't be a field at all, just a location of the context data
	fields[cf_ctx_offset] = generic_ptr();
	st->setBody( fields );

	return te;
}

} //ir_llvm