#include "ir/type_manager_leaf.hpp"
#include "ir/compiler_leaf.hpp"
#include "ir/type.hpp"

#include "lang/var_decl.hpp"
#include "lang/intr_type.hpp"
#include "lang/variable_context.hpp"
#include "lang/scope.hpp"

#define SHOW_DEBUG 0
#include "util/debug.hpp"

namespace ir
{

type_manager_leaf::type_manager_leaf( compiler_leaf & comp )
	: comp( comp )
{
}

bool type_manager_leaf::leaf_type_key::operator<( leaf_type_key const & o ) const
{
	return *leaf_type < *o.leaf_type;
}

shared_ptr<type const> type_manager_leaf::get_fun( shared_ptr<leaf::intr_type const> fun )
{
	//a simple cache of types we've already processed
	auto cached = known_fun_types.if_get( fun->cref );
	if( cached )
		return *cached;

	//check for equivalent type
	leaf_type_key ltk;
	ltk.leaf_type = fun;
	auto lit = leaf_type_cache.find( ltk );
	if( lit != leaf_type_cache.end() ) {
		known_fun_types.set( fun->cref, lit->ir_type );
		return lit->ir_type;
	}
	
	//the type object must be cached prior to recursion
	auto register_out = [&]( shared_ptr<type const> rt ) {
		known_fun_types.set( fun->cref, rt );
		//track for equivalent types
		ltk.ir_type = rt;
		leaf_type_cache.insert( ltk );
	};
	
	//does standard creation/registration
	auto make_out = [&]( type::fundamental_t f ) {
		auto out = make_shared<type>(f);
		if( fun->distinct_name ) {
			STATE_CHECK( fun->name_scope );
			out->name = fun->name_scope->qualify_name( *fun->distinct_name,platform::export_sep );
		}
		register_out(out);
		return out;
	};
	
	switch( fun->get_fun_class() )
	{
		case leaf::intr_type::f_unknown:
		case leaf::intr_type::f_rational:
			PRE_CHECK( false );
			
		case leaf::intr_type::f_function: {
			auto out = make_out(type::f_function_ptr);
			auto & t = fun->cast<leaf::intr_type_function>();
			switch( t.access ) {
				case leaf::intr_type_function::a_bound:
					STATE_CHECK( !t.context );
					out->access = type::a_bound;
					break;
				case leaf::intr_type_function::a_raw:
					STATE_CHECK( !t.context );
					out->access = type::a_raw;
					break;
				case leaf::intr_type_function::a_unbound:
					STATE_CHECK( t.context );
					out->access = type::a_unbound;
					break;
			}
			auto & params = *t.params;
			auto & returns = *t.returns;
			
			for( size_t i=0; i < params.sub.size(); ++i ) {
				auto p = get( params.sub[i].type );
				out->params.push_back( p );
			}
			for( size_t i=0; i < returns.sub.size(); ++i ) {
				auto p = get( returns.sub[i].type );
				out->sub.push_back( p );
			}
			if( t.context ) 
				out->context = get( leaf::extr_type_ref(t.context) );
			
			return out;
		}
		
		case leaf::intr_type::f_integer: {
			auto & t = fun->cast<leaf::intr_type_integer>();
			if( fun->distinct_name ) {
				auto out = make_out(type::f_integer);
				out->trivial_bitsize = t.data_bitsize;
				return out;
			} else {
				//non-distinct must always be the same "type"
				auto out = get_integer(t.data_bitsize);
				register_out(out);
				return out;
			}
		}
		
		case leaf::intr_type::f_binary: {
			auto & t = fun->cast<leaf::intr_type_binary>();
			if( fun->distinct_name ) {
				auto out = make_out(type::f_binary);
				out->trivial_bitsize = t.data_bitsize;
				return out;
			} else {
				//non-distinct must always be the same "type" instance
				auto out = get_binary(t.data_bitsize);
				register_out(out);
				return out;
			}
		}			
		
		case leaf::intr_type::f_float: {
			auto out = make_out(type::f_float);
			auto & t = fun->cast<leaf::intr_type_float>();
			out->trivial_bitsize = t.data_bitsize;
			return out;
		}
		
		case leaf::intr_type::f_boolean: {
			//TODO: must use get_boolean!
			auto out = make_out(type::f_boolean);
			return out;
		}
		
		case leaf::intr_type::f_value_list: {
			auto out = make_out(type::f_tuple);
			auto & t = fun->cast<leaf::intr_type_value_list>();
			//TODO: size == 0 should probably be a special void type (it's not an object like an empty tuple is)
			if( t.sub.size() == 1 ) {
				out = get( t.sub[0].type )->clone();
				break;
			}
			
			out->pack = type::p_normal;
			for( size_t i=0; i < t.sub.size(); ++i )
				out->sub.push_back( get( t.sub[i].type ) );
		
			return out;
		}

		case leaf::intr_type::f_tuple: {
			auto out = make_out(type::f_tuple);
			auto & t = fun->cast<leaf::intr_type_tuple>();
			for( size_t i=0; i < t.sub.size(); ++i ) {
				auto p = get( t.sub[i].type );
				out->sub.push_back( p );
			}
			switch( t.pack ) {
				case leaf::intr_type_tuple::p_normal:
					out->pack = type::p_normal;
					break;
				case leaf::intr_type_tuple::p_abi:
					out->pack = type::p_abi;
					break;
				case leaf::intr_type_tuple::p_byte:
					out->pack = type::p_byte;
					break;
			}
			
			return out;
		}
		
		case leaf::intr_type::f_custom: {
			auto out = make_out(type::f_tuple);
			auto & t = fun->cast<leaf::intr_type_custom>();
			out->pack = type::p_normal;
			auto local = t.context->get_local_decls();
			size_t index = 0;
			for( auto & sref : local ) {
				sref.decl->xref = ++index; //+1 from actual index
				DEBUG( "SET XREF", sref.decl->symbol, sref.decl->xref );
				auto p = get( sref.decl->type );
				out->sub.push_back( p );
			}
			
			if( t.destroy ) {
				out->destroy = comp.obtain_symbol_value_defn( *t.destroy );
			}
			return out;
		}
			
		case leaf::intr_type::f_char: {
			auto out = get_binary(platform::char_size);
			register_out(out);
			return out;
		}
			
		case leaf::intr_type::f_array: {
			auto & t = fun->cast<leaf::intr_type_array>();
			if( t.raw ) {
				auto out = make_out(type::f_raw_pointer);
				out->sub.push_back( get( t.sub ) );
				return out;
			} else {
				auto out = make_out(type::f_array);
				out->sub.push_back( get( t.sub ) );
				return out;
			}
		}
		
		case leaf::intr_type::f_module: {
			STATE_FAIL( "unexpected" );
		}
	}

	STATE_FAIL();
}

shared_ptr<type const> type_manager_leaf::get( leaf::extr_type_ref const & gt )
{
	type_key tk;
	tk.cref = gt.get_fun()->cref;
	tk.shared = gt.is_shared();
	tk.optional = gt.is_optional();
	tk.lvalue = gt.is_lvalue();
	tk.value_ptr = gt.is_value_ptr();
	
	//check cache for this type (cref check is a shortcut)
	if( tk.cref ) {
		auto it = known_type_refs.find(tk);
		if( it != known_type_refs.end() ) {
			return it->cached;
		}
	}
	
	auto out = get_fun( gt.get_fun() );
	tk.cref = gt.get_fun()->cref;
	STATE_CHECK( tk.cref );
	
	if( gt.is_shared() ) {
		auto p = make_shared<type>( type::f_shared_pointer );
		p->sub.push_back( out );
		out = p;
	}
	
	if( gt.is_optional() ) {
		auto p = make_shared<type>( type::f_optional );
		p->sub.push_back( get_boolean() );
		p->sub.push_back( out );
		out = p;
	}
	
	if( gt.is_lvalue() ) {
		auto p = make_shared<type>( type::f_raw_pointer );
		p->sub.push_back( out );
		out = p;
	}
	
	if( gt.is_value_ptr() ) {
		auto p = make_shared<type>( type::f_raw_pointer );
		p->sub.push_back( out );
		out = p;
	}

	tk.cached = out;
	known_type_refs.insert(tk);
	return out;
}

} //eon ir
