#include "lang/class_typer.hpp"
#include "lang/language.hpp"
#include "lang/statement_importer.hpp"

#define SHOW_DEBUG 1
#include "util/debug.hpp"

namespace leaf
{

void check( shared_ptr<type_statement> ts ) {
	STMT_CHECK( *ts, !ts->spec_type );
	STMT_CHECK( *ts, ts->class_block );
}

class_typer::class_typer( shared_ptr<type_statement> ts, typer::context decl_context )
	: source_ts( (check(ts),ts) )
	, decl_context( decl_context )
	, class_context( decl_context.get_owner(), source_ts->class_block->symbols )
{
	has_dtor = false;
	is_ctor_complete = false;
	class_block_done = false;
	
	class_type = make_shared<intr_type_custom>();
	class_type->distinct_name = source_ts->name;
	class_type->distinct = true;
	class_type->context = source_ts->class_block->symbols;
	class_type->local_complete = false;
	
	auto init_block = make_shared<statement_block>(scope_style_t::imperative);
	init_block->symbols->set_parent( source_ts->class_block->symbols );
	init_block->symbols->mark_function_bypass();

	source_ts->class_block->symbols->mark_class_root( class_type, init_block );
	source_ts->class_block->symbols->set_parent( decl_context.sc );
	source_ts->type = extr_type_ref( class_type );
	decl_context.sc->add_fun_type( class_type );
	
	source_ts->init_block = make_shared<multi_statement>();
	class_context.scan_symbols( source_ts->class_block );
}

bool class_typer::process( typer* ) {
	//DEBUG("CLASS-BUILD", source_ts->name);
	if( !class_block_done ) {
		class_block_done = class_context.proc_statement_block( source_ts->class_block, class_context );
	}

	if( !has_dtor ) {
		auto dtor_func = build_dtor();
		if( dtor_func ) {
			auto ddecl = make_shared<var_decl>();
			ddecl->form = var_decl::f_single;
			ddecl->symbol = "destroy_" + source_ts->name;
			ddecl->type = dtor_func->type;
			ddecl->is_export = true;
			class_type->destroy = decl_context.sc->add_single( ddecl );
			decl_context.add_export( ddecl );
			
			auto vs = make_shared<init_statement>();
			vs->complete = true;
			vs->sref = *class_type->destroy;
			vs->value = dtor_func;
			vs->init_refer_value = false;
			vs->undefined = false;
			source_ts->init_block->list.push_back( vs );
		}
		has_dtor = true;
	}
	
	//introduce the ctor into block
	if( !ctor_func ) {
		ctor_func = build_ctor();
		if( ctor_func ) {
			auto cdecl = make_shared<var_decl>();
			cdecl->form = var_decl::f_single;
			cdecl->symbol = "construct_" + source_ts->name;
			cdecl->type = ctor_func->type;
			cdecl->is_export = true; //TODO: should be done only as required
			auto sref = decl_context.sc->add_single( cdecl );
			decl_context.add_export( cdecl );
			
			auto vs = make_shared<init_statement>();
			vs->complete = true;
			vs->sref = sref;
			vs->value = ctor_func;
			vs->init_refer_value = false;
			vs->undefined = false;
			source_ts->init_block->list.push_back( vs );
		}
	}
	
	if( !is_ctor_complete && ctor_func ) {
		is_ctor_complete = complete_ctor();
	}

	if( class_block_done && is_ctor_complete && has_dtor ) {
		class_type->local_complete = true;
		return true;
	} 
	return false;
}

shared_ptr<expr_funcdefn> class_typer::build_ctor() {
	//check for user constructor
	auto cvars = source_ts->class_block->symbols->get_vars( "construct" );
	STMT_CHECK( *source_ts, cvars.size() <= 1 );

	if( cvars.size() == 0 ) {
		auto ctor_func = make_shared<expr_funcdefn>();
		ctor_func->block = class_context.sc->get_init_block();
		auto cf_type = make_shared<intr_type_function>();
		cf_type->params = make_shared<intr_type_value_list>();
		cf_type->returns = make_shared<intr_type_value_list>();
		cf_type->context = class_type;
		cf_type->access = intr_type_function::a_unbound;
		ctor_func->type = extr_type_ref(cf_type);
		ctor_require_import = false;
		return ctor_func;
	}
	
	STATE_CHECK( cvars.size() == 1 );
	STMT_CHECK( *source_ts, cvars[0].decl->source_statement->statement_form() == sf_var_statement );
	auto vs = std::static_pointer_cast<var_statement>( cvars[0].decl->source_statement );
	
	STMT_CHECK( *source_ts, vs->value->form == expression::f_funcdefn );
	auto fd = std::static_pointer_cast<expr_funcdefn>(vs->value);
	//until it's complete we don't know what the ctor looks like
	if( !vs->value->is_complete() ) {
		return shared_ptr<expr_funcdefn>();
	}
	ctor_require_import = true;
	return fd;
}

bool class_typer::complete_ctor() {
	//DEBUG("COMPLETE_CTOR", source_ts->name );
	//so long as class_block is incomplete there might be more items to be added to init_block
	if( !class_block_done ) {
		//DEBUG("!cblock_done");
		return false;
	}
	
	//the init_block may not be finished yet
	auto init_block = class_context.sc->get_init_block();
	if( !class_context.proc_statement_block( init_block, class_context ) ) {	
		//DEBUG("!iblock_done" );
		return false;
	}
	
	if( ctor_require_import ) {
		auto & block = *ctor_func->block;
		statement_importer importer( block.symbols );
		std::vector<shared_ptr<statement>> tlist;
		for( auto & is : class_context.sc->get_init_block()->list ) {
			tlist.push_back( importer.apply_statement( is, class_context.sc ) );
		}
		block.list.insert( block.list.begin(), tlist.begin(), tlist.end() );
	}
	return true;
}

shared_ptr<expr_funcdefn> class_typer::build_dtor() {
	//check for user destructor
	auto dvars = source_ts->class_block->symbols->get_vars( "destroy" );
	STMT_CHECK( *source_ts, dvars.size() <= 1 );
	if( dvars.empty() )
		return shared_ptr<expr_funcdefn>();

	STMT_CHECK( *source_ts, dvars[0].decl->source_statement->statement_form() == sf_var_statement );
	auto vs = std::static_pointer_cast<var_statement>( dvars[0].decl->source_statement );
	
	STMT_CHECK( *source_ts, vs->value->form == expression::f_funcdefn );
	auto fd = std::static_pointer_cast<expr_funcdefn>(vs->value);
		
	return fd;
}

} //eon leaf