#include "lang/conversion_applicator.hpp"
#include "lang/expression.hpp"

#define SHOW_DEBUG 0
#include "util/debug.hpp"
#include "lang/dump.hpp"

namespace leaf
{

shared_ptr<expression> conversion_applicator::apply( type_converter::convert_t const & cost, 
	shared_ptr<expression> expr )
{
	STMT_CHECK( *expr, cost.okay(), cost.explain_error() + " " 
		+ dump::get_extr_type( expr->type ) );
	if( cost.conversions )
		return apply( *cost.conversions, expr );
	return expr;
}

shared_ptr<expression> conversion_applicator::apply_single( cvt_expression & cvt, shared_ptr<expression> expr )
{
	DEBUG( "CVT", typeid(*expr).name(), dump::get_expression(*expr) );
	//could likely work for more than fixate, but those would be optimizations, fixation is required
	if( cvt.form == cvt_expression::c_fixate ) {
		
		size_t inner = expr->num_inner_trans();
		if( inner ) {
			for( size_t i=0; i < inner; ++i ) {
				expr->set_inner_trans( i, apply_single(
					cvt, expr->get_inner_trans( i ) ) );
			}
			expr->construct_result_type();
			return expr;
		} else {
			DEBUG( "UNKNOWN CVT_OPTIONAL", typeid(*expr).name() );
		}
	}
	
	switch( cvt.form )
	{
		case cvt_expression::c_drop_lvalue:
			return drop_lvalue( expr );
		case cvt_expression::c_addr_of: {
			auto addr = make_shared<expr_transform>();
			addr->transform = expr_transform::t_addr_of;
			addr->param = expr;
			addr->construct_result_type();
			return addr;
		}
		case cvt_expression::c_deref: {
			//STATE_CHECK( expr->type->shared );
			auto deref = make_shared<expr_transform>();
			deref->transform = expr_transform::t_deref;
			deref->param = expr;
			deref->construct_result_type();
			return deref;
		}
		case cvt_expression::c_promote: 
		case cvt_expression::c_loss: {
			shared_ptr<intr_type_fixed_number> dst;
			switch( cvt.to_fun ) {
				case cvt_expression::p_float:
					dst = make_shared<intr_type_float>();
					break;
				case cvt_expression::p_integer:
					dst = make_shared<intr_type_integer>();
					break;
				case cvt_expression::p_binary:
					dst = make_shared<intr_type_binary>();
					break;
			}
			dst->data_bitsize = cvt.to_bitsize;
			
			auto promote = make_shared<expr_transform>();
			promote->type = extr_type_ref(dst);
			promote->param = expr;
			promote->transform = cvt.form == cvt_expression::c_promote ?
				expr_transform::t_promote : expr_transform::t_loss;
			return promote;
		}
		case cvt_expression::c_fixate: {
			//fixation results in the removal of the rational constant and conversion
			//to an integer or float value (which is still stored in the "number" so
			//we just need to assign a new type here
			STATE_CHECK( expr->form == expression::f_const_number );
			auto & cn = static_cast<expr_const_number &>(*expr);

			shared_ptr<intr_type_fixed_number> ft;
			switch( cvt.to_fun ) {
				case cvt_expression::p_float:
					ft = make_shared<intr_type_float>();
					cn.src_form = expr_const_number::sf_float;
					break;
				case cvt_expression::p_integer:
					ft = make_shared<intr_type_integer>();
					cn.src_form = expr_const_number::sf_integer;
					break;
				case cvt_expression::p_binary:
					ft = make_shared<intr_type_binary>();
					cn.src_form = expr_const_number::sf_binary;
					break;
			}
			ft->data_bitsize = cvt.to_bitsize;
			expr->type = extr_type_ref(ft);
			return expr;
		}
		case cvt_expression::c_bind_raw_func: {
			//STATE_CHECK( expr->type->shared );
			auto wrap = make_shared<expr_transform>();
			wrap->transform = expr_transform::t_bind_raw_func;
			wrap->param = expr;
			wrap->construct_result_type();			
			return wrap;
		}
		case cvt_expression::c_bind_func: {
			return bind_func( expr );
		}
		case cvt_expression::c_collapse:
			cvt.field = 0;
			//fall-through
		case cvt_expression::c_field: {
			//this is just an optimization now, conversions of function returns force us to make value_lists
			//real things at the compiler level (they become tuples)
			if( expr->form == expression::f_value_list
				|| expr->form == expression::f_tuple ) {
				auto & evl = static_cast<expr_with_sub&>(*expr);
				STMT_CHECK( *expr, cvt.field < evl.sub.size() );
				expr = evl.sub[cvt.field];
			} else {
				auto sub = make_shared<expr_field>();
				sub->base = expr;
				sub->field_index = cvt.field;
				
				intr_type_compound const * vtc = expr->type.if_fun_cast<intr_type_value_list>();
				if( !vtc ) {
					vtc = expr->type.if_fun_cast<intr_type_tuple>();
					STMT_CHECK( *expr, vtc );
				}
				sub->type = vtc->sub[cvt.field].type;
				expr = sub;
			}
			return expr;
		}
		case cvt_expression::c_direct:
			//nothing
			return expr;
		case cvt_expression::c_pack:
		case cvt_expression::c_pack_value: {
			shared_ptr<expr_with_sub> pack;
			shared_ptr<intr_type_compound> pack_type;
			if( cvt.form == cvt_expression::c_pack_value ) {
				pack = make_shared<expr_value_list>();
				auto pt = make_shared<intr_type_value_list>();
				pack_type = pt;
			} else {
				pack = make_shared<expr_tuple>();
				auto pt = make_shared<intr_type_tuple>();
				pt->pack = cvt.pack;
				pack_type = pt;
			}
			pack->type = extr_type_ref(pack_type);
				
			//must avoid double evaluation if the expression may have side-effects
			//the shallow flag assumes the pack's are strict, no out-of-order items!
			expr = create_shared( expr, true );
			
			for( size_t i=0; i < cvt.sub.size(); ++i ) {
				auto sub = apply( *cvt.sub[i].cvt, expr );
				pack_type->push_sub( cvt.sub[i].name, sub->type );
				pack->sub.push_back( sub );
			}
			return pack;
		}
		case cvt_expression::c_optional: {
			STATE_CHECK( !expr->type.is_optional() );
			auto opt = make_shared<expr_transform>();
			opt->transform = expr_transform::t_make_optional;
			opt->param = expr;
			opt->construct_result_type();
			return opt;
		}
		case cvt_expression::c_deref_optional: {
			STATE_CHECK( expr->type.is_optional() );
			auto opt = make_shared<expr_transform>();
			opt->transform = expr_transform::t_deref_optional;
			opt->param = expr;
			opt->construct_result_type();
			return opt;
		}
		case cvt_expression::c_unset_optional: {
			auto d = make_shared<expr_default>();
			d->type = cvt.target_type;
			return d;
		}
		case cvt_expression::c_set_distinct: {
			//TODO: see the cvt_expression todo, this should be only a fun_class
			expr->type.set_fun( cvt.target_type.get_fun() );
			return expr;
		}
	}			
	
	STATE_FAIL("corrupt conversion");
}

shared_ptr<expression> conversion_applicator::apply( cvt_expression & cvt_, shared_ptr<expression> expr )
{
	auto cvt = &cvt_;
	do
	{
		expr = apply_single( *cvt, expr );
		cvt = cvt->then.get();
	} while( cvt );

	return expr;
}

/**
	Conver the expression so it can be used as a shared expression. This shares only the
	parts that actually need to be shared.
*/
shared_ptr<expression> conversion_applicator::create_shared( shared_ptr<expression> expr,
	bool shallow_list )
{
	if( expr->form == expression::f_value_list
		|| expr->form == expression::f_tuple ) {
		
		//in shallow cases we know for sure that we don't have to share the individual parts, this
		//is required for proper fixation in function calls
		if( shallow_list )
			return expr;
		
		STATE_FAIL( "unexpected" ); // will this ever be needed...?
		//this greatly helps reduce wasted temporaries by allowing field and subscript to 
		//bypass a packed structure.
		auto & evl = static_cast<expr_with_sub&>(*expr);
		for( size_t i=0; i < evl.sub.size(); ++i ) {
			evl.sub[i] = create_shared( evl.sub[i] );
		}
	} else {
		//OPT: perhaps also just complex/large expressions should always be shared
		if( expr->may_have_side_effects() ) {
			auto shared_expr = make_shared<expr_transform>();
			shared_expr->transform = expr_transform::t_shared_expr;
			shared_expr->param = expr;
			shared_expr->construct_result_type();
			expr = shared_expr;
		}
	}
	
	return expr;
}

/**
	This function will modify the provided "expr", it does not create a clone or other
	derived expression.
*/
shared_ptr<expression> conversion_applicator::drop_lvalue( shared_ptr<expression> expr )
{
	if( expr->form == expression::f_symbol ) {
		STMT_CHECK( *expr, expr->type.is_lvalue() );
		auto cl = expr->type;
		cl.set_lvalue( false );
		expr->type = cl;
		return expr;
	} else if( expr->form == expression::f_value_list ) {
		auto & vl = static_cast<expr_value_list&>(*expr);
		for( auto & s : vl.sub )
			s = drop_lvalue(s);
		return expr;
	} else if( expr->form == expression::f_field ) {
		auto & ef = static_cast<expr_field&>(*expr);
		ef.base = drop_lvalue( ef.base );
		auto cl = expr->type;
		cl.set_lvalue( false );
		expr->type = cl;
		return expr;
	} else if( expr->form == expression::f_transform ) {
	#if 1
		//TODO: this is perhaps an optimization. There is one test which depends on it, but it correlates
		//to a TODO in the compiler for a context call. Something suspicious
		auto & et = static_cast<expr_transform&>(*expr);
		//these types can only result in lvalue if the inner part is an lvalue
		if( et.transform == expr_transform::t_deref 
			|| et.transform == expr_transform::t_deref_optional ) {
			STMT_CHECK( *expr, et.param->type.is_lvalue() );
			et.param = drop_lvalue( et.param );
			et.type.set_lvalue( false );
			return expr;
		}
	#endif
	}
	
	auto deref = make_shared<expr_transform>();
	deref->transform = expr_transform::t_drop_lvalue;
	deref->param = expr;
	deref->construct_result_type();
	return deref;
}

shared_ptr<expression> conversion_applicator::bind_func( shared_ptr<expression> expr )
{
	auto wrap = make_shared<expr_transform>();
	wrap->transform = expr_transform::t_bind_func;
	wrap->param = expr;
	wrap->construct_result_type();	
	return wrap;
}

} //eon leaf