#include "lang/dump.hpp"
#include "lang/language.hpp"
#include "util/unicode.hpp"

#include <sstream>

#include <boost/algorithm/string.hpp>


namespace leaf
{

dump::dump( std::ostream & out )
	: leaf_util::dump_base( out )
{
	do_validate = true;
	show_full_type = false;
	show_all_types = false;
	show_spec_type = false;
	hide_var_spec_no_fundamental = false;
	keep_variant = false;
	show_spec_implicit = false;
}

dump::dump()
	: leaf_util::dump_base()
{
	do_validate = true;
	show_full_type = false;
	show_all_types = false;
	show_spec_type = false;
	hide_var_spec_no_fundamental = false;
	keep_variant = false;
	show_spec_implicit = false;
}

#define CHECK_NULL( v, msg ) if( !v ) { error( "null "  msg ); return; }

template<typename T>
void dump::validate( T const & o )
{
	if( !do_validate )
		return;
		
	try
	{
		o.validate();
	}
	catch( recoverable_error & re )
	{
		std::string err = boost::current_exception_diagnostic_information();
		boost::trim( err );
		boost::replace_all( err, "\n", std::string("\n") + indent_str() + "// " );
		out << "\e[31m";
		out << indent_str() << "// " << err << std::endl;
		out << "\e[0m";
	}
}

/*static*/ std::string dump::get_extr_type( extr_type_ref const & vt )
{
	std::ostringstream out;
	dump d( out );
	d.dump_intr_type( vt );
	return out.str();
}

/*static*/ std::string dump::get_type_spec( type_spec const & vt )
{
	std::ostringstream out;
	dump d( out );
	d.dump_type_spec( &vt );
	return out.str();
}

/*static*/ std::string dump::get_expression( expression const & expr )
{
	std::ostringstream out;
	dump d( out );
	d.dump_expression( &expr );
	return out.str();
}

void dump::show_type( extr_type_ref type, type_spec const * spec )
{
	if( show_spec_type ) 
		dump_type_spec( spec );
	else
		dump_intr_type( type );
}

void dump::dump_expression( expression const * ex, bool isolated )
{
	if( isolated )
		validate( *ex );
		
	CHECK_NULL(ex, "expression");
	bool omit_type = false;
	bool list_close = false;

	//force value-list notation if type is included to prevent ambiguity
	if( show_all_types && (ex->type.has_fun()) &&
		(ex->form != expression::f_tuple && ex->form != expression::f_value_list
		&& ex->form != expression::f_funcdefn)
		&& !isolated )
	{
		out << "( ";
		list_close = true;
	}
	
	switch( ex->form )
	{
		case expression::f_symbol: {
			auto & es = static_cast<expr_symbol const&>(*ex);
			bool force_type;
			if( es.sref.decl ) {
				switch( es.sref.decl->form ) {
					case var_decl::f_param: out << "params."; break;
					case var_decl::f_return: out << "returns."; break;
					case var_decl::f_capture: out << "captures."; break;
					case var_decl::f_single: force_type = true; break;
					case var_decl::f_multi: break;
				}
			}
			if( force_type && ex->type.has_fun() ) {
				//for multi showing the type is helpful
				omit_type = true;
				out << "( ";
				out << es.symbol;
				out << " : ";
				dump_intr_type( ex->type );
				out << " )";
			} else {
				out << es.symbol;
			}
			break;
		}
		
		case expression::f_const_number: {
			auto & ecn = static_cast<expr_const_number const&>(*ex);
			switch( ecn.src_form ) {
				case expr_const_number::sf_rational: out << ecn.value.to_string(true); break;
				case expr_const_number::sf_integer: out << ecn.integer_value(); break;
				case expr_const_number::sf_binary: out << ecn.integer_value(); break; //TODO: hmm...
				case expr_const_number::sf_float: out << ecn.float_value(); break;
				case expr_const_number::sf_boolean: out << (ecn.boolean_value() ? "true" : "false"); break;
			}
			break;
		}

		case expression::f_text:
		{
			auto & es = static_cast<expr_text const&>(*ex);
			switch( es.what ) {
				case expr_text::w_string:
					out << '\"' << unicode::encode8(es.text) << '\"'; //TODO: proper format
					break;
				case expr_text::w_char:
					out << "\'" << unicode::encode8(es.text) << "\'";
					break;
			}
			break;
		}

		case expression::f_funccall:
		{
			auto & ef = static_cast<expr_funccall const&>(*ex);
			if( ef.context ) {
				dump_expression( ef.context.get(), false );
				out << "~";
			}
			dump_expression( ef.func.get(), false );
			out << " <- ";
			dump_expression( ef.params.get(), false );
			break;
		}

		case expression::f_tuple:
			dump_tuple( static_cast<expr_tuple const*>(ex), '[', ']', ',' );
			break;

		case expression::f_value_list:
			dump_tuple( static_cast<expr_value_list const*>(ex), '(', ')', ',' );
			break;
			
		case expression::f_funcdefn:
		{
			auto & ef = static_cast<expr_funcdefn const&>(*ex);
			omit_type = true;
			show_type( ex->type, ex->spec_type.get() );
			for( auto & vd : ef.param_defaults ) {
				out << endl() << indent_str() 
					<< "[" << vd.first << "] |= ";
				dump_expression( vd.second.get(), true );
			}
			out << " ";
			dump_statement_block( ef.block.get() );
			break;
		}
			
		case expression::f_switch:
		{
			auto & es = static_cast<expr_switch const&>(*ex);
			out << "select(" << endl();
			indent();
			for( auto & s : es.expr_list ) {
				validate(s);
				out << indent_str();
				if( s.cond ) {
					dump_expression( s.cond.get(), false );
					out << " ? ";
				}
				dump_expression( s.value.get(), false );
				out << "," << endl();
			}
			unindent();
			out << indent_str() << ")";
			break;
		}
			
		case expression::f_subscript:
		{
			auto & es = static_cast<expr_subscript const&>(*ex);
			dump_expression( es.base.get(), false );
			out << "#(";
			dump_expression( es.offset.get(), false );
			out << ")";
			break;
		}
			
		case expression::f_block:
			dump_statement_block( static_cast<expr_block const*>(ex)->block.get() );
			break;
			
		case expression::f_transform:
		{
			auto & et = static_cast<expr_transform const&>(*ex);
			switch( et.transform )
			{
				case expr_transform::t_deref:
					out << "⋆";
					dump_expression( et.param.get(), false );
					break;
			
				case expr_transform::t_drop_lvalue:
					out << "drop_lvalue( ";
					dump_expression( et.param.get(), false );
					out << ") ";
					break;
					
				case expr_transform::t_shared_expr:
					out << "shared_expr(";
					dump_expression( et.param.get(), false );
					out << ")" ;
					break;
					
				case expr_transform::t_promote:
					out << "↑";
					dump_expression( et.param.get(), false ); //TODO: force show types (important here!)
					break;
					
				case expr_transform::t_loss:
					out << "↓";
					dump_expression( et.param.get(), false ); //TODO: force show types (important here!)
					break;
					
				case expr_transform::t_make_optional:
					out << "optional(";
					dump_expression( et.param.get(), false );
					out << ")";
					break;
					
				case expr_transform::t_deref_optional:
					out << "∘";
					dump_expression( et.param.get(), false );
					break;
					
				case expr_transform::t_bind_raw_func:
					out << "bind_raw(";
					dump_expression( et.param.get(), false );
					out << ")";
					break;
					
				case expr_transform::t_bind_func:
					out << "bind(";
					dump_expression( et.param.get(), false );
					out << ")";
					break;
					
				case expr_transform::t_addr_of:
					out << "addr_of(";
					dump_expression( et.param.get(), false );
					out << ")";
					break;
					
				case expr_transform::t_code_val:
					out << "code_val(";
					dump_expression( et.param.get(), false );
					out << ")";
					break;
				
				case expr_transform::t_char_val:
					out << "char_val(";
					dump_expression( et.param.get(), false );
					out << ")";
					break;
			}
			break;
		}
		
		case expression::f_field: {
			auto & ef = static_cast<expr_field const&>(*ex);
			if( ef.bind ) {
				out << "bind(";
			}
			dump_expression( ef.base.get(), false );
			if( ef.field_index ) {
				out << "[" << *ef.field_index 
					<< "/*" << ef.field << "*/"
					<< "]";
			} else if( ef.sref ) {
				out << "." << (*ef.sref).decl->symbol;
			} else {
				out << "." << ef.field;
			}
			if( ef.bind ) {
				out << ")";
			}
			break;
		}
		
		case expression::f_special_field: {
			auto & sf = static_cast<expr_special_field const&>(*ex);
			switch( sf.field )
			{
				case expr_special_field::f_array_size:
					dump_expression( sf.base.get(), false );
					out << ".size";
					break;
				case expr_special_field::f_optional_set:
					out << "have(";
					dump_expression( sf.base.get(), false );
					out << ")";
					break;
			}
			break;
		}
		
		case expression::f_new_func: {
			auto & nf = static_cast<expr_new_func const&>(*ex);
			out << "new( ";
			show_type( nf.new_type, nf.new_spec_type.get() );
			out << " )";
			break;
		}
		
		case expression::f_default: {
			omit_type = true;
			out << "default(";
			dump_intr_type( ex->type );
			out << ")";
			break;
		}
		
		case expression::f_conditional: {
			auto & ec = static_cast<expr_conditional const&>(*ex);
			switch( ec.op ) {
				case expr_conditional::o_optional_or:
					dump_expression( ec.cond.get(), false );
					out << " | ";
					dump_expression( ec.b.get(), false );
					break;
				case expr_conditional::o_optional_cond:
					dump_expression( ec.cond.get(), false );
					out << " ? ";
					dump_expression( ec.a.get(), false );
					break;
			}
			break;
		}
		
		case expression::f_binary: {
			auto & eb = static_cast<expr_binary const&>(*ex);
			switch( eb.op ) {
				#define OP( part ) \
					case expr_binary::o_##part: \
						out << "_" #part; \
						break;
				OP(add)
				OP(sub)
				OP(mul)
				OP(div)
				
				OP(mod)
				
				OP(bit_shl)
				OP(bit_shr)
				OP(bit_or)
				OP(bit_and)
				OP(bit_xor)
				
				OP(eq)
				OP(neq)
				OP(lt)
				OP(lte)
				OP(gt)
				OP(gte)
				#undef OP
			}
			out << "(";
			dump_expression( eb.a.get() );
			out << ",";
			dump_expression( eb.b.get() );
			out << ")";
			break;
		}
		
		case expression::f_unary: {
			auto & eb = static_cast<expr_unary const&>(*ex);
			switch( eb.op ) {
				#define OP( part ) \
					case expr_unary::o_##part: \
						out << "_" #part; \
						break;
				OP(neg)
				OP(abs)
				
				OP(bit_cpl)
				#undef OP
			}
			out << "(";
			dump_expression( eb.param.get() );
			out << ")";
			break;
		}
	}
	
	if( !omit_type ) {
		if( !show_spec_type && show_all_types && (ex->type.has_fun()) ) {
			out << " : ";
			dump_intr_type( ex->type );
		} else if( show_spec_type && ex->spec_type ) {
			out << " : ";
			dump_type_spec( ex->spec_type.get() );
		}
	}
	

	if( list_close )
		out << " )";
}

void dump::dump_tuple( expr_with_sub const * ex, char open, char close, char sep )
{
	CHECK_NULL(ex,"tuple")
	
	out << open << " ";
	for( size_t i = 0; i < ex->sub.size(); i++ )
	{
		if( i )
			out << sep << " ";
		dump_expression( ex->sub[i].get(), true );
	}
	if( ex->sub.size() )
		out << " ";
	out << close;
}


void dump::dump_assign_statement( assign_statement const * as )
{
	CHECK_NULL(as,"assign_statement")
	dump_expression( as->left.get(), false );
	switch( as->how ) {
		case assign_statement::h_value: out << " = "; break;
		case assign_statement::h_init_refer: out << " := "; break;
		case assign_statement::h_default: out << " |= "; break;
	}
	dump_expression( as->right.get(), false );
}

template<typename T>
void dump::dump_intr_type_tuple_inner( char o, char c, T const * vt )
{
	CHECK_NULL( vt, "intr_type(T)" )
	
	out << o;
	bool need_space = true; //to deal with empty types
	for( size_t i=0; i < vt->sub.size(); ++i )
	{
		if( i )
			out << ",";
		out << " ";

		auto & sub = vt->sub[i];
		if( sub.name.size() )
			out << sub.name << " : ";
		else
			out << ": ";
		auto pos = out.tellp();
		_dvtti_sub( sub.type );
		need_space = pos != out.tellp();
	}
	if( need_space )
		out << " ";
	out << c;
}

void dump::write_intr_type( intr_type const * vt, bool expand_defined )
{
	CHECK_NULL( vt, "intr_type" )
	std::vector<std::string> markers;
	dump_intr_type( *vt, markers, expand_defined );
}

void dump::dump_intr_type( extr_type_ref const & tr, bool expand_defined )
{
	CHECK_NULL( tr.has_fun(), "intr_type" )
	
	//show all markers/labels in alphabetical order (predictable for tests)
	std::vector<std::string> markers;
	
	//don't show rvalue, just lvalue
	if( tr.is_lvalue() )
		markers.push_back( "lvalue" );
	if( tr.is_shared() )
		markers.push_back( "shared" );
	if( tr.is_value_ptr() )
		markers.push_back( "value_ptr" );
	if( tr.is_optional() )
		markers.push_back( "optional" );
	if( tr.is_cvt_optional() )
		markers.push_back( "convert_optional" );
	if( tr.is_lossy() )
		markers.push_back( "lossy" );
	
	dump_intr_type( *tr.get_fun(), markers, expand_defined );
}

void dump::dump_intr_type( intr_type const & vt, std::vector<std::string> & markers, bool expand_defined )
{
	if( vt.is_distinct() ) {
		if( expand_defined ) {
			out << "type "; //could also be class... actually more likely
		}
		
		if( vt.distinct_name ) {
			out << *vt.distinct_name;
			outline_types.insert( &vt );
		} else {
			out << "*distinct*";
		}
		
		if( expand_defined ) {
			out << " = ";
		}
	} 
	
	if( !vt.is_distinct() || expand_defined ) {
		switch( vt.get_fun_class() )
		{
			case intr_type::f_unknown:
				out << "unknown";
				break;
				
			case intr_type::f_char:
				out << "char";
				break;

			case intr_type::f_tuple:
			{
				auto & tt = vt.cast<intr_type_tuple>();
				switch( tt.pack ) {
					case intr_type_tuple::p_normal: break;
					case intr_type_tuple::p_abi: markers.push_back( "abi_pack"); break;
					case intr_type_tuple::p_byte: markers.push_back( "byte_pack" ); break;
				}
				if( tt.selectable ) {
					markers.push_back( "selectable" );
				}
				dump_intr_type_tuple_inner( '[', ']', &tt );
				break;
			}

			case intr_type::f_value_list:
				dump_intr_type_tuple_inner( '(', ')', &vt.cast<intr_type_value_list>() );
				break;
				
			case intr_type::f_function:
			{
				auto & tf = vt.cast<intr_type_function>();
				dump_intr_type_tuple_inner( '(', ')', tf.params.get() );
				out << " -> ";
				dump_intr_type_tuple_inner( '(', ')', tf.returns.get() );
				switch( tf.access ) {
					case intr_type_function::a_bound: break;
					case intr_type_function::a_raw: markers.push_back( "raw" ); break;
					case intr_type_function::a_unbound: markers.push_back( "unbound" ); break;
				}
				if( tf.context ) {
					out << " context<";
					dump_intr_type( extr_type_ref(tf.context) );
					out << ">";
				}
				break;
			}
				
			case intr_type::f_rational: {
				auto & fr = vt.cast<intr_type_rational>();
				out << "rational";
				if( fr.integer_signed ) {
					markers.push_back( "signed" );
				}
				if( fr.integer_bitsize ) {
					char bits[128];
					sprintf( bits, "%dbit", *fr.integer_bitsize );
					markers.push_back( bits );
				}
				break;
			}
				
			case intr_type::f_integer: {
				auto it = vt.cast<intr_type_integer>();
				out << "integer";
				if( keep_variant || it.data_bitsize != platform::natural_integer_size ) {
					char bits[128];
					sprintf( bits, "%dbit", it.data_bitsize );
					markers.push_back( bits );
				}
				break;
			}
				
			case intr_type::f_binary: {
				auto it = vt.cast<intr_type_binary>();
				out << "binary";
				if( keep_variant || it.data_bitsize != platform::natural_integer_size ) {
					char bits[128];
					sprintf( bits, "%dbit", it.data_bitsize );
					markers.push_back( bits );
				}
				break;
			}
			
			case intr_type::f_float: {
				out << "float";
				auto ft = vt.cast<intr_type_float>();
				auto nf = platform::if_get_float( ft.data_bitsize );
				if( !nf ) {
					error( "invalid-size-variant" );
				} else if( nf->precision != 0 ) {
					if( nf->precision == -1 )
						markers.push_back("low");
					else if( nf->precision == 1 )
						markers.push_back("high" );
					else {
						STATE_FAIL("unexpected now");
					}
				} else if( keep_variant ) {
					markers.push_back( "normal" );
				}
					
				break;
			}
				
			case intr_type::f_boolean:
				out << "boolean";
				break;
		
			case intr_type::f_array: {
				auto & at = vt.cast<intr_type_array>();
				out << "array< ";
				dump_intr_type( at.sub );
				out << " >";
				if( at.raw ) {
					out << " raw";
				}
				break;
			}
			
			case intr_type::f_module:
			case intr_type::f_custom: {
				auto & tc = vt.cast<intr_type_scope>(); 
				out << "{";
				indent();
				for( auto & sr : tc.context->get_local_decls() ) {
					out << std::endl;
					out << indent_str();
					dump_var_decl( sr.decl.get() );
				}
				unindent();
				out << std::endl << "}";
				break;
			}
		}
	}
		
	std::sort( markers.begin(), markers.end() );
	for( size_t i=0; i < markers.size(); ++i ) {
		out << " " << markers[i];
	}
}

void dump::dump_type_spec( const type_spec * vt )
{
	CHECK_NULL( vt, "type_spec" )
	
	//show all markers/labels in alphabetical order (predictable for tests)
	std::vector<std::string> markers;
	
	bool no_fundamental = false;
	if( vt->fun_class.is_set() ) {
		switch( *vt->fun_class )
		{
			case intr_type::f_unknown:
				out << "unknown";
				break;
				
			case intr_type::f_char:
				out << "char";
				break;

			case intr_type::f_tuple:
				dump_intr_type_tuple_inner( '[', ']', vt );
				break;

			case intr_type::f_value_list:
				dump_intr_type_tuple_inner( '(', ')', vt );
				break;
				
			case intr_type::f_function:
				STATE_CHECK( vt->sub.size() == 2 );
				dump_intr_type_tuple_inner( '(', ')', vt->sub[0].type.get() );
				out << " -> ";
				dump_intr_type_tuple_inner( '(', ')', vt->sub[1].type.get() );
				break;
				
			case intr_type::f_rational:
				out << "rational";
				break;
				
			case intr_type::f_integer:
				out << "integer";
				break;
				
			case intr_type::f_binary:
				out << "binary";
				break;
				
			case intr_type::f_float:
				out << "float";
				break;
				
			case intr_type::f_boolean:
				out << "boolean";
				break;
		
			case intr_type::f_array:
				STATE_CHECK( vt->sub.size() == 1 );
				out << "array< ";
				dump_type_spec( vt->sub[0].type.get() );
				out << " >";
				break;
				
			case intr_type::f_custom:
				out << "custom";
				break;
			case intr_type::f_module:
				out << "module";
				break;
		}
	} else if( !hide_var_spec_no_fundamental ) {
		if( vt->fun_class.is_unset() ) {
			if( show_spec_implicit ) {
				out << "?fundamental";
			} else {
				no_fundamental = true;
			}
		} else {
			out << "no-fundamental";
		}
	} else {
		no_fundamental = true;
	}

	if( vt->fun_prototype ) {
		if( !no_fundamental )
			out << " ";
			
		if( !vt->fun_prototype->distinct_name ) {
			out << "!!undefined";
		} else {
			out << std::string( "prototype(" ) << *vt->fun_prototype->distinct_name << ")";
		}
		no_fundamental = false;
	}
	
	if( vt->function_context ) {
		out << " context<";
		dump_intr_type( extr_type_ref(vt->function_context) );
		out << ">";
	}
	
	//typical boolean spec display
	auto any_spec = [&]( util::spec_optional_base const & spec, std::string const & label ) {
		if( spec.is_unset() ) {	
			if( show_spec_implicit )
				markers.push_back( std::string( "?" ) + label );
		} else if( spec.is_default() ) {
			if( show_spec_implicit )
				markers.push_back( std::string( "=" ) + label );
		}
	};
	auto bool_spec = [&]( util::spec_optional<bool> const & spec, std::string const & label ) {
		if( spec.is_set() ) {
			if( *spec )
				markers.push_back( label );
		} else {
			any_spec( spec, label );
		}
	};
	bool_spec( vt->lvalue, "lvalue" );
	bool_spec( vt->shared, "shared" );
	bool_spec( vt->optional, "optional" );
	bool_spec( vt->cvt_optional, "convert_optional" );
	bool_spec( vt->value_ptr, "value_ptr" );
	bool_spec( vt->signed_, "signed" );
	bool_spec( vt->selectable, "selectable" );
	bool_spec( vt->member, "member" );
	bool_spec( vt->lossy, "lossy" );
	
	if( vt->function_access.is_set() ) {
		switch( *vt->function_access ) {
			case intr_type_function::a_bound: break;
			case intr_type_function::a_raw: markers.push_back( "raw" ); break;
			case intr_type_function::a_unbound: markers.push_back( "unbound" ); break;
		}
	} else {
		any_spec( vt->function_access, "access" );
	}
	
	if( vt->pack.is_set() ) {
		switch( *vt->pack ) {
			case intr_type_tuple::p_normal: break;
			case intr_type_tuple::p_abi: markers.push_back( "abi_pack" ); break;
			case intr_type_tuple::p_byte: markers.push_back( "byte_pack" ); break;
		}
	} else {
		any_spec( vt->pack, "pack" );
	}
	
	for( auto & s : vt->symbols )
		markers.push_back( s.c_str() );
			
	if( vt->data_bitsize.is_set() ) {
		char bits[128];
		sprintf( bits, "%dbit", *vt->data_bitsize );
		markers.push_back( bits );
	} else {
		any_spec( vt->data_bitsize, "bitsize" );
	}

	if( vt->precision_variant.is_set() ) {
		switch( *vt->precision_variant ) {
			case -1: markers.push_back( "low" ); break;
			case 0: markers.push_back( "normal" ); break;
			case 1: markers.push_back( "high" ); break;
			default: {
				char bits[128];
				sprintf( bits, "??precision(%d)", *vt->precision_variant );
				markers.push_back( bits );
			}
		}
	} else {
		any_spec( vt->precision_variant, "precision" );
	}
		
	std::sort( markers.begin(), markers.end() );
	for( size_t i=0; i < markers.size(); ++i ) {
		if( i || !no_fundamental )
			out << " ";
		out << markers[i];
	}
}

void dump::dump_var_decl( var_decl const * decl )
{
	CHECK_NULL( decl, "var_decl" );
	
	if( decl->is_export ) {
		out << "@export ";
	}
	
	if( decl->import_type != var_decl::it_none) {
		switch( decl->import_type ) {
			case var_decl::it_none: break;
			case var_decl::it_abi_code: out << "@import("; break;
			case var_decl::it_abi_data: out << "@import_data("; break;
		}
		if( decl->import_module.size() ) {
			out << decl->import_module << ".";
		}
		out << decl->import_name;
		switch( decl->name_form ) {
			case var_decl::nf_bare: out << ",bare"; break;
			case var_decl::nf_leaf: out << ",leaf"; break;
		}
		out << ") ";
	}
	
	switch( decl->form ) {
		case var_decl::f_single: out << "var "; break;
		case var_decl::f_multi: out << "func "; break;
		case var_decl::f_param: out << "param "; break; //internal (not in code directly)
		case var_decl::f_return: out << "return "; break; //internal
		case var_decl::f_capture: out << "capture "; break; //internal
	}
	
	out << decl->symbol;

	if( show_spec_type ) {
		if( decl->spec_type ) {
			out << " : ";
			dump_type_spec( decl->spec_type.get() );
		}
	} else {
		out << " : ";
		dump_intr_type( decl->type );
	}
}

void dump::dump_var_statement( var_statement const * vs )
{
	CHECK_NULL( vs, "var_statement" )
	
	dump_var_decl( vs->decl.get() );
	
	if( vs->value ) {
		if( vs->init_refer_value )
			out << " := ";
		else
			out << " = ";
		dump_expression( vs->value.get(), false );
	}
}

void dump::dump_init_statement( init_statement const * is )
{
	CHECK_NULL( is, "init_statement" )
	auto decl = is->sref.decl;

	out << "init ";
	dump_var_decl( decl.get() );

	if( is->undefined ) 
		out << " = undefined";
		
	if( is->value ) {
		if( is->init_refer_value )
			out << " := ";
		else
			out << " = ";
		dump_expression( is->value.get(), false );
	}
	
	if( is->no_release )
		out << " no_release";
		
	if( is->code ) {
		out << endl();
		indent();
		out << indent_str();
		dump_statement( is->code.get() );
		unindent();
	}
}

void dump::dump_defer_statement( defer_statement const * defer )
{
	CHECK_NULL( defer, "defer_statement" )
	out << (defer->error_only ? "defer_error" : "defer");
	dump_statement_block( defer->block.get() );
}

void dump::dump_do_statement( do_statement const * ds )
{
	CHECK_NULL( ds, "do_statement" )
	if( ds->loop ) {
		out << "loop ";
	} else {
		out << "do ";
	}

	out << "switch {" << endl();
	indent();
	for( auto & v : ds->expr_list ) {
		out << indent_str();
		if( v.cond ) {
			dump_expression( v.cond.get(), true );
			out << " then ";
		} else {
			out << "else";
		}
		dump_statement_block( v.block.get() );
		out << endl();
	}
	unindent();
	out << indent_str() << "}";

	dump_error_statement( *ds );
}

void dump::dump_statement( statement const * st )
{
	CHECK_NULL(st, "statement")
	validate(*st);
	
	switch( st->statement_form() )
	{
		#define DS(type) \
			case sf_##type: \
				dump_##type( static_cast<type const*>(st) ); \
				break;
		DS(assign_statement)
		DS(defer_statement)
		DS(do_statement)
		DS(expression_statement)
		DS(fail_statement)
		DS(for_statement)
		DS(import_statement)
		DS(init_statement)
		DS(loop_flow_statement)
		DS(multi_statement)
		DS(noop_statement)
		DS(return_statement)
		DS(type_statement)
		DS(var_statement)
		#undef DS
	}
}

void dump::dump_outline_types() {
	std::set<intr_type const*> ty;
	ty.swap( outline_types );
	for( auto t : ty ) {
		std::vector<std::string> markers;
		out << std::endl;
		dump_intr_type( *t, markers, true );
	}
}

void dump::dump_statement_block( statement_block const * block )
{
	CHECK_NULL(block, "statement_block")
	out << "{" << endl();
	indent();
	
	auto const & locals = block->symbols->get_local_decls();
	for( auto & local : locals ) {
		out << indent_str();
		dump_var_decl( local.decl.get() );
		out << endl();
	}
	
	for( size_t i=0; i < block->list.size(); ++i ) {
		out << indent_str();
		dump_statement( block->list[i].get() );		
		out << ";" << endl();
	}
	
	unindent();
	out << indent_str() << "}";
}

void dump::dump_multi_statement( multi_statement const * ms )
{
	CHECK_NULL(ms,"multi_statement" );
	for( size_t i=0; i < ms->list.size(); ++i ) {
		if( i ) {
			out << indent_str();
		}
		dump_statement( ms->list[i].get() );
		if( i != (ms->list.size()-1) ) {
			out << ";" << endl();
		}
	}
}

void dump::dump_type_statement( type_statement const * ts )
{
	CHECK_NULL(ts,"type_statement" );
	
	switch( ts->what ) {
		case type_statement::w_prototype:
			out << "type " << ts->name;
			if( show_spec_type ) {
				if( ts->spec_type ) {
					out << " = ";
					dump_type_spec( ts->spec_type.get() );
				}
			} else {
				out << " = ";
				dump_intr_type( ts->type, true );
			}
			break;
			
		case type_statement::w_alias: {
			out << "alias " << ts->name << " = ";
			dump_type_spec( ts->spec_type.get() );
			break;
		}
		
		case type_statement::w_class: {
			out << "class " << ts->name << " ";
			dump_statement_block( ts->class_block.get() );
			break;
		}
	};
	
	if( ts->init_block ) {
		out << " /* INIT{ */" << endl();
		indent();
		out << indent_str();
		dump_multi_statement( ts->init_block.get() );
		unindent();
		out << endl() << indent_str() << "/* } */";
	}
}

void dump::dump_return_statement( return_statement const * rs )
{
	CHECK_NULL(rs,"return_statement" );
	
	out << "return";
	
	if( rs->value ) {
		out << " ";
		dump_expression( rs->value.get(), true );
	}
}

void dump::dump_expression_statement( expression_statement const * es )
{
	CHECK_NULL( es, "expression_statement" );
	dump_expression( es->expr.get(), true );
	dump_error_statement( *es );
}

void dump::dump_error_statement( error_statement const & ds )
{
	if( ds.on_error ) {
		out << " on ";
		switch( ds.on_match ) {
			case fail_level::resume: out << "resume"; break;
			case fail_level::unwind: out << "unwind"; break;
			case fail_level::continue_: out << "continue"; break;
			case fail_level::recover: out << "recover"; break;
			case fail_level::abandon: out << "abandon"; break;
		}
		out << " ";
		dump_statement_block( ds.on_error.get() );
	}
}

void dump::dump_fail_statement( fail_statement const * fs )
{
	CHECK_NULL( fs, "fail_statement" );
	switch( fs->level ) {
		case fail_level::resume: out << "resume"; break;
		case fail_level::unwind: out << "unwind"; break;
		case fail_level::continue_: out << "continue"; break;
		case fail_level::recover: out << "recover"; break;
		case fail_level::abandon: out << "abandon"; break;
	}
}

void dump::dump_loop_flow_statement( loop_flow_statement const * ls )
{
	CHECK_NULL( ls, "loop_flow_statement" );
	switch( ls->flow ) {
		case loop_flow_statement::f_break: out << "break"; break;
		case loop_flow_statement::f_next: out << "next"; break;
	}
}

void dump::dump_noop_statement( noop_statement const * ns )
{
	CHECK_NULL( ns, "noop_statement" );
	out << "noop";
}

void dump::dump_import_statement( import_statement const * is )
{
	CHECK_NULL( is, "import_statement" );
	out << "import " << is->mod_name;
}

void dump::dump_for_statement( for_statement const * fs )
{
	CHECK_NULL( fs, "for_statement" );
	out << "for " << fs->item << " in ";
	dump_expression( fs->iterable.get() );
	out << " ";
	dump_statement_block( fs->block.get() );
}

} //eon leaf