#include "lang/language.hpp"

#include <typeinfo>
	
namespace leaf
{

void expression_statement::validate() const
{
	STATE_CHECK(expr);
	expr->validate();
	if( on_error ) {
		STATE_CHECK( expr->type.is_void() );
	}
}

/*static*/ char const * expression::name_of( form_t f )
{
	switch( f )
	{
		#define EXPRESSION_FORM(tag) case f_##tag: return #tag;
		EXPRESSION_FORMS
		#undef EXPRESSION_FORM
	}
	return "invalid";
}

void expression::validate() const
{
	type.validate();
}

void expr_const_number::validate() const
{
	expression::validate();
	
	switch( src_form )
	{
		case sf_rational:
			STATE_CHECK( type.is_rational() );
			break;
		case sf_integer:
			STATE_CHECK( type.is_integer() );
			break;
		case sf_binary:
			STATE_CHECK( type.is_binary() );
			break;
		case sf_float:
			STATE_CHECK( type.is_float() );
			break;
		case sf_boolean:
			STATE_CHECK( type.is_boolean() );
			break;
	}
}

void expr_symbol::validate() const
{
	expression::validate();
	
	STATE_CHECK( symbol.size() );
	sref.validate();
	
	//TODO: check if the type of the expression is actually compatible with the declaration (not exactly
	//the same since lvalue could be dropped)
}

void expr_with_block::validate() const
{
	expression::validate();
	STATE_CHECK( block );
	block->validate();
}

void expr_subscript::validate() const
{
	expression::validate();
	
	STATE_CHECK( base );
	base->validate();
	STATE_CHECK( base->type.is_array() );
	
	STATE_CHECK( offset );
	offset->validate();
	STATE_CHECK( offset->type.is_integer() );
}

void expr_field::validate() const
{
	expression::validate();
	STATE_CHECK( base );
	base->validate();
	STATE_CHECK( field.size() || field_index );
	STATE_CHECK( field_index || sref );
}

///////////////////////////////////////////////////////////
// expr_switch
void expr_switch::validate() const
{
	expression::validate();
	STATE_CHECK( expr_list.size() );
	for( auto & eli : expr_list ) {
		eli.validate();
		//TODO: verify all value's are the exact same type
	}
}

void expr_switch::expr_list_item::validate() const
{
	if( cond ) { //should only be allowed in last item
		cond->validate();
		STATE_CHECK( cond->type.is_boolean() );
	}
	STATE_CHECK( value );
	value->validate();
}

bool expr_switch::may_have_side_effects() const
{
	for( auto & eli : expr_list ) {
		if( (eli.cond && eli.cond->may_have_side_effects()) || 
			eli.value->may_have_side_effects() )
			return true;
	}
	return false;
}

void expr_switch::construct_result_type()
{
	STATE_CHECK( expr_list.size() );
	type = expr_list[0].value->type;
}

///////////////////////////////////////////////////////////
// expr_with_sub
void expr_with_sub::validate() const
{
	expression::validate();
	for( auto & s : sub ) {
		STATE_CHECK( s );
		s->validate();
	}
}

void expr_funccall::validate() const
{
	expression::validate();
	STATE_CHECK( func );
	func->validate();
	STATE_CHECK( func->type.is_function() );
	STATE_CHECK( params );
	params->validate();
	STATE_CHECK( params->type.is_value_list() );
}

///////////////////////////////////////////////////////////
// expr_transform
void expr_transform::validate() const
{
	expression::validate();
	STATE_CHECK( param );
}

void expr_transform::construct_result_type()
{
	switch( transform ) {
		case t_deref:
			type = param->type;
			type.set_ref_value();
			break;
			
		case t_drop_lvalue:
			type = param->type;
			type.set_lvalue( false );
			break;
			
		case t_shared_expr:
			type = param->type;
			break;
			
		case t_promote:
		case t_loss:
			STMT_FAIL(*this);
			
		case t_make_optional:
			type = param->type;
			type.set_optional(true);
			type.set_cvt_optional(true);
			break;
			
		case t_deref_optional:
			//this intentionally retains lvalue
			type = param->type;
			type.set_optional(false);
			type.set_cvt_optional(false);
			break;
			
		case t_bind_raw_func: {
			auto cl = param->type.fun_cast<intr_type_function>().clone_derived();
			cl->access = intr_type_function::a_bound;
			type = extr_type_ref(cl);
			break;
		}
			
		case t_bind_func: {
			auto cl = param->type.fun_cast<intr_type_function>().clone_derived();
			cl->access = intr_type_function::a_bound;
			cl->context.reset(); //now hidden
			type = extr_type_ref(cl);
			break;
		}
		
		case t_addr_of:  {
			auto arr = param->type.if_fun_cast<intr_type_array>();
			if( arr ) {
				auto ra = make_shared<intr_type_array>();
				ra->sub = arr->sub;
				ra->raw = true;
				type = extr_type_ref(ra);
			} else {
				type = param->type;
				type.set_lvalue(false);
				type.set_ref_value_ptr();
			}
			break;
		}
			
		case t_code_val: {
			static extr_type_ref charint( make_shared<intr_type_integer>(platform::char_size) );
			type = charint;
			break;
		}
			
		case t_char_val: {
			static extr_type_ref char_type( make_shared<intr_type_char>() );
			type = char_type;
			break;
		}
		
	}
}

///////////////////////////////////////////////////////////
// expr_new_func
void expr_new_func::validate() const
{
	expression::validate();
	new_type.validate();
}

void expr_special_field::validate() const
{
	expression::validate();
	STATE_CHECK( base );
	base->validate();
	//TODO: type validation...
}

void expr_binary::validate() const
{
	expression::validate();
	STATE_CHECK(a);
	STATE_CHECK(b);
	a->validate();
	b->validate();
	STATE_CHECK( a->type.same_as( b->type ) );
	if( is_boolean() ) {
		STATE_CHECK( type.is_boolean() );
	} else {
		STATE_CHECK( type.same_as( a->type ) );
	}
		
	STATE_CHECK( op >= min_op_ && op <= max_op_ );
}

void expr_unary::validate() const
{
	expression::validate();
	STATE_CHECK(param);
	param->validate();
	STATE_CHECK( type.same_as( param->type ) );
};

///////////////////////////////////////////////////////////
// expr_conditional
void expr_conditional::validate() const
{
	expression::validate();
	STATE_CHECK(cond);
	STATE_CHECK(a);
	cond->validate();
	a->validate();
	
	STATE_CHECK( op >= min_op_ && op <= max_op_ );
	switch( op ) {
		case o_optional_or:
			STATE_CHECK(b);
			b->validate();
			
			STATE_CHECK( cond->type.is_optional() );
			STATE_CHECK( a->type.same_as( b->type ) );
			STATE_CHECK( type.same_as( a->type ) );
			break;
			
		case o_optional_cond:
			STATE_CHECK( cond->type.is_boolean() );
			STATE_CHECK( type.is_optional() );
			STATE_CHECK( type.get_fun()->same_as( *a->type.get_fun() ) );
			break;
	}
}

size_t expr_conditional::num_inner_trans() const
{ 
	return 1; 
}

shared_ptr<expression> expr_conditional::get_inner_trans( size_t ndx )
{
	STMT_CHECK( *this, ndx == 0 );
	return a; //TODO: this seems wrong, test cases must be incomplete for conversion_applicator::apply_single
}

void expr_conditional::set_inner_trans( size_t ndx, shared_ptr<expression> expr )
{
	STMT_CHECK( *this, ndx == 0 );
	a = expr;
}

void expr_conditional::construct_result_type()
{
	switch( op ) {
		case o_optional_or:
			type = a->type;
			break;
		
		case o_optional_cond:
			type = a->type;
			type.set_optional( true );
			//optional is being constructed, thus it is convertable
			type.set_cvt_optional( true );
			break;
	}
}


} //eon leaf