#include "lang/expression_typer.hpp"
#include "lang/language.hpp"
#include "lang/conversion_applicator.hpp"
#include "lang/variable_context.hpp"

#include "lang/dump.hpp"

#define SHOW_DEBUG 1
#include "util/debug.hpp"

//CPP: There is unfortunately no way to use a function for this pattern
#define COND_PROC_EXPRESSION( expr, ctx, ret_value ) \
	STMT_CHECK( *ret_value, expr ); \
	expr = process_expression( expr, ctx ); \
	if( !expr->is_complete() ) { \
		return ret_value; \
	}
	
namespace leaf
{

namespace
{
	/**
		Find a function which has the lowest call conversion cost from the given parameters.
	*/
	size_t match_function( intr_type_tuple const & select,
		expression const & where, typer::context ctx, intr_type const & params_ )
	{
		struct match_t {
			size_t what;
			type_converter::convert_t cost;
		};
		std::vector<match_t> match;
		
		auto params = params_.if_cast<intr_type_value_list>();
		STMT_CHECK( where, params );

		auto tc = ctx.converter();
		for( size_t i=0; i < select.sub.size(); ++i )
		{
			auto func = select.sub[i].type.if_fun_cast<intr_type_function>();
			if( !func ) {
				continue;
			}
			
			STMT_CHECK( where, func->params );
			auto & func_params = *func->params;
			
			//OPT: should return matching conversions, otherwise caller does agian
			if( func_params.sub.size() < params->sub.size() )
				continue;

			match_t check;
			check.what = i;
			check.cost = tc.convert_function_call( func_params, *params );
			if( check.cost.okay() ) {
				if( match.empty() ) {
					match.push_back( check );
				} else {
					int cmp = match[0].cost.compare_cost( check.cost );
					if( cmp == 0 ) {
						match.push_back( check );
					} else if( cmp > 0 ) {
						match.clear();
						match.push_back( check );
					} //else most costly conversion
				}
			}
		}
		
		if( match.size() == 1 )
			return match[0].what;

		auto fail = [&](char const * message ) {
			std::ostringstream out;
			out << message << std::endl;
			out<< "Params: ";
			leaf::dump d( out );
			d.write_intr_type( params );
			out << std::endl;
			out << "Candidates: " << std::endl;
			for( size_t i=0; i < select.sub.size(); ++i )
			{
				d.write_intr_type( select.sub[i].type );
				out << std::endl;
			}
			STMT_FAIL( where, out.str() );
		};
		
		if( match.size() == 0 ) {
			fail( "no-matching-symbol" );
		} else {
			fail( "multiple-matching-symbols" );
		}
		STATE_FAIL("unreachable");
	}
	
	extr_type_ref resolve_new_func_type( expression const & where,
		extr_type_ref type )
	{
		STMT_CHECK( where, type.is_array() );
		
		auto vt = make_shared<intr_type_function>();
		vt->access = intr_type_function::a_raw;
		
		//size of the array
		auto p = make_shared<intr_type_value_list>();
		auto i = make_shared<intr_type_integer>();
		p->push_sub( "size", extr_type_ref(i) );
		vt->params = p;
		
		//created array
		auto r = make_shared<intr_type_value_list>();
		r->push_sub( type );
		vt->returns = r;
		
		return extr_type_ref(vt);
	}
} //eon anon

shared_ptr<expression> expression_typer::process( shared_ptr<expression> expr )
{
	return process_expression( expr, root_ctx );
}

shared_ptr<expression> expression_typer::process_expression( shared_ptr<expression> expr, 
	typer::context ctx )
{
	//only type things once
	if( expr->is_complete() )
		return expr;
		
	//should the conversion to type_spec be skipped
	bool skip_spec = false;
	shared_ptr<expression> orig = expr;
	
	switch( expr->form )
	{
		case expression::f_symbol:
		{
			auto & es = static_cast<expr_symbol&>(*expr);
			scope::var_vector_t const & match = ctx.sc->get_vars( es.symbol );
			STMT_CHECK( *expr, match.size() != 0, std::string("Unknown symbol: ") + es.symbol );
				
			auto single = [&]( scope_ref sr ) {
				//must become available in our scope
				es.sref = ctx.sc->capture_var(sr);
				auto rtype = es.sref.decl->type;
				rtype.set_lvalue( true );
				es.type = rtype;
			};
			
			if( match.size() == 1 ) {
				//unknown types can't be evaluated yet
				if( !match[0].decl->type.is_complete() )
					return expr;
					
				//DEBUG( "SYMBOL", es.symbol, match[0].decl.get() );
				single( match[0] );
			} else {
				//TODO: can lvalue's be used here (can't since it is a tuple!), I think this is okay for selectables...
				auto select = make_shared<intr_type_tuple>();
				select->selectable = true;
				auto multi = make_shared<expr_tuple>();
				multi->src_loc = expr->src_loc;

				for( auto & sr : match ) {
					auto sym = make_shared<expr_symbol>();
					sym->src_loc = expr->src_loc;
					sym->symbol = sr.decl->symbol;
					sym->sref = sr;
					sym->type = sr.decl->type;
					select->push_sub( sr.decl->type );
					multi->sub.push_back( sym );
				}
				multi->type = extr_type_ref(select);
				expr = multi;
			}
			break;
		}
		
		case expression::f_new_func: {
			auto & nf = static_cast<expr_new_func&>(*expr);
			auto rt = ctx.determine_type( expr, nf.new_spec_type );
			nf.new_type = rt.result;
			nf.type = resolve_new_func_type( *expr, nf.new_type );
			break;
		}

		case expression::f_const_number: {
			auto ecn = static_cast<expr_const_number*>(expr.get());
			shared_ptr<intr_type> st;
			switch( ecn->src_form ) 
			{
				case expr_const_number::sf_rational: {
					auto tr = make_shared<intr_type_rational>();
					if( ecn->value.is_integral() ) {
						if( ecn->value.is_negative() ) {
							tr->integer_bitsize = ecn->value.signed_bits();
							tr->integer_signed = true;
						} else {
							tr->integer_bitsize = ecn->value.unsigned_bits();
							tr->integer_signed = false;
						}
					}
					st = tr;
					break;
				}
				case expr_const_number::sf_integer:
				case expr_const_number::sf_float:
				case expr_const_number::sf_binary:
					STMT_FAIL( *expr, "not expecting integer/float type const" );
					break;
				case expr_const_number::sf_boolean:
					st = make_shared<intr_type_boolean>();
					break;
			}
			expr->type = extr_type_ref(st);
			break;
		}

		case expression::f_text: {
			auto & te = static_cast<expr_text&>(*expr);
			switch( te.what ) {
				case expr_text::w_string: {
					auto at = make_shared<intr_type_array>();
					at->sub = extr_type_ref( make_shared<intr_type_char>() );
					expr->type = extr_type_ref(at);
					break;
				}
				case expr_text::w_char: {
					expr->type =  extr_type_ref( make_shared<intr_type_char>() );
					break;
				}
			}
			break;
		}
			
		case expression::f_tuple:
		{
			auto & et = static_cast<expr_tuple&>(*expr);
			auto tt = make_shared<intr_type_tuple>();
			
			for( size_t i=0; i < et.sub.size(); ++i )
			{
				auto sub = process_expression( et.sub[i], ctx ); 
				if( !sub->is_complete() )
					return expr;
					
				//lvalues cannot be contained inside tuples
				auto cost = ctx.converter().convert_any_nonlvalue( sub->type );
				et.sub[i] = conversion_applicator().apply( cost, sub );
				
				std::string name;
				if( expr->spec_type && i < expr->spec_type->sub.size() )
					name = expr->spec_type->sub[i].name;
				tt->push_sub( name, et.sub[i]->type );
			}
			
			expr->type = extr_type_ref(tt);
			skip_spec = true; //TODO: due to how names are handled
			break;
		}

		case expression::f_value_list:
		{
			auto & evl = static_cast<expr_value_list&>(*expr);
			auto tt = make_shared<intr_type_value_list>();
			expr->type = extr_type_ref(tt);
			for( size_t i=0; i < evl.sub.size(); ++i )
			{
				evl.sub[i] = process_expression( evl.sub[i], ctx );
				tt->push_sub( "", evl.sub[i]->type );
			}
			skip_spec = true; //TODO: should actually be an error, can spec the parts but not the whole expression
			break;
		}

		case expression::f_funccall:
			expr = process_funccall( std::static_pointer_cast<expr_funccall>(expr), ctx );
			break;
			
		case expression::f_funcdefn:
			expr = process_funcdefn( std::static_pointer_cast<expr_funcdefn>(expr), ctx );
			skip_spec = true; //used to construct the function
			break;
			
		case expression::f_block: {
			STATE_FAIL( "no test yet" );
			auto & eb = static_cast<expr_block&>(*expr);
			eb.block->symbols->set_extends( ctx.sc );
			eb.type = extr_type_ref(make_shared<intr_type_value_list>()); //no parts though
			ctx.proc_statement_block( eb.block );
			break;
		}
			
		case expression::f_switch:
			expr = process_switch( std::static_pointer_cast<expr_switch>(expr), ctx );
			break;
			
		case expression::f_subscript:
			expr = process_subscript( std::static_pointer_cast<expr_subscript>(expr), ctx );
			break;
			
		case expression::f_field:
			expr = process_field( std::static_pointer_cast<expr_field>(expr), ctx );
			break;

		case expression::f_conditional:
			expr = process_conditional( std::static_pointer_cast<expr_conditional>(expr), ctx );
			break;
			
		case expression::f_transform:
		case expression::f_special_field:
		case expression::f_default:
		case expression::f_binary:
		case expression::f_unary:
			//may happen in second pass of typing
			break;
	}

	//we couldn't type the expression yet
	if( !expr->is_complete() )
		return expr;
		
	//sanity check
	STMT_CHECK( *expr, expr->type.has_fun(), "failed-to-type-expression" );

	//convert to desired type
	if( !skip_spec && orig->spec_type ) {
		//DEBUG( "MATCHING SPEC_TYPE", dump::get_type_spec( *orig->spec_type),
		//	dump::get_extr_type( expr->type ) );
		auto it = ctx.identifier().infer( orig->spec_type, expr->type, false );
		STMT_CHECK( *orig, it.okay(), it.explain_error() );
		expr = ctx.convert_expression( it.result, expr );
	}
	
	return expr;
}

shared_ptr<expression> expression_typer::process_switch( shared_ptr<expr_switch> expr, 
	typer::context ctx )
{	
	static extr_type_ref cond_type( make_shared<intr_type_boolean>() );
	bool have_else = false;
	
	std::vector<extr_type_ref> types;
	for( auto & item : expr->expr_list ) {
		if( item.cond ) {
			STATE_CHECK( !have_else, "conditions post-else statement" );
			COND_PROC_EXPRESSION( item.cond, ctx, expr );
			item.cond = ctx.convert_expression( cond_type, item.cond );
		} else {
			have_else = true;
		}

		COND_PROC_EXPRESSION( item.value, ctx, expr );
		types.push_back( item.value->type );
	}

	STMT_CHECK( *expr, have_else, "missing an else statement" );

	//types are unified so the switch has a single type result
	auto icost = ctx.converter().unify( &types[0], types.size() );
	STMT_CHECK( *expr, icost.okay(), "no unifying type" );
	//DEBUG( "Unify Type: ", dump::get_extr_type( icost.result ) );
	for( auto & item : expr->expr_list ) {
		item.value = ctx.convert_expression( icost.result, item.value );
	}
	
	expr->construct_result_type();
	return expr;
}

namespace
{
	typedef bool (*check_op_type)( extr_type_ref const & );
	bool check_op_ibf( extr_type_ref const & t ) {
		return t.is_integer() || t.is_binary() || t.is_float();
	}
	bool check_op_ibfx( extr_type_ref const & t ) {
		return t.is_integer() || t.is_binary() || t.is_float() || t.is_boolean();
	}
	bool check_op_if( extr_type_ref const & t ) {
		return t.is_integer() || t.is_float();
	}
	bool check_op_b( extr_type_ref const & t ) {
		return t.is_binary();
	}
	bool check_op_i( extr_type_ref const & t ) {
		return t.is_integer();
	}
	
	struct binary_entry { 
		expr_binary::op_t op; 
		check_op_type check;  
	};
	//construct a static map of symbols to binary builtin ops
	typedef std::map<std::string,binary_entry> bin_op_map_t;
	bin_op_map_t bin_op_map;
	struct bom_init {
		bom_init() {
			#define OP(name,op,type) bin_op_map[#name] = { expr_binary::o_##op, check_op_##type };
			OP(add,add,ibf)
			OP(sub,sub,ibf)
			OP(mul,mul,ibf)
			OP(div,div,ibf)
			
			OP(mod,mod,i)
			
			OP(bit_shl,bit_shl,b)
			OP(bit_shr,bit_shr,b)
			OP(bit_or,bit_or,b)
			OP(bit_and,bit_and,b)
			OP(bit_xor,bit_xor,b)
			
			OP(eq,eq,ibfx)
			OP(neq,neq,ibfx)
			OP(lt,lt,ibfx)
			OP(lte,lte,ibfx)
			OP(gt,gt,ibfx)
			OP(gte,gte,ibfx)
			#undef OP
		}
	} _bom_init;
	
	struct unary_entry { 
		expr_unary::op_t op; 
		check_op_type check;  
	};
	typedef std::map<std::string,unary_entry> un_op_map_t;
	un_op_map_t un_op_map;
	struct uom_init {
		uom_init() {
			#define OP(name,op,type) un_op_map[#name] = { expr_unary::o_##op, check_op_##type };
			OP(neg,neg,if)
			OP(abs,abs,if)
			OP(bit_cpl,bit_cpl,b)
			#undef OP
		}
	} _uom_init;
} //eon anon

/**
	The binary/unary operators cannot be processed as normal functions at this time since they
	have an unlimited number of size variations. Ultimately one would hope that a proper parametric
	function definition would be able to capture the logic being applied here.
	
	@return if null it means it didn't do anything and normal funccall processing should
		continue
*/
shared_ptr<expression> expression_typer::process_funccall_builtin( shared_ptr<expr_funccall> expr, 
	typer::context ctx )
{
	if( expr->params->form != expression::f_value_list 
		|| expr->func->form != expression::f_symbol ) {
		return shared_ptr<expression>();
	}
	auto params = std::static_pointer_cast<expr_value_list>(expr->params);
	auto params_type = params->type.fun_cast<intr_type_value_list>();
	auto & es = static_cast<expr_symbol&>(*expr->func);

	auto bop = bin_op_map.find( es.symbol );
	auto uop = un_op_map.find( es.symbol );
	
	if( es.symbol == "has" ) {
		STMT_CHECK( *params, params->sub.size() == 1 );
		auto cost = ctx.converter().convert_nonlvalue_optional( params_type.sub[0].type );
		params->sub[0] = conversion_applicator().apply( cost, params->sub[0] );
		
		auto have = make_shared<expr_special_field>();
		have->src_loc = expr->src_loc;
		have->field = expr_special_field::f_optional_set;
		have->base = params->sub[0];

		have->type = extr_type_ref(make_shared<intr_type_boolean>());
		return have;
		
	} else if( es.symbol == "lossy" ) { 
		STMT_CHECK( *params, params->sub.size() == 1 );
		auto lt = params->sub[0];
		lt->type.set_lossy( true );
		return lt;
		
	} else if( es.symbol == "code_val" ) {
		static extr_type_ref char_type( make_shared<intr_type_char>() );
		STMT_CHECK( *params, params->sub.size() == 1 );

		auto er = make_shared<expr_transform>();
		er->transform = expr_transform::t_code_val;
		er->param = ctx.convert_expression( char_type, params->sub[0] );
		er->construct_result_type();
		return er;
		
	} else if( es.symbol == "char_val" ) {
		static extr_type_ref charint( make_shared<intr_type_integer>(platform::char_size) );
		
		STMT_CHECK( *params, params->sub.size() == 1 );
		auto er = make_shared<expr_transform>();
		er->transform = expr_transform::t_char_val;
		er->param = ctx.convert_expression( charint, params->sub[0] );
		er->construct_result_type();
		return er;
		
	} else if( bop != bin_op_map.end() ) {
		STMT_CHECK( *params, params->sub.size() == 2 );
		type_converter tc = ctx.converter();
		//DEBUG( dump::get_extr_type(params_type.sub[0].type) );
		//DEBUG( dump::get_extr_type(params_type.sub[1].type) );
		for( int i=0; i < 2; ++i ) {
			auto cost = tc.convert_any_scalar( params_type.sub[i].type );
			if( !cost.okay() )
				return shared_ptr<expression>();
			params->sub[i] = conversion_applicator().apply( cost, params->sub[i] );
		}
		
		auto unify = tc.unify( params->sub[0]->type, params->sub[1]->type );
		STMT_CHECK( *params, unify.okay() );
		
		auto fix = tc.fixate( unify.result );
		STMT_CHECK( *params, fix.okay() );
		STMT_CHECK( *params, bop->second.check( fix.result ), "invalid-binary-op-type" );
				
		for( int i=0; i <2; ++i ) {
			auto cost = tc.convert( fix.result, params->sub[i]->type );
			params->sub[i] = conversion_applicator().apply( cost, params->sub[i] );
		}
		
		auto bin = make_shared<expr_binary>();
		bin->src_loc = expr->src_loc;
		bin->op = bop->second.op;
		bin->a = params->sub[0];
		bin->b = params->sub[1];
		if( bin->is_boolean() )
			bin->type = extr_type_ref(make_shared<intr_type_boolean>());
		else
			bin->type = fix.result;
		//DEBUG( "BOP-TYPE", dump::get_extr_type(*bin->type) );
		return bin;
		
	} else if( uop != un_op_map.end() ) {
		STMT_CHECK( *params, params->sub.size() == 1 );
		//fold negative constant (TODO: should actually do more const folding: to be guaranteed by language)
		if( es.symbol == "neg" && params->sub[0]->type.is_rational() ) {
			auto param = params->sub[0];
			STATE_CHECK( param->form == expression::f_const_number );
			auto & cn = static_cast<expr_const_number&>( *param );
			cn.value = cn.value.neg();
			return param;
		}
		
		auto tc = ctx.converter();
		auto cost = tc.convert_any_scalar( params_type.sub[0].type );
		if( !cost.okay() )
			return shared_ptr<expression>();
		params->sub[0] = conversion_applicator().apply( cost, params->sub[0] );
		
		auto fix = tc.fixate( params->sub[0]->type );
		STMT_CHECK( *params, fix.okay() );
		STMT_CHECK( *params, uop->second.check( fix.result ), "invalid-unary-op-type" );
		cost = tc.convert( fix.result, params->sub[0]->type );
		params->sub[0] = conversion_applicator().apply( cost, params->sub[0] );
		
		auto un = make_shared<expr_unary>();
		un->src_loc = expr->src_loc;
		un->op = uop->second.op;
		un->param = params->sub[0];
		un->type = fix.result;
		return un;
	}
	
	return shared_ptr<expression>();
}

/**
	Functions work with value-lists, but are internall represented by a tuple. So the
	function call operation takes a tuple, but the actual function works with a list.
	The result of the function-call is also a list. This is done to get the desired
	syntax dealing with function calls and multiple-returns.
*/
shared_ptr<expression> expression_typer::process_funccall( shared_ptr<expr_funccall> expr, 
	typer::context ctx )
{
	COND_PROC_EXPRESSION( expr->params, ctx, expr );
	
	//builtin's aren't really function calls, handle them first
	shared_ptr<expression> func_sym = expr->func;
	auto builtin = process_funccall_builtin( expr, ctx );
	if( builtin )	
		return builtin;
	
	shared_ptr<intr_type_value_list const> func_params_type, func_returns_type;
	COND_PROC_EXPRESSION( func_sym, ctx, expr );
	auto cost = ctx.converter().convert_any_callable( func_sym->type );
	expr->func = conversion_applicator().apply( cost, func_sym );
	func_sym = expr->func;
	
	auto * st = func_sym->type.if_fun_cast<intr_type_tuple>();
	if( st && st->selectable ) {
		size_t func_index = match_function( *st, *func_sym, ctx, *expr->params->type.get_fun() );

		if( func_sym->form == expression::f_tuple ) {
			//avoid creating the tuple and just get the indexed item
			//TODO: this is actually an optimization and is not strictly needed (now at least). there are no tests for this behaviour (difficult)
			//TODO: captured srefs will still be captured resulting in unnecessarily large closures
			auto & fst = static_cast<expr_tuple&>(*func_sym);
			STMT_CHECK( *func_sym, func_index < fst.sub.size() );
			func_sym = fst.sub[func_index];
		} else {
			auto sub = make_shared<expr_field>();
			sub->src_loc = func_sym->src_loc;
			sub->base = func_sym;
			sub->field_index = func_index;
			sub->type = st->sub[func_index].type;
			func_sym = sub;
		}
		expr->func = func_sym;
		
		//expression type is that of return value
		func_returns_type = func_sym->type.fun_cast<intr_type_function>().returns;
		func_params_type = func_sym->type.fun_cast<intr_type_function>().params;
	} else {
		auto ft = expr->func->type.if_fun_cast<intr_type_function>();
		STMT_CHECK( *func_sym, ft );
		func_params_type = ft->params;
		func_returns_type = ft->returns;
	}

	//check for member access //TODO: must be based on type, not form
#if 0
	if( func_sym->form == expression::f_field ) {
		auto & field = static_cast<expr_field&>( *func_sym );
		if( field.base->type.is_custom() ) {
			auto cost = ctx.converter().convert_shared_or_unqualified( field.base->type );
			//TODO: auto cost = ctx.converter().convert_retain_lvalue( field.base->type );
			//TODO: This is bad since apply_conversion modifies the provided expression, it doesn't
			//create a clone, so we're modifying the type which is already used above...
			//I did a quick fix in drop_lvalue by processing the field type, which just hides the problem!
			expr->context = conversion_applicator().apply( cost, field.base );
		}
	}
#endif
		
	//convert parameters
	auto pcost = ctx.converter().convert_function_call( *func_params_type,
		expr->params->type.fun_cast<intr_type_value_list>() );
	expr->params = conversion_applicator().apply( pcost, expr->params );
	
	//1-value-list's are always collapsed
	if( func_returns_type->sub.size() == 1 )
		expr->type = func_returns_type->sub[0].type;
	else
		expr->type = extr_type_ref(func_returns_type);
		
	return expr;
}

/**
	Commong code for subscript and field on tuples.
*/
shared_ptr<expression> expression_typer::access_field_index( shared_ptr<expression> ex,
	shared_ptr<expression> base, size_t index,  typer::context /*ctx*/ ) 
{
	bool is_lvalue = base->type.is_lvalue();
	//to support rational fixation this folding of constant lookups to known tuples is required
	if( base->form == expression::f_tuple ) {
		STATE_CHECK( !is_lvalue ); //shouldn't be possible
		auto & tup = static_cast<expr_tuple&>(*base);
		STMT_CHECK( *ex, index < tup.sub.size() );
		return tup.sub[index];
	}
	
	//convert to field access
	auto & bt = base->type.fun_cast<intr_type_tuple>();
	STMT_CHECK( *ex, index < bt.sub.size() );
	
	shared_ptr<expr_field> fex;
	//process_field caller will already have a field
	if( ex->form == expression::f_field ) {
		fex = std::static_pointer_cast<expr_field>(ex);
	} else {
		fex = make_shared<expr_field>();
		fex->src_loc = ex->src_loc;
		fex->base = base;
	}
	
	fex->field_index = index;
	fex->type = bt.sub[index].type;
	if( is_lvalue )
		fex->type.set_lvalue( true );
	return fex;
}

shared_ptr<expression> expression_typer::process_subscript( shared_ptr<expr_subscript> ex, 
	typer::context ctx )
{
	COND_PROC_EXPRESSION( ex->base, ctx, ex );
	COND_PROC_EXPRESSION( ex->offset, ctx, ex );
	
	auto cost = ctx.converter().convert_retain_lvalue( ex->base->type );
	ex->base = conversion_applicator().apply( cost, ex->base );
	
	static extr_type_ref vti( make_shared<intr_type_integer>() );
	ex->offset = ctx.convert_expression( vti, ex->offset );
	
	if( ex->base->type.is_tuple() ) {
		//FEATURE: support constexpr propagation and allow it here
		//TODO: add "const" types then convert here to a const number
		STMT_CHECK( *ex, ex->offset->form == expression::f_const_number );
		auto ecn = static_cast<expr_const_number*>(ex->offset.get());
		STMT_CHECK( *ex, ecn->src_form == expr_const_number::sf_integer );
		auto index = unsigned(ecn->integer_value());

		return access_field_index( ex, ex->base, index, ctx );
		
	} else if( ex->base->type.is_array() ) {
		auto & at = ex->base->type.fun_cast<intr_type_array>();
		auto cl = at.sub;
		if( ex->base->type.is_lvalue() )
			cl.set_lvalue( true );
		ex->type = cl;
		return ex;
		
	} else {
		STMT_FAIL( *ex, "index not supported on this type" );
	}
}

shared_ptr<expression> expression_typer::process_field( shared_ptr<expr_field> field, 
	typer::context ctx )
{
	COND_PROC_EXPRESSION( field->base, ctx, field );
		
	auto cost = ctx.converter().convert_retain_lvalue( field->base->type );
	field->base = conversion_applicator().apply( cost, field->base );
	auto & t = *field->base->type.get_fun();
	auto tuple_t = t.if_cast<intr_type_compound>();
	
	if( t.is_tuple() ) {
		auto index = tuple_t->find_name( field->field );
		STMT_CHECK( *field, index, std::string("no field named: " ) + field->field );
		
		return access_field_index( field, field->base, *index, ctx );
		
	} else if( t.is_array() ) {
		STMT_CHECK( *field, field->field == "size" );
		
		auto sf = make_shared<expr_special_field>();
		sf->src_loc = field->base->src_loc;
		sf->field = expr_special_field::f_array_size;
		sf->base = field->base; //TODO: actually need conversion to array?
		auto cl = sf->base->type;
		cl.set_lvalue( false );
		sf->base->type = cl; //TODO: see above
		sf->type = extr_type_ref(make_shared<intr_type_integer>());
		return sf;
		
	} else if( t.is_scope() ) {
		auto & c = t.cast<intr_type_scope>();
		
		auto vars = c.context->get_vars( field->field );
		if( vars.empty() ) {
			STMT_FAIL( *field, std::string( "unknown-field: " ) + field->field );
		}
		
		auto bind = [&](shared_ptr<expr_field> & ef )->shared_ptr<expression> {
			if( t.is_module() ) {
				auto sym = make_shared<expr_symbol>();
				sym->sref = *field->sref;
				sym->src_loc = field->src_loc;
				sym->symbol = *c.distinct_name + "." + field->sref->decl->symbol; //name is arbitrary
				sym->type = field->sref->decl->type;
				return sym;
			} else {
				//bind member functions (TODO: only "member" functions, that info is lost now, hmmm)
				auto * vtf = ef->type.if_fun_cast<intr_type_function>();
				if( vtf && vtf->is_unbound() ) {
					auto ntf = vtf->clone_derived();
					ntf->access = intr_type_function::a_bound;
					ntf->context.reset();
					ef->type.set_fun( ntf );
					ef->bind = true;
				}
				return ef;
			}
		};
		
		if( vars.size() == 1 ) {
			field->sref = vars[0];
			field->type = field->sref->decl->type;
				
			auto r = bind( field );
			if( field->base->type.is_lvalue() )
				r->type.set_lvalue( true );
			return r;
		} else {
			auto select = make_shared<intr_type_tuple>();
			select->selectable = true;
			auto multi = make_shared<expr_tuple>();
			multi->src_loc = field->src_loc;

			//TODO: the selectable doesn't undergo full conversion during the call, the parts aren't converted...
			field->base = conversion_applicator().drop_lvalue( field->base );
			
			for( auto & sr : vars ) {
				auto sym = make_shared<expr_field>();
				sym->src_loc = field->src_loc;
				sym->field = sr.decl->symbol;
				sym->sref = sr;
				sym->type = sr.decl->type;
				sym->base = field->base; //TODO: should be shared
				
				auto r = bind(sym);
				
				select->push_sub( r->type );
				multi->sub.push_back( r );
			}
			multi->type = extr_type_ref(select);
			return multi;
		}

	} else {
		STMT_FAIL( *field, "type-lacks-fields: " + dump::get_extr_type(field->base->type) );
	}
	
	STMT_FAIL( *field, "unreachable" );
}

shared_ptr<expression> expression_typer::process_funcdefn( shared_ptr<expr_funcdefn> func, 
	typer::context ctx )
{
	//force items with default to be optional before creating the type
	STMT_CHECK( *func, func->spec_type->sub.size() == 2 );
	bool is_member = func->spec_type->member.is_set() && *func->spec_type->member;
	auto & spec_params = *func->spec_type->sub[0].type;
	for( size_t i=0; i < spec_params.sub.size(); ++i ) {
		auto dit = func->param_defaults.find(i);
		if( dit != func->param_defaults.end() ) {
			
			STMT_CHECK( *func, !spec_params.sub[i].type->optional.is_set()
				|| *spec_params.sub[i].type->optional == true );
			spec_params.sub[i].type->optional = true;
		}
	}
	
	auto rt = ctx.identifier().determine( func->spec_type );
	if( rt.unknown_identifier() ) {
		STMT_CHECK( *func, !no_incomplete, rt.explain_error() );
		func->complete = false;
		return func;
	}
	auto func_type = rt.result.if_fun_cast<intr_type_function>();
	STMT_CHECK( *func, func_type );
	func->type = rt.result;
	
	auto & params = *func_type->params;
	STATE_CHECK( params.is_value_list() );
	auto & returns = *func_type->returns;
	STATE_CHECK( returns.is_value_list() );

	statement_block & func_block = *func->block;
	auto funcdefn_parent = ctx.sc->get_defn_parent();
	func_block.symbols->set_parent( funcdefn_parent );
	func_block.symbols->mark_function_root();
	
	std::vector<shared_ptr<statement>> extra_block;
	//introduce a var_statement into the function block for each parameter
	for( size_t i=0; i < params.sub.size(); ++i ) {
		auto & param = params.sub[i];
		STATE_CHECK( param.name.size() );

		auto dit = func->param_defaults.find(i);
		if( dit != func->param_defaults.end() ) {
			//add a default assign into the block
			auto as = make_shared<assign_statement>();
			as->complete = true;
			as->how = assign_statement::h_default;
			auto sym = make_shared<expr_symbol>(); //TODO: a direct reference to the var_decl would be better/safer
			sym->symbol = param.name;
			as->left = sym;
			as->right = dit->second;
			extra_block.push_back( as );
		}
		shared_ptr<var_decl> vs( new var_decl );
		vs->form = var_decl::f_param;
		vs->symbol = param.name;
		vs->type = param.type;
		
		func_block.symbols->add_single( vs, scope::add_in_order );
	}
	//and for each return
	for( size_t i=0; i < returns.sub.size(); ++i )
	{
		auto & ret = returns.sub[i];

		shared_ptr<var_decl> vs( new var_decl );
		vs->form = var_decl::f_return;
		vs->symbol = ret.name; //returns may be anonymous
		vs->type = ret.type;

		auto init = make_shared<init_statement>();
		init->sref = func_block.symbols->add_single( vs, scope::add_in_order );
		init->no_release = true;
		extra_block.push_back( init );
	}
	
	//put extra statements in order at beginning of block
	func_block.list.insert( func_block.list.begin(), extra_block.begin(), extra_block.end() );
	func->param_defaults.clear();

	auto nfunc_type = func_type->clone_derived();
	
	if( funcdefn_parent->is_declarative() || funcdefn_parent->is_dynamic() ) {
		STMT_CHECK( *func, func->block );
		//place onto pending blocks (so this block can see later values)
		ctx.add_pending_block( func->block );
		
		if( funcdefn_parent->is_class_root() ) {
			//TODO: this isn't actually required, you can have non-member functions in a class
			STMT_CHECK( *func, nfunc_type->context );
		} else {
			//STMT_CHECK( *func, captures.empty() ); //can't really check since block isn't processed yet!
			nfunc_type->access = intr_type_function::a_raw;
		}
	} else {
		STMT_CHECK( *func, !nfunc_type->context ); //hmm, should have been a class...
		
		//TODO: this seems like the wrong context (should be from block->symbols?) (probably not used now since proc_statement_block creates a context on those symbols)
		if( !ctx.proc_statement_block( func->block ) ) {
			STMT_FAIL( *func, "unexpected-incomplete-block" )
		}
		auto captures = func_block.symbols->get_captures();
		STMT_CHECK( *func, func->block->is_complete() );
		
		if( captures.empty() ) {
			nfunc_type->access = intr_type_function::a_raw;
		} else {
			//create a custom type for the captures
			nfunc_type->access = intr_type_function::a_unbound;
			auto func_ctx = make_shared<intr_type_custom>();
			func_ctx->distinct = true;
			auto vc = make_shared<managed_variable_context>();
			func_ctx->context = vc;
			
			for( auto cap : captures ) {
				vc->add_managed( cap.second.local );
			}
			nfunc_type->context = func_ctx;
		}
	}
	func->type = extr_type_ref(nfunc_type);
	
	//unbound non-member functions should immediately be captured (syntactically it's messy to wait until
	//a capture is actually needed)
	if( nfunc_type->access == intr_type_function::a_unbound && !is_member ) {
		return conversion_applicator().bind_func( func );
	}
	
	return func;
}

shared_ptr<expression> expression_typer::process_conditional( shared_ptr<expr_conditional> expr,
	typer::context ctx )
{
	static extr_type_ref cond_type( make_shared<intr_type_boolean>() );
	
	COND_PROC_EXPRESSION( expr->cond, ctx, expr );
	
	switch( expr->op ) {
		case expr_conditional::o_optional_or: {
			//DEBUG("OPEN");
			COND_PROC_EXPRESSION( expr->b, ctx, expr );
			type_converter tc = ctx.converter();
			
			//only makes sense if first part is optional
			auto cost = tc.convert_nonlvalue_optional( expr->cond->type );
			expr->cond = conversion_applicator().apply( cost, expr->cond );
			STMT_CHECK( *expr->cond, expr->cond->type.is_optional() );
			
			//DEBUG("COND_TYPE", dump::get_extr_type(expr->cond->type) );
			auto rtype = expr->cond->type;
			rtype.set_optional( false );
			
			auto unify = tc.unify( rtype, expr->b->type );
			STMT_CHECK( *expr, unify.okay() );
			auto fix = tc.fixate( unify.result );
			STMT_CHECK( *expr, fix.okay() );
			
			//DEBUG("FIX_TYPE", dump::get_extr_type(fix.result));
			//the cond and a are supposed to be the same, thus we convert the condition
			//to an optional form of the unified type instead of converting the parts (this
			//allows proper fixation)
			auto ctype = fix.result;
			ctype.set_optional( true );
			cost = tc.convert( ctype, expr->cond->type );
			expr->cond = conversion_applicator().apply( cost, expr->cond );

			//DEBUG("POST_COND", dump::get_extr_type(expr->cond->type));
			//will be used twice, thus must be marked as shared
			expr->cond = conversion_applicator().create_shared( expr->cond );
			
			cost = tc.convert( fix.result, expr->cond->type );
			expr->a = conversion_applicator().apply( cost, expr->cond );
			cost = tc.convert( fix.result, expr->b->type );
			expr->b = conversion_applicator().apply( cost, expr->b );
			
			expr->construct_result_type();
			//DEBUG("CLOSE");
			return expr;
		}
		
		case expr_conditional::o_optional_cond: {
			COND_PROC_EXPRESSION( expr->a, ctx, expr );
			auto cost = ctx.converter().convert_any_nonlvalue( expr->a->type );
			expr->a = conversion_applicator().apply( cost, expr->a );
			
			expr->cond = ctx.convert_expression( cond_type, expr->cond );
			expr->construct_result_type();			
			return expr;
		}
	}
	
	STMT_FAIL(*expr);
}

} //eon leaf