#include "lang/intr_type.hpp"
#include "lang/variable_context.hpp"

#include "util/validation_context.hpp"

#define SHOW_DEBUG 0
#include "util/debug.hpp"

//common pattern for recursive validation
#define VALIDATE_RECURSE(ret_expr) \
	if( vctx.in(this) ) \
		return ret_expr; \
	util::validation_scope vs(vctx,this);

namespace leaf
{

/*static*/ char const* intr_type::name_of( fun_class_t f )
{
	switch( f )
	{
		#define FUN_CLASS_TYPE(s) \
			case f_##s: return #s;
		INTR_TYPE_FUN_CLASSES
		#undef FUN_CLASS_TYPE
	}
	return "corrupt";
}

void extr_type_ref::validate() const
{
	util::validation_context vctx;
	validate_impl( vctx );
}

void extr_type_ref::validate_impl(util::validation_context & vctx) const
{
	STATE_CHECK( fun );
	fun->validate_impl(vctx);
	
	if( is_value_list() ) {
		//value-lists are not real values
		STATE_CHECK( !is_shared() );
		STATE_CHECK( !is_lvalue() );
		STATE_CHECK( !is_optional() );
	}
}

intr_type::intr_type( fun_class_t fun_class )
{
	this->fun_class = fun_class;
	cref = 0;
	//start true since only a few special paths produce incomplete types
	local_complete = true;
	name_scope = nullptr;
}

intr_type::~intr_type()
{
}

void intr_type::validate() const
{
	util::validation_context vctx;
	validate_impl(vctx);
}

void intr_type::validate_impl(util::validation_context &) const
{
	STATE_CHECK( fun_class >= f_unknown && fun_class <= f_custom );
	STATE_CHECK( is_complete() );
}

int intr_type::compare( intr_type const & o, bool logical ) const
{
	util::validation_context vctx;
	return compare_impl( o, logical, vctx );
}

#define CINT( x, y ) { int q = int(x) - int(y); if( q ) return q; }
int intr_type::compare_impl( intr_type const & o, bool /*logical*/, util::validation_context & ) const
{
	CINT( fun_class, o.fun_class )
	return 0;
}

int extr_type_ref::compare( extr_type_ref const & o, bool logical ) const
{
	util::validation_context vctx;
	return compare_impl( o, logical, vctx );
}

int extr_type_ref::compare_impl( extr_type_ref const& o, bool logical, util::validation_context & vctx ) const 
{
	CINT( lvalue, o.lvalue )
	CINT( reference, o.reference )
	CINT( optional, o.optional )
	if( !logical ) {
		CINT( cvt_optional, o.cvt_optional )
	}
	if( !fun && !o.fun )
		return 0;
	if( !fun && o.fun )
		return -1;
	if( fun && !o.fun )
		return 1;
	return fun->compare_impl( *o.fun, logical, vctx );
}

bool intr_type::is_fixated() const
{
	util::validation_context ctx;
	return is_fixated_impl(ctx);
}

bool intr_type::is_complete() const
{ 
	util::validation_context ctx;
	return is_complete_impl(ctx);
}

#define COMPARE_SUPER(t_,o_,logical,vctx) \
	int q = t_::compare_impl( o_, logical, vctx ); \
	if( q ) \
		return q;

///////////////////////////////////////////////////////////
// type_spec
bool type_spec::extrinsic_free() const
{
	return !shared.is_set()
		&& !optional.is_set()
		&& !lvalue.is_set();
}

///////////////////////////////////////////////////////////
// intr_type_rational
int intr_type_rational::compare_impl( intr_type const& o, bool logical, util::validation_context & vctx ) const
{
	COMPARE_SUPER(intr_type_number,o,logical,vctx);
	auto & ov = static_cast<intr_type_rational const&>(o);

	if( integer_signed != ov.integer_signed )
		return integer_signed < ov.integer_signed ? -1 : 1;
		
	bool a = integer_bitsize;
	bool b = ov.integer_bitsize;
	if( a != b )
		return a < b ? -1 : 1;
	if( a )
		return int(*integer_bitsize) - int(*ov.integer_bitsize);
	return 0;
}

///////////////////////////////////////////////////////////
// intr_type_fixed_number
int intr_type_fixed_number::compare_impl( intr_type const & o, bool logical, util::validation_context & vctx ) const
{
	COMPARE_SUPER(intr_type_number,o,logical,vctx)
	auto & ov = static_cast<intr_type_fixed_number const &>(o);
		
	if( data_bitsize != ov.data_bitsize )
		return data_bitsize < ov.data_bitsize ? -1 : 1;
		
	return 0;
}

void intr_type_fixed_number::validate_impl(util::validation_context & vctx) const
{
	intr_type::validate_impl(vctx);
	STATE_CHECK( data_bitsize > 0 );
}

void intr_type_float::validate_impl(util::validation_context & vctx) const
{
	intr_type_fixed_number::validate_impl(vctx);
	STATE_CHECK( platform::if_get_float( data_bitsize ) );
}

///////////////////////////////////////////////////////////
// intr_type_compound
int intr_type_compound::compare_impl( intr_type const & o, bool logical, util::validation_context & vctx ) const
{
	VALIDATE_RECURSE(0)
	COMPARE_SUPER(intr_type,o,logical,vctx)
	auto & co = static_cast<intr_type_compound const&>(o);
	
	if( sub.size() != co.sub.size() )
		return sub.size() < co.sub.size() ? -1 : 1;
	
	for( size_t i=0; i < sub.size(); ++ i )
	{
		if( sub[i].name != co.sub[i].name )
			return sub[i].name < co.sub[i].name ? -1 : 1;
			
		int q = sub[i].type.compare_impl( co.sub[i].type, logical, vctx );
		if( q )
			return q;
	}

	return 0;
}

bool intr_type_compound::is_fixated_impl(util::validation_context & vctx) const
{
	VALIDATE_RECURSE(true)
	
	for( auto & s : sub ) {
		if( !s.type.is_fixated_impl(vctx) )
			return false;
	}
	return true;
}

bool intr_type_compound::is_complete_impl(util::validation_context & vctx) const
{
	VALIDATE_RECURSE(true)
	
	if( !intr_type::is_complete_impl(vctx) )
		return false;
		
	for( auto & s : sub ) {
		if( !s.type.is_complete_impl(vctx) )
			return false;
	}
	
	return true;
}

void intr_type_compound::validate_impl(util::validation_context & vctx) const
{
	VALIDATE_RECURSE()
	
	intr_type::validate_impl(vctx);
	for( auto & s : sub )
		s.type.validate_impl(vctx);
}

boost::optional<size_t> intr_type_compound::find_name( std::string const & name ) const 
{
	for( size_t i=0; i < sub.size(); ++i ) {
		if( name == sub[i].name )
			return i;
	}
	return boost::optional<size_t>();
}

///////////////////////////////////////////////////////////
// intr_type_value_list
//CPP: not really needed, but GCC requires some function to appear somewhere!
void intr_type_value_list::validate_impl(util::validation_context & vctx) const
{
	intr_type_compound::validate_impl(vctx);
}

///////////////////////////////////////////////////////////
// intr_type_tuple
void intr_type_tuple::validate_impl(util::validation_context & vctx) const
{
	intr_type_compound::validate_impl(vctx);
	STATE_CHECK( pack >= p_normal && pack <= p_byte );
}

int intr_type_tuple::compare_impl( intr_type const & o, bool logical, util::validation_context & vctx ) const
{
	COMPARE_SUPER( intr_type_compound, o, logical, vctx );
	
	auto & tt = static_cast<intr_type_tuple const&>(o);
	if( pack != tt.pack )
		return pack < tt.pack ? -1 : 1;
	return 0;
}

void intr_type_tuple::complete_with( intr_type_tuple const & o )
{
	STATE_CHECK( !local_complete );
	selectable = o.selectable;
	pack = o.pack;
	sub = o.sub;
	local_complete = true;
}


///////////////////////////////////////////////////////////
// intr_type_wrap
int intr_type_wrap::compare_impl( intr_type const & o, bool logical, util::validation_context & vctx ) const
{
	COMPARE_SUPER( intr_type, o, logical, vctx )
	
	auto & wo = static_cast<intr_type_wrap const&>(o);
	return sub.compare_impl( wo.sub, logical, vctx );
}

void intr_type_wrap::validate_impl(util::validation_context & vctx) const
{
	VALIDATE_RECURSE()
	intr_type::validate_impl(vctx);
	sub.validate_impl(vctx);
}

///////////////////////////////////////////////////////////
// intr_type_array
int intr_type_array::compare_impl( intr_type const & o, bool logical, util::validation_context & vctx ) const
{
	COMPARE_SUPER( intr_type_wrap, o, logical, vctx )
	
	auto & ao = static_cast<intr_type_array const&>(o);
	if( ao.raw != raw ) {
		return raw < ao.raw ? -1 : 1;
	}
	return 0;
}


///////////////////////////////////////////////////////////
// intr_type_function
int intr_type_function::compare_impl( intr_type const & o, bool logical, util::validation_context & vctx ) const
{
	VALIDATE_RECURSE(0)
	COMPARE_SUPER( intr_type, o, logical, vctx )
	auto & fo = o.cast<intr_type_function>();
	
	if( access != fo.access )
		return access < fo.access ? -1 : 1;
		
	q = params->compare_impl( *fo.params, logical, vctx );
	if( q )
		return q;
		
	q = returns->compare_impl( *fo.returns, logical, vctx );
	if( q )
		return q;
		
	if( context && !fo.context ) {
		return 1;
	} else if( !context && fo.context ) {
		return -1;
	} else if( context && fo.context ) {
		return context->compare_impl( *fo.context, logical, vctx );
	}
	
	return 0;
}

bool intr_type_function::is_fixated_impl(util::validation_context & vctx) const
{
	VALIDATE_RECURSE(true)
	
	if( !params->is_fixated_impl(vctx) || !returns->is_fixated_impl(vctx) )
		return false;
	if( context && !context->is_fixated_impl(vctx) )
		return false;
	return true;
}
	
void intr_type_function::validate_impl(util::validation_context & vctx) const
{
	VALIDATE_RECURSE()
	
	intr_type::validate_impl(vctx);
	params->validate_impl(vctx);
	returns->validate_impl(vctx);
	if( access == a_unbound ) {
		STATE_CHECK(context);
		context->validate_impl(vctx);
		//the ir layer expects these to be only custom types
		STATE_CHECK( context->is_custom() );
	} else {
		STATE_CHECK(!context);
	}
}

///////////////////////////////////////////////////////////
// intr_type_scope
int intr_type_scope::compare_impl( intr_type const & o, bool /*logical*/, util::validation_context & ) const
{
	intptr_t idf = distinct_id() - o.distinct_id();
	if( idf != 0 )
		return idf < 0 ? -1 : 1;
	return 0;
}

void intr_type_scope::validate_impl(util::validation_context & vctx) const
{
	VALIDATE_RECURSE()
	intr_type::validate_impl(vctx);
	STATE_CHECK( context );
}

} //eon leaf