#include "lang/scope.hpp"
#include "lang/var_decl.hpp"

#define SHOW_DEBUG 0
#include "util/debug.hpp"

namespace leaf
{

scope::scope( scope_style_t ss)
{
	scope_style = ss;
	function_root = false;
	function_bypass = false;
	class_root = false;
	base_type = base_type_t(0);
	fref = 0;
}

void scope::set_parent( shared_ptr<scope> parent )
{
	//allow calling set_parent with same parent again (simplifies typer logic)
	PRE_CHECK( !base_type || (base_type == bt_parent && this->base.get() == parent.get()) );
	this->base = parent;
	base_type = bt_parent;
}

void scope::set_extends( shared_ptr<scope> base )
{
	//idempotent
	if( base_type == bt_extends && this->base.get() == base.get() )
		return;	
	PRE_CHECK( !base_type );
	this->base = base;
	base_type = bt_extends;
}

void scope::set_continues( shared_ptr<scope> base )
{
	PRE_CHECK( !base_type );
	this->base = base;
	base_type = bt_continues;
}

void scope::set_namespace( shared_ptr<scope> base, std::string name )
{
	PRE_CHECK( !base_type );
	this->base = base;
	base_type = bt_namespace;
	scope_name = name;
}

void scope::mark_function_root()
{
	PRE_CHECK( !class_root );
	function_root = true;
}

void scope::mark_function_bypass()
{
	PRE_CHECK( !class_root );
	function_root = true;
	function_bypass = true;
}

void scope::mark_class_root( shared_ptr<intr_type_custom> context, shared_ptr<statement_block> init_block )
{
	PRE_CHECK( !function_root )
	class_root = true;
	this->context = context;
	this->init_block = init_block;
}

scope_ref scope::add_single( shared_ptr<var_decl> decl, unsigned flags )
{
	PRE_CHECK( decl->form == var_decl::f_single || decl->form == var_decl::f_param
		|| decl->form == var_decl::f_return );
	PRE_CHECK( !decl->var_scope );
	var_vector_t & vec = var_map[decl->symbol];
	bool is_anon = decl->form == var_decl::f_return && decl->symbol.empty();
	if( !is_anon )
		STATE_CHECK( vec.size() == 0, "duplicate symbol", error_symbol( decl->symbol ) );

	decl->var_scope = this;
	
	scope_ref sr;
	sr.parent_level = 0;
	sr.root_depth = get_level();
	sr.decl = decl;
	
	if( flags & add_in_order )
		in_order.push_back( sr );
	//anonymous cannot be referenced by name
	if( !is_anon )
		vec.push_back( sr );
	
	return sr;
}
		
scope_ref scope::add_multi( shared_ptr<var_decl> decl, unsigned flags )
{
	PRE_CHECK( decl->form == var_decl::f_multi );
	PRE_CHECK( !decl->var_scope );
	var_vector_t & vec = var_map[decl->symbol];
	STATE_CHECK( vec.size() == 0 || vec.front().decl->form == var_decl::f_multi,
		"mixed bound symbol", error_symbol( decl->symbol ) );

	decl->var_scope = this;
	scope_ref sr;
	sr.parent_level = 0;
	sr.root_depth = get_level();
	sr.decl = decl;
	
	if( flags & add_in_order )
		in_order.push_back( sr );
	vec.push_back( sr );
	
	return sr;
}

shared_ptr<intr_type_value_list> scope::get_return_type() const
{
	auto vt = make_shared<intr_type_value_list>();
	for( auto & sr : in_order ) {
		if( sr.decl->form == var_decl::f_return )
			vt->push_sub( sr.decl->symbol, sr.decl->type );
	}
	return vt;
}

scope_ref scope::capture_var( scope_ref sr, scope const * relative_to )
{
	if( relative_to && relative_to != this ) {
		auto p = get_parent().get();
		while( true ) {
			PRE_CHECK( p );
			sr.parent_level++;
			if( p == relative_to )
				break;
			p = p->get_parent().get();
		}
	}
	
	//don't need to capture local and global references
	if( sr.parent_level == 0 || sr.root_depth == 0 )
		return sr;
		
	//capturing is always done in the base
	if( base_type == bt_extends )
		return base->capture_var( sr );
		
	//if already captured
	auto current = capture_map.find(sr.decl.get());
	if( current != capture_map.end() ) {
		return current->second.local;
	}
		
	capture_ref cr;
	cr.local.parent_level = 1;
	cr.local.root_depth = sr.root_depth;
	cr.local.scope_src_decl = reinterpret_cast<intptr_t>(sr.decl.get());
	
	cr.local.decl.reset( new var_decl( *sr.decl ) );
	cr.local.decl->form = var_decl::f_capture;
	cr.local.decl->var_scope = this;
		
	PRE_CHECK( base_type == bt_parent );
	sr.parent_level--;
	cr.src = base->capture_var( sr );

	capture_map[sr.decl.get()] = cr;
	return cr.local;
}

scope_ref scope::uncapture_var( scope_ref sr ) const
{
	//don't need to capture local and global references
	if( sr.parent_level == 0 || sr.root_depth == 0 )
		return sr;
		
	if( base_type == bt_extends )
		return base->uncapture_var( sr );
		
	auto it = capture_map.find(reinterpret_cast<var_decl*>(sr.scope_src_decl));
	PRE_CHECK( it != capture_map.end(), "not-captured" );
	
	PRE_CHECK( base_type == bt_parent );
	auto ret = base->uncapture_var( it->second.src );
	ret.parent_level++;
	return ret;
}


scope_ref scope::import( scope_ref sr )
{
	STATE_CHECK( sr.decl->var_scope );
	auto src = sr.decl->var_scope->uncapture_var( sr );
	return capture_var( src );
}

scope_ref scope::get_var( std::string const & name ) const
{
	var_map_t::const_iterator it = var_map.find( name );
	var_vector_t const * vec = (it == var_map.end()) ? nullptr : &it->second;
	
	if( !vec || vec->empty() )
	{
		if( !base_type ) {
			STATE_FAIL( "unknown symbol", error_symbol( name ) );
		}
		
		scope_ref sr = base->get_var( name );
		if( base_type == bt_parent )
			sr.parent_level++;
		return sr;
	}
		
	if( vec->size() > 1 )
	{
		//QUESTION: if there are also matching symbols in parent, but only 1 here, is it sitll ambiguous?
		STATE_FAIL( "ambiguous symbol", error_symbol( name ) );
	}
	
	return vec->front();
}

scope::var_vector_t scope::get_local_vars( std::string const & name ) const
{
	var_vector_t ret;
	var_map_t::const_iterator it = var_map.find( name );
	if( it != var_map.end() )
		ret = it->second;
	return ret;
}
	
scope::var_vector_t scope::get_local_decls() const
{
	var_vector_t ret;
	for( auto & var_list : var_map ) {	
		for( auto & var : var_list.second ) {
			if( var.parent_level != 0 )
				continue;
			if( var.decl->is_first_class() )
				ret.push_back( var );
		}
	}
	return ret;
}

scope::var_vector_t scope::get_vars( std::string const & name ) const
{
	auto ret = get_local_vars(name);
	//a more local "single" var also hides higher definitions
	//TODO: perhaps any locals, even multis, by default block parent definitions (not extends though)
	if( ret.size() == 1 && ret[0].decl->form != var_decl::f_multi ) {
		return ret;
	}
	
	if( base_type ) {
		var_vector_t const & p = base->get_vars( name );
		for( size_t i=0; i < p.size(); ++i ) {
			scope_ref sr = p[i];
			if( base_type == bt_parent )
				sr.parent_level++;
			ret.push_back( sr );
		}
	}

	return ret;
}

unsigned scope::get_level() const
{
	if( base_type == bt_continues || base_type == bt_namespace )
		return base->get_level();
	else if( base_type )
		return base->get_level() + 1;
	return 0;
}

void scope::add_fun_type( shared_ptr<intr_type const> type )
{
	PRE_CHECK( !type->name_scope );
	PRE_CHECK( type->distinct_name );
	PRE_CHECK( type->distinct );
	auto it = fun_type_map.find( *type->distinct_name );
	PRE_CHECK( it == fun_type_map.end(), "duplicate-type" );
	fun_type_map[*type->distinct_name] = type;
	type->name_scope = this;
}

shared_ptr<intr_type const> scope::if_get_fun_type( std::string const & name, unsigned flags ) const
{
	auto it = fun_type_map.find( name );
	if( it != fun_type_map.end() )
		return it->second;

	if( base_type && !(flags & gf_strict_local) ) 
		return base->if_get_fun_type( name );
	else
		return shared_ptr<intr_type const>();
}

void scope::add_type_alias( std::string const & name, shared_ptr<type_spec const> spec )
{
	auto it = type_alias_map.find( name );
	PRE_CHECK( it == type_alias_map.end(), "duplicate-type-alias" );
	//TODO: should be exclusive with type name?
	type_alias_map[name] = spec;
}

shared_ptr<type_spec const> scope::get_type_alias( std::string const & name, unsigned flags )
{
	auto t = if_get_type_alias( name, flags );
	if( !t )
		PRE_FAIL( std::string( "undefined-type-alias: " ) + name );
	return t;
}

shared_ptr<type_spec const> scope::if_get_type_alias( std::string const & name, unsigned flags )
{
	auto it = type_alias_map.find( name );
	if( it != type_alias_map.end() )
		return it->second;
		
	if( base_type && !(flags & gf_strict_local) )
		return base->if_get_type_alias( name );
	else
		return shared_ptr<type_spec const>();
}


scope const * scope::if_get_function_root() const
{
	if( function_root )
		return this;
	return base->if_get_function_root();
}

scope const & scope::get_function_root() const
{
	if( function_root )
		return *this;
		
	PRE_CHECK( base_type );
	return base->get_function_root();
}

std::string scope::qualify_name( std::string const & local, std::string const & join ) const
{
	if( base_type == bt_namespace ) {
		return base->qualify_name( scope_name + join + local, join );
	} else if( base ) {
		return base->qualify_name( local, join );
	}
	return local;
}

} //eon leaf