#include "lang/serial.hpp"
#include "lang/language.hpp"

#define SHOW_DEBUG 1
#include "util/debug.hpp"

namespace leaf {

///////////////////////////////////////////////////////////
// serial_out
serial_out::serial_out( std::ostream & out ) 
	: out(out)
{
}

void serial_out::add_extr_type( extr_type_ref t )
{
	write( serial_item::t_extr_type );
	t.write(*this);
}

void serial_out::add_var_decl( var_decl const & vd )
{
	auto id = reinterpret_cast<intptr_t>(&vd);
	write( serial_item::t_var_decl );
	write( id );
	write( vd.form );
	write( vd.import_name );
	write( vd.import_type );
	write( vd.import_module );
	write( vd.name_form );
	write( vd.symbol );
	vd.type.write( *this );
}

void serial_out::add_intr_type( intr_type const & t )
{
	auto id = reinterpret_cast<intptr_t>(&t);
	known_intr_types.insert( id );
	write( serial_item::t_intr_type );
	write( id );
	write( t.get_fun_class() );
	t.write(*this);
}

void serial_out::done() 
{
	while( pending_intr_types.size() ) {
		pending_intr_types_t sw;
		pending_intr_types.swap(sw);
		
		for( auto tptr : sw ) {
			add_intr_type( *tptr );
		}
	}
	write( serial_item::t_end_marker );
}

void serial_out::write_intr_type_ref( intr_type const & t )
{
	auto id = reinterpret_cast<intptr_t>(&t);
	write( id );
	write( t.get_fun_class() );
	//each type should only be written once
	if( known_intr_types.find( id ) == known_intr_types.end() ) {
		pending_intr_types.push_back( &t );
		known_intr_types.insert( id );
	}
}

void serial_out::write_intr_type_ref( shared_ptr<intr_type const> t )
{
	if( t ) {
		write_intr_type_ref( *t );
	} else {
		intptr_t id = 0;
		write( id );
	}
}

void serial_out::write( std::string const & t )
{	
	write( t.size() );
	out.write( t.data(), t.size() );
}


///////////////////////////////////////////////////////////
// serian_in
serial_in::serial_in( std::istream & in )
	: in( in )
{
	load_all_items();
}
	
extr_type_ref serial_in::read_primary_extr_type()
{
	STATE_CHECK( loaded_items.size() > 0 );
	STATE_CHECK( loaded_items[0].type == serial_item::t_extr_type );
	return loaded_items[0].v_extr_type_ref;
}

void serial_in::load_all_items()
{	
	bool done = false;
	while( !done ) {
		serial_item sitem;
		read( sitem.type );

		switch( sitem.type ) {
			case serial_item::t_end_marker:
				done = true;
				break;
				
			case serial_item::t_extr_type:
				sitem.v_extr_type_ref.read( *this );
				break;
				
			case serial_item::t_intr_type:
				sitem.v_intr_type = read_intr_type();
				break;
				
			case serial_item::t_var_decl:
				sitem.v_var_decl = read_var_decl();
				break;
				
			default:
				STATE_FAIL( "corrupt-serial" );
		}
		
		if( !done )
			loaded_items.push_back( sitem );
	}
	
	STATE_CHECK( intr_type_pending.empty() );
}

shared_ptr<intr_type> serial_in::read_intr_type()
{
	intptr_t id;
	read(id);
	intr_type::fun_class_t cl;
	read(cl);
		
	shared_ptr<intr_type> res;
	auto p = intr_type_pending.find(id);
	if( p != intr_type_pending.end() ) {
		intr_type_pending.erase( p );
		res = p->second;
	} else {
		res = create_intr_type(cl,id);
	}
		
	STATE_CHECK( res->get_fun_class() == cl );
	res->read(*this);
	return res;
}

shared_ptr<intr_type> serial_in::read_intr_type_ref()
{
	intptr_t id;
	read(id);
	if( !id )
		return shared_ptr<intr_type>();
		
	intr_type::fun_class_t cl;
	read(cl);
	
	auto known = intr_type_known.find(id);
	if( known != intr_type_known.end() ) 
		return known->second;
	
	auto load = create_intr_type(cl,id);
	intr_type_pending[id] = load;
	return load;
}

shared_ptr<intr_type> serial_in::create_intr_type(intr_type::fun_class_t cl, intptr_t id )
{
	auto known = intr_type_known.find(id);
	STATE_CHECK( known == intr_type_known.end() );
	
	shared_ptr<intr_type> load;
	switch( cl ) {
		#define FUN_CLASS_TYPE(s) case intr_type::f_##s: \
			load = make_shared<intr_type_##s>(); \
			break;
		INTR_TYPE_FUN_CLASSES
		#undef FUN_CLASS_TYPE
	}
	intr_type_known[id] = load;
	return load;
}

void serial_in::read( std::string & t )
{
	size_t len;
	read( len );
	t.resize( len );
	in.read( (char*)t.data(), len );
}

shared_ptr<var_decl> serial_in::read_var_decl()
{
	intptr_t id;
	read(id);
	auto vd = make_shared<var_decl>();
	read( vd->form );
	read( vd->import_name );
	read( vd->import_type );
	read( vd->import_module );
	read( vd->name_form );
	read( vd->symbol );
	vd->type.read(*this);
	return vd;
}

} //eon leaf
