#include "lang/serial.hpp"
#include "lang/intr_type.hpp"
#include "lang/variable_context.hpp"
#include "lang/var_decl.hpp"

#define SHOW_DEBUG 1
#include "util/debug.hpp"

namespace leaf {

///////////////////////////////////////////////////////////
// intr_type
void intr_type::write( serial_out & so ) const {
	//the header info, class/pointer is handled by serial_out
	so.write(distinct);
	so.write(distinct_name);
	so.write(local_complete);
}

void intr_type::read( serial_in & si ) {
	si.read(distinct);
	si.read(distinct_name);
	si.read(local_complete);
}

///////////////////////////////////////////////////////////
// intr_type_fixed
void intr_type_fixed_number::write( serial_out & so ) const {
	super::write(so);
	so.write(data_bitsize);
}

void intr_type_fixed_number::read( serial_in & si ) {
	super::read(si);
	si.read(data_bitsize);
}

///////////////////////////////////////////////////////////
// extr_type_ref
void extr_type_ref::write( serial_out & so ) const {
	so.write(lvalue);
	so.write(optional);
	so.write(lossy);
	so.write(cvt_optional);
	so.write(reference);
	
	bool hf = has_fun();
	so.write(hf);
	if( hf )
		so.write_intr_type_ref( *get_fun() );
}

void extr_type_ref::read( serial_in & si ) {
	si.read(lvalue);
	si.read(optional);
	si.read(lossy);
	si.read(cvt_optional);
	si.read(reference);
	
	fun.reset();
	bool hf;
	si.read(hf);
	if( hf )
		fun = si.read_intr_type_ref();
}

///////////////////////////////////////////////////////////
// intr_type_funtion
void intr_type_function::write( serial_out & so ) const {
	super::write(so);
	
	so.write_intr_type_ref( params );
	so.write_intr_type_ref( returns );
	so.write_intr_type_ref( context );
	so.write(access);
}

void intr_type_function::read( serial_in & si ) {
	super::read(si);
	
	params = si.read_intr_type_ref_known<intr_type_value_list>();
	returns = si.read_intr_type_ref_known<intr_type_value_list>();
	context = si.read_intr_type_ref_known<intr_type_custom>();
	si.read(access);
}

///////////////////////////////////////////////////////////
// intr_type_compount
void intr_type_compound::write( serial_out & so ) const {
	super::write(so);
	
	so.write( sub.size() );
	for( auto & sr : sub ) {
		so.write( sr.name );
		sr.type.write( so );
	}
}

void intr_type_compound::read( serial_in & si ) {
	super::read(si);
	
	size_t len;
	si.read( len );
	sub.resize( len );
	for( size_t i=0; i < len; ++i ) {
		si.read( sub[i].name );
		sub[i].type.read(si);
	}
}

///////////////////////////////////////////////////////////
// intr_type_tuple
void intr_type_tuple::write( serial_out & so ) const {
	super::write(so);
	
	so.write(pack);
	so.write(selectable);
}

void intr_type_tuple::read( serial_in & si ) {
	super::read(si);
	
	si.read(pack);
	si.read(selectable);
}

///////////////////////////////////////////////////////////
// intr_type_wrap
void intr_type_wrap::write( serial_out & so ) const {
	super::write(so);
	sub.write(so);
}

void intr_type_wrap::read( serial_in & si ) {
	super::read(si);
	sub.read(si);
}

///////////////////////////////////////////////////////////
// intr_type_array
void intr_type_array::write( serial_out & so ) const {
	super::write(so);
	so.write(raw);
}

void intr_type_array::read( serial_in & si ) {
	super::read(si);
	si.read(raw);
}

///////////////////////////////////////////////////////////
// intr_type_rational
void intr_type_rational::write( serial_out & so ) const {
	super::write(so);
	so.write(integer_bitsize);
	so.write(integer_signed);
}

void intr_type_rational::read( serial_in & si ) {
	super::read(si);
	si.read(integer_bitsize);
	si.read(integer_signed);
}

///////////////////////////////////////////////////////////
// intr_type_scope
void intr_type_scope::write( serial_out & so ) const {
	super::write(so);
	
	auto srs = context->get_local_decls();
	so.write( srs.size() );
	for( auto & sr : srs ) {
		STATE_CHECK( sr.parent_level == 0 );
		STATE_CHECK( sr.decl );
		//TODO: the tiny fraction of var_decl written indicates that perhaps var_decl is not
		//the correct structure for scoped decls...?
		auto & decl = *sr.decl;
		so.write( decl.symbol );
		so.write( decl.form );
		decl.type.write( so );
	}
}

void intr_type_scope::read( serial_in & si ) {
	super::read(si);
	
	size_t len;
	si.read(len);
	auto mctx = make_shared<managed_variable_context>();
	context = mctx;
	for( size_t i=0; i < len; ++i ) {
		auto decl = make_shared<var_decl>();
		si.read( decl->symbol );
		si.read( decl->form );
		decl->type.read(si);
		scope_ref sr;
		sr.parent_level = 0;
		sr.decl = decl;
		//sr.scope ??
		mctx->add_managed( sr );
	}
}

} //eon leaf
