#include "lang/type_converter.hpp"
#include "lang/intr_type.hpp"
#include "lang/scope.hpp"

#define SHOW_DEBUG 0
#include "util/debug.hpp"
#if SHOW_DEBUG
#include "lang/dump.hpp"
#endif

namespace leaf
{

namespace 
{
	cvt_expression const * skip_zero_cost( cvt_expression const * c )
	{
		while( c && (
			c->form == cvt_expression::c_set_distinct ||
			c->form == cvt_expression::c_addr_of ||
			c->form == cvt_expression::c_drop_lvalue ||
			c->form == cvt_expression::c_bind_raw_func ||
			c->form == cvt_expression::c_bind_func ||
			c->form == cvt_expression::c_collapse ||
			c->form == cvt_expression::c_field ||
			c->form == cvt_expression::c_deref_optional ||
			c->form == cvt_expression::c_direct ||
			c->form == cvt_expression::c_optional ||
			c->form == cvt_expression::c_unset_optional) ) {
			c = c->then.get();
		}
		return c;
	}
	
	int compare_cvt_expression( cvt_expression const * a, cvt_expression const * b ) 
	{
		auto next = [&]( bool consume, cvt_expression const * c ) {
			if( consume && c )
				c = c->then.get();
			return skip_zero_cost(c);
		};
		
		a = skip_zero_cost( a );
		b = skip_zero_cost( b );
		
		#define COST(v) (v ? v->form : cvt_expression::c_direct)
		
		bool deref_a = COST(a) == cvt_expression::c_deref;
		bool deref_b = COST(b) == cvt_expression::c_deref;
		if( deref_a != deref_b )
			return deref_a ? 1 : -1;
		a = next( deref_a, a );
		b = next( deref_b, b );

		//TODO: add lossy conversions (uncertain of cost, they must be explicit, but still distinguishable)
		
		bool fix_a = COST(a) == cvt_expression::c_fixate;
		bool fix_b = COST(b) == cvt_expression::c_fixate;
		if( !fix_a && !fix_b ) {
			//no cost
		} else if( !fix_a || !fix_b ) {
			//fixation costs more than not fixating
			return fix_a ? 1 : -1;
		} else if( a->to_fun != b->to_fun ) {
			return a->to_fun < b->to_fun ? -1 : 1;
		} else if( a->to_bitsize != b->to_bitsize ) {
			//natural size preferred
			int natural_size = a->to_fun == cvt_expression::p_float 
				? platform::natural_float_size : platform::natural_integer_size;
			if( a->to_bitsize == natural_size )
				return -1;
			if( b->to_bitsize == natural_size )
				return 1;
				
			//TODO:
			//otherwise the smallest (actually I've decided now there should be no preference beyond natural)
			//return a->to_bitsize < b->to_bitsize ? -1 : 1;
		} 
		a = next( fix_a, a );
		b = next( fix_b, b );
		
		bool promote_a = COST(a) == cvt_expression::c_promote;
		bool promote_b = COST(b) == cvt_expression::c_promote;
		if( !promote_a && !promote_b ) {
			//no cost
		} else if( !promote_a || !promote_b ) {
			//promotion costs more than not promotion
			return promote_a ? 1 : -1;
		} else if( a->to_fun != b->to_fun || a->from_fun != b->from_fun ) {
			bool nox_a = a->no_cross_promote();
			bool nox_b = b->no_cross_promote();
			if( nox_a && nox_b ) {
				//no cost
			} else if( nox_a != nox_b ) {
				//cross promotion costlier than not
				return nox_a ? -1 : 1;
			} else {
				return a->to_fun < b->to_fun ? -1 : 1;
			}
		} else if( a->to_bitsize != b->to_bitsize ) {
			//same promotion prefer lower bitsize
			return a->to_bitsize < b->to_bitsize ? -1 : 1;
		}
		a = next( promote_a, a );
		b = next( promote_b, b );
		
		bool pack_a = COST(a) == cvt_expression::c_pack || COST(a) == cvt_expression::c_pack_value;
		bool pack_b = COST(b) == cvt_expression::c_pack || COST(b) == cvt_expression::c_pack_value;
		if( pack_a || pack_b ) {
			size_t xsub = std::max( a ? a->sub.size() : 0, b ? b->sub.size() : 0 );
			int cmp = 0;
			for( size_t i=0; i < xsub; ++i ) {
				//in sub-matching the exact cost differences are not relevant, only whether each part
				//was less/more costly than the other
				int sub_cmp = compare_cvt_expression(
					(a && i < a->sub.size()) ? a->sub[i].cvt.get() : 0,
					(b && i < b->sub.size()) ? b->sub[i].cvt.get() : 0 );
				if( sub_cmp < 0 )
					cmp--;
				else if( sub_cmp > 0 )
					cmp++;
			}
			if( cmp )
				return cmp;
			//NOTE: packing type conversion doesn't cost anything
		}
		a = next( pack_a, a );
		b = next( pack_b, b );
		return 0;
	}
	
	shared_ptr<cvt_expression> make_cvt( cvt_expression::form_t form ) {
		return make_shared<cvt_expression>(form);
	}
	
	/**
		A simple wrapper to a conversion chain that allows tracking the start and end.
	*/
	struct converter {
		shared_ptr<cvt_expression> start;
		shared_ptr<cvt_expression> current;
	
		/**	
			@param all_none if non-null and a conversion happens then set to false,
				otherwise unchanged
		*/
		void extend( shared_ptr<cvt_expression> then, bool * all_none = nullptr ) {
			if( !then || then->is_none() )
				return;
				
			if( all_none )
				*all_none = false;
				
			if( !start )
				start = then;
			if( current )
				current->then = then;
			//must skip to end of chain
			current = then;
			while( current->then )
				current = current->then;
		}
		
		bool extend( type_converter::convert_t then, bool * all_none = nullptr ) {
			if( !then.okay() )
				return false;
			extend( then.conversions, all_none );
			return true;
		}
		
		#define CVT(type) \
			shared_ptr<cvt_expression> type() { \
				auto t = make_cvt( cvt_expression::c_##type ); \
				extend( t ); \
				return t; \
			}
		CVT(collapse)
		CVT(deref_optional)
		CVT(deref)
		CVT(drop_lvalue)
		CVT(optional)
		CVT(bind_raw_func)
		CVT(bind_func)
		CVT(fixate)
		CVT(promote)
		CVT(loss)
		CVT(field)
		CVT(set_distinct)
		CVT(addr_of)
		#undef CVT
		
		shared_ptr<cvt_expression> promote_loss( bool loss ) {
			auto t = make_cvt( loss ? cvt_expression::c_loss : cvt_expression::c_promote );
			extend(t);
			return t;
		}
	};
	
} //eon anon
int type_converter::convert_t::compare_cost( convert_t const & o ) const
{ 
	return compare_cvt_expression( conversions.get(), o.conversions.get() );
}

type_converter::type_converter( scope & sc )
	: sc( sc )
{
}

type_converter::~type_converter()
{
}

type_converter::convert_t type_converter::convert_list(
	intr_type_compound const& to, intr_type_compound const & from_, 
	unsigned sub_flags, cvt_expression::form_t form )
{
	converter cvt;
	
	intr_type_compound const * from = &from_;
	//1-value-lists within lists can be collapsed. We do only the outermost item here since
	//internal collapsing is done in convert_impl
	if( form == cvt_expression::c_pack_value ) {
		while( true ) {
			if( to.sub.size() == from->sub.size() )
				break;
				
			if( from->sub.size() != 1 || from->sub[0].type.get_fun_class() != intr_type::f_value_list )
				return cerr::mismatch_size;
				
			cvt.collapse();
			from = &from->sub[0].type.fun_cast<intr_type_value_list>();
		}
	} else {
		if( to.sub.size() != from->sub.size() )
			return cerr::mismatch_size;
	}
		
	auto pack = make_cvt( form );
	bool all_none = true;
	//check for pack type mismatch
	if( form == cvt_expression::c_pack ) {
		auto & left = to.cast<intr_type_tuple>();
		auto & right = from->cast<intr_type_tuple>();
		pack->pack = left.pack;
		//force unpack/pack if the packing doesn't match
		if( left.pack != right.pack )
			all_none = false;
	}
	
	for( size_t i=0; i < to.sub.size(); ++i )	{
		converter scvt;
		
		if( !i && from->sub.size() == 1 && form == cvt_expression::c_pack_value ) {
			scvt.collapse();
		} else {
			auto f = scvt.field();
			f->field = i;
		}
		
		auto n = convert_impl( to.sub[i].type, from->sub[i].type, sub_flags );
		if( !scvt.extend( n, &all_none ) )
			return n;
			
		std::string use_name = to.sub[i].name;
		//tuple conversions must match names if not anonymous (do at this level since in non-tuple
		//conversion the names don't matter)
		if( form == cvt_expression::c_pack && to.sub[i].name != from->sub[i].name ) {
			if( from->sub[i].name.empty() ) {
				all_none = false;
			} else {
				return cerr::mismatch_field_name;
			}
		}
		
		pack->sub.push_back( { scvt.start, to.sub[i].name } );
	}
	if( all_none ) {
		return cvt.start;
	}
		
#if 0
	//special reduction to reduce the form `c_pack_value {cvt->c_collapse}` to simply `cvt`
	if( pack->sub.size() == 1 ) {
		shared_ptr<cvt_expression> prev;
		auto last = pack->sub[0];
		while( last->then ) {
			prev = last;
			last = last->then;
		}
			
		if( last->form == cvt_expression::c_collapse ) {
			if( prev ) {
				prev->then.reset();
				pack = pack->sub[0];
			}
		}
	}
#endif
	cvt.extend( pack );
	return cvt.start;
}

type_converter::convert_t type_converter::convert_function_call( intr_type_value_list const & to,
	intr_type_value_list const & from )
{
	convert_t cost;

	if( from.sub.size() > to.sub.size() )
		return cerr::mismatch_size;
			
	auto pack = make_cvt( cvt_expression::c_pack_value );
	
	for( size_t i=0; i < to.sub.size(); ++i )	{
		if( i >= from.sub.size() ) {
			//potential optional parameter matching
			if( !to.sub[i].type.is_optional() )
				return cerr::mismatch_size;
				
			auto f = make_cvt(cvt_expression::c_unset_optional);
			f->target_type = to.sub[i].type;
			pack->sub.push_back( { f, to.sub[i].name } ) ;
		} else {
			auto f = make_cvt(cvt_expression::c_field);
			f->field = i;
			auto n = convert_impl( to.sub[i].type, from.sub[i].type, cf_allow_cvt );
			if( !n.okay() )
				return n;
			f->chain(n.conversions);
			pack->sub.push_back( { f, to.sub[i].name } );
		}
	}
	cost.conversions = pack;
	return cost;
}

type_converter::convert_t type_converter::convert_any_callable( extr_type_ref const & from_ )
{
	extr_type_ref const * from = &from_;
	converter cvt;
	
	while( true ) {
		auto vl = from->if_fun_cast<intr_type_value_list>();
		if( vl && vl->sub.size() == 1 ) {
			cvt.collapse();
			from = &vl->sub[0].type;
			continue;
		}
		break;
	}
	
	if( from->is_lvalue() ) {
		cvt.drop_lvalue();
	}
	if( from->is_optional() ) {
		cvt.deref_optional();
	}
	if( from->is_refer() ) {
		cvt.deref();
	}
		
	auto tuple = from->if_fun_cast<intr_type_tuple>();
	if( !from->is_function() && !(tuple && tuple->selectable) )	
		return cerr::requires_function;
		
	return cvt.start;
}

type_converter::convert_t type_converter::convert_any_common( extr_type_ref const & from_, 
	cerr list_fail, unsigned flags )
{
	extr_type_ref const * from = &from_;
	converter cvt;
	
	while( true ) {
		auto fvl = from->if_fun_cast<intr_type_value_list>();
		if( !fvl )
			break;
		if( !fvl->sub.size() )
			return list_fail;
		
		//collapse	
		if( fvl->sub.size() == 1 ) {
			cvt.collapse( );
			from = &fvl->sub[0].type;
			continue;
		}
		
		//or unpack/convert/pack parts
		bool all_none = true;
		auto pack = make_cvt( cvt_expression::c_pack_value );
		
		for( size_t i=0; i < fvl->sub.size(); ++ i ) {
			auto f = make_cvt(cvt_expression::c_field);
			f->field = i;
			auto n = convert_any_common( fvl->sub[i].type, list_fail, flags );
			if( !n.okay() )
				return n;
			if( f->chain(n.conversions) ) {
				all_none = false;
			}
			pack->sub.push_back( { f, "" } );
		}
		
		if( !all_none )
			cvt.extend( pack );
		return cvt.start;
	}

	//changes must be chained in a certain order, otherwise they aren't valid
	bool allow_cvt = true;
	bool lvalue_cvt = false;
	if( from->is_lvalue() ) {
		if( !(flags & acf_retain_lvalue) ) {
			cvt.drop_lvalue();
		} else {
			lvalue_cvt = true;
		}
	} else if( flags & acf_require_lvalue ) {
		return cerr::requires_lvalue;
	}
	
	if( from->is_optional() ) {
		if( !(flags & acf_retain_optional) ) {
			if( !allow_cvt )
				return cerr::mismatch_optional;
			cvt.deref_optional();
		} else {
			allow_cvt = false;
		}
	} else if( flags & acf_require_optional ) {
		return cerr::requires_optional;
	}
	
	if( from->is_shared() ) {
		if( !(flags & acf_retain_shared) ) {
			if( !allow_cvt )
				return cerr::mismatch_shared;
			cvt.deref();
		} else {
			allow_cvt = false;
		}
	} else if( flags & acf_require_shared ) {
		return cerr::requires_shared;
	}
	
	if( from->is_value_ptr() ) {
		if( !(flags & acf_retain_value_ptr) ) {
			if( !allow_cvt )
				return cerr::mismatch_value_ptr;
			cvt.deref();
		} else {
			allow_cvt = false;
		}
	}

	if( (flags & acf_require_refer) && !from->is_refer() ) {
		return cerr::requires_refer;
	}
	
	if( lvalue_cvt )
		allow_cvt = false;
		
	if( from->is_scalar() ) {
		//nop
	} else if( flags & acf_require_scalar ) {
		return cerr::requires_scalar;
	}
	
	return cvt.start;
}

type_converter::convert_t type_converter::convert_impl( extr_type_ref const & to, 
	extr_type_ref const & from_, unsigned flags )
{
	extr_type_ref const * from = &from_;
	
	DEBUG( "IMPL", dump::get_extr_type(to), dump::get_extr_type(from_) );
	if( to.is_value_list() ) {
		if( !from->is_value_list() )
			return cerr::require_value_list;
		return convert_list( to.fun_cast<intr_type_value_list>(), 
			from->fun_cast<intr_type_value_list>(), flags, cvt_expression::c_pack_value );
	}
 	
	converter cvt;
	bool allow_cvt = flags & cf_allow_cvt;

	//a source value-list of 1 however can just be removed
	while( true ) {
		auto from_vl = from->if_fun_cast<intr_type_value_list>();
		if( from_vl && from_vl->sub.size() == 1 ) {
			cvt.collapse();
			from = &from_vl->sub[0].type;
			continue;
		}
		break;
	}
	auto from_tuple = from->if_fun_cast<intr_type_tuple>();
	auto to_tuple = to.if_fun_cast<intr_type_tuple>();
	if( from_tuple && from_tuple->selectable
		&& (!to_tuple || !to_tuple->selectable) ) {
		return select_convert( to, *from, flags );
	}
	
	bool lvalue_cvt = false;
	bool post_addr_of = false;
	
	auto if_addr_of = [&]() {
		if( from->is_lvalue() ) {
			if( to.is_value_ptr() && !from->is_value_ptr() ) {
				post_addr_of = true;
			} else {
				cvt.drop_lvalue();
			}
		}
	};
	if( flags & cf_to_lvalue ) {
		if( !to.is_lvalue() )
			return cerr::requires_lvalue;
		if_addr_of();
		flags = flags & (~cf_to_lvalue);
	} else if( from->is_lvalue() != to.is_lvalue() ) {
		if( !allow_cvt ) {
			return cerr::mismatch_lvalue;
		} else if( !to.is_lvalue() ) {
			if_addr_of();
		} else {
			return cerr::requires_lvalue;
		}
	} else if( to.is_lvalue() ) {
		//converting to an lvalue always requires an exact match
		lvalue_cvt = true;
	}
	
	bool make_optional = false;
	if( to.is_optional() != from->is_optional() ) {
		if( !allow_cvt ) {
			return cerr::mismatch_optional;
		} else if( lvalue_cvt && to.is_optional() ) {
			return cerr::requires_optional;
		} else  if( !to.is_optional() ) {
			cvt.deref_optional();
		} else {
			//optional has to happen after whatever else changed
			//TODO: this should happen at end of conversion, it's distinct from the deref_optional
			make_optional = true;
		}
	} else if( to.is_optional() ) {
		//need a strict comparison now as optional from type cannot be converted
		//convertable optionals are however exempt form this restriction
		allow_cvt = from->is_cvt_optional();
	}
	
	if( flags & cf_to_refer ) {
		if( !to.is_refer() )
			return cerr::requires_refer;
		if( from->is_refer() )
			cvt.deref();
		flags = flags & (~cf_to_refer);
	} else if( to.is_shared() != from->is_shared() ) {
		if( !allow_cvt ) {
			return cerr::mismatch_shared;
		} else if( !to.is_shared() ) {
			cvt.deref();
		} else {
			return cerr::requires_shared;
		}
	} else if( !post_addr_of && to.is_value_ptr() != from->is_value_ptr() ) {
		//TODO: perhaps complete
		return cerr::mismatch_value_ptr;
	} else if( to.is_refer() ) {
		allow_cvt = false; //requires strict conversion now
	}
	
	if( post_addr_of )
		cvt.addr_of();
		
	//no more lvalue conversion steps
	if( lvalue_cvt )
		allow_cvt = false;
	
	//raw array conversions
	auto from_array = from->if_fun_cast<intr_type_array>(); 
	auto to_array = to.if_fun_cast<intr_type_array>();
	if( to_array && to_array->raw ) {
		if( from_array && from_array->raw ) {
			allow_cvt = false;
		} else if( allow_cvt && from_array && !from_array->raw ) {
			if( !from_array->sub.is_plain() )
				return cerr::invalid_array_cvt;
			if( from_array->sub != to_array->sub )
				return cerr::invalid_array_cvt;
			cvt.addr_of();
			allow_cvt = false;
			//from = &from_array->sub;
		}
	}
	
	bool need_distinct = false;
	if( to.is_distinct() ) {
		//non-distinct types may convert to compatible distinct types
		if( from->is_distinct() ) {
			if( from->get_fun()->distinct_id() != to.get_fun()->distinct_id() )
				return cerr::mismatch_defined;
				
			//skip rest since no conversion is necessary (required for recursion)
			goto skip_check;
		} else {
			//must do after all part conversions/checks
			need_distinct = true;
		}
	}
	
	{
		auto to_function = to.if_fun_cast<intr_type_function>();
		auto from_function = from->if_fun_cast<intr_type_function>();
		if( to_function && from_function && to_function->access != from_function->access ) {
			if( !allow_cvt )
				return cerr::mismatch_function_access;
				
			if( to_function->access == intr_type_function::a_bound
				&& from_function->access == intr_type_function::a_raw ) {
				cvt.bind_raw_func();
			} else if( to_function->access == intr_type_function::a_bound
				&& from_function->access == intr_type_function::a_unbound ) {
				cvt.bind_func();
			} else {
				return cerr::mismatch_function_access;
			}
		}
	}


	if( allow_cvt && to.is_integral() && from->is_rational() ) {
		auto & from_r = from->fun_cast<intr_type_rational>();
		if( !from_r.integer_bitsize )
			return cerr::range_loss;
		if( to.is_binary() && from_r.integer_signed )
			return cerr::mismatch_signed;
			
		auto & to_fnvt = to.fun_cast<intr_type_fixed_number>();
		if( *from_r.integer_bitsize > to_fnvt.data_bitsize )
			return cerr::range_loss;
			
		auto n = cvt.fixate();
		n->to_fun = to.is_integer() ? cvt_expression::p_integer : cvt_expression::p_binary;
		n->to_bitsize = to_fnvt.data_bitsize;
		
	} else if( allow_cvt && to.is_integral() && from->is_float() ) {
		if( !from->is_lossy() )
			return cerr::mismatch_fundamental;
			
		auto & to_fnvt = to.fun_cast<intr_type_fixed_number>();
		
		auto n = cvt.loss();
		n->from_fun = cvt_expression::p_float;
		n->to_fun = to.is_integer() ? cvt_expression::p_integer : cvt_expression::p_binary;
		n->to_bitsize = to_fnvt.data_bitsize;
		
	} else if( allow_cvt && to.is_binary() && from->is_integer() ) {
		if( !from->is_lossy() )
			return cerr::mismatch_fundamental;
			
		auto & to_fnvt = to.fun_cast<intr_type_fixed_number>();
		
		auto n = cvt.loss();
		n->from_fun = cvt_expression::p_integer;
		n->to_fun = cvt_expression::p_binary;
		n->to_bitsize = to_fnvt.data_bitsize;
			
	} else if( allow_cvt &&  to.is_float() && from->is_rational() ) {
		auto & to_fnvt = to.fun_cast<intr_type_fixed_number>();
		//TODO: check that range fits in the target type, we don't care about precision here

		auto n = cvt.fixate();
		n->to_fun = cvt_expression::p_float;
		n->to_bitsize = to_fnvt.data_bitsize;
		
	} else if( allow_cvt &&  to.is_float() && from->is_integral() ) {
		auto & from_nt= from->fun_cast<intr_type_fixed_number>();
		auto & to_nt = to.fun_cast<intr_type_float>();
		
		//-1 since one of those is the sign-bit whtih the float also has in addition to its significand
		auto nf = platform::if_get_float(to_nt.data_bitsize);
		if( !nf )
			return cerr::invalid_size;
		bool loss = (from_nt.data_bitsize-(from->is_integer() ? 1 : 0)) > nf->significand_size;
		if( loss && !from->is_lossy() )
			return cerr::precision_loss;
			
		auto n = cvt.promote_loss(loss);
		n->to_bitsize = to_nt.data_bitsize;
		n->from_fun = from->is_integer() ? cvt_expression::p_integer : cvt_expression::p_binary;
		n->to_fun = cvt_expression::p_float;

	} else if( allow_cvt &&  to.is_integer() && from->is_binary() ) {
		auto & to_fnvt = to.fun_cast<intr_type_fixed_number>();
		auto & from_fnvt = from->fun_cast<intr_type_fixed_number>();
		//requires 1-extra bit to store sign
		bool loss = from_fnvt.data_bitsize >= to_fnvt.data_bitsize;
		if( loss && !from->is_lossy() )
			return cerr::range_loss;
			
		auto n = cvt.promote_loss(loss);
		n->from_fun =cvt_expression::p_binary; 
		n->to_fun = cvt_expression::p_integer;
		n->to_bitsize = to_fnvt.data_bitsize;

	} else 	if( to.get_fun_class() != from->get_fun_class() ) {
		return cerr::mismatch_fundamental;
		
		//past here the fundamental is the same
	} else if( to.is_integral() || to.is_float() ) {
		auto & to_fnvt = to.fun_cast<intr_type_fixed_number>();
		auto & from_fnvt = from->fun_cast<intr_type_fixed_number>();
		if( to_fnvt.data_bitsize != from_fnvt.data_bitsize ) {
		
			if( !allow_cvt )
				return cerr::mismatch_data_size;

			bool loss = to_fnvt.data_bitsize < from_fnvt.data_bitsize;
			if( loss && !from->is_lossy() ) {
				return to.is_float() ? cerr::precision_loss : cerr::range_loss;
			}
			auto n = cvt.promote_loss(loss);
			n->to_bitsize = to_fnvt.data_bitsize;
			if( to.is_float() ) {
				n->from_fun = n->to_fun = cvt_expression::p_float;
			} else if( to.is_binary() ) {
				n->from_fun = n->to_fun = cvt_expression::p_binary;
			} else {
				n->from_fun = n->to_fun = cvt_expression::p_integer;
			}
		}
		
	} else if( to.is_tuple() ) {
		auto ci = convert_list( to.fun_cast<intr_type_tuple>(), 
			from->fun_cast<intr_type_tuple>(), allow_cvt ? cf_allow_cvt : 0, cvt_expression::c_pack );
		if( !cvt.extend( ci ) )
			return ci;
			
	} else if ( to.is_function() ) {
		//here the function itself is being converted, not the parameters for a function
		//call, thus only a very strict comparison can be allowed.
		auto & to_function = to.fun_cast<intr_type_function>();
		auto & from_function = from->fun_cast<intr_type_function>();
		//TODO: creating a extr_type_ref will become problematic
		auto ci = convert_impl( extr_type_ref(to_function.returns), extr_type_ref(from_function.returns), 0 );
		if( !cvt.extend( ci ) )
			return ci;
		ci = convert_impl( extr_type_ref(to_function.params), extr_type_ref(from_function.params), 0 );
		if( !cvt.extend( ci ) )
			return ci;
	}

skip_check:
	if( make_optional )
		cvt.optional();

	if( need_distinct ) {
		auto sd = cvt.set_distinct();
		sd->target_type = to;
	}
	
	return cvt.start;
}

type_converter::resolve_t type_converter::fixate( extr_type_ref const & from )
{
	resolve_t it;
	try {
		it.result = fixate_impl( from );
	} catch( cerr e ) {
		it.error = e;
	}
	return it;
}

/**
	Fixation always picks the type which would be inferred from the same type. This ensures
	consistency. It doesn't always guarantee the best conversion, but here the consistency
	is more important. Such generic fixation is only used when no other type is identified in
	the expression.
*/
extr_type_ref type_converter::fixate_impl( extr_type_ref const & from )
{
	if( from.is_rational() ) {
		auto & trat = from.fun_cast<intr_type_rational>();	
		if( trat.integer_bitsize ) {
			if( *trat.integer_bitsize > platform::natural_integer_size )
				throw cerr::large_rational_integer;
				
			auto n = make_shared<intr_type_integer>();
			n->data_bitsize = platform::natural_integer_size;
			return extr_type_ref(n);
		}
		
		auto n = make_shared<intr_type_float>();
		n->data_bitsize = platform::natural_float_size;
		return extr_type_ref(n);
	}

	if( from.is_value_list() || from.is_tuple() ) {
		shared_ptr<intr_type_compound> comp
			= from.fun_cast<intr_type_compound>().clone_derived();
		for( size_t i=0; i < comp->sub.size(); ++ i ) {
			comp->sub[i].type = fixate_impl( comp->sub[i].type );
		}
		return extr_type_ref(comp);
	}
	
	return from;
}

type_converter::resolve_t type_converter::unify( extr_type_ref const & a, extr_type_ref const & b )
{
	resolve_t ut;
	try {
		ut.result = unify_impl( a, b );
	} catch( cerr e ) {
		ut.error = e;
	}
	return ut;
}

type_converter::resolve_t type_converter::unify( extr_type_ref const * begin, size_t len )
{
	PRE_CHECK( len );
	if( len == 1 ) {
		resolve_t ut;
		ut.result = begin[0];
		return ut;
	}
	if( len == 2 ) {
		return unify( begin[0], begin[1] );
	}
	
	auto a = unify( begin, len / 2 );
	auto b = unify( begin + len/2, len - len/2 );
	if( a.error != cerr::okay )
		return a;
	if( b.error != cerr::okay )
		return b;
	return unify( a.result, b.result );
}

/**
	unify_impl is supposed to unify according to the rules of type conversion. I see no reasonable
	way to share the code from above to ensure that. Thus it'll just have to be covered with
	clear rules, good tests, and ultimately a very thorough specification.
	
	Note however that not all things allowed in type covnersion will be possible in type unification.
	At the moment this specifically includes (as opposed to just not yet done):
		- deref optional (optional is only added here)
*/
extr_type_ref type_converter::unify_impl( extr_type_ref const & a, extr_type_ref const & b )
{
	//DEBUG( "UNIFY", dump::get_extr_type(a), "::", dump::get_extr_type(b));
	//equivalent types are always fine
	if( a == b )
		return a;
	
	auto alist = a.if_fun_cast<intr_type_value_list>();
	auto blist = b.if_fun_cast<intr_type_value_list>();
	if( alist || blist ) {
		//TODO: what about names, must they match?!
		//1-lists can collapse
		if( !alist ) {
			if( blist->sub.size() != 1 )
				throw cerr::mismatch_size;
			return unify_impl( a, extr_type_ref(blist->sub[0].type) );
		}
		if( !blist ) {
			if( alist->sub.size() != 1 )
				throw cerr::mismatch_size;
			return unify_impl( extr_type_ref(alist->sub[0].type), b );
		}
		//if same size unify the components
		if( alist->sub.size() == blist->sub.size() ) {
			auto ulist = make_shared<intr_type_value_list>();
			for( size_t i=0; i < alist->sub.size(); ++ i )
				ulist->push_sub( unify_impl( alist->sub[i].type, blist->sub[i].type ) );
			return extr_type_ref(ulist);
		}
		//again, 1-list can collapse
		if( alist->sub.size() == 1 )
			return unify_impl( extr_type_ref(alist->sub[0].type), b );
		if( blist->sub.size() == 1 )
			return unify_impl( a, extr_type_ref(blist->sub[0].type) );
		
		throw cerr::mismatch_size;
	}
	
	//lvalue's must be exact (but failed previous exact test), but lvalue can be dropped
	if( a.is_lvalue() || b.is_lvalue() ) {
		auto ac = a;
		ac.set_lvalue( false );
		auto bc = b;
		bc.set_lvalue( false );
		return unify_impl( ac, bc );
	}
	
	if( a.is_optional() != b.is_optional() ) {
		if( a.is_optional() ) {
			auto bc = b;
			bc.set_optional( true );
			bc.set_cvt_optional( true );
			return unify_impl( a, bc );
		} else {
			auto ac = a;
			ac.set_optional( true );
			ac.set_cvt_optional( true );
			return unify_impl( ac, b );
		}
	}

	//if both sides are cvt_optional it may be retained (since a conversion can flow through to both)
	if( a.is_cvt_optional() != b.is_cvt_optional() ) {
		auto ac = a;
		ac.set_cvt_optional( false );
		auto bc = b;
		bc.set_cvt_optional( false );
		return unify_impl( ac, bc );
	}
	
	if( a.is_shared() != b.is_shared() ) {
		if( a.is_shared() ) {
			auto ac = a;
			ac.set_ref_value();
			return unify_impl( ac, b );
		} else {
			auto bc = b;
			bc.set_ref_value();
			return unify_impl( a, bc );
		}
	}
	
	auto arat = a.if_fun_cast<intr_type_rational>();
	auto brat = b.if_fun_cast<intr_type_rational>();
	if( a.get_fun_class() != b.get_fun_class() ) {
		//rationals always unify to the other type, even if they could not be unified to that
		//type. This provides a consistency in how types of unified/fixated.
		if( arat && b.is_scalar() ) {
			return b;
		} else if( brat && a.is_scalar() ) {
			return a;
		}
		
		throw cerr::mismatch_fundamental;
	}
	
	if( a.is_integral() || a.is_float() ) {
		auto anum = a.fun_cast<intr_type_fixed_number>();
		auto bnum = b.fun_cast<intr_type_fixed_number>();
		if( anum.data_bitsize < bnum.data_bitsize )
			return b;
		return a;
	}

	auto atuple = a.if_fun_cast<intr_type_tuple>();
	auto btuple = b.if_fun_cast<intr_type_tuple>();
	if( atuple ) {
		if( atuple->sub.size() != btuple->sub.size() )
			throw cerr::mismatch_size;
		
		auto ntuple = make_shared<intr_type_tuple>();
		for( size_t i=0; i < atuple->sub.size(); ++i ) {
			//names can be added but not removed
			std::string use_name = atuple->sub[i].name;
			if( use_name.empty() ) {
				use_name = btuple->sub[i].name;
			} else if( !btuple->sub[i].name.empty() && use_name != btuple->sub[i].name ) {
				throw cerr::mismatch_field_name;
			}
			ntuple->push_sub( use_name, unify_impl( extr_type_ref(atuple->sub[i].type), 
				extr_type_ref(btuple->sub[i].type )) );
		}
		DEBUG( "PT", dump::get_extr_type( extr_type_ref(ntuple) ) );
		return extr_type_ref(ntuple);
	}
	
	if( arat && brat ) {
		//mixing with non-integer becomes non-integer
		if( !arat->integer_bitsize )
			return a;
		if( !brat->integer_bitsize )
			return b;
			
		//otherwise an integer of the larger bitsize
		if( *arat->integer_bitsize > *brat->integer_bitsize )
			return a;
		else
			return b;
	}
	throw cerr::unify_impossible;
}

/**
	Does a conversion by selecting the sub element which has the best conversion cost.
*/
type_converter::convert_t type_converter::select_convert( extr_type_ref const & to, 
	extr_type_ref const & from, unsigned flags )
{
	auto & from_tuple = from.fun_cast<intr_type_tuple>();

	std::vector<convert_t> match;
	converter cvt;
	
	//TODO: this pattern is repeated, make generic
	if( from.is_lvalue() ) {
		cvt.drop_lvalue();
	}
	if( from.is_optional() ) {
		cvt.deref_optional();
	}
	if( from.is_refer() ) {
		cvt.deref();
	}
	
	for( size_t i=0; i < from_tuple.sub.size(); ++i ) {
		auto & tref = from_tuple.sub[i];
		auto cost = convert_impl( to, tref.type, flags );
		if( !cost.okay() )
			continue;

		if( !match.empty() ) {
			int cmp = match[0].compare_cost( cost );
			if( cmp > 0 ) { //better match
				match.clear();
			} else if( cmp < 0 ) { //same cost
				continue;
			}
		}
		
		//add field access
		auto field = make_shared<cvt_expression>(cvt_expression::c_field);
		field->field = i;
		field->then = cost.conversions;
		cost.conversions = field;
		match.push_back( cost );
	}
	
	if( match.empty() )
		return cerr::select_none;
	if( match.size() > 1 )
		return cerr::select_multiple;
	
	cvt.extend( match[0] );
	return cvt.start;
}


} //eon leaf