#include "lang/type_identifier.hpp"
#include "lang/intr_type.hpp"
#include "lang/scope.hpp"
#include "lang/var_decl.hpp"
#include "lang/variable_context.hpp"
#include "platform/platform.hpp"

#include <map>

#define SHOW_DEBUG 1
#include "util/debug.hpp"
#include "lang/dump.hpp"

namespace leaf
{

type_identifier::resolve_t type_identifier::determine( shared_ptr<type_spec const> vt, unsigned flags )
{
	if( !vt )
		return cerr::null_spec;
	return determine( *vt, flags );
}

namespace
{
	typedef std::map<std::string, type_spec> spec_map_t;
	spec_map_t spec_map;
		
	struct predefined_types_init
	{
		predefined_types_init() {
			{
				type_spec vts;
				vts.fun_class = intr_type::f_binary;
				vts.data_bitsize = 8;
				spec_map["octet"] = vts;
			}
			{
				type_spec vts;
				vts.fun_class = intr_type::f_binary;
				vts.data_bitsize = platform::abi_ptr_size;
				spec_map["abi_ptr"] = vts;
			}
			{
				type_spec vts;
				vts.fun_class = intr_type::f_binary;
				vts.data_bitsize = 8;
				spec_map["abi_char"] = vts;
			}
			{
				type_spec vts;
				vts.pack = intr_type_tuple::p_abi;
				spec_map["abi_pack"] = vts;
			}
			{
				type_spec vts;
				vts.pack = intr_type_tuple::p_byte;
				spec_map["byte_pack"] = vts;
			}
			
			simple( "rational", intr_type::f_rational );
			simple( "float", intr_type::f_float );
			simple( "integer", intr_type::f_integer );
			simple( "binary", intr_type::f_binary );
			simple( "char", intr_type::f_char );
			simple( "boolean", intr_type::f_boolean );
			
			native_int( "short", platform::abi_short_size );
			native_int( "int", platform::abi_int_size );
			native_int( "long", platform::abi_long_size );
			native_int( "long_long", platform::abi_long_long_size );
			
			float_( "abi_float", platform::abi_float_size );
			float_( "abi_double", platform::abi_double_size );
			float_( "abi_long_double", platform::abi_long_double_size );
		}
		
		/**
			The simple types by default have all these properties set to the default
			rather than being left just undefined. Explicit typing can override, but
			if not set the default is used. This is what allows "integer" on its own
			to be a full type, rather than just a fundamental declaration.
		*/
		void simple( char const * name, intr_type::fun_class_t f ) {
			type_spec vts;
			vts.fun_class = f;
			vts.optional.revert_default();
			vts.cvt_optional.revert_default();
			vts.value_ptr.revert_default();
			vts.shared.revert_default();
			vts.lvalue.revert_default();
			vts.signed_ = false;
			vts.pack.revert_default();
			vts.data_bitsize.revert_default();
			vts.precision_variant.revert_default();
			vts.lossy.revert_default();
			spec_map[name] = vts;
		}
		
		void float_( char const * name, int bit_size ) {
			auto nf = platform::if_get_float( bit_size );
			STATE_CHECK(nf);
			type_spec vts;
			vts.fun_class = intr_type::f_float;
			vts.precision_variant = nf->precision;
			spec_map[name] = vts;
		}
		
		void native_int( char const * name_bit, int bit_size ) {
			type_spec vts;
			vts.data_bitsize = bit_size;
			vts.fun_class = intr_type::f_integer;
			char name[64];
			sprintf( name, "abi_%s", name_bit );
			spec_map[name] = vts;
			
			//also the unsigned variant
			sprintf( name, "abi_u%s", name_bit );
			spec_map[name] = vts;
		}
		
	} predefined_types_init_;
	
	template<typename T>
	bool merge( T & to, T const & from ) {
		if( from.is_set() ) {
			if( to.is_set() )
				return false;
			to = from;
		} else if( from.is_default() ) {
			if( to.is_unset() )
				to = from;
		}
		
		return true;
	}
} //eon anon

cerr_base type_identifier::expand_spec( type_spec & vt, type_spec const & from )
{
	vt = from;
	return expand_spec( vt );
}

shared_ptr<intr_type const> type_identifier::resolve_type_name( 
	variable_context const & where, std::string const & name )
{
	DEBUG( "RESOLVE", name, &where );
	auto dot = name.find_first_of( '.' );
	if( dot != std::string::npos ) {
		auto ns_name = name.substr(0,dot);
		auto ns = where.get_vars( ns_name );
		DEBUG( "RESOLVE-NS", ns_name, ns.size() );
		if( ns.size() != 1 || !ns[0].is_set() || !ns[0].decl->type.has_fun() ) {	
			return shared_ptr<intr_type const>();
		}
		
		auto mtype = ns[0].decl->type.if_fun_cast<intr_type_module>();
		if( !mtype ) {
			return shared_ptr<intr_type const>();
		}
		
		return resolve_type_name( *mtype->context, name.substr( dot+1 ) );
	} else {
		return where.if_get_fun_type(name);
	}
}

shared_ptr<intr_type const> type_identifier::resolve_type_name( std::string const & name ) {
	return resolve_type_name( sc, name ); 
}

cerr_base type_identifier::expand_spec( type_spec & vt )
{
	//nothing to translate
	if(vt.symbols.size() )	{		
		for( auto & s : vt.symbols ) {
			auto alias = sc.if_get_type_alias( s );

			auto pt = spec_map.find( s );
			if( pt != spec_map.end() || alias ) {
				//merge the types
				auto & vts = alias ? *alias : pt->second;
				//DEBUG( "SPEC", dump::get_type_spec( vts ) );
				
				#define MERGE(field, err) if( !merge( vt.field, vts.field ) ) { return cerr_base(cerr::err, s); }
				MERGE(fun_class, redundant_fundamental)
				MERGE(data_bitsize,redundant_bitsize)
				MERGE(precision_variant,redundant_bitsize)
				MERGE(pack,redundant_pack)
				MERGE(lvalue,redundant_lvalue)
				MERGE(shared,redundant_shared)
				MERGE(signed_,redundant_signed)
				MERGE(function_access,redundant_function_access)
				MERGE(optional, redundant_optional)
				MERGE(cvt_optional,redundant_optional)
				MERGE(value_ptr,redundant_value_ptr)
				MERGE(selectable, redundant_selectable)
				MERGE(member, redundant_member)
				MERGE(lossy, redundant_lossy)
				#undef MERGE
				
				if( vts.sub.size() ) { 
					//TODO: should check fun_class, may be redundant to have sub parts
					if( vt.sub.size() ) 
						return cerr::redundant_fundamental;
					vt.sub = vts.sub;
				}
				
				if( vts.fun_prototype ) {
					if( vt.fun_prototype )
						return cerr_base( cerr::redundant_fundamental, s );
					vt.fun_prototype = vts.fun_prototype;
				}
				
				//DEBUG( "SPECOUT", dump::get_type_spec( vt ) );
				
			} else if( s == "signed" ) {
				vt.signed_ = true;
			} else {
				if( vt.fun_class.is_set() )
					return cerr_base( cerr::redundant_fundamental, s );
				auto sym = resolve_type_name( s );
				if( !sym )
					return cerr_base( cerr::unknown_identifier, s );
				vt.fun_prototype = sym;
				vt.optional.revert_default_if_unset();
				vt.cvt_optional.revert_default_if_unset();
				vt.shared.revert_default_if_unset();
				vt.value_ptr.revert_default_if_unset();
				vt.lvalue.revert_default_if_unset();
				vt.signed_.revert_default_if_unset();
				vt.pack.revert_default_if_unset();
				vt.data_bitsize.revert_default_if_unset();
				vt.function_access.revert_default_if_unset();
				vt.precision_variant.revert_default_if_unset();
				vt.selectable.revert_default_if_unset();
				vt.member.revert_default_if_unset();
				vt.lossy.revert_default_if_unset();
			}
		}
	}
	
	for( size_t i=0; i < vt.sub.size(); ++i ) {
		STATE_CHECK( vt.sub[i].type );
		auto err = expand_spec( *vt.sub[i].type );
		if( !err.okay() )
			return err;
	}
	
	vt.symbols.clear();
	return cerr::okay;
}

type_identifier::resolve_t type_identifier::determine( type_spec const & vt_, unsigned flags )
{
	type_spec vt;
	auto ex = expand_spec( vt, vt_ );
	if( !ex.okay() )
		return resolve_t(ex);
		
	if( !vt.fun_class.is_set() )
		vt.fun_class = intr_type::f_unknown;

	shared_ptr<intr_type const> ret = vt.fun_prototype; //may be null
	
	if( !ret ) {
		//GCC: appears wrong here, there is no way fundamental is not initialized (see if above)
		#pragma GCC diagnostic push
		#pragma GCC diagnostic ignored "-Wuninitialized"
		auto fund = *vt.fun_class;
		#pragma GCC diagnostic pop
		switch( fund ) {
			case intr_type::f_unknown:
				ret = make_shared<intr_type_unknown>();
				break;
				
			case intr_type::f_boolean:
				ret = make_shared<intr_type_boolean>();
				break;
				
			case intr_type::f_char:
				ret = make_shared<intr_type_char>();
				break;

			case intr_type::f_tuple: {
				auto tt = make_shared<intr_type_tuple>();
				if( vt.pack.is_set() )
					tt->pack = *vt.pack;
				if( vt.selectable.is_set() )
					tt->selectable = *vt.selectable;
					
				for( size_t i=0; i < vt.sub.size(); ++i ) {
					auto pt = determine( vt.sub[i].type, flags | df_sub_type );
					if( !pt.okay() )
						return pt;
					tt->push_sub( vt.sub[i].name, pt.result );
				}
				ret = tt;
				break;
			}
			
			case intr_type::f_value_list: {
				auto tt = make_shared<intr_type_value_list>();
				for( size_t i=0; i < vt.sub.size(); ++i ) {
					auto pt = determine( vt.sub[i].type, flags ); //do not add df_sub_type here (it may already be applied however)
					if( !pt.okay() )
						return pt;
					tt->push_sub( vt.sub[i].name, pt.result );
				}
				ret = tt;
				break;
			}
			
			case intr_type::f_function: {
				auto ft = make_shared<intr_type_function>();
				if( vt.function_access.is_set() ) {
					ft->access = *vt.function_access;
					ft->context = vt.function_context;
				} else if( vt.member.is_set() && *vt.member ) {
					ft->access = intr_type_function::a_unbound;
					ft->context = sc.get_class_scope_context();
				}
				if( vt.sub.size() != 2 )
					return cerr::mismatch_size;
				
				auto params = determine( vt.sub[0].type, df_sub_type );
				if( !params.okay() )
					return params;
				if( !params.result.is_value_list() )
					return cerr::require_value_list;
				ft->params = std::static_pointer_cast<intr_type_value_list const>(params.result.get_fun());
				
				auto returns = determine( vt.sub[1].type, df_sub_type );
				if( !returns.okay() )
					return returns;
				if( !returns.result.is_value_list() )
					return cerr::require_value_list;
				ft->returns = std::static_pointer_cast<intr_type_value_list const>(returns.result.get_fun());
				
				ret = ft;
				break;
			}
			
			case intr_type::f_array: {
				if( vt.sub.size() != 1 )
					return cerr::mismatch_size;
				auto at = make_shared<intr_type_array>();
				if( vt.function_access.is_set() && *vt.function_access == intr_type_function::a_raw ) {
					at->raw = true;
				}
				auto rt = determine( vt.sub[0].type, df_sub_type );
				if( !rt.okay() )
					return rt;
				at->sub = rt.result;
				ret = at;
				break;
			}

			case intr_type::f_integer:
			case intr_type::f_binary: {
				shared_ptr<intr_type_fixed_number> it;
				if( fund == intr_type::f_integer )
					it = make_shared<intr_type_integer>();
				else
					it = make_shared<intr_type_binary>();
					
				if( vt.data_bitsize.is_set() ) {
					it->data_bitsize = *vt.data_bitsize;
				} else {
					it->data_bitsize = platform::natural_integer_size;
				}
				ret = it;
				break;
			}
			
			case intr_type::f_float: {
				auto ft = make_shared<intr_type_float>();
				if( vt.data_bitsize.is_set() ) {
					if( vt.precision_variant.is_set() )
						return cerr::redundant_bitsize;
					ft->data_bitsize = *vt.data_bitsize;
					if( !platform::if_get_float( ft->data_bitsize ) )
						return cerr::unknown_variant;
				} else if( vt.precision_variant.is_set() ) {
					auto nf = platform::if_get_float_precision( *vt.precision_variant );
					if( !nf )
						return cerr::unknown_variant;
					ft->data_bitsize = nf->size;
				} else {
					ft->data_bitsize = platform::natural_float_size;
				}
				ret = ft;
				break;
			}
			
			case intr_type::f_rational: {
				auto rt = make_shared<intr_type_rational>();
				if( vt.data_bitsize.is_set() )
					rt->integer_bitsize = *vt.data_bitsize;
				if( vt.signed_.is_set() )
					rt->integer_signed = *vt.signed_;
				ret = rt;
				break;
			}
			
			case intr_type::f_module:
			case intr_type::f_custom: {
				STATE_FAIL( "not-expected-here" );
			}
		}
	}
	STATE_CHECK(ret);

	extr_type_ref tr(ret);
	
	//TODO: if defined the lvalue/shared cannot be removed (well, clarify...)
	if( vt.lvalue.is_set() )
		tr.set_lvalue( *vt.lvalue );
	if( vt.shared.is_set() ) {
		if( *vt.shared ) {
			tr.set_ref_shared();
		} else if( tr.is_shared() ) {
			tr.set_ref_value();
		}
	}
	if( vt.optional.is_set() )
		tr.set_optional( *vt.optional );
	if( vt.cvt_optional.is_set() )
		tr.set_cvt_optional( *vt.cvt_optional );
	if( vt.value_ptr.is_set() ) {
		if( *vt.value_ptr ) {
			tr.set_ref_value_ptr();
		} else if( tr.is_value_ptr() ) {
			tr.set_ref_value();
		}
	}
	if( vt.lossy.is_set() )
		tr.set_lossy( *vt.lossy );
		
	return tr;
}

/**
	Infer the variable type from the given spec and the variable to be assigned from it.
	Note that this doesn't consider conversion and may result in a variable type which
	cannot be assigned the given value: it's meant to follow some basic very predictable
	rules.
*/
type_identifier::resolve_t type_identifier::infer( shared_ptr<type_spec const> spec_, 
	extr_type_ref const & assign, bool init_refer )
{
	static type_spec const empty_spec;
	type_spec const & spec = spec_ ? *spec_ : empty_spec;
	return infer( spec, assign, init_refer );
}

cerr_base infer_spec_impl( type_spec & spec, extr_type_ref const & assign, bool init_refer )
{
	//DEBUG( "INFER", dump::get_extr_type( assign ) );
	
	//track if these fields are relevant based on the assign type. 
	bool rel_pack = false, rel_precision_variant = false, rel_function_access = false, rel_signed = false,
		rel_bitsize = false;
	
	//size-equivalent tuples are handled piece-wise (defined items are not subject to tuple conversion)
	bool use_assign = false;
	auto * atuple = assign.is_distinct() ? nullptr : assign.if_fun_cast<intr_type_tuple>();
	if( spec.fun_prototype ) {
		//leave the prototype in place, it dominates over assign's fun_type
	} else if( atuple && atuple->selectable && 
		(!spec.selectable.is_set() || *spec.selectable == false) ) {
		//selectable's only influence inference if matched in the spec_type
	} else if( spec.fun_class.is_set() ) {
	
		if( *spec.fun_class == intr_type::f_tuple 
			&& atuple && atuple->sub.size() == spec.sub.size() ) {
		
			for( size_t i=0; i < spec.sub.size(); ++i ) {
				auto err = infer_spec_impl( *spec.sub[i].type, atuple->sub[i].type, false );
				if( !err.okay() )
					return err;
				if( spec.sub[i].name.empty() )
					spec.sub[i].name = atuple->sub[i].name;
			}
		}
	} else {
		//fundamental not-set
		use_assign = true;
		if( assign.is_distinct() ) {
			spec.fun_prototype = assign.get_fun();
		} else switch( assign.get_fun_class() ) {
		
			case intr_type::f_rational: {
				//rationals never exist as actual variables (TODO: once "const" is introduced this may change)
				auto & rt = assign.fun_cast<intr_type_rational>();
				
				if( !rt.integer_bitsize ) {
					spec.fun_class = intr_type::f_float;
				} else {
					spec.fun_class = intr_type::f_integer;
				}
				break;
			}
			case intr_type::f_value_list:
			case intr_type::f_tuple: {
				if( spec.sub.empty() ) {
					auto & comp = assign.fun_cast<intr_type_compound>();
					spec.fun_class = atuple ? intr_type::f_tuple : intr_type::f_value_list;
					//each part of the tuple must be inferred (this similar to the above when the spec matches)
					spec.sub.resize( comp.sub.size() );
					for( size_t i=0; i < comp.sub.size(); ++i ) {
						spec.sub[i].type = make_shared<type_spec>();
						auto lit = infer_spec_impl( *spec.sub[i].type, comp.sub[i].type, false );
						if( !lit.okay() )
							return lit;
						spec.sub[i].name = comp.sub[i].name;
					}
				}
				break;
			}
			case intr_type::f_function: {
				auto & func = assign.fun_cast<intr_type_function>();

				spec.fun_class = intr_type::f_function;
				//if( func.access == intr_type_function::a_raw )
					spec.function_access = func.access;
				spec.function_context = func.context;
				spec.sub.resize( 2 );
				
				for( size_t i=0; i < 2; ++i ) {
					spec.sub[i].type = make_shared<type_spec>();
					auto lit = infer_spec_impl( *spec.sub[i].type, extr_type_ref(i ? func.returns : func.params), false );
					if( !lit.okay() )
						return lit;
						
				}
				break;
			}
			case intr_type::f_array: {
				auto & arr = assign.fun_cast<intr_type_array>();
				spec.fun_class = intr_type::f_array;
				if( arr.raw ) {
					spec.function_access = intr_type_function::a_raw; //TODO: need to rename/reshuffle access terms
				}
				spec.sub.resize(1);
				spec.sub[0].type = make_shared<type_spec>();
				auto err = infer_spec_impl( *spec.sub[0].type, arr.sub, false );
				if( !err.okay() )
					return err;
				break;
			}
			case intr_type::f_unknown:
				STATE_FAIL( "unexpected" );
				
			case intr_type::f_integer:
			case intr_type::f_binary: {
				auto & num = assign.fun_cast<intr_type_fixed_number>();
				spec.fun_class = assign.get_fun_class();
				if( spec.data_bitsize.is_unset() ) {
					if( num.data_bitsize == platform::natural_integer_size )
						spec.data_bitsize.revert_default();
					else
						spec.data_bitsize = num.data_bitsize;
				}
				rel_bitsize = true;
				break;
			}
			case intr_type::f_float: {
				//TODO: hmm
			}
			case intr_type::f_boolean:
			case intr_type::f_char:
				spec.fun_class = assign.get_fun_class();
				break;
				
			case intr_type::f_module:
			case intr_type::f_custom: {
				STATE_FAIL( "not-expected-here" );
			}
		}
	}

	//lvalue can never be inferred
	if( spec.lvalue.is_unset() )
		spec.lvalue = false;

	//init_refer requires a shared type
	if( init_refer ) {
		if( spec.shared.is_set() && !*spec.shared ) {
			return cerr::mismatch_shared;
		}
		spec.shared = true;
	}
	
	//only copy assigned flags here if the fundamental wasn't set
	if( use_assign ) {
		if( spec.optional.is_unset() )
			spec.optional = assign.is_optional();
		if( spec.shared.is_unset() )
			spec.shared = assign.is_shared();
		if( spec.value_ptr.is_unset() )
			spec.value_ptr = assign.is_value_ptr();
	}
	
	//non-relevant fields which aren't set become the default
	if( !rel_bitsize && spec.data_bitsize.is_unset() )
		spec.data_bitsize.revert_default();
	if( !rel_pack && spec.pack.is_unset() )
		spec.pack.revert_default();
	if( !rel_precision_variant && spec.precision_variant.is_unset() )
		spec.precision_variant.revert_default();
	if( !rel_function_access && spec.function_access.is_unset() )
		spec.function_access.revert_default();
	if( !rel_signed && spec.signed_.is_unset() )
		spec.signed_.revert_default();
		
	return cerr::okay;
}

type_identifier::resolve_t type_identifier::infer( type_spec const & spec_, 
	extr_type_ref const & assign, bool init_refer )
{
	auto spec = infer_spec( spec_, assign, init_refer );
	if( !spec.okay() )
		return resolve_t(spec);
		
	auto ret = determine( *spec.result );
	if( !ret.okay() )
		return resolve_t(ret);
	return resolve_t(ret.result);
}

type_identifier::resolve_spec_t type_identifier::infer_spec( type_spec const & spec_, 
	extr_type_ref assign, bool init_refer )
{
	resolve_t it;
	it.error = cerr::okay;

	//decompose value-list
	while( true ) {
		auto vl = assign.if_fun_cast<intr_type_value_list>();
		if( !vl )
			break;
			
		if( vl->sub.size() != 1 ) {
			return cerr::disallow_value_list;
		}
		assign = extr_type_ref(vl->sub[0].type);
	}
	
	//we need to work with an expanded spec
	auto spec = make_shared<type_spec>();
	auto err = expand_spec( *spec, spec_ );
	if( !err.okay() )
		return resolve_spec_t(err);
	err = infer_spec_impl( *spec, assign, init_refer );
	if( !err.okay() )
		return resolve_spec_t(err);
	return spec;
}

} //eon leaf