#include "lang/typer.hpp"
#include "lang/language.hpp"
#include "lang/type_converter.hpp"
#include "lang/type_identifier.hpp"
#include "lang/variable_context.hpp"
#include "lang/conversion_applicator.hpp"
#include "lang/expression_typer.hpp"
#include "lang/class_typer.hpp"
#include "lang/statement_visitor.hpp"
#include "lang/expression_visitor.hpp"
#include "lang/serial.hpp"
#include "lang/module_loader.hpp"

#include "lang/dump.hpp"

#define SHOW_DEBUG 0
#include "util/debug.hpp"
		
#define RETURN_IF_INCOMPLETE( obj, stmt ) \
	if( !(obj).is_complete() ) { \
		STMT_CHECK( *stmt, !no_incomplete, "incomplete-typing" ); \
		stmt->complete = false; \
		return stmt; \
	}
	
namespace leaf
{

///////////////////////////////////////////////////////////
// context
typer::context::context( typer * owner, shared_ptr<scope> const & sc ) 
	: sc( sc )
	, owner( owner )
{
}

shared_ptr<expression> typer::context::proc_expression( shared_ptr<expression> expr ) const
{
	return expression_typer( *this, owner->no_incomplete ).process( expr );
}

void typer::context::add_pending_block( shared_ptr<statement_block> block ) const
{
	//pushing an empty block would mess with progress tracking (completed_statements)
	if( block->list.empty() ) {
		return;
	}
	owner->pending_jobs.push_back( make_shared<job_statement_block>(block) );
}

shared_ptr<expression> typer::context::convert_expression( extr_type_ref const & target, 
	shared_ptr<expression> source ) const
{
	auto cost = converter().convert( target, source->type );
	return conversion_applicator().apply( cost, source );
}

shared_ptr<expression> typer::context::convert_expression( type_converter::unary_convert target,
	shared_ptr<expression> source ) const
{
	auto cost = (converter().*target)( source->type );
	return conversion_applicator().apply( cost, source );
}

shared_ptr<expression> typer::context::convert_expression( type_converter::binary_convert how,
	shared_ptr<expression> target, shared_ptr<expression> source ) const
{
	auto cost = (converter().*how)( target->type, source->type );
	STMT_CHECK( *source, cost.okay(), cost.explain_error(),
		typer::tag_cerr_source_type( dump::get_extr_type(source->type) ),
		typer::tag_cerr_target_type( dump::get_extr_type(target->type) )
	)
	return conversion_applicator().apply( cost, source );
}

///////////////////////////////////////////////////////////
// job_statement_block
bool typer::job_statement_block::process( typer * ty ) {
	auto ctx = context( ty, sb->symbols );
	return ty->proc_statement_block( sb, ctx );
}

///////////////////////////////////////////////////////////
// typer
typer::typer( shared_ptr<scope> global, shared_ptr<module_loader> mloader )
	: global( global )
	, mloader( mloader )
{
	no_incomplete = false;
}

void typer::full_add_block( shared_ptr<statement_block> block ) 
{
	try {
		scan_symbols( block );
		pending_jobs.push_back( make_shared<job_statement_block>(block) );
	} catch( recoverable_error & re ) {
		re << error_typer(true);
		throw;
	}
}

void typer::full_complete_type() 
{
	try {
		complete_type();
	} catch( recoverable_error & re ) {
		re << error_typer(true);
		throw;
	}
}

void typer::full_type( shared_ptr<statement_block> block )
{
	full_add_block( block );
	full_complete_type();
}

shared_ptr<expression> typer::full_type( shared_ptr<expression> expr, unsigned flags )
{
	try
	{
		auto ctx = context(this,global);
		auto result = ctx.proc_expression( expr );
		complete_type();
		
		if( flags & tef_convert_inferred ) {
			auto it = type_identifier(*global).infer( shared_ptr<type_spec>(), result->type, false );
			STMT_CHECK( *expr, it.okay(), it.explain_error() );
			auto cost = type_converter(*global).convert( it.result, result->type );
			result = conversion_applicator().apply( cost, result );
		}
		
		return result;
	}
	catch( recoverable_error & re )
	{
		re << error_typer(true);
		throw;
	}
}

/**
	In a declarative scope the statements and blocks have no particular order. This allows
	forward referencing of both variables and types. Rather than do a fixed 2-pass approach
	we just keep looping so long as progress is being made.
*/
void typer::complete_type()
{
	auto proc_jobs = [&,this]() {
		job_vector_t jobs;
		jobs.swap( pending_jobs );
			
		for( auto & j : jobs ) {
			if( !j->process( this ) ) {
				pending_jobs.push_back( j );
			}
		}
	};
	
	while( pending_jobs.size() )
	{
		completed_statements = 0;
		proc_jobs();
		
		//if nothing happened let it run once more to report errors
		if( !completed_statements ) {
			//TODO: need an error reporter to always report this
			std::cout << "No progress being made" << std::endl;
			no_incomplete = true;
			proc_jobs();
			STATE_FAIL( "no progress and no error?!" );
		}
	}
}

shared_ptr<statement> typer::proc_assign_statement( shared_ptr<assign_statement> as, 
	context ctx )
{
	STMT_CHECK( *as, !ctx.sc->is_declarative() );
	as->left = ctx.proc_expression( as->left );
	as->right = ctx.proc_expression( as->right );
	RETURN_IF_INCOMPLETE( *as->left, as )
	RETURN_IF_INCOMPLETE( *as->right, as )
		
	//DEBUG( "LT", dump::get_extr_type( as->left->type ) );
   	//DEBUG( "RT", dump::get_extr_type( as->right->type ) );
	auto tc = ctx.converter();
	
	//a reference init means the left must resolve to a shared lvalue first (which allows optionals to work)
	type_converter::unary_convert uhow;
	switch( as->how ) {
		case assign_statement::h_value:
		case assign_statement::h_default:
			uhow = &type_converter::convert_any_lvalue;
			break;
		case assign_statement::h_init_refer:
			uhow = &type_converter::convert_any_refer;
			break;
	}
	as->left = ctx.convert_expression( uhow, as->left );
	
	type_converter::binary_convert bhow;
	switch( as->how ) {
		case assign_statement::h_value:
			bhow = &type_converter::convert_to_lvalue;
			break;
		case assign_statement::h_init_refer:
			bhow = &type_converter::convert_to_refer;
			break;
		case assign_statement::h_default:
			STMT_CHECK( *as, as->left->type.is_optional() );
			bhow = &type_converter::convert_to_lvalue;
			break;
	}
	as->right = ctx.convert_expression( bhow, as->left, as->right );
	as->complete = true;
	return as;
}

shared_ptr<statement> typer::proc_var_statement( shared_ptr<var_statement> vs, 
	context ctx )
{
	STMT_CHECK( *vs, vs->decl );
	vs->decl->source_statement = vs;

	//DEBUG( "VAR", vs->decl->symbol );
	
	//an init_block needs to be evaluated in a different context (TODO: perhaps our context should
	//hold that optional context rather than constructing it)
	auto value_ctx = ctx;
	if( ctx.sc->has_init_block() ) {
		value_ctx.sc = ctx.sc->get_init_block()->symbols;
	}
	
	if( vs->value ) {
		vs->value = value_ctx.proc_expression( vs->value );
		RETURN_IF_INCOMPLETE( vs->value->type, vs )
		
		auto it = ctx.identifier().infer( vs->decl->spec_type, vs->value->type, vs->init_refer_value );
		STMT_CHECK( *vs, it.okay(), it.explain_error() );
		vs->decl->type = extr_type_ref(it.result);
		vs->value->sug_name = vs->decl->symbol;
	} else {
		auto rt = ctx.identifier().determine( vs->decl->spec_type );
		if( rt.unknown_identifier() /*|| !rt.result.is_complete()*/ ) {
			//DEBUG( "VS", vs->decl->symbol, "U/C", rt.unknown_identifier(), rt.result.is_complete() );
			STMT_CHECK( *vs, !no_incomplete, rt.explain_error() );
			vs->complete = false;
			return vs;
		}
		vs->decl->type = extr_type_ref(rt.result);
	}
	STMT_CHECK( *vs, vs->decl->type.is_fixated() );
	//DEBUG( "var_type", dump::get_extr_type( vs->decl->type ) );
	
	if( vs->value ) {
		//DEBUG( "VALUE_TYPE", dump::get_extr_type( vs->value->type ) );
		
		auto tc = value_ctx.converter();
		type_converter::convert_t cost = 
			vs->init_refer_value 
				? tc.convert_to_refer( vs->decl->type, vs->value->type )
				: tc.convert( vs->decl->type, vs->value->type );
		vs->value = conversion_applicator().apply( cost, vs->value );
	}

	if( vs->init_refer_value )
		STMT_CHECK( *vs, vs->decl->type.is_shared() );

	//the symbol is not available until after the var_statement
	auto init = make_shared<init_statement>();
	init->src_loc = vs->src_loc;
	if( vs->sref.is_set() ) {
		init->sref = vs->sref;
	} else {
		//TODO: same code as in symbol_scanner
		switch( vs->decl->form )
		{
			case var_decl::f_single:
				init->sref = ctx.sc->add_single( vs->decl );
				break;
			
			case var_decl::f_multi:
				init->sref = ctx.sc->add_multi( vs->decl );
				break;
				
			case var_decl::f_param:
			case var_decl::f_return:
			case var_decl::f_capture:
				STMT_FAIL(*vs);
		}
	}

	//TODO: rought hack since we lack statics at the moment
	if( ctx.sc->has_init_block() && (vs->decl->symbol == "construct" 
		|| vs->decl->symbol == "destroy") ) {
		auto noop = make_shared<noop_statement>();
		noop->complete = true;
		return noop;
	}
	
	init->value = vs->value;
	init->init_refer_value = vs->init_refer_value;
	
	shared_ptr<statement> ret;
	if( ctx.sc->has_init_block() ) {
		init->no_release = true;
		init->sref = ctx.sc->get_init_block()->symbols->capture_var( init->sref, ctx.sc.get() );
		ctx.sc->get_init_block()->list.push_back( init );
		
		ret = make_shared<noop_statement>();
		ret->complete = true;
	} else {
		ret = proc_init_statement( init, ctx );
	}
	
	//export handling
	if( vs->decl->is_export ) {
		STMT_CHECK( *vs, ctx.sc->get_level() == 0, "exports-only-in-root" );
		var_exports.push_back( vs->decl );
	}

	return ret;
}

/**
	init is generated by the typer itself in var_statement. This function assumes that is how it
	was created. This separation was only done to deal with incomplete recursive types.
*/
shared_ptr<statement> typer::proc_init_statement( shared_ptr<init_statement> is, context ctx )
{
 	if( is->complete )
 		return is;

	STMT_CHECK( *is, is->sref.is_set() );
	auto type = is->sref.decl->type;
	if( type.is_optional() ) {
		//nop
	} else if( is->value ) {
		//nop
	} else if( type.is_custom() ) {
		//find constructor
		is->undefined = true;
		auto & custom = is->sref.decl->type.fun_cast<intr_type_custom>();
		auto ctor_name = std::string("construct_") + *custom.distinct_name;
		DEBUG( "CONSTRUCT", custom.name_scope );
		STMT_CHECK( *is, custom.name_scope );
		auto ctors = custom.name_scope->get_vars( ctor_name );
		//auto ctors = ctx.sc->get_vars( ctor_name );
		if( ctors.size() != 1 ) {
			STMT_CHECK( *is, !no_incomplete, std::string( "incomplete-type: " ) + ctor_name );
			//leave incomplete
			return is;
		}
		
		//add call to the constructor
		auto es = make_shared<expression_statement>();
		es->complete = true;
		auto fc = make_shared<expr_funccall>();
		es->expr = fc;
		fc->type = extr_type_ref(make_shared<intr_type_value_list>());
		is->code = es;
	
		auto sym = make_shared<expr_symbol>();
		sym->symbol = is->sref.decl->symbol;
		sym->sref = is->sref;
		sym->type = is->sref.decl->type;
		auto cost = ctx.converter().convert_shared_or_unqualified( sym->type );
		fc->context = conversion_applicator().apply( cost, sym );
		
		fc->params = ctx.proc_expression( make_shared<expr_value_list>() );
		
		auto ctor = make_shared<expr_symbol>();
		ctor->symbol = ctor_name;
		ctor->sref = ctors[0];
		ctor->type = ctor->sref.decl->type;
		fc->func = ctor;
	}
	is->complete = true;
	return is;
}

shared_ptr<statement> typer::proc_statement( shared_ptr<statement> s, context ctx )
{
	bool was_complete = s->complete;
	
	//DEBUG( "STATEMENT: ", typeid( *s.get() ).name() );
	shared_ptr<statement> result;
	switch( s->statement_form() )
	{
		#define STM(type) \
			case sf_##type: \
				result = proc_##type( std::static_pointer_cast<type>(s), ctx ); \
				break;
		STM(assign_statement)
		STM(defer_statement)
		STM(do_statement)
		STM(expression_statement)
		STM(fail_statement)
		STM(for_statement)
		STM(import_statement)
		STM(init_statement)
		STM(loop_flow_statement)
		STM(return_statement)
		STM(type_statement)
		STM(var_statement)
		#undef STM

		//not expected in input (generated by typer)
		case sf_noop_statement:
		case sf_multi_statement:
			result = s;
			break;
	}
	STMT_CHECK( *s, result );
	
	if( !was_complete && result->complete ) {
		completed_statements++;
	}
	return result;
}

shared_ptr<statement> typer::proc_do_statement( shared_ptr<do_statement> ds, context ctx )
{
	STMT_CHECK( *ds, !ctx.sc->is_declarative() );
	static extr_type_ref cond_type( make_shared<intr_type_boolean>() );
	
	for( size_t i=0; i < ds->expr_list.size(); ++i )
	{
		auto & ex = ds->expr_list[i];
		if( ex.cond ) {
			//TODO: This should check for completeness, but I have no situation where it can happen yet
			ex.cond = ctx.proc_expression( ex.cond );
			ex.cond = ctx.convert_expression( cond_type, ex.cond );
		}
		
		ex.block->symbols->set_extends( ctx.sc );
		if( !proc_statement_block( ex.block ) )
			return ds; //incomplete
	}
	
	return proc_error_statement( ds, ctx );
}

shared_ptr<statement> typer::proc_error_statement( shared_ptr<error_statement> ds, context ctx )
{
	STMT_CHECK( *ds, !ctx.sc->is_declarative() );
	if( ds->on_error ) {
		STMT_CHECK( *ds, !(ds->get_flow() & (flow_return | flow_loop_break | flow_loop_next)), 
			"terminator-handled" );
		ds->on_error->symbols->set_extends( ctx.sc );
		if( !proc_statement_block( ds->on_error ) )
			return ds; //incomplete
	}
	
	ds->complete = true;
	return ds;
}

bool typer::proc_statement_block( shared_ptr<statement_block> block )
{
	auto ctx = context( this, block->symbols );
	return proc_statement_block( block, ctx );
}

/**
	Declarative statement blocks  can be processed piece-wise. Any incomplete statement
	is simply moved to the end and tried again later. After enough passes this should result
	in an appropriate ordering of complete statements.
	TODO: it should be detected when a loop happens and no ordering is possible.
	
	Sequence blocks can only be processed in order. Any incomplete statement stops the
	processing and pushes the whole block on the pending list.
*/
bool typer::proc_statement_block( shared_ptr<statement_block> block, context ctx )
{
	if( ctx.sc->is_declarative() ) {
		std::vector<shared_ptr<statement>> done, after;
		
		for( size_t i=0; i < block->list.size(); ++i ) {	
			auto st = proc_statement( block->list[i], ctx );
			//DEBUG( "BS", typeid(*st).name(), st->complete );
			if( st->complete ) {
				done.push_back(st);
			} else {
				after.push_back(st);
			}
		}
		
		if( after.size() ) {
			done.insert( done.end(), after.begin(), after.end() );
		}
		block->list.swap( done );
		if( after.size() ) {
			return false;
		}
		
	} else {
		//TODO: catch "resume" being used in non-handler
		for( size_t i=0; i < block->list.size(); ++i ) {	
			if( i && !(block->list[i-1]->get_flow() & flow_sequence) ) 
				STMT_FAIL( *block->list[i], "code-post-terminator" );
			block->list[i] = proc_statement( block->list[i], ctx );
			if( !block->list[i]->complete ) {
				return false;
			}
		}
		block->update_flow();
	}
	
	return true;
}


shared_ptr<statement> typer::proc_defer_statement( shared_ptr<defer_statement> ds, context ctx )
{
	STMT_CHECK( *ds, !ctx.sc->is_declarative() );
	ds->block->symbols->set_extends( ctx.sc );
	if( proc_statement_block( ds->block ) ) {
		ds->complete = true;
	}
	return ds;
}

shared_ptr<statement> typer::proc_type_statement( shared_ptr<type_statement> ts, context ctx )
{
	if( ts->complete )
		return ts;
	//DEBUG( "TYPE", ts->name );
		
	if( ts->what == type_statement::w_prototype ) {
		auto cur_type = ctx.sc->if_get_fun_type( ts->name, scope::gf_strict_local );
		shared_ptr<intr_type_tuple> inc_tuple;
		
		if( !cur_type ) {
			//introduce incomplete type first to allow recursion
			inc_tuple = make_shared<intr_type_tuple>();
			inc_tuple->local_complete = false;
			inc_tuple->distinct_name = ts->name;
			inc_tuple->distinct = true;
			ctx.sc->add_fun_type( inc_tuple );
		} else {
			STMT_CHECK( *ts, cur_type->is_tuple() );
			STMT_CHECK( *ts, cur_type->is_distinct() );
			//TODO: is this const_cast okay? (probably, all refers to this type expect incomplete)
			inc_tuple = std::const_pointer_cast<intr_type_tuple>(
				std::static_pointer_cast<intr_type_tuple const>(cur_type) );
		}

		if( ts->spec_type ) {
			STMT_CHECK( *ts, !inc_tuple->local_complete );
			
			//'type' statements create new classes from plain tuples
			STMT_CHECK( *ts, ts->spec_type->extrinsic_free() );
			STMT_CHECK( *ts, ts->spec_type->fun_class.is_set() && 
				*ts->spec_type->fun_class == intr_type::f_tuple );
				
			auto nt = ctx.identifier().determine( ts->spec_type );
			if( nt.okay() ) {
				STMT_CHECK( *ts, !nt.result.get_fun()->distinct_name );
				auto nt_tuple = nt.result.if_fun_cast<intr_type_tuple>();
				STMT_CHECK( *ts, nt_tuple );
				inc_tuple->complete_with( *nt_tuple );
			} else if( nt.unknown_identifier() ) {
				//DEBUG( "UNKNOWN", ts->name );
				STMT_CHECK( *ts, !no_incomplete, nt.explain_error() );
				ts->complete = false; //TODO: shouldn't be needed (false by default)
				return ts;
			} else {
				STMT_FAIL( *ts, nt.explain_error() );
			}
		} else {
			//not necessary, but has cleaner dump
			auto noop = make_shared<noop_statement>();
			noop->complete = true;
			return noop;
		}
		
		//expected in dumps
		ts->type = extr_type_ref(inc_tuple);
	
	} else if( ts->what == type_statement::w_alias ) {
		auto err = ctx.identifier().expand_spec( *ts->spec_type );
		if( err.unknown_identifier() ) {
			STMT_CHECK( *ts, !no_incomplete, err.explain_error() );
			//aliases cannot be partially constructure, they only exist in the scope once fully definable
			return ts;
		}
		//DEBUG( "ALIAS", ts->name, dump::get_type_spec( *ts->spec_type ) );
		ctx.sc->add_type_alias( ts->name, ts->spec_type );
	} else if( ts->what == type_statement::w_class ) {
		auto ct = make_shared<class_typer>( ts, ctx );
		pending_jobs.push_back( ct );
	}
	
	ts->complete = true;
	return ts;
}

shared_ptr<statement> typer::proc_return_statement( shared_ptr<return_statement> rs, context ctx )
{
	STMT_CHECK( *rs, !ctx.sc->is_declarative() );
	if( !rs->value ) {
		rs->complete = true;
		return rs;
	}
		
	//TODO: This should check for completeness, but I have situation where it can happen yet
	rs->value = ctx.proc_expression( rs->value );

	auto & func_scope = ctx.sc->get_function_root();
	auto return_type = func_scope.get_return_type();
	
	//convert 1-value-list return to value for type conversion
	extr_type_ref use_return_type;
	if( return_type->sub.size() == 1 )
		use_return_type = return_type->sub[0].type;
	else
		use_return_type = extr_type_ref(return_type);
	
	rs->value = ctx.convert_expression( use_return_type, rs->value );
	rs->complete = true;
	return rs;
}

shared_ptr<statement> typer::proc_fail_statement( shared_ptr<fail_statement> fs, context ctx )
{
	STMT_CHECK( *fs, !ctx.sc->is_declarative() );
	fs->complete = true;
	return fs;
}

shared_ptr<statement> typer::proc_for_statement( shared_ptr<for_statement> fs, context /*ctx*/ )
{
	static extr_type_ref cond_type( make_shared<intr_type_boolean>() );
	
	//all declarations are local to a new block
	auto outer = make_shared<do_statement>();

	auto iterable = make_shared<var_statement>();
	iterable->decl = make_shared<var_decl>();
	iterable->decl->form = var_decl::f_single;
	iterable->decl->symbol = std::string("_") + fs->item + "_iterator";
	iterable->value = fs->iterable;
	
	do_statement::expr_list_item eli;
	eli.block = make_shared<statement_block>( scope_style_t::imperative );
	eli.block->list.push_back( iterable );
	outer->expr_list.push_back( eli );
	
	//the inner block is the looping part
	auto cond_sym = make_shared<expr_symbol>();
	cond_sym->symbol = iterable->decl->symbol;
	
	auto cond_func = make_shared<expr_field>();
	cond_func->base = cond_sym;
	cond_func->field = "is_more";
	
	auto cond_call = make_shared<expr_funccall>();
	cond_call->func = cond_func;
	cond_call->params = make_shared<expr_value_list>();
	
	auto inner = make_shared<do_statement>();
	inner->loop = true;
	do_statement::expr_list_item inner_eli;
	inner_eli.block = make_shared<statement_block>( scope_style_t::imperative );
	inner_eli.cond = cond_call;
	inner->expr_list.push_back( inner_eli );
	
	eli.block->list.push_back( inner );
	
	//create the item access
	auto item_func = make_shared<expr_field>();
	item_func->base = cond_sym;
	item_func->field = "get";
	
	auto item_call = make_shared<expr_funccall>();
	item_call->func = item_func;
	item_call->params = make_shared<expr_value_list>();
	
	auto item = make_shared<var_statement>();
	item->decl = make_shared<var_decl>();
	item->decl->form = var_decl::f_single;
	item->decl->symbol = fs->item;
	item->value = item_call;
	
	inner_eli.block->list.push_back( item );
	
	//step to next value
	auto step_func = make_shared<expr_field>();
	step_func->base = cond_sym;
	step_func->field = "step";
	
	auto step_call = make_shared<expr_funccall>();
	step_call->func = step_func;
	step_call->params = make_shared<expr_value_list>();
	
	auto stat = make_shared<expression_statement>();
	stat->expr = step_call;
	inner_eli.block->list.push_back( stat );
	
	//add the block
	auto body = make_shared<do_statement>();
	do_statement::expr_list_item body_eli;
	body_eli.block = fs->block;
	body->expr_list.push_back( body_eli );
	inner_eli.block->list.push_back( body );

	return outer;
}

shared_ptr<statement> typer::proc_loop_flow_statement( shared_ptr<loop_flow_statement> ls, 
	context ctx )
{
	STMT_CHECK( *ls, !ctx.sc->is_declarative() );
	ls->complete = true;
	return ls;
}

shared_ptr<statement> typer::proc_expression_statement( shared_ptr<expression_statement> es, 
	context ctx )
{
	STMT_CHECK( *es, !ctx.sc->is_declarative() );
	es->expr = ctx.proc_expression( es->expr );
	RETURN_IF_INCOMPLETE( *es->expr, es )
	return proc_error_statement( es, ctx );
}

shared_ptr<statement> typer::proc_import_statement( shared_ptr<import_statement> is, context ctx )
{
	if( is->complete )
		return is;
		
	STATE_CHECK( mloader );
	libraries.push_back( is->mod_name );

	auto sin = mloader->resolve_import( is->mod_name );
	
	//the imported module becomes a variable with a distinct type
	auto mdecl = make_shared<var_decl>();
	mdecl->form = var_decl::f_single;
	mdecl->symbol = is->mod_name;
	mdecl->source_statement = is;
	ctx.sc->add_single( mdecl );
	
	auto mscope = scope::create( scope_style_t::declarative );
	mscope->set_namespace( ctx.sc, is->mod_name );
	
	auto mtype = make_shared<intr_type_module>();
	mtype->distinct_name = std::string("module_") + is->mod_name; //name is arbitrary
	mtype->distinct = true;
	mtype->local_complete = true;
	mtype->context = mscope;
	mdecl->type = extr_type_ref(mtype);
	
	for( auto & sitem : sin->get_all_items() ) {
		switch( sitem.type ) {
			case leaf::serial_item::t_invalid:
			case leaf::serial_item::t_end_marker:
			case leaf::serial_item::t_extr_type:
				break;
				
			case leaf::serial_item::t_intr_type:
				//TODO: only those marked as exported should be available for direct use)
				//TODO: hmm, perhaps should be in mscope and vairables use scoped lookup
				if( sitem.v_intr_type->is_distinct() ) {
					DEBUG( "IMPORT", *sitem.v_intr_type->distinct_name, mscope.get(), mdecl->symbol );
					mscope->add_fun_type( sitem.v_intr_type );
				}
				break;
				
			case leaf::serial_item::t_var_decl:
				//TODO: should export itself set the fields appropriately for import?
				auto decl = sitem.v_var_decl;
				decl->import_name = decl->symbol;
				decl->import_module = is->mod_name;
				decl->name_form = var_decl::nf_leaf;
				decl->import_type = var_decl::it_abi_data;
				
				//inside a scope, so same name is okay
				decl->symbol = decl->import_name;
				
				switch( decl->form ) {
					case var_decl::f_single:
						mscope->add_single( decl );
						break;
					
					case var_decl::f_multi:
						mscope->add_multi( decl );
						break;
						
					case var_decl::f_param:
					case var_decl::f_return:
					case var_decl::f_capture:
						STMT_FAIL(*is);
				}
				//DEBUG( "IMPORT", sitem.v_var_decl->symbol );
				break;
		}
	}
	
	is->complete = true;
	return is;
}


namespace {
	struct symbol_scanner : public stat_expr_transformer, public expression_transformer {
		typer & owner;
		
		symbol_scanner( typer & owner )
			: stat_expr_transformer( static_cast<expression_transformer&>(*this) )
			, expression_transformer( false )
			, owner( owner ) {
		}
		
		shared_ptr<statement> istmt_var( shared_ptr<var_statement> vs, shared_ptr<scope> sc ) {
			switch( vs->decl->form ) {
				case var_decl::f_single:
					vs->sref = sc->add_single( vs->decl );
					break;
				case var_decl::f_multi:
					vs->sref = sc->add_multi( vs->decl );
					break;
					
				case var_decl::f_param:
				case var_decl::f_return:
				case var_decl::f_capture:
					break;
			}
			return vs;
		}
	};
} //eon anon

/**
	Supporting overloads requires all possible symbols to be known the moment a symbol is
	encountered. This requires pre-scanning all the code for symbols. If this didn't happen
	an overload may resolve to a type from previous declarations, but may not include the
	correct one. This is then just extended to work with all symbols in a declaration scope.
*/
void typer::scan_symbols( shared_ptr<statement_block> block )
{
	symbol_scanner ss( *this );
	ss.apply_statement_block( block );
}

} //eon leaf