#include "parser/intr_type_parser.hpp"

#define SHOW_DEBUG 0
#include "util/debug.hpp"

#include "regex_patterns.hpp"

namespace leaf
{

	namespace
	{
		bool is_single_end_char( src_char_t c ) {
			return c == ',' || c == ')' || c == '}' || c == '>' || c == ';' || c == ']' || c == '\n';
		}
		
		bool is_inline_break_char( src_char_t c ) {
			return c == '(' || c == '{' || c == '[';
		}
	}

shared_ptr<type_spec> intr_type_parser::parse_tuple( source & src, src_char_t end )
{
	auto vt = make_shared<type_spec>();
	extend_with_tuple( vt, src, end );
	return vt;
}
	
void intr_type_parser::extend_with_tuple( shared_ptr<type_spec> vt, source & src, src_char_t end )
{
	STATE_CHECK( vt->fun_class.is_unset() );
	STATE_CHECK( vt->sub.empty() );
	vt->fun_class = intr_type::f_tuple;

	static boost::regex re_symbol( SYMBOL_REGEX );

	bool at_name = true, linger_sep = false;
	std::string cur_name;
	
	while( true )
	{
		STATE_CHECK( src.has_next(source::skip_line_ends), "unexpected end of type tuple" );

		//check for completion
		char nc = src.peek_next();
		if( is_single_end_char(nc) ) {
			if( cur_name.size() ) {
				//names without types are also okay
				vt->push_sub( cur_name, make_shared<type_spec>() );
			} else if( linger_sep ) {
				//simple separators also denote an empty entry
				vt->push_sub( make_shared<type_spec>() );
			}
			
			if( nc == end )	{
				src.next();
				break;
			} else 	if( nc == ',' ) {
				src.next();
				at_name = true;
				linger_sep = false;
				cur_name.clear();
				continue;
			}
			src.throw_unexpected( "type-end-char" );
		}
		
		static boost::regex re_op( OP_REGEX );
		if( src.match( re_op ) ) {
			//is a separator required (this comes before the re_symbol check since a symbol is optional)
			if( src.cur_match() == ":" ) {
				linger_sep = true;
				if( !at_name )
					src.throw_unexpected( "type-separator" );
					
				at_name = false;
				continue;
			} else {
				src.throw_unexpected( "incomplete type tuple" );
			}
		}

		if( at_name )
		{
			if( cur_name.size() )
				src.throw_expecting( "type separator" );
			if( !src.match( re_symbol ) )
				src.throw_expecting( "symbol" );
			
			cur_name = src.cur_match();
		}
		else
		{
			auto s = parse_single(src, pf_line_cont);
			vt->push_sub( cur_name, s );
			cur_name = "";
			linger_sep = false;
		}
	}
}

shared_ptr<type_spec> intr_type_parser::parse_signature( source & src )
{
	auto t_params = parse_tuple( src, ')' );
	t_params->fun_class = intr_type::f_value_list;

	src.skip_all_space();
	static boost::regex re_arrow( "->\\s*\\(" );
	if( !src.match( re_arrow ) )
		return t_params; //nope, just a value list

	auto t_return = parse_tuple( src, ')' );
	t_return->fun_class = intr_type::f_value_list;

	auto sig = make_shared<type_spec>();
	sig->fun_class = intr_type::f_function;
	sig->push_sub( t_params );
	sig->push_sub( t_return );
	return sig;
}

shared_ptr<type_spec> intr_type_parser::parse_single( source & src, unsigned flags, bool * empty )
{
	auto vt = make_shared<type_spec>();
	bool line_cont = flags & pf_line_cont;
	bool first_token = true;
	bool end_close_paren = false;
	while( true )
	{
		if( !src.has_next( (end_close_paren || line_cont) ? source::skip_line_ends : 0) )
			break;

		src_char_t nc = src.peek_next();
		if( end_close_paren ) {
			//in this mode only a close-paren ends the type (though other end characters will
			//still be flagged as invalid usually
			if( nc == ')' )
				break;
		} else {
			static boost::regex re_op( OP_REGEX );
			if( src.peek_match(re_op) || is_single_end_char(nc) )
				break;
			if( (flags & pf_inline) && is_inline_break_char(nc)
				&& !(first_token && nc == '(') /*see below*/) {
				break;
			}
		}

		if( first_token )
		{
			first_token = false;
			if( nc == '(' )
			{
				src.next();
				if( flags & pf_inline ) {
					//inline starting with open-paren is terminated by matching close-paren
					end_close_paren = true;
					continue;
				} else {
					vt = parse_signature( src );
				}
				continue;
			}
			
		}
		
		if( nc == '[' )
		{
			src.next();
			extend_with_tuple( vt, src, ']' );
			continue;
		}

		static boost::regex re_fund( "(array)" EXCLUDE_SYMBOL_REGEX_CHAR );
		static boost::regex re_marker( "(lvalue|rvalue|shared|raw|unbound|"
			"high|low|normal|optional|selectable|member|convert_optional|value_ptr|"
			"lossy)" EXCLUDE_SYMBOL_REGEX_CHAR );
		static boost::regex re_symbol( TYPE_SYMBOL_REGEX );
		static boost::regex re_bitsize( "([0-9]+)(bit)" EXCLUDE_SYMBOL_REGEX_CHAR );

		if( src.match( re_fund ) ) {
			STATE_CHECK( vt->fun_class.is_unset() );
			std::string f( src.cur_match() );
			if( f == "array" ) {
				vt->fun_class = intr_type::f_array;
				STATE_CHECK( src.has_all_next() );
				STATE_CHECK( src.next() == '<' );
				vt->push_sub( "", parse_single( src ) );
				STATE_CHECK( src.has_all_next() );
				STATE_CHECK( src.next() == '>' );
			} else {
				STATE_FAIL( "unexpected fundamental match" );
			}
		} else if( src.match( re_marker ) ) {
			std::string f( src.cur_match() );
			if( f == "lvalue" ) {
				STATE_CHECK( !vt->lvalue.is_set() );
				vt->lvalue = true;
			} else if( f == "rvalue" ) {
				STATE_CHECK( !vt->lvalue.is_set() );
				vt->lvalue = false;
			} else if( f == "shared" ) {
				STATE_CHECK( !vt->shared.is_set() );
				vt->shared = true;
			} else if( f == "raw" ) {
				STATE_CHECK( !vt->function_access.is_set() ); 
				vt->function_access = intr_type_function::a_raw;
			} else if( f == "unbound" ) {
				STATE_CHECK( !vt->function_access.is_set() );
				vt->function_access = intr_type_function::a_unbound;
			} else if( f == "optional" ) {
				STATE_CHECK( !vt->optional.is_set() );
				vt->optional = true;
			} else if( f == "convert_optional" ) {
				STATE_CHECK( !vt->cvt_optional.is_set() );
				vt->cvt_optional = true;
			} else if( f == "value_ptr" ) {
				STATE_CHECK( !vt->value_ptr.is_set() );
				vt->value_ptr = true;
			} else if( f == "high" ) {
				STATE_CHECK( !vt->precision_variant.is_set() );
				vt->precision_variant = 1;
			} else if( f == "low" ) {
				STATE_CHECK( !vt->precision_variant.is_set() );
				vt->precision_variant = -1;
			} else if( f == "normal" ) {
				STATE_CHECK( !vt->precision_variant.is_set() );
				vt->precision_variant = 0;
			} else if( f == "selectable" ) {
				STATE_CHECK( !vt->selectable.is_set() );
				vt->selectable = true;
			} else if( f == "member" ) {
				STATE_CHECK( !vt->member.is_set() );
				vt->member = true;
			} else if( f == "lossy" ) {
				STATE_CHECK( !vt->lossy.is_set() );
				vt->lossy = true;
			}
			else
				STATE_CHECK( false );
		} else if( src.match( re_bitsize ) ) {
			STATE_CHECK( !vt->data_bitsize.is_set() );
			vt->data_bitsize = atoi( src.cur_match(1).c_str() );
		} else if( src.match( re_symbol ) ) {
			vt->symbols.push_back( src.cur_match() );
		} else {
			src.throw_invalid( "type syntax" );
		}
	}

	//an empty type is valid -- simply unknown
	if( empty )
		*empty = first_token;
		
	return vt;
}

intr_type_parser::intr_type_parser()
{
}

/*static*/ shared_ptr<type_spec> intr_type_parser::parse( source & src )
{
	intr_type_parser p;
	return p.parse_single( src );
}

/*static*/ shared_ptr<type_spec> intr_type_parser::parse_inline( source & src)
{
	intr_type_parser p;
	return p.parse_single( src, pf_inline );
}

/*static*/ shared_ptr<type_spec> intr_type_parser::if_parse_inline( source & src )
{
	intr_type_parser p;
	bool empty;
	auto r = p.parse_single( src, pf_inline, &empty );
	if( empty )
		return shared_ptr<type_spec>();
	return r;
}

} //eon leaf
