#include "parser/node_converter.hpp"
#include "parser/node_parser.hpp"
#include "parser/node.hpp"

#include "lang/language.hpp"
#include "lang/dump.hpp"
#include "util/unicode.hpp"

#define SHOW_DEBUG 1
#include "util/debug.hpp"

#define NODE_CHECK( node, cond, ... )	\
	if( !(cond) ) { auto x = error::create_exception<recoverable_error>( __VA_ARGS__ ); \
		x << error::tag_error_condition( #cond ); \
		x << util::source_location::error( (node).src_loc ); \
		BOOST_THROW_EXCEPTION( x ); }
#define NODE_FAIL( node, ... )	{ \
	auto x = error::create_exception<recoverable_error>( __VA_ARGS__ ); \
	x << util::source_location::error( (node).src_loc ); \
	BOOST_THROW_EXCEPTION( x ); }

#define EXPECT_FUNCTION( src ) \
	NODE_CHECK( (src), (src).form == node::f_funccall ); \
	NODE_CHECK( (src), (src).sub.size() == 2 ); \
	auto & symbol = *(src).sub[0]; \
	NODE_CHECK( symbol, !symbol.text.empty() ); \
	auto & params = *(src).sub[1]; \
	NODE_CHECK( params, params.form == node::f_value_list );

#define STD_FUNCTION( src, sym ) \
	EXPECT_FUNCTION( src ); \
	NODE_CHECK( symbol, symbol.text == sym );

namespace leaf
{

/*static*/ shared_ptr<expression> node_converter::expression_convert( node const & src )
{
	try {
		node_converter nc;
		return nc.convert_expression( src );
	} catch( recoverable_error & re ) {
		re << error_converter(true);
		throw;
	}
}

/*static*/ shared_ptr<statement> node_converter::convert( node const & src )
{
	try {
		node_converter nc;
		return nc.convert_statement( src );
	} catch( recoverable_error & re ) {
		re << error_converter(true);
		throw;
	}
}

/*static*/ shared_ptr<statement_block> node_converter::convert_block( node const & src,
	scope_style_t scope_style )
{
	try {
		node_converter nc;
		
		NODE_CHECK( src, src.form == node::f_block );
		return nc.convert_block_inner( src, scope_style );
	} catch( recoverable_error & re ) {
		re << error_converter(true);
		throw;
	}
}


shared_ptr<statement> node_converter::convert_statement( node const & src )
{
	auto * primary = &src;
	if( src.form == node::f_expr_list ) {
		if( src.sub.empty() ) {
			return make_shared<noop_statement>();
		}
		primary = src.sub[0].get();
	}
	
	//special declarations/flow
	if( primary->form == node::f_symbol ) {
		auto & decl = *primary;
		if( decl.text == "class" ) {
			return convert_class_statement(src);
		} else if( decl.text == "type" ) {
			return convert_type_statement( src, false );
		} else if( decl.text == "alias" ) {
			return convert_type_statement( src, true );
		} else if( decl.text == "var" || decl.text == "func" ) {
			return convert_var_statement( src );
		} else if( decl.text == "do" || decl.text == "loop" ) {
			return convert_do_statement( src );
		} else if( decl.text == "defer" ) {
			return convert_defer_statement( src, false );
		} else if( decl.text == "defer_error" ) {
			return convert_defer_statement( src, true );
		} else if( decl.text == "import" ) {
			return convert_import_statement( src );
		} else if( decl.text == "assign" ) {
			return convert_assign_statement( src, ah_value );
		} else if( decl.text == "assign_value" ) {
			return convert_assign_statement( src, ah_init_refer );
		} else if( decl.text == "assign_default" ) {
			return convert_assign_statement( src, ah_default );
		} else if( decl.text == "return" ) {
			return convert_return_statement( src );
		} else if( decl.text == "resume" ) {
			return convert_resume_statement( src );
		} else if( decl.text == "break" ) {
			return convert_break_statement( src, false );
		} else if( decl.text == "next" ) {
			return convert_break_statement( src, true );
		} else if( decl.text == "unwind" ) {
			return convert_fail_statement( src, fail_level::unwind );
		} else if( decl.text == "continue" ) {
			return convert_fail_statement( src, fail_level::continue_ );
		} else if( decl.text == "recover" ) {
			return convert_fail_statement( src, fail_level::recover );
		} else if( decl.text == "abandon" ) {
			return convert_fail_statement( src, fail_level::abandon );
		} else if( decl.text == "for" ) {
			return convert_for_statement( src );
		}
	}
	
	if( src.form == node::f_funccall ) {
		//some pseudo-functions become statements
		NODE_CHECK( src, src.sub.size() == 2 );
		if( src.sub[0]->form == node::f_symbol ) {
			auto & symbol = src.sub[0]->text;
			
			if( symbol == "assign" )
				return convert_assign_statement( src, ah_value );
			if( symbol == "assign_value" )
				return convert_assign_statement( src, ah_init_refer );
			if( symbol == "assign_default" )
				return convert_assign_statement( src, ah_default );
			//some decls may also look like functions with a trailing value-list
			if( symbol == "return" )
				return convert_return_statement( src );
		}
	} else if( src.form == node::f_on ) {
		return convert_on_statement( src );
	}
	return convert_expression_statement( src );	
}

void node_converter::convert_on_block( node const & match, node const & block, error_statement & ex )
{
	NODE_CHECK( match, match.form == node::f_symbol );
	auto & ons = match.text;
	if( ons == "unwind" ) {
		ex.on_match = fail_level::unwind;
	} else if( ons == "continue" ) {
		ex.on_match = fail_level::continue_;
	} else if( ons == "recover" ) {
		ex.on_match = fail_level::recover;
	} else if( ons == "abandon" ) {
		ex.on_match = fail_level::abandon;
	} else {
		NODE_FAIL( match, "invalid-on-match" );
	}
	ex.on_error = convert_block_inner( block, scope_style_t::imperative );
}

shared_ptr<statement> node_converter::convert_on_statement( node const & src )
{
	NODE_CHECK( src, src.sub.size() == 3 );
	auto ex = convert_expression_statement( *src.sub[0] );
	convert_on_block( *src.sub[1], *src.sub[2], *ex );
	return ex;
}

shared_ptr<expression_statement> node_converter::convert_expression_statement(
	node const & src )
{
	auto es = make_shared<expression_statement>();
	es->expr = convert_expression( src );
	es->src_loc = es->expr->src_loc;
	return es;
}

shared_ptr<expression> node_converter::convert_expression( node const & src )
{
	shared_ptr<expression> ex;
	
	bool omit_type = false;
	switch( src.form )
	{
		case node::f_symbol: {
			if( src.text == "new" ) {
				ex = convert_new_expression( src );
			} else {
				auto es = make_shared<expr_symbol>();
				es->symbol = src.text;
				ex = es;
			}
			break;
		}
			
		case node::f_const_number: {
			auto & ncn = static_cast<node_const_number const&>(src);
			auto ecn = make_shared<expr_const_number>();
			ecn->src_form = expr_const_number::sf_rational;
			//TODO: need an explicit high-prec float still
			ecn->value = ncn.value;
			ex = ecn;
			break;
		}
			
		case node::f_const_boolean: {
			auto ecn = make_shared<expr_const_number>();
			if( src.text == "true" )
				ecn->val_boolean = true;
			else if( src.text == "false" )
				ecn->val_boolean = false;
			else	
				NODE_FAIL( src, "invalid boolean constant" );
			ecn->src_form = expr_const_number::sf_boolean;
			ex = ecn;
			break;
		}

		case node::f_value_list: {
			auto evl = make_shared<expr_value_list>();
			convert_sub( *evl, src );
			ex = evl;
			break;
		}

		case node::f_tuple: {
			auto et = make_shared<expr_tuple>();
			convert_sub( *et, src );
			ex = et;
			if( et->spec_type )
				et->spec_type->fun_class = intr_type::f_tuple;
			break;
		}

		case node::f_funccall: {
			//special functions
			NODE_CHECK( src, src.sub.size() == 2 );
			if( src.sub[0]->form == node::f_symbol ) {
				auto & symbol = src.sub[0]->text;
				
				if( symbol == "field" )
					return convert_field_expression( src );
				if( symbol == "select" )
					return convert_select_expression( src );
				if( symbol == "ifelse" )
					return convert_ifelse_expression( src );
				if( symbol == "optional_or" )
					return convert_optional_or_expression( src );
				if( symbol == "optional_cond" )
					return convert_optional_cond_expression( src );
			}
				
			auto ef = make_shared<expr_funccall>();
			ef->func = convert_expression( *src.sub[0] );
			ef->params = convert_expression( *src.sub[1] );
			ex = ef;
			break;
		}

		case node::f_funcdefn: {
			ex = convert_funcdefn_expression( src );
			omit_type = true;
			break;
		}
			
		case node::f_block: {
			auto eb = make_shared<expr_block>();
			eb->block = convert_block_inner( src, scope_style_t::imperative );
			ex = eb;
			break;
		}

		case node::f_const_string: {
			auto es = make_shared<expr_text>();
			es->what = expr_text::w_string;
			es->text = unicode::decode8(src.text);
			ex = es;
			break;
		}
		
		case node::f_const_char: {
			auto es = make_shared<expr_text>();
			es->what = expr_text::w_char;
			es->text = unicode::decode8(src.text);
			NODE_CHECK( src, es->text.size() == 1 );
			ex = es;
			break;
		}
			
		case node::f_subscript: {
			NODE_CHECK( src, src.sub.size() == 2 );
			auto es = make_shared<expr_subscript>();
			es->base = convert_expression( *src.sub[0] );
			es->offset = convert_expression( *src.sub[1] );
			ex = es;
			break;
		}

		case node::f_type:
		case node::f_named:
		case node::f_on:
		case node::f_expr_list:
			NODE_FAIL( src, "node type should not appear here" );
	}

	NODE_CHECK( src, ex);
	if( src.type && !omit_type ) {
		NODE_CHECK( src, !ex->spec_type );
		ex->spec_type = src.type;
	}
	ex->src_loc = src.src_loc;
	return ex;
}


shared_ptr<expression> node_converter::convert_field_expression( node const & src )
{
	NODE_CHECK( src, src.form == node::f_funccall );
	NODE_CHECK( src, src.sub.size() == 2 );
	NODE_CHECK( src, src.sub[0]->text == "field" );
	
	auto & params = *src.sub[1];
	NODE_CHECK( src, params.form == node::f_value_list );
	NODE_CHECK( src, params.sub.size() == 2 );
	
	auto fe = make_shared<expr_field>();
	fe->src_loc = src.src_loc;
	fe->base = convert_expression( *params.sub[0] );
	
	NODE_CHECK( src, params.sub[1]->form == node::f_symbol );
	fe->field = params.sub[1]->text;
	
	return fe;
}

void node_converter::convert_sub( expr_with_sub & out, node const & src )
{
	NODE_CHECK( src, !out.spec_type );
	out.spec_type = make_shared<type_spec>();
	bool need_type = false;
	
	for( size_t i=0; i < src.sub.size(); ++i ) {
		type_spec::sub_ref sr;
		
		auto & item = src.sub[i];
		if( item->form == node::f_funccall && item->sub.size() == 2
			&& item->sub[0]->text == "assign" ) {
			auto & params = *item->sub[1];
			NODE_CHECK( src, params.sub.size() == 2 );
			auto ex = convert_expression( *params.sub[1] );
			sr.name = params.sub[0]->text;
			need_type = true;
			out.sub.push_back( ex );
		} else {
			out.sub.push_back( convert_expression( *item ) );
		}
		out.spec_type->sub.push_back( sr );
	}
	
	if( !need_type )
		out.spec_type.reset();
}


shared_ptr<expression> node_converter::convert_ifelse_expression( node const & src )
{
	STD_FUNCTION( src, "ifelse" )
	NODE_CHECK( src, params.sub.size() == 3 );

	auto ex = make_shared<expr_switch>();
	
	expr_switch::expr_list_item eli;
	eli.cond = convert_expression( *params.sub[0] );
	eli.value = convert_expression( *params.sub[1] );
	ex->expr_list.push_back( eli );
	
	//const_bool true
	expr_switch::expr_list_item eli_else;
	eli_else.value = convert_expression(*params.sub[2]);
	ex->expr_list.push_back( eli_else );
	
	return ex;
}

shared_ptr<expression> node_converter::convert_optional_or_expression( node const & src )
{
	STD_FUNCTION( src, "optional_or" );
	NODE_CHECK( src, params.sub.size() == 2 );
	
	auto ex = make_shared<expr_conditional>();
	ex->op = expr_conditional::o_optional_or;
	ex->cond = convert_expression( *params.sub[0] );
	//a is not set here, implied from cond
	ex->b = convert_expression( *params.sub[1] );
	return ex;
}

shared_ptr<expression> node_converter::convert_optional_cond_expression( node const & src )
{
	STD_FUNCTION( src, "optional_cond" );
	NODE_CHECK( src, params.sub.size() == 2 );
	
	auto ex = make_shared<expr_conditional>();
	ex->op = expr_conditional::o_optional_cond;
	ex->cond = convert_expression( *params.sub[0] );
	ex->a = convert_expression( *params.sub[1] );
	//b is not used
	return ex;
}

shared_ptr<expression> node_converter::convert_select_expression( node const & src )
{
	STD_FUNCTION( src, "select" );
	
	auto ex = make_shared<expr_switch>();
	ex->src_loc = src.src_loc;

	for( auto & ni : params.sub ) {
		expr_switch::expr_list_item eli;
		
		//in thsi form the conditional has special meaning
		if( ni->form == node::f_funccall 
			&& ni->sub[0]->form == node::f_symbol
			&& ni->sub[0]->text == "optional_cond" ) {
			STATE_CHECK( ni->sub[1]->sub.size() == 2 );
			eli.cond = convert_expression( *ni->sub[1]->sub[0] );
			eli.value = convert_expression( *ni->sub[1]->sub[1] );
		} else {
			eli.value = convert_expression( *ni );
		}
		ex->expr_list.push_back( eli );
	}
	
	return ex;
}

shared_ptr<do_statement> node_converter::convert_do_statement( node const & list )
{
	NODE_CHECK( list, list.sub.size() == 2 );
	auto & src = *list.sub[1];
	
	shared_ptr<expression> value;
	
	if( src.form == node::f_on ) {
		value = convert_expression( *src.sub[0] );
	} else {
		value = convert_expression( src );
	}
	
	shared_ptr<do_statement> ds( new do_statement );
	ds->src_loc = src.src_loc;
	//TODO: eventually this must be cleaned up to accept anything and let the typer
	//deal with it. (It would require block's as closures first though...)
	if( value->form == expression::f_switch ) {
		auto & es = static_cast<expr_switch&>(*value);
		for( auto & item : es.expr_list ) {
			
			do_statement::expr_list_item eli;
			eli.cond = item.cond;
			NODE_CHECK( src, item.value->form == expression::f_block );
			eli.block = static_cast<expr_block&>(*item.value).block;
			ds->expr_list.push_back( eli );
		}
		
	} else if( value->form == expression::f_conditional ) {
		auto & ec = static_cast<expr_conditional&>(*value);
		do_statement::expr_list_item eli;
		if( ec.op == expr_conditional::o_optional_cond ) {
			eli.cond = ec.cond;
			NODE_CHECK( src, ec.a->form == expression::f_block );
			eli.block = static_cast<expr_block&>(*ec.a).block;
		} else {
			NODE_FAIL( src, "expecting optional_cond" );
		}
		ds->expr_list.push_back( eli );
	} else if( value->form == expression::f_block ) {
		do_statement::expr_list_item eli;
		eli.block = static_cast<expr_block*>(value.get())->block;
		ds->expr_list.push_back( eli );
	} else {
		NODE_FAIL( src, "expecting switch or block" );
	}
	
	if( src.form == node::f_on ) {
		convert_on_block( *src.sub[1], *src.sub[2], *ds );
	}
	
	if( list.sub[0]->text == "loop" ) {
		ds->loop = true;
	}
	
	return ds;
}

shared_ptr<statement_block> node_converter::convert_block_inner( node const & src,
	scope_style_t scope_style )
{
	shared_ptr<statement_block> block( new statement_block( scope_style ) );
	
	for( size_t i=0; i < src.sub.size(); ++i )
	{
		shared_ptr<statement> sub = convert( *src.sub[i] );
		if( scope_style == scope_style_t::declarative )
			NODE_CHECK( *src.sub[i], sub->is_declarative(), "expect-declarative-block" );

		block->list.push_back( sub );
	}

	return block;
}

/**
	An "assign funcall" or "symbol" form can convert to a var_statement.
*/
shared_ptr<var_statement> node_converter::convert_var_statement( node const & list )
{
	NODE_CHECK( list, list.sub.size() == 2 );
	auto & src = *list.sub[1];

	node const* symbol_node = &src;
	node const* value_node = nullptr;
	
	bool init_refer_value = false;
	if( src.form == node::f_funccall )
	{
		NODE_CHECK( src, src.sub.size() == 2 );
		
		if( src.sub[0]->text == "assign" )
			init_refer_value = false;
		else if( src.sub[0]->text == "assign_value" )
			init_refer_value = true;
		else
			NODE_FAIL( src, "invalid input node" );
		
		NODE_CHECK( src, src.sub[0]->form == node::f_symbol );
		NODE_CHECK( src, src.sub[1]->sub.size() == 2 );
		symbol_node = src.sub[1]->sub[0].get();
		value_node = src.sub[1]->sub[1].get();
	}
	
	NODE_CHECK( src, symbol_node->form == node::f_symbol );

	shared_ptr<var_statement> vs( new var_statement );
	vs->src_loc = src.src_loc;
	vs->init_refer_value = init_refer_value;
	vs->decl.reset( new var_decl );
	
	vs->decl->form = (list.sub[0]->text == "var") ? var_decl::f_single : var_decl::f_multi;

	for( auto it = list.annotations.begin(); it != list.annotations.end(); ++it ) {
		if( it->first == "import" || it->first == "import_data" ) {
			auto & sub = it->second->sub;
			NODE_CHECK( *it->second, sub.size() >= 1 && sub.size() <= 2 );
			vs->decl->import_type = it->first == "import" ? var_decl::it_abi_code : var_decl::it_abi_data;
			
			//split name into module.name if needed
			auto & sym = *sub[0];
			NODE_CHECK( sym, sym.form == node::f_const_string );
			size_t dot = sym.text.find_first_of( '.' );
			if( dot != std::string::npos ) {
				vs->decl->import_module = sym.text.substr(0,dot);
				vs->decl->import_name = sym.text.substr(dot+1);
			} else {
				vs->decl->import_name = sym.text;
			}
			
			//extract optional form
			if( sub.size() > 1 ) {
				auto & form = *sub[1];
				NODE_CHECK( form, form.form == node::f_symbol );
				if( form.text == "leaf" ) {
					vs->decl->name_form = var_decl::nf_leaf;
				} else if( form.text == "bare" ) {
					vs->decl->name_form = var_decl::nf_bare;
				} else {
					NODE_FAIL( form, "import-unknown-name-form" );
				}
			}
		} else if( it->first == "export" ) {
			vs->decl->is_export = true;
		} else {
			NODE_FAIL( *it->second, "unknown annotation" );
		}
	}

	vs->decl->symbol = symbol_node->text;
	vs->decl->spec_type = symbol_node->type;

	if( value_node )
		vs->value = convert_expression( *value_node );
	
	return vs;
}

shared_ptr<assign_statement> node_converter::convert_assign_statement( node const & src, 
	assign_how_t how )
{
	NODE_CHECK( src, src.form == node::f_funccall );
	NODE_CHECK( src, src.sub.size() == 2 );
	
	node const & func = *src.sub[0];
	NODE_CHECK( func, func.form == node::f_symbol );

	node const & params = *src.sub[1];
	NODE_CHECK( params, params.form == node::f_value_list );
	NODE_CHECK( params, params.sub.size() == 2 );

	shared_ptr<assign_statement> as( new assign_statement );
	as->src_loc = src.src_loc;
	as->left = convert_expression(*params.sub[0]);
	switch( how ) {
		case ah_value: as->how = assign_statement::h_value; break;
		case ah_init_refer: as->how = assign_statement::h_init_refer; break;
		case ah_default: as->how = assign_statement::h_default; break;
	}
	
	as->right = convert_expression(*params.sub[1]);
	return as;
}

shared_ptr<defer_statement> node_converter::convert_defer_statement( node const & list, bool error_only )
{
	NODE_CHECK( list, list.sub.size() == 2 );
	auto & src = *list.sub[1];
	
	shared_ptr<expression> value = convert_expression( src );
	
	NODE_CHECK( src, value->form == expression::f_block );
	
	shared_ptr<defer_statement> defer( new defer_statement );
	defer->src_loc = src.src_loc;
	defer->error_only = error_only;
	defer->block = static_cast<expr_block*>(value.get())->block;
	
	return defer;
}

shared_ptr<type_statement> node_converter::convert_type_statement( node const & src, bool alias )
{
	NODE_CHECK( src, src.sub.size() == 2 );
	NODE_CHECK( src, src.sub[1]->form == node::f_symbol );
	
	auto ts = make_shared<type_statement>();
	ts->src_loc = src.src_loc;
	ts->name = src.sub[1]->text;
	ts->what = alias ? type_statement::w_alias : type_statement::w_prototype;
	ts->spec_type = src.sub[1]->type; //may be null
		
	return ts;
}

shared_ptr<type_statement> node_converter::convert_class_statement( node const & src )
{
	NODE_CHECK( src, src.sub.size() == 3 );
	NODE_CHECK( src, src.sub[1]->form == node::f_symbol );
	NODE_CHECK( src, src.sub[2]->form == node::f_block );
	auto ts = make_shared<type_statement>();
	ts->src_loc = src.src_loc;
	ts->name = src.sub[1]->text;
	ts->what = type_statement::w_class;
	ts->class_block = convert_block_inner( *src.sub[2], scope_style_t::declarative );
	return ts;
}


shared_ptr<return_statement> node_converter::convert_return_statement( node const & src )
{
	NODE_CHECK( src, src.sub.size() <= 2, "too-many-return-values" );
	auto rs = make_shared<return_statement>();
	rs->src_loc = src.src_loc;	
	if( src.sub.size() == 2 ) {
		rs->value = convert_expression( *src.sub[1] );
	}
	return rs;
}

shared_ptr<expression> node_converter::convert_new_expression( node const & src )
{
	NODE_CHECK( src, src.sub.size() == 1 );
	NODE_CHECK( src, src.sub[0]->form == node::f_type );
	
	auto nf = make_shared<expr_new_func>();
	nf->new_spec_type = make_shared<type_spec>( *src.sub[0]->type );
	return nf;
}

shared_ptr<statement> node_converter::convert_resume_statement( node const & src )
{
	NODE_CHECK( src, src.sub.size() == 0 );
	
	auto rs = make_shared<fail_statement>( fail_level::resume );
	rs->src_loc = src.src_loc;
	return rs;
}

shared_ptr<statement> node_converter::convert_fail_statement( node const & src, fail_level level )
{
	NODE_CHECK( src, src.sub.size() == 0 );
	
	auto fs = make_shared<fail_statement>( level );
	fs->src_loc = src.src_loc;
	return fs;
}


shared_ptr<statement> node_converter::convert_break_statement( node const & src, bool cont )
{
	NODE_CHECK( src, src.sub.size() == 0 );
	
	auto bs = make_shared<loop_flow_statement>();
	bs->src_loc = src.src_loc;
	bs->flow = cont ? loop_flow_statement::f_next : loop_flow_statement::f_break;
	return bs;
}

/**
	Conver the node into a type. This works since the node contains
	as much information as the type -- though it must valid that only type-correct
	specs have been used.
*/
shared_ptr<type_spec> node_converter::convert_to_type( shared_ptr<node> params,
	sig_assigns_t * assigns )
{
	auto type = make_shared<type_spec>();
	type->fun_class = intr_type::f_value_list;

	for( size_t i=0; i < params->sub.size(); ++i )
	{
		auto ex = params->sub[i];
		
		if( ex->form == node::f_symbol ) {
			//
		} else if( ex->form == node::f_funccall ) {
			STD_FUNCTION( *ex, "assign" );
			NODE_CHECK( params, assigns, "default value not allowed here" );
			NODE_CHECK( params, params.sub.size() == 2 );
			ex = params.sub[0];
		
			sig_assign sa;
			sa.index = i;
			sa.value = convert_expression( *params.sub[1] );
			assigns->push_back(sa);
		} else {
			NODE_FAIL( *params, "invalid type syntax" );
		}
		
		auto add = make_shared<type_spec>();
		if( ex->type ) {
			*add = *ex->type;
		} else {
			add->fun_class = intr_type::f_unknown;
		}

		type->push_sub( ex->text, add );
	}
	
	return type;
}

/**
	Convert what we though was an node into a type.
*/
shared_ptr<type_spec> node_converter::convert_to_type( 
	shared_ptr<type_spec> type,
	shared_ptr<node> params,
	shared_ptr<node> returns, sig_assigns_t & param_assigns )
{
	NODE_CHECK( *params, type->fun_class.is_unset() );
	NODE_CHECK( *params, type->sub.empty() );
	type->fun_class = intr_type::f_function;
	type->push_sub( convert_to_type( params, &param_assigns ) );
	type->push_sub( convert_to_type( returns, nullptr ) );
	
	return type;
}

shared_ptr<expression> node_converter::convert_funcdefn_expression( node const & src )
{
	NODE_CHECK( src, src.sub.size() == 3 );
	
	auto ef = make_shared<expr_funcdefn>();
	sig_assigns_t assigns;
	if( src.type ) {
		ef->spec_type = make_shared<type_spec>( *src.type );
	} else {
		ef->spec_type = make_shared<type_spec>();
	}
	convert_to_type( ef->spec_type, src.sub[0], src.sub[1], assigns );
	ef->block = convert_block_inner( *src.sub[2], scope_style_t::imperative );
	
	for( auto & sa : assigns ) {
		NODE_CHECK( src, sa.index < ef->spec_type->sub[0].type->sub.size() );
		ef->param_defaults[sa.index] = sa.value;
	}
	
	return ef;
}

shared_ptr<statement> node_converter::convert_import_statement( node const & list )
{
	NODE_CHECK( list, list.sub.size() == 2 );
	auto & src = *list.sub[1];
	
	NODE_CHECK( src, src.sub.size() == 0 );
	NODE_CHECK( src, src.form == node::f_symbol );
	
	auto is = make_shared<import_statement>();
	is->mod_name = src.text;
	
	return is;
}

shared_ptr<for_statement> node_converter::convert_for_statement( node const  & list )
{
	NODE_CHECK( list, list.sub.size() == 5 );
	NODE_CHECK( list, list.sub[2]->text == "in" );
	auto & item = *list.sub[1];
	auto & iterable = *list.sub[3];
	auto & block = *list.sub[4];
	
	auto fs = make_shared<for_statement>();
	NODE_CHECK( item, item.form == node::f_symbol );
	fs->item = item.text;
	fs->iterable = convert_expression( iterable );
	fs->block = convert_block( block, scope_style_t::imperative );
	
	return fs;
}

} //eon leaf