#include "parser/node_parser.hpp"
#include "parser/intr_type_parser.hpp"

#define SHOW_DEBUG 0
#include "util/debug.hpp"

#include "regex_patterns.hpp"
#include "util/unicode.hpp"

namespace
{
	enum precedence_level
	{
		pl_on = 10,
		pl_assign = 50,
		pl_then = 75,
		pl_compare = 100,
		pl_binary = 200,
		pl_unary = 300,
		//refer to unit_tests/block/tuple/call_prec.lfb to see equal precidence justification
		pl_function = 400,
		pl_field = 400,
	};
} //eon anon

namespace leaf
{

typedef boost::basic_regex<src_char_t> regex;

/*static*/ shared_ptr<node> node_parser::parse_as_block( source & src )
{
	try {
		return parse_compound( src, 0, 0, ct_block );
	} catch( recoverable_error & re ) {
		re << error_parser(true);
		throw;
	}
}

/*static*/ shared_ptr<node> node_parser::parse_single( source & src )
{
	try {
		source orig = src;
		
		node_parser ep;
		return ep.parse( src, line_ends_terminate );
	} catch( recoverable_error & re ) {
		re << error_parser(true);
		throw;
	}
}

void node_parser::push_op( op const & o )
{
	DEBUG( "push_op: ", o.text );

	is_value = false;

	while( true )
	{
		if( op_stack.empty() )
		{
			op_stack.push_back( o );
			return;
		}

		//if higher precendence then just push and keep processing
		if( o.precedence > op_stack.back().precedence )
		{
			op_stack.push_back( o );
			return;
		}
		//if equal precedence then check left/right associativity
		if( o.precedence == op_stack.back().precedence
			&& o.right_assoc )
		{
			op_stack.push_back( o );
			return;
		}

		pop_op();
		is_value = false;
	}
}

shared_ptr<node> node_parser::pop_out()
{
	STATE_CHECK( out_stack.size() > 0 );
	shared_ptr<node> ex = out_stack.back();
	out_stack.pop_back();
	return ex;
}

void node_parser::pop_op()
{
	op o = op_stack.back();
	op_stack.pop_back();

	DEBUG( "pop_op: ", o.text );

	auto fex = make_shared<node>( node::f_funccall, o.src_loc );
	fex->sub.resize( 2 );

	fex->sub[0] = make_shared<node>( node::f_symbol, o.src_loc );
	fex->sub[0]->text = o.text;
	
	switch( o.form )
	{
		case op::f_ternary:
		{
			STATE_CHECK( out_stack.size() >= 3, "ternary operator requires three parameters" );

			auto p3 = pop_out();
			auto p2 = pop_out();
			auto p1 = pop_out();
			
			if( o.text == "on" ) {	
				fex->form = node::f_on;
				fex->sub.resize(3);
				fex->sub[0] = p1;
				fex->sub[1] = p2;
				fex->sub[2] = p3;
			} else {
				auto part = make_shared<node>( node::f_value_list, p1->src_loc );
				part->sub.resize(3);
				part->sub[0] = p1;
				part->sub[1] = p2;
				part->sub[2] = p3;
				fex->sub[1] = part;
			}
			push_out(fex);
			break;
		}
		
		case op::f_binary:
		{
			//normalize to a function
			STATE_CHECK( out_stack.size() >= 2, "binary operator requires two parameters",
				util::source_location::error( o.src_loc ) );

			auto p2 = pop_out();
			auto p1 = pop_out();

			//convert special functions into switch
			//the ?| becomes ternary when used together
			if( o.text == "optional_or" && p1->form == node::f_funccall
				&& p1->sub.size() > 0 && p1->sub[0]->text == "optional_cond" ) {
				DEBUG( "TCELSE", p1->form, p1->sub[0]->text, p1->sub.size() );
				p1->sub[0]->text = "ifelse";
				p1->sub[1]->sub.push_back( p2 );
				fex = p1;
			} else if( o.text == "named" ) {
				fex->form = node::f_named;
				fex->text = "named";
				fex->sub[0] = p1;
				fex->sub[1] = p2;
			} else if( o.text == "subscript" ) {
				fex->form = node::f_subscript;
				fex->text = "subscript";
				fex->sub[0] = p1;
				fex->sub[1] = p2;
			} else {
				auto part = make_shared<node>( node::f_value_list, p1->src_loc );
				part->sub.resize( 2 );
				part->sub[1] = p2;
				part->sub[0] = p1;
				fex->sub[1] = part;
			}
			push_out( fex );
			break;
		}

		case op::f_unary:
		{
			fex->sub[1] = make_shared<node>( node::f_value_list, o.src_loc );
			auto p = pop_out();
			fex->sub[1]->sub.resize( 1 );
			fex->sub[1]->sub[0] = p;

			push_out( fex );
			break;
		}

		case op::f_funccall:
		{
			STATE_CHECK( out_stack.size() >= 2, "function operator requires function and tuple" );
				
			fex->sub[1] = pop_out();
			fex->sub[0] = pop_out();

			push_out( fex );
			break;
		}
			
		case op::f_subscript:
		{
			STATE_CHECK( out_stack.size() >=2, "subscript requires base and offset" );
			fex->form = node::f_subscript;
			fex->sub[1] = pop_out();
			fex->sub[0] = pop_out();
			push_out( fex );
			break;
		}
	}
}

void node_parser::push_out( shared_ptr<node> ex )
{
	DEBUG( "push_out: ", ex->text );
	out_stack.push_back( ex );
	is_value = true;
}

namespace
{
	/**
		A static table of binary operators. Note how awkward it is to do this type of
		thing in C++.
	*/
	struct operator_map
	{
		struct op
		{
			char const * symbol;
			char const * name;
			unsigned precedence;
			bool right_assoc;
		};

		struct bstrcmp
		{
			bool operator()( char const * a, char const * b ) const
			{ return strcmp( a, b ) < 0; }
		};

		typedef std::map<char const*, op,bstrcmp> ops_map_t;
		ops_map_t ops_map;

		operator_map( std::initializer_list<op> base );

		op const & get( char const * text ) const {
			ops_map_t::const_iterator it = ops_map.find( text );
			STATE_CHECK( it != ops_map.end(), "unknown operator",
					node_parser::error_symbol( text ) );
			return it->second;
		}
		
		bool has( char const * text ) const {
			return ops_map.find( text ) != ops_map.end();
		}

	};

	operator_map::operator_map( std::initializer_list<op> base )
	{
		for( auto it = base.begin(); it != base.end(); ++it )
			ops_map[it->symbol] = *it;
	}
	
	operator_map binary_operators({
		{ "=", "assign", pl_assign, false }, //though in C it is right associative I don't want to support chaining of = (such as a=b=c;)
		{ ":=", "assign_value", pl_assign, false },
		{ "|=", "assign_default", pl_assign, false },

		{ ".", "field", pl_field, false },
		{ "#", "subscript", pl_field, false },
		
		{ "?", "optional_cond", pl_then, true },
		{ "|", "optional_or", pl_then-1, true },

		{ "==", "eq", pl_compare, false },
		{ "!=", "neq", pl_compare, false },
		{ "<", "lt", pl_compare, false },
		{ ">", "gt", pl_compare, false },
		{ "<=", "lte", pl_compare, false },
		{ ">=", "gte", pl_compare, false },
		
		{ "+", "add", pl_binary, false },
		{ "-", "sub", pl_binary, false },
		{ "++", "concat", pl_binary, false },
		{ "*", "mul", pl_binary + 10, false },
		{ "/", "div", pl_binary + 10, false },
		{ "^", "pow", pl_binary + 50, true },
	});

	operator_map unary_operators({
		{ "-", "neg", pl_unary, true },
		{ "else", "else", pl_then, true },
	});
} //eon anon

node_parser::op node_parser::binary_op( std::string const & text, util::source_location loc )
{
	op o( loc );
	o.form = op::f_binary;

	if( text == "<-" ) {
		o.precedence = pl_function;
		o.form = op::f_funccall;
		o.text = text;
	} else {
		auto  & sop = binary_operators.get( text.c_str() );
		o.precedence = sop.precedence;
		o.right_assoc = sop.right_assoc;
		o.text = sop.name;
	}

	return o;
}

node_parser::op node_parser::unary_op( std::string const & text, util::source_location loc )
{
	op o( loc );
	o.form = op::f_unary;

	auto & sop = unary_operators.get( text.c_str() );
	o.precedence = sop.precedence;
	o.right_assoc = sop.right_assoc;
	o.text = sop.name;

	return o;
}

/*static*/ shared_ptr<node> node_parser::parse_compound( source & src,
	src_char_t start, src_char_t end, compound_type_t ct )
{
	DEBUG( "parse_tuple", end );
	auto tuple = make_shared<node>( 
		ct == ct_block ? node::f_block : 
		ct == ct_value_list ? node::f_value_list : node::f_tuple, 
		src.cur_location() );

	if( start && (!src.has_next(source::skip_line_ends) || src.next() != start) )
		src.throw_expecting( "tuple" );
	
	unsigned flags = 0;
	if( ct == ct_block )
		flags |= line_ends_terminate;
		
	while( true )
	{
		if( !src.has_next() ) {
			if( !end )
				break;
			src.throw_unexpected( "end of tuple" );
		}

		//check for completion
		char nc = src.peek_next();
		if( nc == end ) {
			src.next();
			break;
		}
		
		//was it an item, and what is hte block type
		if( nc == ';' || (ct == ct_block && nc == '\n') ) {
			if( ct != ct_block )
				src.throw_invalid( "statement combined in tuple" );
			tuple->form = node::f_block;
			
			src.next();
			src.skip_space(source::skip_line_ends);
			continue;
		}
		if( nc == ',' ) {
			if( ct == ct_block )
				src.throw_invalid( "expression combined in statement block" );
			
			src.next();
			src.skip_space(source::skip_line_ends);
			continue;
		}
		
		if( nc == '\n' ) {
			src.next();
			continue;
		}

		node_parser np;
		tuple->sub.push_back( np.parse( src, flags ) );
	}

	return tuple;
}

/**
	Parsing of a single statement. Though some parts are recursive this outer
	wrapper has a specific form distinct to that (allowing additional markers
	like declarators and annotations).
	
	TODO: This function is too big.
*/
shared_ptr<node> node_parser::parse( source & src, unsigned flags )
{
	reset();
	
	src.skip_space(source::skip_line_ends);
	
	//pre-annotaitons
	node::annotate_map_t anno;
	parse_annotations( src, anno );
	//TODO: perhaps skip line ends again?
	src.skip_space();
	auto decl_loc = src.cur_location();
	
	auto push_value = [&]( shared_ptr<node> const & value ) {
		//pending stack must be collapsed to allow multiple expressions
		if( is_value ) {
			while( op_stack.size() )
				pop_op();
		}
		push_out( value );
	};
	
	while( true ) {
		if( !src.has_next() ) {
			break;
		}

		src_char_t nc = src.peek_next();
		
		//a line-end terminates the current expression
		if( nc == '\n' ) {
			src.next();
			if( (flags & line_ends_terminate) )
				break;
			continue;
		}
		
		if( nc == ',' || nc == ';' || nc == '@' || nc == ']' || nc == ')' || nc == '}' )
			break;
		
		//open block
		if( nc == '{' ) {
			push_value( parse_block( src ) );
			continue;
		} else if( nc == '[' ) {
			push_value( parse_tuple( src ) );
			continue;
		}
		
		//indicate string
		if( nc == '\"' || nc == '\'' ) {
			auto ex = make_shared<node>( node::f_const_string, src.cur_location() );
			ex->text = parse_string( src, nc );
			if( nc == '\'' ) {
				ex->form = node::f_const_char;
				if( unicode::len8( ex->text ) != 1 )
					src.throw_expecting( "char constant" );
			}
			push_value( ex );
			continue;
		}
		
		//consume all operator like symbols into one, this prevents visual ambiguity and allows for operator extension later
		//paired symbols are however consume one-at-a-time
		bool is_operator = src.match( leaf_parse::re_operator );
		
		//indicate type
		if( is_operator && src.cur_match() == ":" ) {
			if( out_stack.size() == 0 ) {
				//introduce anonymous typed node
				push_value( make_shared<node>(node::f_symbol, src.cur_location() ) );
			}
			out_stack.back()->type = intr_type_parser::parse( src );
			continue;
		}

		static regex re_symbol( SYMBOL_REGEX );
		//specifically disallow leading/trailing dot (may lead to ambiguity with other operators)
		static regex re_number( "[0-9_]+(\\.[0-9_]+)?([eE][+-]?[0-9]+)?" );
		static regex re_base_number( "0([xbo])([0-9a-fA-F_]+(\\.[0-9a-fA-F_]+)?)" );
		static regex re_bool( "true|false" EXCLUDE_SYMBOL_REGEX_CHAR );
		
		static regex re_special( "(new|on)" EXCLUDE_SYMBOL_REGEX_CHAR );
		
		//this could be a function call, or a value
		if( is_operator && src.cur_match() == "(" ) {
		
			if( is_value ) {
				//is it a function call (this is treated as a binary operator of the function and its parameters as a tuple)
				op o( src.cur_location() );
				o.form = op::f_funccall;
				o.precedence = pl_function;
				o.text = "<-";
				push_op( o );
			}
			
			//parses like a tuple but is a value-list
			shared_ptr<node> ex = parse_compound( src, 0, ')', ct_value_list );
			push_value( ex );
		}
		else if( !is_operator && src.match( re_special ) ) {
			if( src.cur_match() == "on" ) {
				op o( src.cur_location() );
				o.form = op::f_ternary;
				o.precedence = pl_on;
				o.right_assoc = false;
				o.text = "on";
				push_op( o );
				
				if( !src.has_next() || !src.match( re_symbol ) )
					src.throw_expecting( "error class" );
				auto sym = make_shared<node>( node::f_symbol, src.cur_location() );
				sym->text = src.cur_match();
				push_out( sym );
				
				auto block = parse_block( src );
				push_out( block );
				
				pop_op();
			} else if( src.cur_match() == "new" ) {
				auto nt = make_shared<node>( node::f_type, src.cur_location() );
				nt->type = intr_type_parser::parse_inline( src );
				
				auto fex = make_shared<node>( node::f_symbol, src.cur_location() );
				fex->text = "new";
				fex->sub.push_back( nt );
				push_value(fex);
			
			} else {
				STATE_FAIL( "failed regex" );
			}
		}
		//otherwise we could expect a prefix unary operator. Use the same regex however to
		//also avoid operator ambiguity
		else if( is_operator )
		{
			DEBUG( "OP: ", src.cur_match() );
			if( src.cur_match() == "->" ) {
				auto func = make_shared<node>( node::f_funcdefn, src.cur_location() );
				
				//specified or default empty params
				if( out_stack.size() && out_stack.back()->form == node::f_value_list ) {
					func->sub.push_back( out_stack.back() ); 
					out_stack.pop_back();
				} else {
					auto params = make_shared<node>( node::f_value_list, src.cur_location() );
					func->sub.push_back( params );
				}
				
				src.skip_space();
				if( src.peek_next() == '(' ) {
					auto returns = parse_value_list( src );
					func->sub.push_back( returns );
				} else {
					auto returns = make_shared<node>( node::f_value_list, src.cur_location() );
					func->sub.push_back( returns );
				}
				
				func->type = intr_type_parser::if_parse_inline( src );
				
				auto block = parse_block( src );
				func->sub.push_back( block );
				push_out( func ); //not push_value
			} else if( is_value && binary_operators.has( src.cur_match().c_str() ) ) {
				//is_value also checked to allow shared prefix operators (like -)
				push_op( binary_op( src.cur_match(), src.cur_location() ) );
			} else {
				push_op( unary_op( src.cur_match(), src.cur_location() ) );
			}
		}
		//if not then an actual value
		else
		{
			if( src.match( re_bool ) ) { //must be before symbol
				auto ex = make_shared<node>( node::f_const_boolean, src.cur_location());
				ex->text = src.cur_match();
				push_value( ex );
			} else if( src.match( re_symbol ) ) {
				auto ex = make_shared<node>( node::f_symbol, src.cur_location());
				ex->text = src.cur_match();
				push_value( ex );
			} else if( src.match( re_base_number ) ) {
				auto ex = make_shared<node_const_number>(src.cur_location());
				unsigned f;
				char c = src.cur_match(1)[0];
				if( c == 'x' )
					ex->value = util::number::parse_hexadecimal( src.cur_match(2), f );
				else if( c == 'o'  )
					ex->value = util::number::parse_octal( src.cur_match(2), f );
				else if( c == 'b'  )
					ex->value = util::number::parse_binary( src.cur_match(2), f );
				ex->was_float = f & util::number::pdf_had_decimal;
				push_value( ex );
			} else if( src.match( re_number ) ) {
				auto ex = make_shared<node_const_number>(src.cur_location());
				unsigned f;
				ex->value = util::number::parse_decimal( src.cur_match(), f );
				ex->was_float = f & util::number::pdf_had_decimal;
				push_value( ex );
			} else {
				src.throw_expecting( "value" );
			}
		}
	}

	while( op_stack.size() )
		pop_op();

	shared_ptr<node> result;
	
	if( out_stack.size() == 1 ) {
		//simple single node result
		result = out_stack[0];
	} else {
		//expression list result
		result = make_shared<node>( node::f_expr_list, decl_loc );
		result->sub = out_stack;
	}
	
	//parse_annotations(src, anno);
	result->annotations.swap(anno);
	
	//from start
	result->src_loc = decl_loc;
	return result;
}

void node_parser::parse_annotations( source & src, node::annotate_map_t & anno ) 
{
	//check annotations
	while( src.has_next() && src.peek_next() == '@' ) {
		static regex re_anno( "@(" SYMBOL_REGEX ")" );
		if( !src.match( re_anno ) )
			src.throw_expecting( "annotation" );
		std::string name = src.cur_match( 1 );

		//value-list used for annotation value
		if( src.has_next() && src.peek_next() == '(' ) {
			anno[name] = parse_value_list( src );
		} else {
			anno[name] = make_shared<node>(node::f_value_list, src.cur_location() );
		}
	}
}

/**
	The delimiters and escape sequences are handled as individual characters even
	though we may are parsing utf-8 files. This is okay since all these ASCII chars
	are represented exactly the same in utf-8, with no ambiguity.
	
	The supported escapes are a common set coming from C,C++,Java,Python.
*/
/*static*/ std::string node_parser::parse_string( source & src, src_char_t end )
{
	if( !src.has_next() )
		src.throw_expecting( "string open-quotes" );
	src.next();
	
	//parse fixed length hex sequences
	auto parse_hex = [&](int len) {
		unsigned val = 0;
		for( int i=0; i < len; ++i ) {
			val <<= 4;
			auto v = src.safe_next();
			if( v >= '0' && v <= '9' ) {
				val += (v-'0');
			} else if( v >= 'a' && v <= 'f' ) {
				val += (v-'a') + 10;
			} else if( v >= 'A' && v <= 'F' ) {
				val += (v-'A') + 10;
			} else {
				src.throw_invalid( "escape-sequence" );
			}
		}
		return val;
	};
	
	std::string ret;
	while( src.has_raw_next() )
	{
		src_char_t n = src.next();
		if( n == end )
			return ret;

		if( n == '\\' ) {
			n = src.safe_next();
			if( n == '0' ) {
				ret += '\0';
			} else if( n == '\'' ) {
				ret += char(0x27);
			} else if( n == '\"' ) {
				ret += char(0x22);
			} else if( n == '\\' ) {
				ret += char(0x5c);
			} else if( n == 'a' ) {
				ret += char(0x07);
			} else if( n == 'b' ) {
				ret += char(0x08);
			} else if( n == 'f' ) {
				ret += char(0x0c);
			} else if( n == 'n' ) {
				ret += char(0x0a);
			} else if( n == 'r' ) {
				ret += char(0x0d);
			} else if( n == 't' ) {
				ret += char(0x09);
			} else if( n == 'v' ) {
				ret += char(0x0b);
			} else if( n == 'x' ) {
				ret += unicode::encode8(parse_hex(2));
			} else if( n == 'u' ) {
				ret += unicode::encode8(parse_hex(4));
			} else if( n == 'U' ) {
				ret += unicode::encode8(parse_hex(8));
			} else {
				src.throw_unexpected( "escape-sequence" );
			}
		} else {
			ret += n;
		}
	}

	src.throw_expecting( "string close-quotes" );
	return "not reached";
}

} //eon leaf