#include <boost/test/unit_test.hpp>

#include "lang/language.hpp"
#include "parser/intr_type_parser.hpp"
#include "lang/typer.hpp"
#include "lang/dump.hpp"
#include "lang/type_converter.hpp"
#include "lang/type_identifier.hpp"

#include "parser/node_parser.hpp"
#include "parser/node_converter.hpp"

#include <iostream>

#define SHOW_DEBUG 0
#include "util/debug.hpp"

using namespace leaf;

BOOST_AUTO_TEST_SUITE( expr_type_test )

/**
	A test framework to handle the scope and expression parsing.
*/
struct tester
{
	shared_ptr<scope> sc;

	tester()
		: sc( scope::create( scope_style_t::imperative ) )
	{ }
	
	void add( char const * name, char const* type_code )
	{
		shared_ptr<var_decl> vs( new var_decl );
		vs->symbol = name;
		vs->form = var_decl::f_single;
		
		auto src = source::create( type_code, std::strlen( type_code ) );
		auto rt = type_identifier(*sc).determine( intr_type_parser::parse( *src ) );
		STATE_CHECK( rt.okay(), rt.explain_error() );
		vs->type = rt.result;
		
		sc->add_single( vs );
	}
	
	void bind( char const * name, char const* type_code )
	{
		shared_ptr<var_decl> vs( new var_decl );
		vs->symbol = name;
		vs->form = var_decl::f_multi;
		
		auto src = source::create( type_code, std::strlen( type_code ) );
		auto rt = type_identifier(*sc).determine( intr_type_parser::parse( *src ) );
		STATE_CHECK( rt.okay(), rt.explain_error() );
		vs->type = rt.result;
		
		sc->add_multi( vs );
	}

	shared_ptr<expression> parse_expr( char const * code )
	{
		auto src = source::create( code, std::strlen( code ) );

		auto n = node_parser::parse_single( *src );
		return node_converter::expression_convert( *n );
	}
	
	void operator()( char const * code )
	{
		typer t(sc);
		t.full_type( parse_expr( code ) );
	}
	
	void operator()( char const * code, char const * expect, unsigned flags = 0 )
	{
		try
		{
			shared_ptr<expression> ex = parse_expr( code );
			#if SHOW_DEBUG
			{
				leaf::dump d( std::cout );
				d.set_validate(false);
				d.set_show_spec_type( true );
				d.write_expression( *ex );
			}
			#endif
			typer t(sc);
			ex = t.full_type( ex, flags );
			leaf::dump d;
			d.write_intr_type( ex->type );
			std::string result = d.str();
			if( result != expect )
			{
				std::cerr
					<< "Failed: " << code << std::endl
					<< "Expect: " << expect << std::endl
					<< "Actual: " << result << std::endl
					;
				BOOST_FAIL( "Failed" );
			}
		}
		catch( ... )
		{
			std::cerr << "Failed: " << code << std::endl;
			throw;
		}

		BOOST_CHECK(true);
	}

	void fail( char const* code, char const * reason )
	{
		shared_ptr<expression> ex = parse_expr( code );
		try
		{
			typer t(sc);
			t.full_type( ex );
			BOOST_FAIL( "Should not have typed: " << code );
		}
		catch( recoverable_error & re )
		{
			std::string what = re.what();
			if( what.find_first_of( reason ) != what.npos )
				BOOST_CHECK( true );
			else
				BOOST_FAIL( what );
		}
	}
	
};

BOOST_AUTO_TEST_CASE( basic )
{
	tester t;

	t( "1", "rational 1bit" );
	//t( "-1", "rational 2bit signed" ); //TODO: need constant folding first

	t.add( "a", "integer" );
	t( "a", "integer lvalue" );

	t.add( "b", "float" );
	t( "b", "float lvalue" );
	
	//lvalue is always lost in tuples
	t( "[a,b]", "[ : integer, : float ]" );
	t( "[[a],1]","[ : [ : integer ], : rational 1bit ]" );

	//but preserved in value-lists
	t( "(a,b)", "( : integer lvalue, : float lvalue )" );
	
	//typing does not collapse 1-value-lists, only type-conversion does
	t( "(a)", "( : integer lvalue )" );
}

BOOST_AUTO_TEST_CASE( boolean )
{
	tester t;
	
	t( "true", "boolean" );
	t( "false", "boolean" );
	
	t.add( "a", "boolean" );
	t( "a", "boolean lvalue" );
}

BOOST_AUTO_TEST_CASE( binary_function )
{
	tester t;

	t.add( "madd", "(:integer,:integer)->(:integer)" );
	t( "madd(1,2)", "integer" );

	t.add( "a", "integer" );
	t.add( "b", "integer" );
	t( "madd(madd(a,3),b)", "integer" );

	t.add( "foo", "(:integer,:integer)->(:float,:integer)" );
	t("foo(madd(1,2),b)","( : float, : integer )" );
}

BOOST_AUTO_TEST_CASE( core_ops )
{
	tester t;
	t( "1+2", "integer" );
}

/**
 	For rationals a normal "float" is the preferred conversion to "low/high". For fixed floats the
 	closest higher matching float. When mixed in signature the preferrence aspect is the lowest
 	priority.
*/
BOOST_AUTO_TEST_CASE( float_preference )
{
	tester t;

	t.add( "fs", "float low" );
	t.add( "fn", "float normal" );
	t.add( "fl", "float high" );
	
	t.bind( "fo", "(:float)->(:float)" );
	t.bind( "fo", "(:float low)->(:float low)" );
	t.bind( "fo", "(:float high)->(:float high)" );
	t( "fo( 1.5 )", "float" );
	t( "fo(fs)", "float low" );
	t( "fo(fl)", "float high" );
	t( "fo(fn)", "float" );
	
	t.bind( "foo", "(:float,:float)->(:float)" );
	t.bind( "foo", "(:float low,:float low)->(:float low)" );
	t.bind( "foo", "(:float high,:float high)->(:float high)" );
	
	t( "foo( 3.5, 4.5 )", "float" );
	//t( "foo( fs, 4.5 )", "float low" ); //undecided now, this is ambiguous in the current rules
	t( "foo( 3.5, fl )", "float high" );
	t( "foo( fs, fn )", "float" );
	t( "foo( fs, fl )", "float high" );
}

BOOST_AUTO_TEST_CASE( integer_preference )
{
	tester t;

	//use size variations large enough to hopefully be platform independent
	t.add( "is", "integer 16bit" );
	t.add( "in", "integer" );
	t.add( "il", "integer 128bit" );
	
	t.bind( "fo", "(:integer)->(:integer)" );
	t.bind( "fo", "(:integer 16bit)->(:integer 16bit)" );
	t.bind( "fo", "(:integer 128bit)->(:integer 128bit)" );
	t( "fo( 1 )", "integer" );
	t( "fo(is)", "integer 16bit" );
	t( "fo(il)", "integer 128bit" );
	t( "fo(in)", "integer" );
	
	t.bind( "foo", "(:integer,:integer)->(:integer)" );
	t.bind( "foo", "(:integer 16bit,:integer 16bit)->(:integer 16bit)" );
	t.bind( "foo", "(:integer 128bit,:integer 128bit)->(:integer 128bit)" );
	
	t( "foo( 3, 4 )", "integer" );
	//t( "foo( is, 4 )", "integer 16bit" ); //undecided now, currently ambiguous
	t( "foo( 3, il )", "integer 128bit" );
	t( "foo( is, in )", "integer" );
	t( "foo( is, il )", "integer 128bit" );
}

BOOST_AUTO_TEST_CASE( unsigned_preference )
{
	tester t;
	
	t.add( "un", "binary" );
	t.add( "in", "integer" );
	
	t.bind( "num", "(:integer)->(:integer)" );
	t.bind( "num", "(:binary)->(:binary)" );

	t( "num(13)", "integer" ); //integer conversion is preferred
	t( "num(un)", "binary" );
	t( "num(in)", "integer" );
	
	t.bind( "under", "(:binary)->(:binary)" );
	t( "under(13)", "binary" );
	
	t.add( "unbit32", "binary 32bit" );
	t( "num(unbit32)", "binary" );
}

BOOST_AUTO_TEST_CASE( rational_match )
{
	tester t;
	t.add( "madd", "(:float,:float)->(:float)" );
	t( "madd(1,2)", "float" ); //rationals kept until conversion to float
}

BOOST_AUTO_TEST_CASE( multi_match )
{
	tester t;
	t.bind( "madd", "( : integer, : integer)->( : integer)" );
	t.bind( "madd", "( : float, : float)->( : float)" );
	t( "madd(1,2)", "integer" );
	t( "madd(1.0,2.0)", "integer" ); //rationals are best match for integer still .0 does not change the value
	t( "madd(1.1,2)", "float" ); //one matching float will be enough to get float version
	t( "madd(1,2.1)", "float" );
	
 	t.bind( "promote", "( : integer)->( : float)" );
 	t( "madd(promote(1),2.5)", "float" );
 	
 	t.bind( "trunc", "(:float)->(:integer)" );
 	t( "madd(1,trunc(2.5))", "integer" );
}

BOOST_AUTO_TEST_CASE( funcdefn )
{
	tester t;
	//testing the anonymous name, not the raw bit
	t( "()->(:integer){}", "( ) -> ( : integer ) raw" );
}

BOOST_AUTO_TEST_CASE( tuple_names )
{
	tester t;
	t( "[ a =12 ]", "[ a : rational 4bit ]" );
	t( "[ 3, b = 4.5, c= [ 0 ] ]", "[ : rational 2bit, b : rational, c : [ : rational 1bit ] ]" );
}

BOOST_AUTO_TEST_CASE( strings )
{
	tester t;
	t( " \"hello\" ", "array< char >" );
	
	t( " 'b' ", "char" );
}

BOOST_AUTO_TEST_SUITE_END()
