#include <boost/test/unit_test.hpp>

#include "lang/language.hpp"
#include "lang/typer.hpp"
#include "lang/dump.hpp"

#include "parser/node_parser.hpp"
#include "parser/node_converter.hpp"
#include "parser/intr_type_parser.hpp"
#include "parser/dump.hpp"

#include <iostream>

#define SHOW_DEBUG 0
#include "util/debug.hpp"

using namespace leaf;

BOOST_AUTO_TEST_SUITE( full_type_test )

/**
	A test framework to handle the scope and expression parsing.
*/
struct tester
{
	shared_ptr<statement_block> sblock;
	
	tester()
	{
	}

	void block( char const * code )
	{
		auto src = source::create( code, std::strlen( code ) );

		shared_ptr<node> block = node_parser::parse_as_block( *src );
		DEBUG( node_dump::dump_flat( *block ) );
		sblock = node_converter::convert_block( *block, scope_style_t::imperative );
		typer t;
		t.full_type( sblock );
	}
	
	void operator()( char const * code, char const * expect )
	{
		auto src = source::create( code, std::strlen( code ) );

		shared_ptr<node> n = node_parser::parse_single( *src );
		auto ex = node_converter::expression_convert( *n );
		typer t( sblock->symbols );
		ex = t.full_type( ex );
		
		leaf::dump d;
		d.write_intr_type( ex->type );
		std::string result = d.str();
		if( result != expect )
		{
			std::cerr
				<< "Failed: " << code << std::endl
				<< "Expect: " << expect << std::endl
				<< "Actual: " << result << std::endl
				;
			BOOST_FAIL( "Failed" );
		}

		BOOST_CHECK(true);
	}
};


BOOST_AUTO_TEST_CASE( assign )
{
	tester t;
	t.block( "@import(\"add_integer\") func add : (:integer,:integer)->(:integer);"
		"var a : integer;"
		"var b: integer;"
		);
	t( "a+b", "integer" );
	t( "add", "( : integer, : integer ) -> ( : integer ) lvalue");
}

BOOST_AUTO_TEST_CASE( chain )
{
	tester t;
	t.block(
		"@import(\"add_integer\") func add : (:integer,:integer)->(:integer);"
		"func sum = (a:integer,b:integer)->(c:integer) { c = a /*TODO:+ b*/; };"
		"func rand=()->(c:integer) {  };"  );
		
	t( "sum(1,2)", "integer" );
	t( "sum(rand(),rand())", "integer" );
}

BOOST_AUTO_TEST_CASE( overload )
{
	tester t;
	t.block( "@import(\"sum\") func sum : (:integer,:integer)->(:integer); "
		"func sum = (a:float,b:float)->(c:float) { }; ");
	t( "sum(1,2)", "integer" );
	t( "sum(1.1,2.0)", "float" ); //note that .0 is not enough to match a float, it must actually be a non-integer
}

BOOST_AUTO_TEST_CASE( one )
{
	tester t;
	t.block( "func one = (a:integer)->() { };" );
	t( "one(12)", "( )" );
}

BOOST_AUTO_TEST_SUITE_END()
