#include <boost/test/unit_test.hpp>

#include "lang/scope.hpp"
#include "lang/var_decl.hpp"

using namespace leaf;

#define SHOW_DEBUG 1
#include "util/debug.hpp"

BOOST_AUTO_TEST_SUITE( scope_test )

shared_ptr<var_decl> vs_single( std::string const & name )
{
	shared_ptr<var_decl> vs( new var_decl );
	vs->symbol = name;
	vs->form = var_decl::f_single;
	return vs;
}

shared_ptr<var_decl> vs_multi( std::string const & name )
{
	shared_ptr<var_decl> vs( new var_decl );
	vs->symbol = name;
	vs->form = var_decl::f_multi;
	return vs;
}

BOOST_AUTO_TEST_CASE( basic )
{
	auto s = scope::create( scope_style_t::imperative /*doens't matter here*/ );
	
	s->add_single( vs_single( "a" ) );
	scope_ref sr = s->get_var( "a" );
	BOOST_CHECK_EQUAL( sr.decl->symbol, "a" );
}

BOOST_AUTO_TEST_CASE( parent )
{
	auto sa=scope::create( scope_style_t::imperative ), 
		sb = scope::create( scope_style_t::imperative ), 
		sc = scope::create( scope_style_t::imperative );
	sb->set_parent( sa );
	sc->set_parent( sb );
	
	BOOST_CHECK_EQUAL( 0, sa->get_level() );
	BOOST_CHECK_EQUAL( 2, sc->get_level() );
	
	sa->add_single( vs_single( "a" ) );
	
	scope_ref sr = sc->get_var( "a" );
	BOOST_CHECK_EQUAL( 2, sr.parent_level );
	BOOST_CHECK_EQUAL( 0, sr.root_depth );

	//decl overrides one from parent
	sb->add_single( vs_single( "a" ) );
	sr = sc->get_var( "a" );
	BOOST_CHECK_EQUAL( 1, sr.parent_level );
	BOOST_CHECK_EQUAL( 1, sr.root_depth );
	
	//multi's can exist at multiple levels (might be crazy though) //TODO: perhaps only nearest should resolve?
	sa->add_multi( vs_multi( "b" ) );
	sb->add_multi( vs_multi( "b" ) );
	sc->add_multi( vs_multi( "b" ) );
	scope::var_vector_t all = sc->get_vars( "b" );
	BOOST_REQUIRE_EQUAL( 3, all.size() );
	//the order shouldn't really be guaranteed
	BOOST_CHECK_EQUAL( 0, all[0].parent_level );
	BOOST_CHECK_EQUAL( 1, all[1].parent_level );
	BOOST_CHECK_EQUAL( 2, all[2].parent_level );
	
	all = sc->get_local_vars( "b" );
	BOOST_CHECK_EQUAL( 1, all.size() );
	all = sc->get_local_vars( "a" );
	BOOST_CHECK_EQUAL( 0, all.size() );
}

bool same_ref( scope_ref a, scope_ref b ) {
	return a.parent_level == b.parent_level
		&& a.root_depth == b.root_depth
		&& a.decl.get() == b.decl.get();
}

BOOST_AUTO_TEST_CASE( extends )
{
	auto sa = scope::create( scope_style_t::imperative ), 
		sb = scope::create( scope_style_t::imperative ), 
		sc = scope::create( scope_style_t::imperative ), 
		sd = scope::create( scope_style_t::imperative );
	sb->set_parent( sa );
	sc->set_parent( sb );
	sd->set_extends( sc );
	
	BOOST_CHECK_EQUAL( 2, sc->get_level() );
	BOOST_CHECK_EQUAL( 3, sd->get_level() );
	
	sa->add_single( vs_single( "a" ) );
	auto sr = sd->get_var( "a" );
	BOOST_CHECK_EQUAL( 0, sr.root_depth );
	BOOST_CHECK_EQUAL( 2, sr.parent_level );
	
	sc->add_single( vs_single( "b" ) );
	sr = sd->get_var( "b" );
	BOOST_CHECK_EQUAL( 2, sr.root_depth );
	BOOST_CHECK_EQUAL( 0, sr.parent_level );
	
	sd->add_single( vs_single( "c" ) );
	BOOST_CHECK_EQUAL( 0, sc->get_vars( "c" ).size() );
	
	sr = sa->get_var( "a" );
	auto cr = sd->capture_var( sr );
	BOOST_CHECK( same_ref( sr, cr ) );
	sb->add_single( vs_single("d") );
	sr = sd->get_var( "d" );
	cr = sd->capture_var( sr );
	BOOST_CHECK_EQUAL( var_decl::f_capture, cr.decl->form );
	BOOST_CHECK_EQUAL( sc.get(), cr.decl->var_scope );
	sr = sc->capture_var( sr );
	BOOST_CHECK( same_ref( cr, sr ) );
}

BOOST_AUTO_TEST_CASE( continues )
{
	auto sa = scope::create( scope_style_t::imperative ), 
		sb = scope::create( scope_style_t::imperative );
	sb->set_continues( sa );
	
	sa->add_single( vs_single( "a" ) );
	auto sr = sb->get_var( "a" );
	BOOST_CHECK_EQUAL( 0, sr.root_depth );
	BOOST_CHECK_EQUAL( 0, sr.parent_level );
}

BOOST_AUTO_TEST_CASE( capture )
{
	auto sa = scope::create( scope_style_t::imperative ), 
		sb = scope::create( scope_style_t::imperative ), 
		sc = scope::create( scope_style_t::imperative ), 
		sd = scope::create( scope_style_t::imperative );
	sb->set_parent( sa );
	sc->set_parent( sb );
	sd->set_parent( sc );
	
	sb->add_single( vs_single( "a" ) );
	scope_ref sr = sd->get_var( "a" );
	auto local = sd->capture_var( sr );
	
	{
		scope::capture_map_t const & cm = sd->get_captures();
		BOOST_REQUIRE_EQUAL( 1,  cm.size() );
	}
	
	{
		scope::capture_map_t const & cm = sc->get_captures();
		BOOST_REQUIRE_EQUAL( 1,  cm.size() );
	}

	{
		scope::capture_map_t const & cm = sb->get_captures();
		BOOST_REQUIRE_EQUAL( 0,  cm.size() );
	}

	//test uncapture
	auto src = sd->uncapture_var( local );
	BOOST_CHECK_EQUAL( sr.parent_level, src.parent_level );
	BOOST_CHECK_EQUAL( sr.root_depth, src.root_depth );
	BOOST_CHECK_EQUAL( sr.decl.get(), src.decl.get() );
	
	//capture from a parent scope_ref
	sb->add_single( vs_single("b") );
	sr = sb->get_var( "b" );
	local = sc->capture_var( sr, sb.get() );
	BOOST_CHECK_EQUAL( 1, local.root_depth );
	BOOST_CHECK_EQUAL( 1, local.parent_level );
	BOOST_CHECK_EQUAL( sc.get(), local.decl->var_scope );
	
}

BOOST_AUTO_TEST_CASE( namespace_basic )
{
	auto sa = scope::create( scope_style_t::imperative ), 
		sb = scope::create( scope_style_t::imperative );
	sb->set_namespace( sa, "sub" );
	
	sa->add_single( vs_single("only_a") );
	sb->add_single( vs_single("only_b") );
	sa->add_single( vs_single("both") );
	sb->add_single( vs_single("both") );
	
	auto sr = sb->get_var( "only_a" );
	BOOST_CHECK_EQUAL( sa.get(), sr.decl->var_scope);
	BOOST_CHECK_EQUAL( 0, sr.parent_level );
	
	sr = sb->get_var( "both" );
	BOOST_CHECK_EQUAL( sb.get(), sr.decl->var_scope );
	sr = sa->get_var( "both" );
	BOOST_CHECK_EQUAL( sa.get(), sr.decl->var_scope );
	
	auto vars = sa->get_vars( "only_b" );
	BOOST_CHECK_EQUAL( 0, vars.size() );
	
	vars = sb->get_vars( "both" );
	BOOST_CHECK_EQUAL( 1, vars.size() );
}

BOOST_AUTO_TEST_SUITE_END()
