#include <boost/test/unit_test.hpp>

#include "lang/type_converter.hpp"
#include "lang/type_identifier.hpp"
#include "lang/intr_type.hpp"
#include "parser/intr_type_parser.hpp"
#include "lang/scope.hpp"
#include "lang/dump.hpp"

#include <sstream>

#include "util/debug.hpp"

using namespace leaf;

/**
	Abstracts some common behaviour of many of the tests.
*/
struct expect_test
{	
	struct part {
		std::string label, value;
	};
	std::vector<part> parts;
	
	expect_test(){
	}
	void add( std::string const & label, std::string const & value ) {
		part p;
		p.label = label;
		p.value = value;
		parts.push_back(p);
	}
	void add( std::string const & value ) {
		char buf[32];
		sprintf( buf, "%d", int(parts.size()) );
		add( buf, value );
	}
	
	void fail( std::string const & msg ) {
		header( msg );
		BOOST_REQUIRE( false );
	}
	void except( ) {
		fail( std::string("Exception: ") + boost::current_exception_diagnostic_information() );
		throw;
	}
	void check( bool cond, std::string const & msg, char const * prefix = "" ) {
		if( !cond ) {
			fail( std::string( prefix ) + ": " + msg );
		} else {
			BOOST_CHECK(true);
		}
	}
	
	void check_actual( std::string const & actual, std::string const & expect ) {
		if( actual == expect ) {
			BOOST_CHECK( true );
			return;
		}
			
		header( "mismatched" );
		std::cout 
			<< "Expect: " << expect << std::endl
			<< "Actual: " << actual << std::endl;
		BOOST_CHECK(false);
	}
	
	void header( std::string const & msg ) {
		for( auto & p : parts )
			std::cout << p.label << ": " << p.value << std::endl;
		std::cout << msg << std::endl;
	}
};

struct type_converter_tester
{
	shared_ptr<scope> sc;
	type_converter tcvt;
	type_identifier tid;
	
	type_converter_tester()
		: sc( scope::create(scope_style_t::declarative) )
		, tcvt( *sc )
		, tid( *sc )
	{ }
	
	extr_type_ref parse( char const * code )
	{
		auto s = source::create( code, std::strlen(code) );
		auto vts = intr_type_parser::parse( *s );
		auto vt = tid.determine( vts );
		STATE_CHECK(vt.okay(),vt.explain_error());
		DEBUG( code, " / ", dump::get_type_spec(*vts), " / ", dump::get_extr_type(vt.result) );
		return vt.result;
	}

	/**
		Adds a new fundamental class type. This is expected to be a tuple and mimics what is
		done in typer::proc_type_statement to support recursion.
	*/
	void add_defined_type( std::string name, char const * code,
		shared_ptr<scope> where = shared_ptr<scope>() )
	{
		if( !where ) {
			where = sc;
		}
		
		auto inc_tuple = make_shared<intr_type_tuple>();
		inc_tuple->local_complete = false;
		inc_tuple->distinct_name = name;
		inc_tuple->distinct = true;
		where->add_fun_type( inc_tuple );
		
		auto raw = parse( code );
		PRE_CHECK( !raw.get_fun()->distinct_name );
		auto nt_tuple = raw.if_fun_cast<intr_type_tuple>();
		PRE_CHECK( nt_tuple );
		inc_tuple->complete_with( *nt_tuple );
	}
	
	void add_type_alias( std::string name, char const * code )
	{
		auto s = source::create( code, std::strlen(code) );
		auto spec = intr_type_parser::parse( *s );
		auto err = tid.expand_spec( *spec );
		PRE_CHECK( err.okay(), err.explain_error() );
		sc->add_type_alias( name, spec );
	}
	
	void dump( std::ostream & out, cvt_expression const& ce )
	{
		bool force_sub = false;
		switch( ce.form )
		{
			#define CODE(tag) \
				case cvt_expression::tag: \
					out << #tag; break;
			CODE(c_addr_of)
			CODE(c_drop_lvalue)
			CODE(c_deref)
			CODE(c_bind_raw_func)
			CODE(c_bind_func)
			CODE(c_pack_value)
			CODE(c_direct)
			CODE(c_collapse)
			CODE(c_optional)
			CODE(c_deref_optional)
			CODE(c_unset_optional)
			#undef CODE
			
			case cvt_expression::c_set_distinct: {
				out << "c_set_distinct(" << *ce.target_type.get_fun()->distinct_name << ")";
				break;
			}
			
			case cvt_expression::c_pack: {
				out << "c_pack";
				switch( ce.pack ) {
					case intr_type_tuple::p_normal: break;
					case intr_type_tuple::p_abi: out << "(abi)"; break;
					case intr_type_tuple::p_byte: out << "(byte)"; break;
				}
				force_sub = true;
				break;
			}
			case cvt_expression::c_promote:
			case cvt_expression::c_fixate: 
			case cvt_expression::c_loss: {
				if( ce.form == cvt_expression::c_promote )
					out << "c_promote(";
				else if( ce.form == cvt_expression::c_fixate )
					out << "c_fixate(";
				else
					out << "c_loss(";

				if( ce.to_fun == cvt_expression::p_float ) {
					auto nf = platform::if_get_float( ce.to_bitsize );
					STATE_CHECK(nf);
					switch( nf->precision ) {
						case -1: out << "low"; break;
						case 1: out << "high"; break;
						case 0: out << "normal";  break;
						default: STATE_FAIL();
					}
				} else { 
					if( ce.to_bitsize != platform::natural_integer_size )
						out << ce.to_bitsize << "bit";
					else
						out << "normal";
				}
				out << " ";
				
				auto fun_print = [&]( cvt_expression::fundamental_t p ) {
					switch( p ) {
						case cvt_expression::p_integer:
							out << "integer";
							break;
						case cvt_expression::p_binary:
							out << "binary";
							break;
						case cvt_expression::p_float:
							out << "float";
							break;
					}
				};
				
				if( ce.form == cvt_expression::c_fixate ) {
					fun_print( ce.to_fun );
				} else {
					fun_print( ce.from_fun );
					if( ce.from_fun != ce.to_fun ) {
						out << "_to_";
						fun_print( ce.to_fun );
					}
				}
				out << ")";
				break;
			}

			case cvt_expression::c_field:
				out << "[" << ce.field << "]";
				break;
		}
		
		if( ce.sub.size() || force_sub ) {
			out << " {";
			for( size_t i=0; i < ce.sub.size(); ++i ) 
			{
				if( i )
					out << " ";
				if( !ce.sub[i].name.empty() )
					out << ce.sub[i].name << "=";
				dump( out, *ce.sub[i].cvt );
			}
			out << "}";
		}
		
		if( ce.then ) {
			out << "->";
			dump( out, *ce.then );
		}
	}
};
