#define SHOW_DEBUG 0
#include "type_converter_common.hpp"

#include "lang/serial.hpp"
#include <sstream>

BOOST_AUTO_TEST_SUITE( type_converter_test )

struct tester : public type_converter_tester
{
	typedef type_converter::convert_t (type_converter::*from_func)( extr_type_ref const & );
	typedef type_converter::convert_t (type_converter::*to_from_func)( extr_type_ref const &,
		extr_type_ref const & );
		
	from_func use_from_func;
	to_from_func use_to_from_func;
	
	tester( from_func use_from_func )
		: use_from_func( use_from_func )
		, use_to_from_func( nullptr ) 
	{ }
	tester( to_from_func use_to_from_func )
		: use_from_func( nullptr )
		, use_to_from_func( use_to_from_func )
	{ }
	tester()
		: use_from_func( nullptr )
		, use_to_from_func( &type_converter::convert )
	{ }
	
	void operator()( char const * sto, char const * sfrom, 
		char const * expect )
	{
		expect_test ex;
		if( sto )
			ex.add( "To", sto );
		ex.add( "From", sfrom );
		
		type_converter::convert_t cost;

		extr_type_ref to, from;
		if( sto )
			to = parse( sto );
		from = parse( sfrom );
		
		//do serialization tests on all tested types, good way to get coverage
		if( sto )
			test_serial( to );
		test_serial( from );

		//DEBUG( leaf::dump::get_extr_type(*to), "==", leaf::dump::get_extr_type(*from));
		try {
			if( use_to_from_func )
				cost = (tcvt.*use_to_from_func)( to, from );
			else
				cost = (tcvt.*use_from_func)( from );
		} catch( ... ) {
			ex.except();
		}
		
		ex.check( cost.okay(), cost.explain_error() );
			
		std::ostringstream out;
		if( cost.conversions )
			dump( out, *cost.conversions );
		ex.check_actual( out.str(), expect );
	}
	void operator()( char const * sfrom, char const * expect ) {
		(*this)( nullptr, sfrom, expect ); 
	}
	
	void operator()( char const * sto, char const * sfrom, cerr expect )
	{
		expect_test ex;
		if( sto )
			ex.add( "To", sto );
		ex.add( "From", sfrom );
		
		extr_type_ref to, from;
		if( sto )
			to = parse( sto );
		from = parse( sfrom );

		if( sto )
			test_serial( to );
		test_serial( from );

		type_converter::convert_t cost;
		try {
			if( use_to_from_func )
				cost = (tcvt.*use_to_from_func)( to, from );
			else
				cost = (tcvt.*use_from_func)( from );
		} catch( ... ) {
			ex.except();
		}
		ex.check( cost.error == expect, cost.explain_error(), "wrong error" );
	}
	void operator()( char const * sfrom, cerr expect ) {
		(*this)( nullptr, sfrom, expect );
	}
	
	void test_serial( extr_type_ref t ) {
		std::ostringstream outs;
		leaf::serial_out out( outs );
		out.add_extr_type( t );
		out.done();
		
		std::istringstream ins( outs.str() );
		leaf::serial_in in( ins );
		auto load = in.read_primary_extr_type();
		
		if( t != load ) {
			std::cout << "Serialization failed" << std::endl;
			std::cout << "Orig: " << leaf::dump::get_extr_type( t ) << std::endl;
			std::cout << "Load: " << leaf::dump::get_extr_type( load ) << std::endl;
			BOOST_CHECK(false);
		}
	}
};

BOOST_AUTO_TEST_CASE( basic )
{
	tester t;
	t( "integer", "integer", "" );
	t( "integer shared", "integer shared", "" );
	t( "integer lvalue", "integer lvalue", "" );
	t( "integer lvalue shared", "integer lvalue shared", "" );
	t( "integer shared", "integer lvalue shared", "c_drop_lvalue" );
	t( "integer", "integer shared lvalue", "c_drop_lvalue->c_deref" );
	
 	t( "integer", "integer lvalue",	 "c_drop_lvalue" );
 	t( "integer", "integer shared", "c_deref" );
	t( "[]", "[] lvalue", "c_drop_lvalue" );
	
}

BOOST_AUTO_TEST_CASE( functions )
{
	tester t;
	t( "()->()", "()->()", "" );
	t( "()->()", "()->() raw lvalue", "c_drop_lvalue->c_bind_raw_func" );
	t( "()->()", "()->() raw shared lvalue", "c_drop_lvalue->c_deref->c_bind_raw_func" );
	t( "(:integer)->()", "(:integer)->()", "" );
	t( "()->()", "()->() raw", "c_bind_raw_func" );
	
	t( "()->()", "()->() unbound", "c_bind_func" );
}

BOOST_AUTO_TEST_CASE( floats )
{
	tester t;
	t( "float", "float", "" );
	t( "float high", "float", "c_promote(high float)" );
	t( "float", "rational", "c_fixate(normal float)" );
	t( "float", "integer 10bit", "c_promote(normal integer_to_float)" );
	t( "float low", "integer", cerr::precision_loss );
	t( "float", "binary 10bit", "c_promote(normal binary_to_float)" );
	t( "float low", "binary", cerr::precision_loss );
	
	//unfortunately these are tied directly to the standard float sizes(though 64-bit is almost universal)
	t( "float", "binary 53bit", "c_promote(normal binary_to_float)" );
	t( "float", "binary 54bit", cerr::precision_loss );
	t( "float", "integer 54bit", "c_promote(normal integer_to_float)" );
	t( "float", "integer 55bit", cerr::precision_loss );
}

BOOST_AUTO_TEST_CASE( lvalue_strict )
{
	tester t;
	t( "float lvalue", "integer lvalue", cerr::mismatch_fundamental );
	t( "lvalue [ : integer ]", "lvalue [ : integer 2bit ]", cerr::mismatch_data_size );
	t( "lvalue [ : float shared ]", "lvalue [ : float ]", cerr::mismatch_shared );
	t( "lvalue shared []", "lvalue shared []", "" );
	
	//contrast to retained, optional wraps shared so you can't drop the shared here
	t( "integer optional lvalue", "integer optional shared lvalue", cerr::mismatch_shared );
}

BOOST_AUTO_TEST_CASE( lvalue_retain )
{
	tester t;
	t( "float lvalue", "lvalue optional float", "c_deref_optional" );
	t( "float lvalue", "lvalue shared float", "c_deref" );
	t( "integer lvalue", "lvalue integer optional shared", "c_deref_optional->c_deref" );
	t( "integer shared lvalue", "integer shared optional lvalue", "c_deref_optional" );
}

BOOST_AUTO_TEST_CASE( tuples )
{
	tester t;
	t( "[]", "[]", "" );
	t( "[:[]]", "[:[]]", "" );
	t( "[:integer,:float]", "[:integer,:float]", "" );
	t( "[:integer]", "[:integer lvalue]", "c_pack {[0]->c_drop_lvalue}" );
	t( "[:integer,:float]", "[:integer lvalue,:float shared]", 
		"c_pack {[0]->c_drop_lvalue [1]->c_deref}" );
	t( "[:integer,:float]", "[:integer,:float shared]",
		"c_pack {[0] [1]->c_deref}" );
	t( "[:[:integer,:float]]", "[:[:integer,:float shared]]",
		"c_pack {[0]->c_pack {[0] [1]->c_deref}}" );
		
	//packing styles
	t( "[] abi_pack", "[]", "c_pack(abi) {}" );
	t( "[:integer]", "[:integer] byte_pack", "c_pack {[0]}" );
}

BOOST_AUTO_TEST_CASE( tuple_names )
{
	tester t;
	t( "[ a : integer ]", "[ :integer ]", "c_pack {a=[0]}" );
	t( "[ : integer]", "[ a : integer ]", cerr::mismatch_field_name );
}

BOOST_AUTO_TEST_CASE( value_lists )
{
	tester t;
 	t( "(:integer)", "(:integer)", "" );
 	
	t( "integer", "(:(:integer))", "c_collapse->c_collapse" );
	t( "integer", "(:integer)", "c_collapse" );
	t( "integer", "(:integer lvalue shared)", "c_collapse->c_drop_lvalue->c_deref" );

 	t( "(:integer,:float)", "(:(:(:integer),:float))", "c_collapse->c_pack_value {[0]->c_collapse [1]}" );
 	t( "(:(:integer),:float)", "(:(:integer),:float)", "" );
 	t( "(:(:integer),:float)", "(:(:(:integer),:float))", "c_collapse" );

 	//NOTE: The optimization that makes the shorted form is removed from type_converter. I'd to
 	//avoid that the "typer" and further actually require that optimization.
	//t( "(:integer)", "(:(:integer))", "c_collapse" );
	t( "(:integer)", "(:(:integer))", "c_pack_value {c_collapse->c_collapse}" );
 	//t( "(:(:integer),:float)", "(:(:(:integer)),:float)", "c_pack_value {[0]->c_collapse [1]}" );
 	t( "(:(:integer),:float)", "(:(:(:integer)),:float)", "c_pack_value {[0]->c_pack_value {c_collapse->c_collapse} [1]}" );
 	
 	//1-value-list collapsing is one-direction only
 	t( "(:integer)", "integer", cerr::require_value_list );
 	
 	//lvalue should be preserved
 	t( "integer lvalue", "(:integer lvalue)", "c_collapse" );
 	t( "integer lvalue", "(:(:integer lvalue))", "c_collapse->c_collapse" );
}

BOOST_AUTO_TEST_CASE( invalid )
{
	tester t;
 	t( "integer", "float", cerr::mismatch_fundamental );
 	t( "boolean", "[]", cerr::mismatch_fundamental );
 	t( "integer shared", "integer", cerr::requires_shared );
 	t( "integer lvalue", "integer", cerr::requires_lvalue );
 	t( "[:integer]", "[:integer,:float]", cerr::mismatch_size );
 	t( "()->() raw", "()->()", cerr::mismatch_function_access );
 	t( "integer lvalue", "integer shared", cerr::requires_lvalue );
 	t( "integer lvalue", "float lvalue", cerr::mismatch_fundamental );
 	t( "[:integer] shared", "[:integer shared] shared", cerr::mismatch_shared );
 	
 	
	t( "()->()", "(:integer)->()", cerr::mismatch_size );
	t( "()->()", "()->(:integer)", cerr::mismatch_size );
	t( "(:integer)->()", "()->()", cerr::mismatch_size );
	t( "()->(:integer)", "()->()", cerr::mismatch_size );
	t( "(:integer)->()", "(:integer shared)->()", cerr::mismatch_shared );
}

BOOST_AUTO_TEST_CASE( assign )
{
	tester t( &type_converter::convert_to_lvalue );
	t( "integer lvalue", "integer", "" );
	t( "integer lvalue", "integer lvalue", "c_drop_lvalue" );
	t( "integer lvalue", "integer shared", "c_deref" );
	t( "integer", "integer", cerr::requires_lvalue );
	t( "(:integer lvalue)", "(:integer)", "" );
	t( "(:integer)", "(:integer)", cerr::requires_lvalue );
	t( "(:integer)->(:integer) lvalue", "()->() raw", cerr::mismatch_size );
	
	t( "integer optional lvalue", "integer", "c_optional" );                                                                                                                                                                                                                            
 	t.add_defined_type( "point", "[ x : float, y : float ]" );
 	t( "point lvalue", "point", "" );
	t( "point lvalue optional", "point", "c_optional" );
	
	t( "integer value_ptr lvalue", "integer value_ptr", "" );
	t( "integer value_ptr lvalue", "integer", cerr::mismatch_value_ptr );
	
	//with lossy
	t( "integer 8bit lvalue", "integer lossy", "c_loss(8bit integer)" );
	t( "integer 8bit lvalue", "integer lossy lvalue", "c_drop_lvalue->c_loss(8bit integer)" );
}                                                                         

BOOST_AUTO_TEST_CASE( to_refer )
{
	tester t( &type_converter::convert_to_refer );
	t( "integer shared", "integer shared", "c_deref" );
	t( "float shared", "float", "" );
	t( "boolean", "boolean", cerr::requires_refer );
	t( "integer shared", "integer shared lvalue", "c_drop_lvalue->c_deref" );
	
	t( "integer value_ptr", "integer value_ptr", "c_deref" );
	t( "float value_ptr", "float", "" );
	t( "boolean", "boolean", cerr::requires_refer );
	//t( "integer value_ptr", "integer value_ptr lvalue", "c_drop_lvalue->c_deref" );
}

BOOST_AUTO_TEST_CASE( function )
{
	tester t( &type_converter::convert_any_callable );
	t( "()->()", "" );
	t( "()->() lvalue", "c_drop_lvalue" );
	t( "()->() shared", "c_deref" );
	t( "( :(:integer)->(:boolean) )", "c_collapse" );
	t( "()->() optional", "c_deref_optional" );
	
	t( "( : [ ] selectable )", "c_collapse" );
	
	t( "integer", cerr::requires_function );
	t( "( :()->(), :()->() )", cerr::requires_function );
}

BOOST_AUTO_TEST_CASE( any_refer )
{
	tester t( &type_converter::convert_any_refer );
	t( "shared integer", "" );
	t( "(:shared integer)", "c_collapse" );
	t( "optional integer shared", "c_deref_optional" );
	t( "shared lvalue", "c_drop_lvalue" );
	t( "( : shared integer, : shared integer lvalue, : shared integer optional )",
		"c_pack_value {[0] [1]->c_drop_lvalue [2]->c_deref_optional}" );
	t( "( : shared integer, : shared float )", "" );//no pack expected if all sub-parts are empty

	t( "value_ptr integer", "" );
	t( "(:value_ptr integer)", "c_collapse" );
	t( "optional integer value_ptr", "c_deref_optional" );
	t( "value_ptr lvalue", "c_drop_lvalue" );
	t( "( : value_ptr integer, : value_ptr integer lvalue, : value_ptr integer optional )",
		"c_pack_value {[0] [1]->c_drop_lvalue [2]->c_deref_optional}" );
	t( "( : value_ptr integer, : value_ptr float )", "" );//no pack expected if all sub-parts are empty
	
	t( "value_ptr integer lvalue", "c_drop_lvalue" );
	
	t( "optional integer", cerr::requires_refer );
	t( "integer", cerr::requires_refer );
	t( "()", cerr::requires_shared ); //TODO: error is suspicious (expect refer instead)
}

BOOST_AUTO_TEST_CASE( any_lvalue )
{
	tester t( &type_converter::convert_any_lvalue );
	t( "integer lvalue", "" );
	t( "( : integer lvalue, : lvalue float)", "" );
	t( "( : integer lvalue )", "c_collapse" );
	
	t( "integer lvalue value_ptr", "" );
	
	t( "integer", cerr::requires_lvalue );
	t( "()", cerr::requires_lvalue );
}

BOOST_AUTO_TEST_CASE( nonlvalue_optional )
{
	tester t( &type_converter::convert_nonlvalue_optional );
	t( "optional integer", "" );
	t( "shared optional integer", "" );
	t( "optional lvalue float", "c_drop_lvalue" );
	t( "(:optional boolean)", "c_collapse" );
	
	t( "integer", cerr::requires_optional );
}

BOOST_AUTO_TEST_CASE( any_scalar )
{
	tester t( &type_converter::convert_any_scalar );
	t( "integer", "" );
	t( "binary", "" );
	t( "optional integer", "c_deref_optional" );
	t( "shared float", "c_deref" );
	t( "(:optional shared integer)", "c_collapse->c_deref_optional->c_deref" );
	
	//rationals are also scalar
	t( "rational lvalue", "c_drop_lvalue" );
	t( "optional rational 15bit", "c_deref_optional" );
	
	//booleans are also scalar
	t( "boolean lvalue", "c_drop_lvalue" );
}

BOOST_AUTO_TEST_CASE( unqualified )
{
	tester t( &type_converter::convert_unqualified );
	t( "integer", "" );
	t( "shared integer", "c_deref" );
	t( "( : float optional lvalue shared )",  "c_collapse->c_drop_lvalue->c_deref_optional->c_deref" );
}

BOOST_AUTO_TEST_CASE( defined )
{
	tester t;
	t.add_defined_type( "point", "[ x : float, y : float ]" );
	t.add_defined_type( "complex", "[ r: float, i : float ]" );
	
	//OPT: ideally this could just be "c_set_distinct(point)" since the parts are already compatible
	t( "point", "[ : float, : float ]", "c_pack {x=[0] y=[1]}->c_set_distinct(point)" );
	t( "point", "[ x : float, y : float ]", "c_set_distinct(point)" );
	t( "point", "[ a : float, b : float ]", cerr::mismatch_field_name );
	t( "point", "[ : integer 8bit, y:integer 7bit ]", 
		"c_pack {"
			"x=[0]->c_promote(normal integer_to_float) "
			"y=[1]->c_promote(normal integer_to_float)"
		"}->c_set_distinct(point)" ); //integer promotion
	
	t( "point", "complex", cerr::mismatch_defined );
	
 	t( "point", "point shared", "c_deref" );
 	t( "point", "point lvalue", "c_drop_lvalue" );
 	t( "point", "point shared lvalue", "c_drop_lvalue->c_deref" );
 	t( "point", "( :point lvalue )", "c_collapse->c_drop_lvalue" );
 	
 	t( "point", "[ : float, : float ] lvalue", "c_drop_lvalue->"
 		"c_pack {x=[0] y=[1]}->c_set_distinct(point)" );

	t( "[ a : point ]", "[ : [ : float, : float ] ]",
		"c_pack {a=[0]->c_pack {x=[0] y=[1]}->c_set_distinct(point)}" );
	
	t.add_defined_type( "rect", "[ ul: point, br: point ]" );
	t( "rect", "[ :point, :point ]", "c_pack {ul=[0] br=[1]}->c_set_distinct(rect)" );
	//OPT: also ideally this would be just be "c_set_distinct(rect)"
	t( "rect", "[:[x:float,:float],br:[:float,:float]]", 
		"c_pack {ul=[0]->c_pack {x=[0] y=[1]}->c_set_distinct(point) br=[1]->c_pack {x=[0] y=[1]}->c_set_distinct(point)}->c_set_distinct(rect)" );
}

BOOST_AUTO_TEST_CASE( array )
{
	tester t;
	t( "array<integer>", "array<integer>", "" );
	t( "array<integer>", "array<integer> lvalue", "c_drop_lvalue" );
}

BOOST_AUTO_TEST_CASE( promote )
{
	tester t;
	t( "integer", "integer 3bit", "c_promote(normal integer)" );
	t( "integer 60bit", "integer 11bit", "c_promote(60bit integer)" );
	t( "integer 128bit", "integer", "c_promote(128bit integer)" );
	t( "integer 200bit", "integer 200bit", "" );
	//deref must come first
	t( "integer 60bit", "shared integer 8bit", "c_deref->c_promote(60bit integer)" );
	
	//references to types cannot be promoted
	t( "shared integer", "shared integer 8bit", cerr::mismatch_data_size );
	t( "lvalue integer", "lvalue integer 8bit", cerr::mismatch_data_size );
	
	t( "binary", "binary 16bit", "c_promote(normal binary)" );
	t( "binary 100bit", "binary 73bit", "c_promote(100bit binary)" );
	
	//binary to signed (note that integer->binary is not a promotion, would be a truncation)
	t( "integer", "binary", cerr::range_loss );
	t( "integer 41bit", "binary 40bit", "c_promote(41bit binary_to_integer)" );
}

BOOST_AUTO_TEST_CASE( optional )
{
	tester t;
	t( "integer", "optional integer", "c_deref_optional" );
	t( "integer optional", "integer", "c_optional" );
	t( "integer optional", "integer optional", "" );
	
	t( "integer optional lvalue", "integer lvalue", cerr::requires_optional );
	t( "integer lvalue", "integer optional lvalue", "c_deref_optional" ); //may retain lvalue in deref
	t( "integer shared optional", "integer shared", "c_optional" );
	t( "integer optional", "integer shared", "c_deref->c_optional" );
	t( "integer optional", "integer shared optional", cerr::mismatch_shared );
	t( "integer optional", "integer shared", "c_deref->c_optional" );
	t( "integer optional", "integer shared lvalue", "c_drop_lvalue->c_deref->c_optional" );
	t( "integer shared optional", "integer shared lvalue", "c_drop_lvalue->c_optional" );
	
	t( "integer", "integer optional shared", "c_deref_optional->c_deref" );
	t( "integer shared", "integer optional shared", "c_deref_optional" );
	//TODO: should the following be supported, both removing and adding an optional?
	//t( "integer optional", "integer optional shared", "c_deref_optiona->c_deref->c_optional" );
}

BOOST_AUTO_TEST_CASE( fixation )
{
	tester t;
	//NOTE: these tests are somewhat unfortunately tied to the platform size... may need to fix on limited platforms
	t( "integer", "rational 29bit", "c_fixate(normal integer)" );
	t( "optional integer 6bit", "rational 2bit", "c_fixate(6bit integer)->c_optional" );
	t( "integer 100bit", "rational 16bit optional", "c_deref_optional->c_fixate(100bit integer)" ); //though an optional rational is not actually allowed
	t( "binary", "rational 32bit", "c_fixate(normal binary)");
	t( "binary 32bit", "rational 32bit", "c_fixate(32bit binary)");
	t( "binary 32bit", "rational 32bit", "c_fixate(32bit binary)");
	
	t( "float", "rational", "c_fixate(normal float)" );
	
	t( "binary", "rational signed 15bit", cerr::mismatch_signed );
	
}

BOOST_AUTO_TEST_CASE( convert_optional )
{
	tester t;
	t( "optional integer", "rational 3bit optional convert_optional", "c_fixate(normal integer)" );
}

BOOST_AUTO_TEST_CASE( shared_or_unqualified )
{
	tester t( &type_converter::convert_shared_or_unqualified );
	t( "float", "" );
	t( "lvalue integer", "c_drop_lvalue" );
	t( "shared optional [ : integer ]", "c_deref_optional" );
}

BOOST_AUTO_TEST_CASE( selectable )
{
	tester t;
	t( "float", "[ : integer, : float ] selectable", "[1]" );
	t( "integer 20bit", "[ :integer 21bit, : integer, : integer 10bit ] selectable", "[2]->c_promote(20bit integer)" );
	t( "float optional", "[ : float, : rational ] selectable", "[0]->c_optional" );
	t( "()->(:integer)", "[ :()->(x:integer) raw, :()->(x:float) ] selectable", "[0]->c_bind_raw_func" );
	
	//TODO: should add a generic assert method which ensures we don't take fields of optionals/lvalues!
	t( "float", "[ : integer, : float ] selectable lvalue", "c_drop_lvalue->[1]" );
	t( "integer", "[ : integer shared, : float ] selectable optional shared", 
		"c_deref_optional->c_deref->[0]->c_deref" );
		
	//if target is selectable source doesn't select
	t( "[ : integer, : float ] selectable", "[ : integer, : float ] selectable", "" );
}

BOOST_AUTO_TEST_CASE( value_ptr )
{
	tester t;
	t( "float value_ptr", "float lvalue", "c_addr_of" );
	t( "integer value_ptr", "integer optional lvalue", "c_deref_optional->c_addr_of" );
	t( "integer value_ptr", "integer shared lvalue", "c_deref->c_addr_of" );
	
	t( "integer value_ptr", "integer value_ptr", "" );
	t( "integer value_ptr", "integer value_ptr lvalue", "c_drop_lvalue" );
	
	t( "integer value_ptr", "integer", cerr::mismatch_value_ptr );
}

BOOST_AUTO_TEST_CASE( array_ptr )
{
	tester t;
	t( "array<float> raw", "array<float>", "c_addr_of" );
	t( "array<float> raw", "array<float optional>", cerr::invalid_array_cvt );
	t( "array<float> raw", "array<rational>", cerr::invalid_array_cvt );
	t( "array<float> raw", "optional array<float>", "c_deref_optional->c_addr_of" ) ;
	t( "array<float> raw", "shared lvalue array<float>", "c_drop_lvalue->c_deref->c_addr_of" ) ;
	
	t( "array<float> raw", "float", cerr::mismatch_fundamental );
	
	t( "array<float> raw", "array<float> raw", "" );
}

/**
	Be aware in this test the recursive type is not actually valid in Leaf. The type
	system supports it, but it creates an infinitely sized structure! This isn't the
	concern of the type system, thus it should always work here.
*/
BOOST_AUTO_TEST_CASE( recursive )
{
	tester t;
	t.add_defined_type( "point", "[ x : integer, pt : point ]" );
	
	t( "point", "point", "" );
	t( "point optional", "(:point)", 
		"c_collapse->c_optional" );
	t( "point", "[ : integer, : point ] shared", 
		"c_deref->c_pack {x=[0] pt=[1]}->c_set_distinct(point)" );
	t( "point", "(:[ : integer, : point lvalue ])", 
		"c_collapse->c_pack {x=[0] pt=[1]->c_drop_lvalue}->c_set_distinct(point)" );
	t( "point", "[ : integer, : [ : integer, : point ] ]", 
		"c_pack {x=[0] pt=[1]->c_pack {x=[0] pt=[1]}->c_set_distinct(point)}->c_set_distinct(point)" );
}

BOOST_AUTO_TEST_CASE( lossy ) 
{
	tester t;
	t( "integer 8bit", "integer lossy", "c_loss(8bit integer)" );
	t( "float low", "float lossy", "c_loss(low float)" );
	t( "integer", "float lossy", "c_loss(normal float_to_integer)");
	t( "float low", "integer 128bit lossy", "c_loss(low integer_to_float)" );
	t( "integer", "binary lossy", "c_loss(normal binary_to_integer)" );
	t( "binary", "integer lossy", "c_loss(normal integer_to_binary)" );
	t( "integer 10bit", "integer 20bit lossy", "c_loss(10bit integer)" );

	t( "integer 8bit", "integer lossy", "c_loss(8bit integer)" );
	
	//no lossy conversion is needed here, normal fixation
	t( "float", "rational 91bit lossy", "c_fixate(normal float)" );
	//it doesn't seem to make sense to allow lossy rational conversion, either the rational is a valid
	//size or its guaranteed to be a loss (not likely wanted).
	t( "integer 30bit", "rational 40bit lossy", cerr::range_loss);
}

BOOST_AUTO_TEST_SUITE_END()
