#define SHOW_DEBUG 0
#include "type_converter_common.hpp"

BOOST_AUTO_TEST_SUITE( type_converter_unify )

struct tester_unify : public type_converter_tester
{
	void operator()( char const * a, char const * b, char const * expect ) 
	{
		//order shouldn't matter, so try both
		impl(a,b,expect);
		impl(b,a,expect);
	}
	
	void impl( char const * a, char const * b, char const * expect ) 
	{
		expect_test ex;
		ex.add( "A", a );
		ex.add( "B", b );
		
		auto atype = parse(a);
		auto btype = parse(b);
		
		auto unity = tcvt.unify( atype, btype );
		ex.check( unity.okay(), unity.explain_error() );
		unity.result.validate();
		
		ex.check_actual( leaf::dump::get_extr_type(unity.result), expect );
	}

	void operator()( std::vector<char const*> a, char const * expect ) 
	{
		expect_test ex;
			
		extr_type_ref types[a.size()];
		for( size_t i=0; i < a.size(); i++ ) {
			ex.add( a[i] );
			types[i] = parse(a[i]);
		}
		
		auto unity = tcvt.unify( types, a.size() );
		ex.check( unity.okay(), unity.explain_error() );
		unity.result.validate();
		
		ex.check_actual( leaf::dump::get_extr_type(unity.result), expect );
	}

	void operator()( char const * a, char const * b, cerr expect ) 
	{
		expect_test ex;
		ex.add( a );
		ex.add( b );
		
		auto atype = parse(a);
		auto btype = parse(b);
		
		auto unity = tcvt.unify( atype, btype );
		ex.check( unity.error == expect, unity.explain_error(), "unexpected error" );
		BOOST_CHECK(true);
	}
	
};

BOOST_AUTO_TEST_CASE( basic )
{
	tester_unify t;
	t( "integer 16bit", "integer", "integer" );
	t( "integer 8bit", "integer 90bit", "integer 90bit" );
	
	//rationals
	t( "rational", "rational", "rational" );
	t( "rational 17bit", "rational 15bit", "rational 17bit" );
	t( "rational 30bit", "rational", "rational" );
	//TODO: signed
	
	//rational/fixed_number
	t( "rational 17bit", "integer", "integer" );
	t( "rational 1bit", "integer 8bit", "integer 8bit" );
	
	//shared can be dereferenced
	t( "integer shared", "integer", "integer" );
	t( "integer 10bit shared", "integer", "integer" );

	//equivalent (sub-)types are always okay even if unify can't deal with them
	t.add_defined_type( "point", "[ x : float, y : float ]" );
	t( "point", "point", "point" );
	t( "optional point", "point", "point optional" );
	t( "point shared", "optional point", "point optional" );

	//lvalues
	t( "integer lvalue", "integer lvalue", "integer lvalue" );
	t( "integer lvalue", "integer 8bit lvalue", "integer" );
	t( "lvalue optional integer", "lvalue shared integer", "integer optional" );
	
	//value-list
	t( "(:integer shared)", "integer", "integer" );
	t( "integer", "(:integer optional)", "integer optional" );
	t( "(:(:integer 16bit))", "(:integer 8bit)", "( : integer 16bit )" );
	t( "(:integer,:float lvalue)", "(:(:integer,:float lvalue))", "( : integer, : float lvalue )" );
	t( "integer lvalue", "(:integer lvalue)", "integer lvalue" );
	
	//tuples
	t( "[ : integer ]", "[ : integer ]", "[ : integer ]" );
	t( "[ : integer shared ]", "[ : integer optional ]", "[ : integer optional ]" );
	t( "[:float optional]", "[a:float]", "[ a : float optional ]" );
	
	//failing cases
	t( "float", "integer", cerr::mismatch_fundamental );
	t( "[:integer]", "[]", cerr::mismatch_size );
	t( "(:integer)", "(:integer,:integer)", cerr::mismatch_size );
	t( "[a:float]", "[b:float]", cerr::mismatch_field_name );
	t( "rational", "[:float]", cerr::mismatch_fundamental );
	
	//TODO: I think there is some errors in how an optional rational might be handled, or just optional
	//since two non-compatible fundamentals can be converted and both are optional the conversion
	//chain may not work
}

BOOST_AUTO_TEST_CASE( binary )
{
	tester_unify t;
	t( "binary", "binary", "binary" );
	t( "binary 20bit", "binary 10bit", "binary 20bit" );
	
	//larger types won't be introduce by unification
	t( "integer 16bit", "binary 16bit", cerr::mismatch_fundamental );
	//unification is not meant to resolve what may be ambiguity, that is, while these can be
	//converted it isn't clear they should unify as such
	t( "integer 32bit", "binary 8bit", cerr::mismatch_fundamental );
	t( "float", "binary 8bit", cerr::mismatch_fundamental );
}

BOOST_AUTO_TEST_CASE( vector )
{
	tester_unify t;
	t( {"integer", "integer", "integer 80bit"}, "integer 80bit" );
	t( {"integer"}, "integer" );
	t( {"integer 16bit", "integer shared 8bit", "integer optional", "integer 128bit", "integer shared" },
		"integer 128bit optional" );
	t( { "optional integer lvalue", "shared integer lvalue", "integer 4bit lvalue" }, "integer optional" );
}

BOOST_AUTO_TEST_CASE( optionals )
{
	tester_unify t;
	
	//optional always adds, never removes
	t( "optional integer", "integer", "integer optional" );
	t( "optional integer", "integer 200bit", "integer 200bit optional" );
	t( "float", "optional float", "float optional" );

	//TODO: two mismtached optionals can still be unified ?
	//t( "optional integer tiny", "optional float", "optional float" );
	//t( "optional integer", "optional rational 3bit", "optional integer" );
}

BOOST_AUTO_TEST_CASE( convert_optional )
{
	tester_unify t;
	
	t( "optional integer", "optional integer convert_optional", "integer optional" );
	t( "optional 7bit integer", "optional integer 4bit convert_optional", "integer 7bit optional" );
	
	//since the intenger is bare now it is conversion which adds the optional, thus it is also convertible
	t( "integer", "optional integer convert_optional", "integer convert_optional optional" );
	
	t( "optional integer convert_optional", "optional integer convert_optional",
		"integer convert_optional optional" );
	t( "optional integer 1bit convert_optional", "optional integer 3bit convert_optional",
		"integer 3bit convert_optional optional" );
		
	t.add_defined_type( "point", "[ x : float, y : float ]" );
	t( "optional point", "point optional convert_optional", "point optional" );
		
}

BOOST_AUTO_TEST_SUITE_END()