#define SHOW_DEBUG 0
#include "type_converter_common.hpp"

#include "platform/platform.hpp"
#include <cstdio>
#include <stdarg.h>

BOOST_AUTO_TEST_SUITE( type_identifier_determine )

struct tester : public type_converter_tester
{
	bool keep_variant;
	tester()
	{
		keep_variant = false;
	}
	
	void operator()( char const * spec, char const * expect_fmt, ... ) 
	{
		char expect[1024];
		va_list ap;
		va_start(ap, expect_fmt);
		std::vsprintf( expect, expect_fmt, ap );
		
		expect_test ex;
		ex.add( "Spec", spec );
	
		auto spectype = parse(spec);
		leaf::dump d;
		d.set_keep_variant( keep_variant );
		d.write_intr_type( spectype );
		ex.check_actual( d.str(), expect );
		
		va_end(ap);
	}
	
	void operator()( char const * spec, cerr expect )
	{
		expect_test ex;
		ex.add( "Spec", spec );
		
		type_identifier::resolve_t result;
		try {
			auto s = source::create( spec, std::strlen(spec) );
			auto vt_spec = intr_type_parser::parse( *s );
			result = tid.determine( vt_spec );
		} catch( ... ) {
			ex.except();
		}
		
		ex.check( result.error == expect, result.explain_error(), "wrong error" );
	}
};

BOOST_AUTO_TEST_CASE( basic )
{
	tester t;
	t( "integer ", "integer" );
	t( "float lvalue", "float lvalue" );
	t( "boolean", "boolean" );
	t( "", cerr::okay ); //TODO: uhm...
	t( "array<char>", "array< char >" );
	t( "shared integer", "integer shared" );
	t( "float normal", "float" );
	t( "integer 17bit", "integer 17bit" );
	t( "rational optional", "rational optional" );
	t( "binary", "binary" );
	t( "binary 90bit", "binary 90bit" );
	t( "no_such_symbol", cerr::unknown_identifier );
}

BOOST_AUTO_TEST_CASE( alt_names )
{
	tester t;
	t( "octet", "binary 8bit" );
}

BOOST_AUTO_TEST_CASE( native_types )
{
	tester t;
	t.keep_variant = true;
	t( "abi_int", "integer %dbit", platform::abi_int_size );
	t( "abi_short", "integer %dbit", platform::abi_short_size );
	t( "abi_long", "integer %dbit", platform::abi_long_size );
	t( "abi_long_long", "integer %dbit", platform::abi_long_long_size );
	//NOTE: this is not likely truly platform independent
	t( "abi_float", "float low" );
	t( "abi_double", "float normal" );
	t( "abi_long_double", "float high" );
	
	//TODO: these are temporary until appropriate types appear
	t( "abi_char", "binary %dbit", platform::abi_char_size );
	t( "abi_ptr", "binary %dbit", platform::abi_ptr_size );
}

BOOST_AUTO_TEST_CASE( tuple )
{
	tester t;
	t( "[:integer,:float]",
		"[ : integer, : float ]" );
	t( "[]", "[ ]" );
	t( "[:]", cerr::okay );
	t( "[:[:integer],:[],:[:float,:integer]]",
		"[ : [ : integer ], : [ ], : [ : float, : integer ] ]" );
		
	t( "abi_pack [ : integer, : float]", "[ : integer, : float ] abi_pack" );
	t( "[ ] abi_pack", "[ ] abi_pack" );
}

BOOST_AUTO_TEST_CASE( tuple_deco )
{
	tester t;
	t( "lvalue []", "[ ] lvalue" );
	
	t( "selectable [ :integer, : float ]", "[ : integer, : float ] selectable" );
}

BOOST_AUTO_TEST_CASE( value_list )
{
	tester t;
	t( "()" , "( )" );
	t( "(:integer)", "( : integer )" );
	t( "(a:integer)", "( a : integer )" );
	t( "(a)", cerr::okay );
}

BOOST_AUTO_TEST_CASE( signature )
{
	tester t;
	t( "(a:integer)->()", "( a : integer ) -> ( )" );
	t( "(a:float,)->()", "( a : float ) -> ( )" );
	t( "(:integer,:integer)->()", "( : integer, : integer ) -> ( )" );
	t( "()->() raw", "( ) -> ( ) raw" );
	//TODO: member?
}

BOOST_AUTO_TEST_CASE( lossy )
{
	tester t;
	t( "integer lossy", "integer lossy" );
}

//ensure that these fail
BOOST_AUTO_TEST_CASE( broken )
{
	tester t;
	t( "integer float", cerr::redundant_fundamental ); //only one fundamental allowed
	
	//these are disallowed due to reserved words, though they are in a different namespace
	//TODO: test_noparse( "{integer}" )
	//TODO: test_noparse( "{integer,float}" )
}

BOOST_AUTO_TEST_CASE( array )
{
	tester t;
	t( "array<integer>", "array< integer >" );
	t( "array<shared integer>", "array< integer shared >" );
	t( "shared array<array<float shared>>", "array< array< float shared > > shared" );
	t( "[:array<[:integer,:float]>]", "[ : array< [ : integer, : float ] > ]" );
	
	t( "array<integer> raw", "array< integer > raw" );
	t( "array<array<integer> raw> raw", "array< array< integer > raw > raw" );
}

BOOST_AUTO_TEST_CASE( classes )
{
	tester t;
	auto ctx = make_shared<intr_type_custom>();
	ctx->distinct = true;
	ctx->distinct_name = "happy";
	t.sc->mark_class_root( ctx, shared_ptr<statement_block>() );
	t( "()->() member", "( ) -> ( ) context<happy> unbound" );
	
	//only the outermost function should be implicitly a member function
	t( "()->( : ()->() ) member", "( ) -> ( : ( ) -> ( ) ) context<happy> unbound" );
}

BOOST_AUTO_TEST_CASE( alias )
{
	tester t;
	t.add_type_alias( "sinter", "integer 8bit" );
	t( "sinter", "integer 8bit" );
	t( "shared sinter", "integer 8bit shared" );
	t( "sinter float", cerr::redundant_fundamental );
	t( "[ : sinter ]", "[ : integer 8bit ]" );
	
	t.add_type_alias( "ptr", "optional shared" );
	t( "float ptr", "float optional shared" );
	
	t.add_type_alias( "point", "[ x : integer, y : integer ]" );
	t( "point", "[ x : integer, y : integer ]" );
	
	t( "[ : sinter, pt : point ] ptr", "[ : integer 8bit, pt : [ x : integer, y : integer ] ] optional shared" );
}


BOOST_AUTO_TEST_CASE( alias_distinct )
{
	tester t;
	t.add_defined_type( "point", "[ x : float, y : float ]" );
	t.add_type_alias( "point_shared", "point shared" );
	t( "point_shared", "point shared" );
}

BOOST_AUTO_TEST_CASE( scoped )
{
	tester t;
	auto ns = scope::create(scope_style_t::declarative);
	ns->set_namespace( t.sc, "branch" );
	t.add_defined_type( "cone", "[ a : integer ]" );
	t( "branch.cone", "branch.cone" );
}

BOOST_AUTO_TEST_SUITE_END()