#define SHOW_DEBUG 1
#include "type_converter_common.hpp"

BOOST_AUTO_TEST_SUITE( type_identifier_infer )

struct tester_infer : public type_converter_tester
{
	bool init_refer;
	
	tester_infer() {
		init_refer = false;
	}
	
	void operator()( char const * spec, char const * assign, 
		char const * expect )
	{
		expect_test ex;
		ex.add( "Spec", spec );
		ex.add( "Assign", assign );
		
		type_identifier::resolve_spec_t result;
		
		auto s = source::create( spec, std::strlen(spec) );
		auto vt_spec = intr_type_parser::parse( *s );
		auto vt_assign = parse( assign );

		try {
			result = tid.infer_spec( *vt_spec, vt_assign, init_refer );
		} catch( ... ) {
			ex.except();
		}
		
		ex.check( result.okay(), result.explain_error() );
		ex.check_actual( dump::get_type_spec( *result.result ), expect );
	}
	
	void operator()( char const * spec, char const * assign, cerr expect )
	{
		expect_test ex;
		ex.add( "Spec", spec );
		ex.add( "Assign", assign );
		
		auto s = source::create( spec, std::strlen(spec) );
		auto vt_spec = intr_type_parser::parse( *s );
		auto vt_assign = parse( assign );

		type_identifier::resolve_spec_t result;
		try {
			result = tid.infer_spec( *vt_spec, vt_assign, init_refer );
		} catch( ... ) {
			ex.except();
		}

		ex.check( result.error == expect, result.explain_error(), "wrong error" );
	}
};

BOOST_AUTO_TEST_CASE( basic )
{
	tester_infer t;
	t( "", "integer", "integer" );
	t( "shared", "float", "float shared" );
	t( "value_ptr", "float", "float value_ptr" );
	t( "float", "float shared", "float" );
	t( "integer", "integer 16bit", "integer" );
	t( "float", "float low", "float" );
	t( "", "integer lvalue 16bit", "integer 16bit" );
	t( "", "(:(:[:integer,:float]))", "[ : integer, : float ]" );
	t( "optional", "boolean", "boolean optional" );
	t( "boolean", "optional boolean", "boolean" );
	t( "binary lvalue", "binary", "binary lvalue" );
	t( "binary", "binary lvalue", "binary" );
	
	t( "", "(:integer,:integer)", cerr::disallow_value_list );
	
	t( "", "rational 19bit", "integer" );
	t( "", "rational", "float" );
	t( "integer 16bit", "rational 88bit", "integer 16bit" ); //note that conversion is not checked in inference (the assingment may not be possible)
	
	//since a fundamental is provider the modifiers are not taken on the outermost type
	t( "[:]", "[ : integer ] lvalue", "[ : integer ]" );
	t( "[:]", "[ : integer ] shared", "[ : integer ]" );
}

BOOST_AUTO_TEST_CASE( functions )
{
	tester_infer t;
	t( "()->() raw", "()->()", "( ) -> ( ) raw" );
	t( "()->()", "()->() raw", "( ) -> ( )" );
	t( "", "()->() unbound", "( ) -> ( ) unbound" ); //NOTE: this isn't really a valid type though, it needs a context!
	t( "()->()", "()->()", "( ) -> ( )" );
}

//came from examples which weren't working
BOOST_AUTO_TEST_CASE( practical )
{
	tester_infer t;
	t( "float low", "float low", "float low" );
	t( "", "( a: optional boolean )->()", "( a : boolean optional ) -> ( )" );
	t( "", "( a: optional boolean value_ptr )->()", "( a : boolean optional value_ptr ) -> ( )" );
	t( "", "shared float", "float shared" );
	t( "", "value_ptr boolean", "boolean value_ptr" );
	
	t( "binary 5bit", "rational 3bit", "binary 5bit" );
	t( "integer 11bit", "binary 10bit", "integer 11bit" );
	t( "float", "integer 15bit", "float" );
}

BOOST_AUTO_TEST_CASE( infer_defined )
{
	tester_infer t;
	t.add_defined_type( "point", "[ x : float, y : float ]" );

	//TODO: really need a better way to handle user types
	t( "", "point", "prototype(point)" );
	t( "optional", "point", "prototype(point) optional" );
	t( "integer", "point", "integer" );
	t( "point", "[:float,:float]", "prototype(point)" );
	
	t( "", "array<point>", "array< prototype(point) >" );
}

BOOST_AUTO_TEST_CASE( infer_array )
{
	tester_infer t;
	t( "", "array<integer>", "array< integer >" );
	t( "", "array<integer> raw", "array< integer > raw" );
}

BOOST_AUTO_TEST_CASE( infer_init_refer )
{
	tester_infer t;
	t.init_refer = true;
	t( "", "integer", "integer shared" );
	
	//TODO: need that spec_type has shared set to "false" by default, not undefined
	//t( "integer", "integer", cerr::mismatch_shared ); //since non-specific shared is off by default
}

BOOST_AUTO_TEST_CASE( infer_tuple )
{
	tester_infer t;
	t( "[:]", "[ : integer ]", "[ : integer ]" );
	t( "[a,:]", "[:float,b:boolean]", "[ a : float, b : boolean ]" );
	t( "[:optional,:128bit]", "[:integer,:integer]", "[ : integer optional, : integer 128bit ]" );
	
	t( "", "[ : rational 2bit ]", "[ : integer ]" );
	t( "optional", "[ : rational 2bit, : [ ] ]", "[ : integer, : [ ] ] optional" );
	
	t( "[ : integer, : float ] abi_pack", "[ : rational 4bit, : rational 8bit ]",
		"[ : integer, : float ] abi_pack" );
}

BOOST_AUTO_TEST_CASE( selectable )
{
	tester_infer t;
	//nothing can be inferred from selectables, so it becomes a 'determine' call
	t( "integer", "selectable [:float,:integer]", "integer" );
	t( "", "selectable [:float,:integer]", "" );
	t( "selectable", "[ : float, : integer ]", "[ : float, : integer ] selectable" );
	
	//selectable can however be retained if marked in the target
	t( "selectable", "[ : float, : integer ] selectable", "[ : float, : integer ] selectable" );
	
	//not a useful (nor valid now) type, but infer doesn't care
	t( "selectable", "integer", "integer selectable" );
	t( "selectable", "[]", "[ ] selectable" );
}

BOOST_AUTO_TEST_CASE( class_members )
{
	tester_infer t;
	auto ctx = make_shared<intr_type_custom>();
	ctx->distinct = true;
	ctx->distinct_name = "happy";
	t.sc->mark_class_root( ctx, shared_ptr<statement_block>() );
	
	t( "", "()->() member", "( ) -> ( ) context<happy> unbound" );
}

BOOST_AUTO_TEST_SUITE_END()