#include "util/number.hpp"
#include "core_types.hpp"

#include <boost/regex.hpp>

#define SHOW_DEBUG 1
#include "util/debug.hpp"

namespace util
{

number::number()
{
}

number::number( number const & o )
{
	value = o.value;
}

number::number(mpq_class ovalue)
{
	value = ovalue;
}

number & number::operator=(number const & o)
{
	value = o.value;
	return *this;
}

namespace 
{
	template<typename T>
	mpz_class parse_bitz( T begin, T end, int base, int & digits )
	{
		digits = 0;
		if( begin == end )
			return mpz_class( 0 );
		//strip separators (just used for visual clarification)
		char buf[end - begin + 1];
		char * c = buf;
		while( begin != end ) {
			if( *begin != '_' ) {
				*c++ = *begin;
				digits++;
			}
			begin++;
		}
		*c = 0;
		return mpz_class( buf, base );
	}
} //eon anon

/*static*/ number number::parse_base( std::string const & str, unsigned & meta, int base )
{
	meta = pdf_is_exact;
	
	//this is intentionally far more lenient than the one in node_parser so this class is flexible
	//note that due to conflict on 'e' hex can't have exponent, for simplicity/clarity is thus only
	//accepted on exact decimal representation (which I think is a common expectation)
	static boost::regex re_decimal( "([+-]?)([0-9_]*)(\\.([0-9_]*))?([eE]([+-]?)([0-9]+))?" );
	static boost::regex re_other( "([+-]?)([0-9a-fA-F_]*)(\\.([0-9a-fA-F_]*))?" );
	bool is_decimal = base == 10;
	boost::match_results<std::string::const_iterator> what;
	if( !boost::regex_match( str, what, is_decimal ? re_decimal : re_other ) )
		PRE_FAIL( std::string( "invalid number: " + str ) );
		
	mpz_class part;
	mpq_class combine;
	bool negative = false;
	
	if( what[1].matched && *what[1].first == '-' )
		negative = true;
		
	int digits;
	if( what[2].matched )
		combine = parse_bitz( what[2].first, what[2].second, base, digits );
	
	if( what[3].matched ) {
		meta |= pdf_had_decimal;
		part = parse_bitz( what[4].first, what[4].second, base, digits );
		if( digits ) {
			mpz_class div(base);
			mpz_pow_ui( div.get_mpz_t(), div.get_mpz_t(), digits );
			combine += mpq_class( part, div );
		}
	}
	
	if( is_decimal && what[5].matched ) {
		meta |= pdf_had_exponent;
		mpz_class e = parse_bitz( what[7].first, what[7].second, base, digits );
		mpz_class ebase(base), ebase_exp;
		
		//TODO: check that "e" is constrained to unsigned long, check some limit as well
		mpz_pow_ui( ebase_exp.get_mpz_t(), ebase.get_mpz_t(), e.get_ui() );
		
		if( *what[6].first == '-' )
			combine /= ebase_exp;
		else
			combine *= ebase_exp;
	}

	if( negative )
		combine *= -1;
	combine.canonicalize();
	return number( combine );
}

std::string number::to_string(bool force_denom) const
{
	std::string ret = value.get_str(10);
	if( is_integral() && force_denom )
		ret += "/1";
	return ret;
}

bool number::is_negative() const
{
	return sgn(value) < 0;
}

bool number::is_integral() const
{
	//assume canonical form already
	return value.get_den() == 1;
}

unsigned number::signed_bits() const
{
	mpz_class ival( value );
	return unsigned( mpz_sizeinbase( ival.get_mpz_t(), 2 ) + 1 );
}

unsigned number::unsigned_bits() const
{
	mpz_class ival( value );
	return unsigned( mpz_sizeinbase( ival.get_mpz_t(), 2 ) );
}

bool number::fits_signed( unsigned bits ) const
{
	if( bits < 2 )
		return 0;
	bits--; //drop sign
	
	mpz_class base(2);
	mpz_class max;
	mpz_pow_ui( max.get_mpz_t(), base.get_mpz_t(), bits );
	
	mpz_class ival( value );
	if( sgn(ival) < 0 ) {
		return ival >= (-max-1);
	} else {
		return ival <= max;
	}
}

bool number::fits_unsigned( unsigned bits ) const
{
	mpz_class base(2);
	mpz_class max;
	mpz_pow_ui( max.get_mpz_t(), base.get_mpz_t(), bits );
	
	mpz_class ival( value );
	return ival <= max;
}

//TODO: All the below conversions do not support the large integer types, 64-bit, 128-bit (where possible)
//or long double!!! (well get_si does work on platforms where long is 64bit)
intmax_t number::as_int() const
{
	return mpz_class(value.get_num() / value.get_den()).get_si();
}

uintmax_t number::as_uint() const
{
	return mpz_class(value.get_num() / value.get_den()).get_ui();
}

double number::as_double() const
{
	return value.get_d();
}

number number::neg() const 
{
	number r;
	mpq_neg( r.value.get_mpq_t(), value.get_mpq_t() );
	return r;
}

} //eon util
